Package: Immutables
Type: Package
Title: Fast and Functional Data Structures
URL: https://oneilsh.github.io/Immutables/,
        https://github.com/oneilsh/Immutables
BugReports: https://github.com/oneilsh/Immutables/issues
Version: 1.1.0
Authors@R: c(
    person(given = "Shawn T.", family = "O'Neil", email = "shawn@tislab.org", role = c("aut", "cre"))
  )
Description: Provides fast, side-effect free data structures, including
    catenable named lists, priority queues, double-ended queues, ordered
    sequences, and interval indices. Implementation is based on the
    finger-tree data structure of Hinze and Paterson (2006)
    <doi:10.1017/S0956796805005769>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: coro, Rcpp, lambda.r
LinkingTo: Rcpp
Suggests: bench, covr, dplyr, ggplot2, ggtext, igraph, IRanges, knitr,
        microbenchmark, pkgdown, pkgload, rmarkdown, rprojroot,
        rstackdeque, rticles, S4Vectors, scales, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-08-17 18:33:52 UTC; oneilsh
Author: Shawn T. O'Neil [aut, cre]
Maintainer: Shawn T. O'Neil <shawn@tislab.org>
Repository: CRAN
Date/Publication: 2026-08-21 17:50:08 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-08-21 19:05:25 UTC; unix
Archs: Immutables.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("$",flexseq)
S3method("$",interval_index)
S3method("$",ordered_sequence)
S3method("$",priority_queue)
S3method("$<-",flexseq)
S3method("$<-",interval_index)
S3method("$<-",ordered_sequence)
S3method("$<-",priority_queue)
S3method("[",flexseq)
S3method("[",interval_index)
S3method("[",ordered_sequence)
S3method("[",priority_queue)
S3method("[<-",flexseq)
S3method("[<-",interval_index)
S3method("[<-",ordered_sequence)
S3method("[<-",priority_queue)
S3method("[[",flexseq)
S3method("[[",interval_index)
S3method("[[",ordered_sequence)
S3method("[[",priority_queue)
S3method("[[<-",flexseq)
S3method("[[<-",interval_index)
S3method("[[<-",ordered_sequence)
S3method("[[<-",priority_queue)
S3method(add_monoids,default)
S3method(add_monoids,flexseq)
S3method(add_monoids,interval_index)
S3method(add_monoids,ordered_sequence)
S3method(add_monoids,priority_queue)
S3method(as.list,flexseq)
S3method(as.list,interval_index)
S3method(as.list,ordered_sequence)
S3method(as.list,priority_queue)
S3method(as_flexseq,default)
S3method(as_flexseq,flexseq)
S3method(as_flexseq,interval_index)
S3method(as_flexseq,ordered_sequence)
S3method(as_flexseq,priority_queue)
S3method(c,flexseq)
S3method(c,ordered_sequence)
S3method(c,priority_queue)
S3method(coro::as_iterator,flexseq)
S3method(coro::as_iterator,priority_queue)
S3method(fapply,flexseq)
S3method(fapply,interval_index)
S3method(fapply,ordered_sequence)
S3method(fapply,priority_queue)
S3method(insert,default)
S3method(insert,interval_index)
S3method(insert,ordered_sequence)
S3method(insert,priority_queue)
S3method(length,flexseq)
S3method(length,interval_index)
S3method(length,ordered_sequence)
S3method(length,priority_queue)
S3method(locate_by_predicate,flexseq)
S3method(merge,flexseq)
S3method(merge,interval_index)
S3method(merge,ordered_sequence)
S3method(merge,priority_queue)
S3method(print,flexseq)
S3method(print,interval_index)
S3method(print,ordered_sequence)
S3method(print,priority_queue)
S3method(split_around_by_predicate,flexseq)
S3method(split_around_by_predicate,priority_queue)
S3method(split_at,flexseq)
S3method(split_at,priority_queue)
S3method(split_by_predicate,flexseq)
S3method(split_by_predicate,priority_queue)
S3method(str,flexseq)
S3method(unlist,flexseq)
export(add_monoids)
export(as_flexseq)
export(as_interval_index)
export(as_ordered_sequence)
export(as_priority_queue)
export(count_between)
export(count_key)
export(elements_between)
export(fapply)
export(flexseq)
export(get_measure)
export(get_measures)
export(insert)
export(insert_at)
export(interval_index)
export(locate_by_predicate)
export(loop)
export(lower_bound)
export(max_endpoint)
export(max_key)
export(max_priority)
export(measure_monoid)
export(min_endpoint)
export(min_key)
export(min_priority)
export(ordered_sequence)
export(peek_all_containing)
export(peek_all_key)
export(peek_all_max)
export(peek_all_min)
export(peek_all_overlaps)
export(peek_all_point)
export(peek_all_within)
export(peek_at)
export(peek_back)
export(peek_containing)
export(peek_front)
export(peek_key)
export(peek_max)
export(peek_min)
export(peek_overlaps)
export(peek_point)
export(peek_within)
export(plot_structure)
export(pop_all_containing)
export(pop_all_key)
export(pop_all_max)
export(pop_all_min)
export(pop_all_overlaps)
export(pop_all_point)
export(pop_all_within)
export(pop_at)
export(pop_back)
export(pop_containing)
export(pop_front)
export(pop_key)
export(pop_max)
export(pop_min)
export(pop_overlaps)
export(pop_point)
export(pop_within)
export(priority_queue)
export(push_back)
export(push_front)
export(split_around_by_predicate)
export(split_at)
export(split_by_predicate)
export(upper_bound)
export(validate_name_state)
export(validate_tree)
import(lambda.r)
importFrom(Rcpp,evalCpp)
importFrom(coro,loop)
importFrom(graphics,par)
importFrom(graphics,plot.new)
importFrom(graphics,polygon)
importFrom(graphics,strheight)
importFrom(graphics,strwidth)
useDynLib(Immutables, .registration = TRUE)

# Immutables <img src="man/figures/logo.png" align="right" height="139" alt="" />

<!-- Uncomment after CRAN submission:
[![CRAN status](https://www.r-pkg.org/badges/version/Immutables)](https://CRAN.R-project.org/package=Immutables)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/Immutables)](https://CRAN.R-project.org/package=Immutables)
-->
[![Lifecycle: experimental](https://lifecycle.r-lib.org/reference/figures/lifecycle-experimental.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/oneilsh/immutables/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/oneilsh/immutables/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/oneilsh/immutables/branch/main/graph/badge.svg)](https://app.codecov.io/gh/oneilsh/immutables)
[![pkgdown](https://github.com/oneilsh/immutables/actions/workflows/pkgdown.yaml/badge.svg)](https://github.com/oneilsh/immutables/actions/workflows/pkgdown.yaml)
[![docs](https://img.shields.io/badge/docs-pkgdown-blue)](https://oneilsh.github.io/immutables/)
[![DOI](https://zenodo.org/badge/154557443.svg)](https://doi.org/10.5281/zenodo.19686096)

`Immutables` implements several immutable, or persistent, data
structures for R: operations return modified copies while remaining fast 
and true to R's side-effect-free functional nature with `list`-like semantics.


- [`flexseq`](https://oneilsh.github.io/immutables/articles/flexseq.html)s provide list-like operations:
   - indexed and named element access
   - push/pop/peek from either end for double-ended queue behavior
   - efficient insertion, splitting, and concatenation
- [`priority_queue`](https://oneilsh.github.io/immutables/articles/priority-queues.html)s associate items with priority values and provide
  min and max peek/pop by priority and fast insertion.
- [`ordered_sequence`](https://oneilsh.github.io/immutables/articles/ordered-sequences.html)s associate items with key values and keep the elements in
  sorted order by key. These may be similarly be inserted/popped/peeked by key
  value as well as position. Keys may be duplicated, with first-in-first-out order within key groups.
- [`interval_index`](https://oneilsh.github.io/immutables/articles/interval-indices.html)es store items associated with interval ranges, supporting
  point as well as interval overlaps/contains/within queries. Items are kept
  in start-order enabling ordered sequence operations and sweep-line
  algorithms.

## Speed and Flexibility

Backed by monoid-annotated finger trees as described by
[Hinze and Paterson](https://doi.org/10.1017/S0956796805005769), most
operations are constant time, amortized constant time, or $O(\log(n))$.
Core functions are implemented in C++ (via Rcpp) for speed, with matching
pure-R reference implementations using `lambda.r` syntax similar to the paper's
Haskell.

The [developer API](https://oneilsh.github.io/immutables/articles/developer-api.html)
supports the addition of custom structures via combinations of monoids and
measures. See the [full documentation](https://oneilsh.github.io/immutables/)
for vignettes and reference, including a [gallery of worked examples](https://oneilsh.github.io/immutables/articles/gallery.html)
and [benchmarks](https://oneilsh.github.io/immutables/articles/benchmarks.html).

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

loaded <- tryCatch({
  pkgload::load_all(quiet = TRUE)
  TRUE
}, error = function(e) FALSE)

if(!loaded) {
  library(Immutables)
}

library(microbenchmark)
library(ggplot2)

# Two env-var gates control this vignette:
#   IMMUTABLES_RUN_SLOW=true       -> re-run all benchmarks and save
#                                     results to inst/extdata/benchmarks.rds
#   IMMUTABLES_SAVE_FIGURES=true   -> also write pub-ready figures to
#                                     paper/figures/*.pdf
# When neither is set (CRAN, ordinary pkgdown build) the vignette loads
# cached results and renders plots without re-timing.
run_benchmarks <- identical(Sys.getenv("IMMUTABLES_RUN_SLOW"), "true")
save_figures   <- identical(Sys.getenv("IMMUTABLES_SAVE_FIGURES"), "true")
repeats <- 8L

# Accumulator populated either by the benchmark chunks (when
# run_benchmarks) or by the loader chunk below (otherwise).
results_list <- list()

bench_one <- function(rows, impl, op, n, repeats, setup, bench) {
  gc(FALSE)
  time_us <- numeric(repeats)
  for(i in seq_len(repeats)) {
    state <- setup()
    gc(FALSE)
    time_us[[i]] <- microbenchmark(bench(state), times = 1L)$time[[1L]] / 1000
  }
  push_back(rows, data.frame(impl = impl, op = op, n = n, time_us = time_us))
}

## ----load-cached, eval=!run_benchmarks, include=FALSE-------------------------
.cache_path <- system.file("extdata", "benchmarks.rds", package = "Immutables")
if(nzchar(.cache_path) && file.exists(.cache_path)) {
  results_list <- readRDS(.cache_path)
}

## ----sequence-bench, eval=run_benchmarks, cache=FALSE-------------------------
# repeats <- 6L
# 
# sequence_sizes <- 2^(10 + 0:6)
# rows <- flexseq()
# 
# for(n in sequence_sizes) {
#   cat("Sequence ops, size ", n, "\n")
#   vals <- function() as.list(sprintf("v_%06d", seq_len(n)))
#   mid <- as.integer(n / 2)
#   flex_setup <- function() list(s = as_flexseq(vals()), mid = mid)
#   list_setup <- function() list(s = vals(), mid = mid)
#   pair_flex  <- function() list(a = as_flexseq(vals()), b = as_flexseq(vals()))
#   pair_list  <- function() list(a = vals(), b = vals())
# 
#   rows <- bench_one(rows, "flexseq", "append",  n, repeats, flex_setup,
#     function(st) push_back(st$s, "z"))
#   rows <- bench_one(rows, "flexseq", "prepend", n, repeats, flex_setup,
#     function(st) push_front(st$s, "z"))
#   rows <- bench_one(rows, "flexseq", "get middle", n, repeats, flex_setup,
#     function(st) st$s[[st$mid]])
#   rows <- bench_one(rows, "flexseq", "replace middle", n, repeats, flex_setup,
#     function(st) { s <- st$s; s[[st$mid]] <- "y"; s })
#   rows <- bench_one(rows, "flexseq", "remove middle", n, repeats, flex_setup,
#     function(st) pop_at(st$s, st$mid)$remaining)
#   rows <- bench_one(rows, "flexseq", "concatenate", n, repeats, pair_flex,
#     function(st) c(st$a, st$b))
#   rows <- bench_one(rows, "flexseq", "split at middle", n, repeats, flex_setup,
#     function(st) split_at(st$s, st$mid))
# 
#   rows <- bench_one(rows, "base R list", "append",  n, repeats, list_setup,
#     function(st) c(st$s, list("z")))
#   rows <- bench_one(rows, "base R list", "prepend", n, repeats, list_setup,
#     function(st) c(list("z"), st$s))
#   rows <- bench_one(rows, "base R list", "get middle", n, repeats, list_setup,
#     function(st) st$s[[st$mid]])
#   rows <- bench_one(rows, "base R list", "replace middle", n, repeats, list_setup,
#     function(st) { s <- st$s; s[[st$mid]] <- "y"; s })
#   rows <- bench_one(rows, "base R list", "remove middle", n, repeats, list_setup,
#     function(st) st$s[-st$mid])
#   rows <- bench_one(rows, "base R list", "concatenate", n, repeats, pair_list,
#     function(st) c(st$a, st$b))
#   rows <- bench_one(rows, "base R list", "split at middle", n, repeats, list_setup,
#     function(st) list(
#       left = st$s[seq_len(st$mid - 1L)],
#       value = st$s[[st$mid]],
#       right = st$s[(st$mid + 1L):n]
#     ))
# }
# 
# results_list$sequence <- do.call(rbind, as.list(rows))

## ----sequence-plot, fig.width=9, fig.height=7---------------------------------
if(!is.null(results_list$sequence)) {
  seq_results <- results_list$sequence
  seq_results$time_ms <- seq_results$time_us / 1000
  sorted_sizes <- sort(unique(seq_results$n))
  pow_labels <- lapply(sorted_sizes, function(s) bquote(2^.(log2(s)))) |> as.character()
  seq_results$n_cat <- factor(seq_results$n, levels = sorted_sizes)

  p_sequence <- ggplot(seq_results, aes(x = n_cat, y = time_ms, color = impl)) +
    geom_point(position = position_jitter(width = 0.15, height = 0)) +
    facet_wrap(~ op, scales = "free_y") +
    scale_x_discrete(labels = pow_labels) +
    labs(
      title = "Sequence Operations",
      x = "Number of elements",
      y = "Time (ms)",
      color = "Implementation"
    ) +
    theme_bw() +
    theme(plot.title = element_text(hjust = 0.5), legend.position = "bottom")
  print(p_sequence)
} else {
  knitr::asis_output("*Benchmark results not yet generated. Run `data-raw/generate_publication_results.R` to populate.*")
}

## ----queue-bench, eval=run_benchmarks, cache=FALSE----------------------------
# queue_sizes <- 2^(10 + 0:7)
# rows <- flexseq()
# 
# for(n in queue_sizes) {
#   items <- function() as.list(rep("queue_item", n))
#   flex_setup <- function() list(q = as_flexseq(items()))
#   rsd_setup  <- function() list(q = rstackdeque::as.rpqueue(items()))
#   list_setup <- function() list(q = items())
# 
#   rows <- bench_one(rows, "flexseq",     "enqueue", n, repeats, flex_setup,
#     function(st) push_back(st$q, "d"))
#   rows <- bench_one(rows, "flexseq",     "dequeue", n, repeats, flex_setup,
#     function(st) pop_front(st$q)$remaining)
#   rows <- bench_one(rows, "rstackdeque", "enqueue", n, repeats, rsd_setup,
#     function(st) rstackdeque::insert_back(st$q, "d"))
#   rows <- bench_one(rows, "rstackdeque", "dequeue", n, repeats, rsd_setup,
#     function(st) rstackdeque::without_front(st$q))
#   rows <- bench_one(rows, "base R list", "enqueue", n, repeats, list_setup,
#     function(st) c(st$q, list("d")))
#   rows <- bench_one(rows, "base R list", "dequeue", n, repeats, list_setup,
#     function(st) st$q[-1L])
# }
# 
# results_list$queue <- do.call(rbind, as.list(rows))

## ----queue-plot, fig.width=8, fig.height=4.8----------------------------------
if(!is.null(results_list$queue)) {
  queue_results <- results_list$queue
  queue_results$time_ms <- queue_results$time_us / 1000
  queue_results$n_cat <- factor(queue_results$n, levels = sort(unique(queue_results$n)))

  p_queue <- ggplot(queue_results, aes(x = n_cat, y = time_ms, color = impl)) +
    geom_boxplot() +
    facet_wrap(~ op, scales = "free_y") +
    labs(
      title = "Queue Operations",
      x = "Number of elements",
      y = "Time (ms)",
      color = "Implementation"
    ) +
    theme_bw() +
    theme(plot.title = element_text(hjust = 0.5), legend.position = "bottom")
  print(p_queue)
} else {
  knitr::asis_output("*Benchmark results not yet generated.*")
}

## ----pq-bench, eval=run_benchmarks, cache=FALSE-------------------------------
# pq_sizes <- c(100, 500, 1000, 5000, 10000, 50000)
# rows <- flexseq()
# 
# set.seed(42)
# max_pq <- max(pq_sizes)
# all_pq_vals <- sprintf("val_%06d", seq_len(max_pq))
# all_pq_pri  <- runif(max_pq)
# 
# for(n in pq_sizes) {
#   pv <- as.list(all_pq_vals[seq_len(n)])
#   pw <- all_pq_pri[seq_len(n)]
#   pq_setup   <- function() list(pq = as_priority_queue(pv, priorities = pw))
#   base_setup <- function() list(v = all_pq_vals[seq_len(n)], p = pw)
# 
#   rows <- bench_one(rows, "priority_queue", "insert",   n, repeats, pq_setup,
#     function(st) insert(st$pq, "val_new", 0.5))
#   rows <- bench_one(rows, "priority_queue", "peek min", n, repeats, pq_setup,
#     function(st) peek_min(st$pq))
#   rows <- bench_one(rows, "priority_queue", "pop min",  n, repeats, pq_setup,
#     function(st) pop_min(st$pq)$remaining)
#   rows <- bench_one(rows, "priority_queue", "peek max", n, repeats, pq_setup,
#     function(st) peek_max(st$pq))
#   rows <- bench_one(rows, "priority_queue", "pop max",  n, repeats, pq_setup,
#     function(st) pop_max(st$pq)$remaining)
# 
#   rows <- bench_one(rows, "base R", "insert",   n, repeats, base_setup,
#     function(st) list(values = c(st$v, "val_new"), priorities = c(st$p, 0.5)))
#   rows <- bench_one(rows, "base R", "peek min", n, repeats, base_setup,
#     function(st) st$v[which.min(st$p)])
#   rows <- bench_one(rows, "base R", "pop min",  n, repeats, base_setup,
#     function(st) { i <- which.min(st$p); list(values = st$v[-i], priorities = st$p[-i]) })
#   rows <- bench_one(rows, "base R", "peek max", n, repeats, base_setup,
#     function(st) st$v[which.max(st$p)])
#   rows <- bench_one(rows, "base R", "pop max",  n, repeats, base_setup,
#     function(st) { i <- which.max(st$p); list(values = st$v[-i], priorities = st$p[-i]) })
# }
# 
# results_list$pq <- do.call(rbind, as.list(rows))

## ----pq-plot, fig.width=9, fig.height=7---------------------------------------
if(!is.null(results_list$pq)) {
  pq_results <- results_list$pq
  pq_results$time_ms <- pq_results$time_us / 1000
  pq_medians <- aggregate(time_us ~ impl + op + n, data = pq_results, FUN = median)
  pq_medians$time_ms <- pq_medians$time_us / 1000

  p_pq <- ggplot(pq_results, aes(x = n, y = time_ms, color = impl)) +
    geom_point(alpha = 0.25, size = 1.2, position = position_jitter(width = 0.03)) +
    geom_line(data = pq_medians, linewidth = 0.6) +
    geom_point(data = pq_medians, size = 1.8) +
    facet_wrap(~ op, scales = "free_y") +
    scale_x_log10(labels = scales::label_comma()) +
    scale_y_log10(labels = scales::label_comma()) +
    labs(
      title = "Priority Queue Operations",
      x = "Number of elements",
      y = "Time (ms)",
      color = "Implementation"
    ) +
    theme_bw() +
    theme(plot.title = element_text(hjust = 0.5), legend.position = "bottom")
  print(p_pq)
} else {
  knitr::asis_output("*Benchmark results not yet generated.*")
}

## ----ivx-bench, eval=run_benchmarks, cache=FALSE------------------------------
# ivx_sizes <- c(100, 500, 1000, 5000, 10000, 50000)
# rows <- flexseq()
# 
# set.seed(123)
# max_ivx <- max(ivx_sizes)
# all_starts <- sort(sample.int(max_ivx * 10L, max_ivx))
# all_widths <- sample.int(100L, max_ivx, replace = TRUE)
# all_ends   <- all_starts + all_widths
# all_vals   <- sprintf("interval_%06d", seq_len(max_ivx))
# 
# qpt   <- all_starts[as.integer(max_ivx / 2)] + 10L
# qlo   <- all_starts[as.integer(max_ivx * 0.4)]
# qhi   <- all_starts[as.integer(max_ivx * 0.5)]
# ins_s <- all_starts[as.integer(max_ivx / 2)]
# ins_e <- ins_s + 50L
# 
# has_iranges <- requireNamespace("IRanges", quietly = TRUE) &&
#   requireNamespace("S4Vectors", quietly = TRUE)
# 
# for(n in ivx_sizes) {
#   starts <- all_starts[seq_len(n)]
#   ends   <- all_ends[seq_len(n)]
#   vals   <- all_vals[seq_len(n)]
# 
#   ivx_setup  <- function() list(ix = as_interval_index(as.list(vals), start = starts, end = ends, default_query_bounds = "[]"))
#   df_setup   <- function() list(df = data.frame(start = starts, end = ends, value = vals, stringsAsFactors = FALSE))
# 
#   rows <- bench_one(rows, "interval_index", "insert", n, repeats, ivx_setup,
#     function(st) insert(st$ix, "interval_new", ins_s, ins_e))
#   rows <- bench_one(rows, "interval_index", "point query", n, repeats, ivx_setup,
#     function(st) peek_point(st$ix, qpt, bounds = "[]"))
#   rows <- bench_one(rows, "interval_index", "all point matches", n, repeats, ivx_setup,
#     function(st) peek_all_point(st$ix, qpt, bounds = "[]"))
#   rows <- bench_one(rows, "interval_index", "overlap query", n, repeats, ivx_setup,
#     function(st) peek_all_overlaps(st$ix, qlo, qhi, bounds = "[]"))
# 
#   rows <- bench_one(rows, "base R", "insert", n, repeats, df_setup,
#     function(st) rbind(st$df, data.frame(start = ins_s, end = ins_e, value = "interval_new", stringsAsFactors = FALSE)))
#   rows <- bench_one(rows, "base R", "point query", n, repeats, df_setup,
#     function(st) {
#       hits <- which(st$df$start <= qpt & qpt <= st$df$end)
#       if(length(hits)) st$df$value[hits[1L]] else NULL
#     })
#   rows <- bench_one(rows, "base R", "all point matches", n, repeats, df_setup,
#     function(st) st$df[st$df$start <= qpt & qpt <= st$df$end, , drop = FALSE])
#   rows <- bench_one(rows, "base R", "overlap query", n, repeats, df_setup,
#     function(st) st$df[st$df$start <= qhi & st$df$end >= qlo, , drop = FALSE])
# 
#   if(has_iranges) {
#     ir_setup <- function() list(
#       ir = IRanges::IRanges(start = starts, end = ends),
#       v  = vals
#     )
# 
#     rows <- bench_one(rows, "IRanges", "insert", n, repeats, ir_setup,
#       function(st) list(
#         ir = c(st$ir, IRanges::IRanges(start = ins_s, end = ins_e)),
#         v  = c(st$v, "interval_new")
#       ))
#     rows <- bench_one(rows, "IRanges", "point query", n, repeats, ir_setup,
#       function(st) {
#         hits <- S4Vectors::subjectHits(IRanges::findOverlaps(IRanges::IRanges(start = qpt, width = 1L), st$ir))
#         if(length(hits)) st$v[hits[1L]] else NULL
#       })
#     rows <- bench_one(rows, "IRanges", "all point matches", n, repeats, ir_setup,
#       function(st) {
#         st$v[S4Vectors::subjectHits(IRanges::findOverlaps(IRanges::IRanges(start = qpt, width = 1L), st$ir))]
#       })
#     rows <- bench_one(rows, "IRanges", "overlap query", n, repeats, ir_setup,
#       function(st) {
#         st$v[S4Vectors::subjectHits(IRanges::findOverlaps(IRanges::IRanges(start = qlo, end = qhi), st$ir))]
#       })
#   }
# }
# 
# results_list$ivx <- do.call(rbind, as.list(rows))

## ----ivx-plot, fig.width=9, fig.height=6--------------------------------------
if(!is.null(results_list$ivx)) {
  ivx_results <- results_list$ivx
  ivx_results$time_ms <- ivx_results$time_us / 1000
  ivx_medians <- aggregate(time_us ~ impl + op + n, data = ivx_results, FUN = median)
  ivx_medians$time_ms <- ivx_medians$time_us / 1000

  p_ivx <- ggplot(ivx_results, aes(x = n, y = time_ms, color = impl)) +
    geom_point(alpha = 0.25, size = 1.2, position = position_jitter(width = 0.03)) +
    geom_line(data = ivx_medians, linewidth = 0.6) +
    geom_point(data = ivx_medians, size = 1.8) +
    facet_wrap(~ op, scales = "free_y") +
    scale_x_log10(labels = scales::label_comma()) +
    scale_y_log10(labels = scales::label_comma()) +
    labs(
      title = "Interval Index Queries",
      x = "Number of elements",
      y = "Time (ms)",
      color = "Implementation"
    ) +
    theme_bw() +
    theme(plot.title = element_text(hjust = 0.5), legend.position = "bottom")
  print(p_ivx)
} else {
  knitr::asis_output("*Benchmark results not yet generated.*")
}

## ----save-results, eval=run_benchmarks, include=FALSE-------------------------
# # Persist this run so subsequent builds (CRAN, pkgdown) render the same
# # figures without re-timing. Path resolves from knit_root_dir, which the
# # generator script sets to the repo root.
# saveRDS(results_list, "inst/extdata/benchmarks.rds")

## ----save-figures, eval=save_figures, include=FALSE---------------------------
# dir.create("paper/figures", showWarnings = FALSE, recursive = TRUE)
# if(exists("p_sequence")) ggsave("paper/figures/benchmarks-sequence.pdf", p_sequence, width = 9, height = 7)
# if(exists("p_queue"))    ggsave("paper/figures/benchmarks-queue.pdf",    p_queue,    width = 8, height = 4.8)
# if(exists("p_pq"))       ggsave("paper/figures/benchmarks-pq.pdf",       p_pq,       width = 9, height = 7)
# if(exists("p_ivx"))      ggsave("paper/figures/benchmarks-ivx.pdf",      p_ivx,      width = 9, height = 6)

#SO

# Query-spec builders return a plan consumed by .ivx_run_relation_query():
# - lower: optional lower start-bound value for candidate windowing (NULL = no bound).
# - lower_strict: FALSE => start >= lower, TRUE => start > lower.
# - upper: optional upper start-bound value for candidate windowing (NULL = no bound).
# - upper_strict: FALSE => start <= upper, TRUE => start < upper.
# - no_match_subtree(node): prune predicate over subtree measures; TRUE means
#   the subtree cannot contain matches and traversal can skip it.
# - leaf_match(entry): exact entry-level relation test run on remaining candidates.

# Point relation spec.
# Dispatches on `match_at`:
#   "interval" -> entry interval contains query point (under `bounds`).
#   "start"    -> entry$start == point (coordinate equality; bounds ignored).
#   "end"      -> entry$end   == point (coordinate equality; bounds ignored).
#   "either"   -> entry$start == point OR entry$end == point (bounds ignored).
# **Inputs:** `qp` normalized point query list; scalar `bounds` (or NULL for
#   non-"interval" modes); bounds-flag list `flags` (or NULL); scalar `match_at`.
# **Outputs:** query-spec list(lower/lower_strict/upper/upper_strict/no_match_subtree/leaf_match).
# **Used by:** peek_point(), peek_all_point(), pop_point(), pop_all_point().
.ivx_spec_point <- function(qp, bounds, flags, match_at = "interval") {
  switch(
    match_at,
    interval = .ivx_spec_point_interval(qp, bounds, flags),
    start    = .ivx_spec_point_start(qp),
    end      = .ivx_spec_point_end(qp),
    either   = .ivx_spec_point_either(qp),
    stop("Unknown match_at mode: ", match_at)
  )
}

# `match_at = "interval"` — interval-containment query under `bounds`.
.ivx_spec_point_interval <- function(qp, bounds, flags) {
  include_start <- isTRUE(flags$include_start)
  include_end <- isTRUE(flags$include_end)

  leaf_match <- if(.ivx_is_fast_endpoint_type(qp$endpoint_type)) {
    function(e) {
      left_ok <- if(include_start) isTRUE(qp$value >= e$start) else isTRUE(qp$value > e$start)
      right_ok <- if(include_end) isTRUE(qp$value <= e$end) else isTRUE(qp$value < e$end)
      isTRUE(left_ok && right_ok)
    }
  } else {
    function(e) .ivx_contains_point(e$start, e$end, qp$value, bounds, qp$endpoint_type)
  }

  list(
    lower = NULL,
    lower_strict = FALSE,
    upper = qp$value,
    upper_strict = FALSE,
    no_match_subtree = function(node) {
      m <- node_measure(node, ".ivx_max_end")
      if(!isTRUE(m$has)) {
        return(TRUE)
      }
      cmp <- .ivx_compare_scalar_fast(m$end, qp$value, endpoint_type = qp$endpoint_type)
      if(include_end) {
        cmp < 0L
      } else {
        cmp <= 0L
      }
    },
    leaf_match = leaf_match
  )
}

# `match_at = "start"` — entries whose start coordinate equals point.
# Tree is start-sorted, so windowing alone isolates matches.
.ivx_spec_point_start <- function(qp) {
  leaf_match <- function(e) {
    isTRUE(.ivx_compare_scalar_fast(e$start, qp$value, endpoint_type = qp$endpoint_type) == 0L)
  }
  list(
    lower = qp$value,
    lower_strict = FALSE,
    upper = qp$value,
    upper_strict = FALSE,
    no_match_subtree = function(node) FALSE,
    leaf_match = leaf_match
  )
}

# `match_at = "end"` — entries whose end coordinate equals point.
# Tree is not end-sorted; skip start windowing, prune via min/max-end monoids.
.ivx_spec_point_end <- function(qp) {
  leaf_match <- function(e) {
    isTRUE(.ivx_compare_scalar_fast(e$end, qp$value, endpoint_type = qp$endpoint_type) == 0L)
  }
  list(
    lower = NULL,
    lower_strict = FALSE,
    upper = NULL,
    upper_strict = FALSE,
    no_match_subtree = function(node) {
      mmax <- node_measure(node, ".ivx_max_end")
      mmin <- node_measure(node, ".ivx_min_end")
      if(!isTRUE(mmax$has) || !isTRUE(mmin$has)) {
        return(TRUE)
      }
      if(.ivx_compare_scalar_fast(mmax$end, qp$value, endpoint_type = qp$endpoint_type) < 0L) {
        return(TRUE)
      }
      if(.ivx_compare_scalar_fast(mmin$end, qp$value, endpoint_type = qp$endpoint_type) > 0L) {
        return(TRUE)
      }
      FALSE
    },
    leaf_match = leaf_match
  )
}

# `match_at = "either"` — entries whose start or end equals point.
# Start > point rules out both start == point and end == point (since end >= start),
# so we can upper-window by start. End pruning still uses max_end.
.ivx_spec_point_either <- function(qp) {
  leaf_match <- function(e) {
    isTRUE(.ivx_compare_scalar_fast(e$start, qp$value, endpoint_type = qp$endpoint_type) == 0L) ||
      isTRUE(.ivx_compare_scalar_fast(e$end, qp$value, endpoint_type = qp$endpoint_type) == 0L)
  }
  list(
    lower = NULL,
    lower_strict = FALSE,
    upper = qp$value,
    upper_strict = FALSE,
    no_match_subtree = function(node) {
      mmax <- node_measure(node, ".ivx_max_end")
      if(!isTRUE(mmax$has)) {
        return(TRUE)
      }
      .ivx_compare_scalar_fast(mmax$end, qp$value, endpoint_type = qp$endpoint_type) < 0L
    },
    leaf_match = leaf_match
  )
}

# Runtime: O(1).
# Overlap relation spec (entry overlaps query interval under current bounds).
# **Inputs:** `q` normalized interval query list; scalar `bounds`; bounds-flag list `flags`.
# **Outputs:** query-spec list(lower/lower_strict/upper/upper_strict/no_match_subtree/leaf_match).
# **Used by:** peek_overlaps(), pop_overlaps().
.ivx_spec_overlaps <- function(q, bounds, flags) {
  touching_is_overlap <- isTRUE(flags$include_start) && isTRUE(flags$include_end)

  leaf_match <- if(.ivx_is_fast_endpoint_type(q$endpoint_type)) {
    function(e) {
      a_before_b <- if(touching_is_overlap) isTRUE(e$end < q$start) else isTRUE(e$end <= q$start)
      b_before_a <- if(touching_is_overlap) isTRUE(q$end < e$start) else isTRUE(q$end <= e$start)
      !isTRUE(a_before_b || b_before_a)
    }
  } else {
    function(e) .ivx_overlaps_interval(e$start, e$end, q$start, q$end, bounds, q$endpoint_type)
  }

  list(
    lower = NULL,
    lower_strict = FALSE,
    upper = q$end,
    upper_strict = !isTRUE(touching_is_overlap),
    no_match_subtree = function(node) {
      m <- node_measure(node, ".ivx_max_end")
      if(!isTRUE(m$has)) {
        return(TRUE)
      }
      cmp <- .ivx_compare_scalar_fast(m$end, q$start, endpoint_type = q$endpoint_type)
      if(touching_is_overlap) {
        cmp < 0L
      } else {
        cmp <= 0L
      }
    },
    leaf_match = leaf_match
  )
}

# Runtime: O(1).
# Containing relation spec (entry contains query interval).
# **Inputs:** `q` normalized interval query list; scalar `bounds`; bounds-flag list `flags`.
# **Outputs:** query-spec list(lower/lower_strict/upper/upper_strict/no_match_subtree/leaf_match).
# **Used by:** peek_containing(), pop_containing().
.ivx_spec_containing <- function(q, bounds, flags) {
  touching_is_overlap <- isTRUE(flags$include_start) && isTRUE(flags$include_end)

  leaf_match <- if(.ivx_is_fast_endpoint_type(q$endpoint_type)) {
    function(e) {
      contains <- isTRUE(e$start <= q$start) && isTRUE(e$end >= q$end)
      if(!contains) {
        return(FALSE)
      }
      a_before_b <- if(touching_is_overlap) isTRUE(e$end < q$start) else isTRUE(e$end <= q$start)
      b_before_a <- if(touching_is_overlap) isTRUE(q$end < e$start) else isTRUE(q$end <= e$start)
      !isTRUE(a_before_b || b_before_a)
    }
  } else {
    function(e) {
      .ivx_overlaps_interval(e$start, e$end, q$start, q$end, bounds, q$endpoint_type) &&
        .ivx_contains_interval(e$start, e$end, q$start, q$end, q$endpoint_type)
    }
  }

  list(
    lower = NULL,
    lower_strict = FALSE,
    upper = q$start,
    upper_strict = FALSE,
    no_match_subtree = function(node) {
      m <- node_measure(node, ".ivx_max_end")
      if(!isTRUE(m$has)) {
        return(TRUE)
      }
      .ivx_compare_scalar_fast(m$end, q$end, endpoint_type = q$endpoint_type) < 0L
    },
    leaf_match = leaf_match
  )
}

# Runtime: O(1).
# Within relation spec (entry is within query interval).
# **Inputs:** `q` normalized interval query list; scalar `bounds`; bounds-flag list `flags`.
# **Outputs:** query-spec list(lower/lower_strict/upper/upper_strict/no_match_subtree/leaf_match).
# **Used by:** peek_within(), pop_within().
.ivx_spec_within <- function(q, bounds, flags) {
  touching_is_overlap <- isTRUE(flags$include_start) && isTRUE(flags$include_end)

  leaf_match <- if(.ivx_is_fast_endpoint_type(q$endpoint_type)) {
    function(e) {
      within <- isTRUE(q$start <= e$start) && isTRUE(q$end >= e$end)
      if(!within) {
        return(FALSE)
      }
      a_before_b <- if(touching_is_overlap) isTRUE(e$end < q$start) else isTRUE(e$end <= q$start)
      b_before_a <- if(touching_is_overlap) isTRUE(q$end < e$start) else isTRUE(q$end <= e$start)
      !isTRUE(a_before_b || b_before_a)
    }
  } else {
    function(e) {
      .ivx_overlaps_interval(e$start, e$end, q$start, q$end, bounds, q$endpoint_type) &&
        .ivx_contains_interval(q$start, q$end, e$start, e$end, q$endpoint_type)
    }
  }

  list(
    lower = q$start,
    lower_strict = FALSE,
    upper = q$end,
    upper_strict = !isTRUE(touching_is_overlap),
    no_match_subtree = function(node) {
      mmin <- node_measure(node, ".ivx_min_end")
      mmax <- node_measure(node, ".ivx_max_end")
      if(!isTRUE(mmin$has) || !isTRUE(mmax$has)) {
        return(TRUE)
      }

      # No interval in subtree can satisfy end <= q.end.
      if(.ivx_compare_scalar_fast(mmin$end, q$end, endpoint_type = q$endpoint_type) > 0L) {
        return(TRUE)
      }

      # Even the largest end is too far left to overlap q.
      cmp <- .ivx_compare_scalar_fast(mmax$end, q$start, endpoint_type = q$endpoint_type)
      if(touching_is_overlap) {
        cmp < 0L
      } else {
        cmp <= 0L
      }
    },
    leaf_match = leaf_match
  )
}

#SO

# Internal insert engine preserving interval ordering by start.
# **Inputs:** `x` interval_index; `entry` canonical entry; optional `endpoint_type`.
# **Outputs:** updated interval_index with inserted entry.
# **Used by:** insert.interval_index().
.ivx_insert_entry <- function(x, entry, endpoint_type = NULL) {
  ep <- if(is.null(endpoint_type)) .ivx_endpoint_type_state(x) else endpoint_type
  ms <- resolve_tree_monoids(x, required = TRUE)
  can_use_cpp_key_insert <- isTRUE(
    .ivx_supports_oms_key_type(ep) &&
      .ft_cpp_can_use(ms) &&
      !is.null(ms[[".oms_max_key"]])
  )

  out <- if(length(x) == 0L) {
    # First insert may need to upgrade required monoids once endpoint type is known.
    # Use one append path here; C++ ordered insert is kept for non-empty trees.
    base <- x
    if(.ivx_supports_oms_key_type(ep) && is.null(ms[[".oms_max_key"]])) {
      monoids0 <- attr(x, "monoids", exact = TRUE)
      user0 <- monoids0[setdiff(
        names(monoids0),
        c(".size", ".named_count", ".ivx_max_start", ".ivx_max_end", ".ivx_min_end", ".oms_max_key")
      )]
      ms0 <- .ivx_merge_monoids(if(length(user0) == 0L) NULL else user0, endpoint_type = ep)
      base <- empty_tree(monoids = ms0)
    }
    .ft_push_back_impl(base, entry, context = "insert()")
  } else if(can_use_cpp_key_insert) {
    .ft_cpp_oms_insert(x, entry, ms, ep)
  } else {
    s <- split_by_predicate(
      x,
      function(v) isTRUE(v$has) && .ivx_compare_scalar(v$start, entry$start, v$endpoint_type) > 0L,
      ".ivx_max_start"
    )
    left_plus <- .ft_push_back_impl(s$left, entry, context = "insert()")
    concat_trees(left_plus, s$right)
  }

  .ivx_wrap_like(x, out, endpoint_type = ep)
}

# Runtime: O(n) over tree size for any non-trivial update (rebind/recompute pass).
# Interval-specific add_monoids guard for reserved monoid names.
# **Inputs:** `t` interval_index-like tree; user `monoids`; scalar logical `overwrite`.
# **Outputs:** interval_index tree with merged monoids.
# **Used by:** public add_monoids() S3 dispatch.
#' @method add_monoids interval_index
#' @export
#' @noRd
add_monoids.interval_index <- function(t, monoids, overwrite = FALSE) {
  if(length(monoids) > 0L) {
    bad <- intersect(names(monoids), c(".size", ".named_count", ".ivx_max_start", ".ivx_max_end", ".ivx_min_end", ".oms_max_key"))
    if(length(bad) > 0L) {
      stop("Reserved monoid names cannot be supplied for interval_index: ", paste(bad, collapse = ", "))
    }
  }
  add_monoids.flexseq(t, monoids, overwrite = overwrite)
}

# Runtime: O(log n) near split point depth.
# Public insert method for interval_index.
# **Inputs:** `x` interval_index; payload `element`; scalar `start`/`end`; optional `name`.
# **Outputs:** updated interval_index.
# **Used by:** users/tests.
#' @method insert interval_index
#' @export
#' @noRd
insert.interval_index <- function(x, element, start, end, name = NULL, ...) {
  .ivx_assert_index(x)

  norm <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  entry <- .ivx_make_entry(element, norm$start, norm$end)

  if(!is.null(name)) {
    entry <- .ft_set_name(entry, name)
  }

  .ivx_insert_entry(x, entry, endpoint_type = norm$endpoint_type)
}

#SO

#' @method as_flexseq interval_index
#' @export
# Runtime: O(n) from list materialization + linear rebuild.
# Public cast to plain flexseq while dropping interval-specific behavior.
# **Inputs:** `x` interval_index.
# **Outputs:** flexseq.
# **Used by:** users/tests.
as_flexseq.interval_index <- function(x) {
  entries <- as.list(x)
  .as_flexseq_build.default(entries, monoids = NULL)
}

# Runtime: O(n).
# Coerce to payload list in interval order.
# **Inputs:** `x` interval_index.
# **Outputs:** base list of payload items with names preserved.
# **Used by:** users/tests.
#' Coerce Interval Index to List
#'
#' @method as.list interval_index
#' @param x An `interval_index`.
#' @param ... Unused.
#' @return A plain list of payload elements in interval order.
#' @details
#' This returns payload values only.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(3, 1, 2), end = c(4, 2, 3))
#' as.list(ix)
#' @export
as.list.interval_index <- function(x, ...) {
  .ivx_assert_index(x)
  entries <- as.list.flexseq(x, ...)
  out <- lapply(entries, function(e) e$value)
  names(out) <- names(entries)
  out
}

# Runtime: O(1).
# Length method backed by cached `.size` measure.
# **Inputs:** `x` interval_index.
# **Outputs:** scalar integer length.
# **Used by:** users/tests/base generics.
#' Interval Index Length
#'
#' @method length interval_index
#' @param x An `interval_index`.
#' @return Number of indexed intervals.
#' @details
#' Uses cached size metadata and runs in O(1).
#' @examples
#' ix <- interval_index("a", "b", start = c(1, 3), end = c(2, 5))
#' length(ix)
#'
#' length(interval_index())
#' @export
length.interval_index <- function(x) {
  as.integer(node_measure(x, ".size"))
}

#' Merge Two Interval Indices
#'
#' Returns a new `interval_index` containing every entry from both inputs,
#' preserving start-position order. On tied start positions, `x`'s entries
#' precede `y`'s (left-biased FIFO).
#'
#' @method merge interval_index
#' @param x An `interval_index`.
#' @param y An `interval_index`.
#' @param ... Unused.
#' @return A new `interval_index` of size `length(x) + length(y)`.
#' @details
#' The merge runs in O(m + n) via a zipper-style traversal on interval
#' starts, with a fast path to O(log(min(m, n))) when the start ranges are
#' disjoint.
#'
#' Both indices must share the same endpoint type and the same `bounds`
#' convention (e.g. `"[)"` half-open vs. `"[]"` closed), and the same
#' monoid set. Mismatches error.
#'
#' The reserved monoids `.ivx_min_end` / `.ivx_max_end` recompute
#' automatically on the merged tree, so `min_endpoint()` / `max_endpoint()`
#' and interval-relation queries work immediately on the result.
#'
#' Both inputs are left unmodified.
#' @examples
#' a <- interval_index("A1", "A2", start = c(1, 5), end = c(4, 8))
#' b <- interval_index("B1", "B2", start = c(3, 7), end = c(6, 10))
#' m <- merge(a, b)
#' as.list(m)
#' @export
# Runtime: O(m + n); O(log(min(m, n))) on the disjoint fast path.
merge.interval_index <- function(x, y, ...) {
  if(!inherits(y, "interval_index")) {
    stop("Both arguments to `merge()` must be interval_indices.")
  }

  if(length(x) == 0L) return(y)
  if(length(y) == 0L) return(x)

  ep_x <- attr(x, "ivx_endpoint_type", exact = TRUE)
  ep_y <- attr(y, "ivx_endpoint_type", exact = TRUE)
  if(!identical(ep_x, ep_y)) {
    stop("Cannot merge interval_indices with different endpoint types.")
  }
  b_x <- attr(x, "ivx_bounds", exact = TRUE)
  b_y <- attr(y, "ivx_bounds", exact = TRUE)
  if(!identical(b_x, b_y)) {
    stop("Cannot merge interval_indices with different bounds conventions.")
  }
  if(!identical(sort(names(attr(x, "monoids", exact = TRUE))),
                sort(names(attr(y, "monoids", exact = TRUE))))) {
    stop("Cannot merge interval_indices with different monoid sets.")
  }

  nxs <- length(x); nys <- length(y)

  # Materialize entries for zipper and boundary checks.
  xs <- .ft_to_list(x)
  ys <- .ft_to_list(y)

  # Disjoint fast paths on start values.
  x_max_start <- xs[[nxs]]$start
  y_min_start <- ys[[1L]]$start
  if(.ivx_compare_scalar(x_max_start, y_min_start, ep_x) <= 0L) {
    return(.ivx_wrap_like(x, concat_trees(x, y)))
  }
  y_max_start <- ys[[nys]]$start
  x_min_start <- xs[[1L]]$start
  if(.ivx_compare_scalar(y_max_start, x_min_start, ep_x) < 0L) {
    return(.ivx_wrap_like(x, concat_trees(y, x)))
  }

  # General zipper merge on pre-sorted (by start) entries.
  merged <- vector("list", nxs + nys)
  i <- 1L; j <- 1L; k <- 1L
  while(i <= nxs && j <= nys) {
    if(.ivx_compare_scalar(xs[[i]]$start, ys[[j]]$start, ep_x) <= 0L) {
      merged[[k]] <- xs[[i]]; i <- i + 1L
    } else {
      merged[[k]] <- ys[[j]]; j <- j + 1L
    }
    k <- k + 1L
  }
  while(i <= nxs) { merged[[k]] <- xs[[i]]; i <- i + 1L; k <- k + 1L }
  while(j <= nys) { merged[[k]] <- ys[[j]]; j <- j + 1L; k <- k + 1L }

  ms <- attr(x, "monoids", exact = TRUE)
  out <- .ivx_tree_from_ordered_entries(merged, ms)
  .ivx_wrap_like(x, out)
}

#SO

#' @method as_flexseq ordered_sequence
#' @export
# Runtime: O(n) from list materialization + linear rebuild.
as_flexseq.ordered_sequence <- function(x) {
  entries <- as.list(x)
  .as_flexseq_build.default(entries, monoids = NULL)
}

#' Coerce Ordered Sequence to List
#'
#' @method as.list ordered_sequence
#' @param x An `ordered_sequence`.
#' @param ... Unused.
#' @return A plain list of elements in key order.
#' @details
#' Returns payload elements only (keys are omitted) in canonical key order.
#' If entries are named, names are preserved on the returned list.
#' @examples
#' x <- ordered_sequence("a", "b", "c", keys = c(2, 1, 3))
#' as.list(x)
#' @export
# Runtime: O(n).
as.list.ordered_sequence <- function(x, ...) {
  .oms_assert_set(x)
  .oms_extract_items(as.list.flexseq(x, ...))
}

#' Ordered Sequence Length
#'
#' @method length ordered_sequence
#' @param x An `ordered_sequence`.
#' @return Integer length.
#' @details
#' Uses cached size metadata and runs in O(1).
#' @examples
#' x <- ordered_sequence("a", "b", keys = c(2, 1))
#' length(x)
#'
#' length(ordered_sequence())
#' @export
# Runtime: O(1).
length.ordered_sequence <- function(x) {
  .oms_assert_set(x)
  as.integer(node_measure(x, ".size"))
}

#SO

# Runtime: O(log n) near split point depth.
# not using public pop/peek because those don't return priority (for peek) or name (for pop)
.pq_extract_entry <- function(q, monoid_name) {
  .pq_assert_queue(q)
  if(length(q) == 0L) {
    stop("Cannot extract from an empty priority_queue.")
  }

  target <- node_measure(q, monoid_name)
  pred <- function(v) .pq_measure_equal(v, target)
  ctx <- resolve_named_monoid(q, monoid_name)
  ms <- ctx$monoids
  mr <- ctx$monoid
  s <- if(.ft_cpp_can_use(ms)) {
    .ft_cpp_split_tree(q, pred, ms, monoid_name, mr$i)
  } else {
    split_tree_impl_fast(pred, mr$i, q, ms, mr, monoid_name)
  }

  rest <- concat_trees(s$left, s$right)
  rest <- .pq_wrap_like(q, rest)
  list(entry = s$value, remaining = rest)
}

#' Print a Priority Queue Summary
#'
#' Prints a compact summary with priority range and a head/tail preview.
#'
#' @param x A `priority_queue`.
#' @param max_elements Maximum number of elements shown in the preview.
#' @param show_custom_monoids Logical; show attached non-default monoids and
#'   their root cached measures.
#' @param ... Passed through to per-element `print()`.
#' @return Invisibly returns `x`.
#' @examples
#' q <- priority_queue(one = 1, two = 2, three = 3, priorities = c(20, 30, 10))
#' print(q, max_elements = 4)
#' sum_item <- measure_monoid(`+`, 0, function(entry) as.numeric(entry$value))
#' q3 <- add_monoids(priority_queue(1, 2, priorities = c(2, 1)), list(sum_item = sum_item))
#' print(q3, max_elements = 0, show_custom_monoids = TRUE)
#'
#' q2 <- priority_queue(1, 2, 3, priorities = c(2, 1, 3))
#' print(q2, max_elements = 3)
#'
#' print(priority_queue())
#' @export
# Runtime: O(k log n), where k = shown elements.
print.priority_queue <- function(x, max_elements = 4L, show_custom_monoids = FALSE, ...) {
  .pq_assert_queue(x)
  max_elements <- .ft_validate_print_max_elements(max_elements)
  show_custom <- .ft_validate_show_custom_monoids(show_custom_monoids)

  n <- as.integer(node_measure(x, ".size"))
  nn <- as.integer(node_measure(x, ".named_count"))
  named <- isTRUE(nn > 0L)
  cat(if(named) "Named" else "Unnamed", " priority_queue with ", n, " element", if(n == 1L) "" else "s", ".\n", sep = "")
  if(n > 0L) {
    minm <- node_measure(x, ".pq_min")
    maxm <- node_measure(x, ".pq_max")
    cat("Minimum priority: ", .ft_format_scalar(minm$priority), ", Maximum priority: ", .ft_format_scalar(maxm$priority), "\n", sep = "")
  }
  if(show_custom) {
    .ft_print_custom_monoids(x, excluded_names = c(".size", ".named_count", ".pq_min", ".pq_max"))
  }
  if(n == 0L || max_elements == 0L) {
    return(invisible(x))
  }

  cat("\nElements (by priority):\n\n")

  preview <- .pick_preview_sizes(n, max_elements)
  head_entries <- vector("list", length(preview$head))
  q_head <- x
  if(length(preview$head) > 0L) {
    for(i in seq_along(head_entries)) {
      out <- .pq_extract_entry(q_head, ".pq_min")
      head_entries[[i]] <- out$entry
      q_head <- out$remaining
    }
  }

  tail_entries <- vector("list", length(preview$tail))
  q_tail <- x
  if(length(preview$tail) > 0L) {
    for(i in seq_along(tail_entries)) {
      out <- .pq_extract_entry(q_tail, ".pq_max")
      tail_entries[[i]] <- out$entry
      q_tail <- out$remaining
    }
    tail_entries <- rev(tail_entries)
  }

  excluded_names <- c(".size", ".named_count", ".pq_min", ".pq_max")
  for(entry in head_entries) {
    nm <- .ft_get_name(entry)
    if(named && !is.null(nm)) {
      cat("$", nm, " (priority ", .ft_format_scalar(entry$priority), ")\n", sep = "")
    } else {
      cat("(priority ", .ft_format_scalar(entry$priority), ")\n", sep = "")
    }
    print(entry$value, ...)
    .ft_print_elem_custom_monoids(x, entry, excluded_names, show_custom)
    cat("\n")
  }

  .ft_print_skipped(preview$skipped)

  for(entry in tail_entries) {
    nm <- .ft_get_name(entry)
    if(named && !is.null(nm)) {
      cat("$", nm, " (priority ", .ft_format_scalar(entry$priority), ")\n", sep = "")
    } else {
      cat("(priority ", .ft_format_scalar(entry$priority), ")\n", sep = "")
    }
    print(entry$value, ...)
    .ft_print_elem_custom_monoids(x, entry, excluded_names, show_custom)
    cat("\n")
  }
  invisible(x)
}

#SO

# Runtime: O(1).
.pq_priority_type_state <- function(x) {
  attr(x, "pq_priority_type", exact = TRUE)
}

# Runtime: O(1).
# Normalize/validate one priority scalar and enforce queue-wide type consistency.
# Used by: .pq_make_entry() for constructors and insert paths.
.pq_resolve_priority <- function(priority, priority_type = NULL, arg_name = "priority") {
  # 1) Normalize to package canonical scalar representation.
  norm <- .ft_normalize_scalar_orderable(priority, arg_name = arg_name)

  # 2) Enforce single priority domain for a queue.
  resolved_type <- if(is.null(priority_type)) {
    norm$value_type
  } else if(identical(priority_type, norm$value_type)) {
    priority_type
  } else {
    stop("Incompatible priority type for this priority_queue.")
  }

  list(priority = norm$value, priority_type = resolved_type)
}

# Runtime: O(1).
# Assert queue class/shape at API boundaries.
# Used by: queue ops, print, and apply entry points.
.pq_assert_queue <- function(q) {
  if(!inherits(q, "priority_queue") || !is_structural_node(q)) {
    stop("`q` must be a priority_queue.")
  }
  invisible(TRUE)
}

# Runtime: O(1).
# Build one canonical entry (`list(value, priority)` + optional name) while
# threading the queue priority_type invariant.
# Used by: constructors and insert paths.
.pq_make_entry <- function(value, priority, priority_type = NULL, name = NULL) {
  resolved <- .pq_resolve_priority(priority, priority_type = priority_type)
  out <- list(value = value, priority = resolved$priority)
  list(
    entry = .ft_set_name(out, name),
    priority_type = resolved$priority_type
  )
}

# Runtime: O(1) under fixed monoid set.
# Merge user monoids with required queue monoids and reject reserved-name writes.
# Used by: constructors and fapply rebuild paths.
.pq_merge_monoids <- function(monoids = NULL) {
  user <- if(is.null(monoids)) list() else monoids
  if(length(user) > 0L) {
    bad <- intersect(names(user), c(".size", ".named_count", ".pq_min", ".pq_max"))
    if(length(bad) > 0L) {
      stop(paste0("Reserved monoid names cannot be supplied for priority queues: ", paste(bad, collapse = ", ")))
    }
  }
  c(user, .pq_required_monoids())
}

# Runtime: O(1).
# Attach class + PQ metadata to a structural tree.
# Used by: constructors and .pq_wrap_like().
.as_priority_queue <- function(x, priority_type = NULL) {
  if(!is_structural_node(x)) {
    stop("Expected a structural tree node.")
  }
  class(x) <- unique(c("priority_queue", "flexseq", setdiff(class(x), "list")))
  attr(x, "pq_next_seq") <- NULL
  attr(x, "pq_priority_type") <- priority_type
  x
}

# Runtime: O(1).
# Rewrap a newly-built tree using queue metadata from a template queue.
# Used by: insert/pop/apply/print helpers after structural operations.
.pq_wrap_like <- function(template, tree, priority_type = NULL) {
  resolved_type <- if(is.null(priority_type)) .pq_priority_type_state(template) else priority_type
  .as_priority_queue(tree, priority_type = resolved_type)
}

#SO

# Runtime: O(1).
.ft_cpp_enabled <- function() {
  isTRUE(getOption("immutables.use_cpp", TRUE))
}

# Runtime: O(1).
.ft_cpp_eligible_monoids <- function(monoids) {
  if(is.null(monoids) || !is.list(monoids) || is.null(names(monoids))) {
    return(FALSE)
  }
  TRUE
}

# Runtime: O(1).
.ft_cpp_can_use <- function(monoids) {
  .ft_cpp_enabled() && .ft_cpp_eligible_monoids(monoids)
}

# Runtime: O(1).
.ft_cpp_can_use_oms_insert <- function(monoids, key_type) {
  .ft_cpp_can_use(monoids) && .oms_is_cpp_key_type(key_type)
}

# Runtime: O(log n) near right edge.
.ft_cpp_add_right <- function(t, el, monoids) {
  .Call("ft_cpp_append_right", t, el, monoids, PACKAGE = "Immutables")
}

# Runtime: O(log n) near right edge.
.ft_cpp_add_right_named <- function(t, el, name, monoids) {
  .Call("ft_cpp_append_right_named", t, el, name, monoids, PACKAGE = "Immutables")
}

# Runtime: O(log n) near left edge.
.ft_cpp_add_left <- function(t, el, monoids) {
  .Call("ft_cpp_prepend_left", t, el, monoids, PACKAGE = "Immutables")
}

# Runtime: O(log n) near left edge.
.ft_cpp_add_left_named <- function(t, el, name, monoids) {
  .Call("ft_cpp_prepend_left_named", t, el, name, monoids, PACKAGE = "Immutables")
}

# Runtime: O(n), n = number of elements.
.ft_cpp_tree_from <- function(elements, monoids) {
  .Call("ft_cpp_tree_from", elements, monoids, PACKAGE = "Immutables")
}

# Runtime: O(n), n = number of elements.
.ft_cpp_tree_from_prepared <- function(elements, names, monoids) {
  .Call("ft_cpp_tree_from_prepared", elements, names, monoids, PACKAGE = "Immutables")
}

# Runtime: O(n) over ordered element count.
.ft_cpp_tree_from_sorted <- function(elements, monoids) {
  .Call("ft_cpp_tree_from_sorted", elements, monoids, PACKAGE = "Immutables")
}

# Runtime: O(log n1 + log n2) in the depths/sizes of both input trees.
.ft_cpp_concat <- function(x, y, monoids) {
  .Call("ft_cpp_concat", x, y, monoids, PACKAGE = "Immutables")
}

# Runtime: O(log n) near insertion point depth.
.ft_cpp_oms_insert <- function(x, entry, monoids, key_type) {
  .Call("ft_cpp_oms_insert", x, entry, monoids, key_type, PACKAGE = "Immutables")
}

# Runtime: O(c_pred log n), where c_pred is the cost of one predicate call.
# Structural traversal visits O(log n) nodes; predicate work is per visited node,
.ft_cpp_locate <- function(t, predicate, monoids, monoid_name, accumulator) {
  .Call("ft_cpp_locate", t, predicate, monoids, monoid_name, accumulator, PACKAGE = "Immutables")
}

# Runtime: O(log n * c_pred), where c_pred is the cost of one predicate call.
# As with locate, O(log n) traversal steps each may invoke predicate evaluation.
.ft_cpp_split_tree <- function(t, predicate, monoids, monoid_name, accumulator) {
  .Call("ft_cpp_split_tree", t, predicate, monoids, monoid_name, accumulator, PACKAGE = "Immutables")
}

# Runtime: O(log n). Callback-free index-based split; no R<->C++ predicate overhead.
# idx is a 1-based integer position.
.ft_cpp_split_at_index <- function(t, idx, monoids) {
  .Call("ft_cpp_split_at_index", t, as.integer(idx), monoids, PACKAGE = "Immutables")
}

# Runtime: O(n) worst-case, O(k) until first matched name.
.ft_cpp_find_name_position <- function(t, name) {
  .Call("ft_cpp_find_name_position", t, name, PACKAGE = "Immutables")
}

# Runtime: O(log n) near queried index depth.
.ft_cpp_get_by_index <- function(t, idx) {
  .Call("ft_cpp_get_by_index", t, as.integer(idx), PACKAGE = "Immutables")
}

# Runtime: O(k log n), where k = length(idx_vec).
.ft_cpp_get_many_by_index <- function(t, idx_vec) {
  .Call("ft_cpp_get_many_by_index", t, as.integer(idx_vec), PACKAGE = "Immutables")
}

# Runtime: O(n) in tree size.
.ft_cpp_name_positions <- function(t) {
  .Call("ft_cpp_name_positions", t, PACKAGE = "Immutables")
}

.gc_stress_enabled <- function() {
  # Never run on CRAN even if the opt-in env var is somehow set on a CRAN
  # machine: `NOT_CRAN` is conventionally "true" only in interactive / CI
  # contexts, so treating a missing / non-"true" value as "this is CRAN"
  # gives belt-and-suspenders protection alongside the skip_on_cran() call
  # inside `.skip_if_no_gc_stress()`.
  if(!identical(Sys.getenv("NOT_CRAN"), "true")) return(FALSE)
  gc_stress <- tolower(Sys.getenv("IMMUTABLES_GC_STRESS", unset = "0"))
  gc_stress %in% c("1", "true", "yes", "on")
}

.skip_if_no_gc_stress <- function() {
  testthat::skip_on_cran()
  if(!isTRUE(.gc_cpp_run$enabled)) {
    testthat::skip("Set IMMUTABLES_GC_STRESS=1 to run GC torture regression.")
  }
}

.run_gc_stress_suite <- function() {
  if(!.gc_stress_enabled()) {
    return(list(enabled = FALSE, setup_error = NULL, results = list()))
  }

  results <- list()
  setup_error <- NULL

  tryCatch({
    old_cpp <- getOption("immutables.use_cpp")
    options(immutables.use_cpp = FALSE)
    on.exit({
      if(is.null(old_cpp)) {
        options(immutables.use_cpp = NULL)
      } else {
        options(immutables.use_cpp = old_cpp)
      }
    }, add = TRUE)

    base_plain <- as_flexseq(as.list(1:8))
    ms <- attr(base_plain, "monoids", exact = TRUE)
    base_names <- paste0("k", seq_len(8))

    x_oms <- as_ordered_sequence(list("aa", "bb", "c", "ddd"), keys = c(2, 2, 1, 3))
    ms_oms <- attr(x_oms, "monoids", exact = TRUE)
    key_type_oms <- attr(x_oms, "oms_key_type", exact = TRUE)
    x_ivx <- as_interval_index(list("a", "b", "c"), start = c(1, 2, 2), end = c(3, 2, 4), default_query_bounds = "[]")
    q_num <- priority_queue("a", "b", "c", priorities = c(2, 1, 3))
    q_date <- priority_queue(
      "d1", "d2", "d3",
      priorities = as.Date(c("2020-01-03", "2020-01-01", "2020-01-02"))
    )
    x_ord_num <- as_ordered_sequence(list("aa", "bb", "cc"), keys = c(2, 1, 2))
    x_ord_date <- as_ordered_sequence(
      list("aa", "bb", "cc"),
      keys = as.Date(c("2020-01-02", "2020-01-01", "2020-01-02"))
    )

    options(immutables.use_cpp = TRUE)

    old_torture <- gctorture2(1, inhibit_release = FALSE)
    on.exit(gctorture2(old_torture, inhibit_release = FALSE), add = TRUE)

    step <- function(label, expr) {
      results[[label]] <<- tryCatch(
        {
          force(expr)
          NULL
        },
        error = function(e) conditionMessage(e)
      )
      invisible(NULL)
    }

    t <- NULL
    step("ft_cpp_tree_from", t <- .ft_cpp_tree_from(as.list(1:8), ms))
    step("ft_cpp_tree_from_prepared", .ft_cpp_tree_from_prepared(as.list(1:8), base_names, ms))
    step("ft_cpp_tree_from_sorted", .ft_cpp_tree_from_sorted(as.list(1:8), ms))

    if(is.null(results[["ft_cpp_tree_from"]])) {
      step("ft_cpp_add_right", t <- .ft_cpp_add_right(t, 9, ms))
      step("ft_cpp_add_left", t <- .ft_cpp_add_left(t, 0, ms))
    } else {
      results[["ft_cpp_add_right"]] <- "skipped because ft_cpp_tree_from failed"
      results[["ft_cpp_add_left"]] <- "skipped because ft_cpp_tree_from failed"
    }

    step("ft_cpp_concat", .ft_cpp_concat(base_plain, base_plain, ms))
    step("ft_cpp_locate", .ft_cpp_locate(base_plain, function(v) v >= 4, ms, ".size", 0))
    step("ft_cpp_split_tree", .ft_cpp_split_tree(base_plain, function(v) v >= 4, ms, ".size", 0))

    step("ft_cpp_oms_insert", .ft_cpp_oms_insert(x_oms, list(value = "new", key = 2), ms_oms, key_type_oms))

    step("insert", insert(x_oms, "newer", key = 2))
    step("insert_interval_index", insert(x_ivx, "newer", start = 2, end = 5))
    step("peek_point_interval_index", peek_all_point(x_ivx, 2))
    step("pop_point_interval_index", pop_all_point(x_ivx, 2))
    step("peek_overlaps_interval_index", peek_all_overlaps(x_ivx, 2, 3))
    step("peek_containing_interval_index", peek_all_containing(x_ivx, 2, 3))
    step("peek_within_interval_index", peek_all_within(x_ivx, 2, 3))

    step("pq_numeric_peek_min_max", {
      invisible(peek_min(q_num))
      invisible(peek_max(q_num))
    })
    step("pq_numeric_pop_min_insert", {
      p1 <- pop_min(q_num)
      q2 <- insert(p1$remaining, "x", priority = 1)
      invisible(peek_min(q2))
      invisible(peek_max(q2))
    })
    step("pq_date_peek_pop_insert", {
      invisible(peek_min(q_date))
      invisible(peek_max(q_date))
      p1 <- pop_min(q_date)
      q2 <- insert(p1$remaining, "d4", priority = as.Date("2020-01-05"))
      invisible(peek_min(q2))
      invisible(peek_max(q2))
    })
    step("ordered_bounds_numeric", {
      invisible(lower_bound(x_ord_num, 2))
      invisible(upper_bound(x_ord_num, 2))
    })
    step("ordered_bounds_date", {
      k <- as.Date("2020-01-02")
      invisible(lower_bound(x_ord_date, k))
      invisible(upper_bound(x_ord_date, k))
    })
  }, error = function(e) {
    setup_error <<- conditionMessage(e)
  })

  list(enabled = TRUE, setup_error = setup_error, results = results)
}

.gc_cpp_run <- .run_gc_stress_suite()

.expect_step_ok <- function(label) {
  .skip_if_no_gc_stress()

  if(!is.null(.gc_cpp_run$setup_error)) {
    testthat::fail(sprintf("GC stress suite setup failed: %s", .gc_cpp_run$setup_error))
  }

  err <- .gc_cpp_run$results[[label]]
  if(is.null(err)) {
    testthat::succeed()
  } else {
    testthat::fail(sprintf("%s failed: %s", label, err))
  }
}

testthat::test_that("C++ tree builders survive GC torture", {
  .expect_step_ok("ft_cpp_tree_from")
  .expect_step_ok("ft_cpp_tree_from_prepared")
  .expect_step_ok("ft_cpp_tree_from_sorted")
})

testthat::test_that("C++ core tree ops survive GC torture", {
  .expect_step_ok("ft_cpp_add_right")
  .expect_step_ok("ft_cpp_add_left")
  .expect_step_ok("ft_cpp_concat")
  .expect_step_ok("ft_cpp_locate")
  .expect_step_ok("ft_cpp_split_tree")
})

testthat::test_that("C++ OMS primitives survive GC torture", {
  .expect_step_ok("ft_cpp_oms_insert")
})

testthat::test_that("OMS public APIs survive GC torture", {
  .expect_step_ok("insert")
})

testthat::test_that("interval_index public APIs survive GC torture", {
  .expect_step_ok("insert_interval_index")
  .expect_step_ok("peek_point_interval_index")
  .expect_step_ok("pop_point_interval_index")
  .expect_step_ok("peek_overlaps_interval_index")
  .expect_step_ok("peek_containing_interval_index")
  .expect_step_ok("peek_within_interval_index")
})

testthat::test_that("priority_queue public APIs survive GC torture", {
  .expect_step_ok("pq_numeric_peek_min_max")
  .expect_step_ok("pq_numeric_pop_min_insert")
  .expect_step_ok("pq_date_peek_pop_insert")
})

testthat::test_that("ordered bounds survive GC torture", {
  .expect_step_ok("ordered_bounds_numeric")
  .expect_step_ok("ordered_bounds_date")
})

#' Immutables: Fast and Functional Data Structures
#'
#' The `Immutables` R package implements several immutable, or persistent, data
#' structures: operations return modified copies while remaining fast and true
#' to R's side-effect-free functional nature.
#'
#' - `flexseq`s provide list-like operations: indexed and named element access;
#'   push/pop/peek from either end for double-ended queue behavior; insertion,
#'   splitting, and concatenation.
#' - `priority_queue`s associate items with priority values and provide
#'   min and max peek/pop by priority and fast insertion.
#' - `ordered_sequence`s associate items with key values and keep the elements in
#'   sorted order by key. These may be similarly be inserted/popped/peeked by key
#'   value as well as position. Keys may be duplicated, with first-in-first-out order within key groups.
#' - `interval_index`es store items associated with interval ranges, supporting
#'   point as well as interval overlaps/contains/within queries. Items are kept
#'   in start-order enabling ordered sequence operations and sweep-line
#'   algorithms.
#'
#' Backed by monoid-annotated 2-3 fingertrees as described by
#' Hinze and Paterson, most operations are constant time, amortized constant time, or $O(\\log(n))$. Core functions are implemented in C++ (via Rcpp) for speed, with matching pure-R reference implementations using `lambda.r` syntax to match the paper.
#'
#' Finally, the developer API supports the addition of custom structures via combinations of monoids and measures.
#'
#' - Hinze, R. and Paterson, R. (2006), *Finger trees: a simple general-purpose data structure*. \doi{10.1017/S0956796805005769}
#'
#' @keywords internal
#' @import lambda.r
#' @importFrom Rcpp evalCpp
#' @useDynLib Immutables, .registration = TRUE
"_PACKAGE"

#SO

# Reference concat primitive from Hinze/Paterson finger trees:
# concatenate two trees with an explicit bridge list `ts`.
# Runtime: O(k log n) in the general internal-helper sense where k = length(ts)
# and n is receiving-tree size in repeated edge-insert paths. In package use,
# k is effectively bounded by small digit/node bridge sizes, so this is
# effectively O(log n).
if(FALSE) app3 <- function(e, ts, xs, monoids) NULL
app3(e, ts, xs, monoids) %::% Empty : list : FingerTree : list : FingerTree
app3(e, ts, xs, monoids) %as% add_all_left(xs, ts, monoids)

# Runtime: O(k log n) in the general internal-helper sense where k = length(ts)
# and n is receiving-tree size in repeated edge-insert paths. In package use,
# k is effectively bounded by small digit/node bridge sizes, so this is
# effectively O(log n).
app3(xs, ts, e, monoids) %::% FingerTree : list : Empty : list : FingerTree
app3(xs, ts, e, monoids) %as% add_all_right(xs, ts, monoids)

# Runtime: O(k log n) in the general internal-helper sense where k = length(ts)
# and n is receiving-tree size in repeated edge-insert paths. In package use,
# k is effectively bounded by small digit/node bridge sizes, so this is
# effectively O(log n).
app3(x, ts, xs, monoids) %::% Single : list : FingerTree : list : FingerTree
app3(x, ts, xs, monoids) %as% {
  add_left(add_all_left(xs, ts, monoids), .subset2(x, 1), monoids)
}

# Runtime: O(k log n) in the general internal-helper sense where k = length(ts)
# and n is receiving-tree size in repeated edge-insert paths. In package use,
# k is effectively bounded by small digit/node bridge sizes, so this is
# effectively O(log n).
app3(xs, ts, x, monoids) %::% FingerTree : list : Single : list : FingerTree
app3(xs, ts, x, monoids) %as% {
  add_right(add_all_right(xs, ts, monoids), .subset2(x, 1), monoids)
}

# Runtime: O(log n + k) in balanced usage, where k = length(ts), with recursion
# depth proportional to concatenation spine depth.
app3(xs, ts, ys, monoids) %::% Deep : list : Deep : list : FingerTree
app3(xs, ts, ys, monoids) %as% {
  mid_ts <- measured_nodes(
    c(as.list(.subset2(xs, "suffix")), ts, as.list(.subset2(ys, "prefix"))),
    monoids
  )
  measured_deep(
    .subset2(xs, "prefix"),
    app3(.subset2(xs, "middle"), mid_ts, .subset2(ys, "middle"), monoids),
    .subset2(ys, "suffix"),
    monoids
  )
}

# Runtime: O(log(min(nx, ny))) for balanced trees in normal package use
# (bounded bridge width); more generally follows app3's k-sensitive behavior.
if(FALSE) concat <- function(xs, ys, monoids) NULL
concat(xs, ys, monoids) %::% FingerTree : FingerTree : list : FingerTree
concat(xs, ys, monoids) %as% {
  app3(xs, list(), ys, monoids)
}

# Convert bridge elements into Node2/Node3 only (never Node1), matching the
# reference finger-tree middle-layer shape invariant.
# Runtime: O(k), where k = length(l).
if(FALSE) measured_nodes <- function(l, monoids) NULL
measured_nodes(l, monoids) %::% list : list : list
measured_nodes(l, monoids) %as% {
  if(length(l) < 2L) {
    stop("measured_nodes requires at least two elements.")
  }
  if(length(l) == 2) { return(list(
    measured_node2( l[[1]], l[[2]], monoids )
  ))}
  if(length(l) == 3) { return(list(
    measured_node3( l[[1]], l[[2]], l[[3]], monoids )
  ))}
  if(length(l) == 4) { return(list(
    measured_node2( l[[1]], l[[2]], monoids ),
    measured_node2( l[[3]], l[[4]], monoids )
  ))}

  first = measured_node3( l[[1]], l[[2]], l[[3]], monoids )
  rest = measured_nodes(l[4:length(l)], monoids)
  rest = c(list(first), rest)
  return(rest)
}

#SO

# Runtime: O(1).
# Build one canonical ordered entry payload.
# Used by: .oms_build_from_items() constructor assembly loop.
.oms_make_entry <- function(value, key_value, name = NULL) {
  out <- list(value = value, key = key_value)
  nm <- .ft_normalize_name(name)
  if(!is.null(nm)) {
    out <- .ft_set_name(out, nm)
  }
  out
}

# Runtime: O(n log n) for stable merge sort by key.
# Stable sort fallback when base order() cannot safely order key objects.
# Used by: .oms_order_entries() for non-primitive/custom key classes.
.oms_merge_sort_indices <- function(idx, entries, key_type) {
  n <- length(idx)
  if(n <= 1L) {
    return(idx)
  }

  mid <- as.integer(n %/% 2L)
  left <- .oms_merge_sort_indices(idx[seq_len(mid)], entries, key_type)
  right <- .oms_merge_sort_indices(idx[(mid + 1L):n], entries, key_type)

  out <- integer(n)
  i <- 1L
  j <- 1L
  k <- 1L
  while(i <= length(left) && j <= length(right)) {
    cmp <- .oms_compare_key(entries[[left[[i]]]]$key, entries[[right[[j]]]]$key, key_type)
    if(cmp <= 0L) {
      out[[k]] <- left[[i]]
      i <- i + 1L
    } else {
      out[[k]] <- right[[j]]
      j <- j + 1L
    }
    k <- k + 1L
  }

  while(i <= length(left)) {
    out[[k]] <- left[[i]]
    i <- i + 1L
    k <- k + 1L
  }
  while(j <= length(right)) {
    out[[k]] <- right[[j]]
    j <- j + 1L
    k <- k + 1L
  }
  out
}

# Runtime: O(n log n) stable by key and FIFO on ties.
# Sort entries by key while preserving input order across ties.
# Used by: .oms_build_from_items() before tree construction.
.oms_order_entries <- function(entries, key_type) {
  if(length(entries) <= 1L) {
    return(entries)
  }

  idx <- seq_along(entries)
  ord <- if(identical(key_type, "numeric")) {
    order(vapply(entries, function(e) e$key, numeric(1)), idx)
  } else if(identical(key_type, "character")) {
    order(vapply(entries, function(e) e$key, character(1)), idx)
  } else if(identical(key_type, "logical")) {
    order(vapply(entries, function(e) e$key, logical(1)), idx)
  } else {
    keys <- lapply(entries, function(e) e$key)
    ord_try <- tryCatch(order(do.call(c, keys), idx), error = function(e) NULL)
    if(is.null(ord_try) || length(ord_try) != length(entries)) {
      .oms_merge_sort_indices(idx, entries, key_type)
    } else {
      ord_try
    }
  }
  entries[ord]
}

# Runtime: O(n) in entry count.
# Normalize outer names into ft_name entry attrs, then clear list names.
# Used by: .oms_tree_from_ordered_entries() to preserve name indexing contract.
.oms_prepare_entry_names <- function(entries) {
  if(length(entries) == 0L) {
    return(entries)
  }
  nms <- names(entries)
  if(is.null(nms)) {
    return(entries)
  }

  out <- entries
  for(i in seq_along(out)) {
    nm <- .ft_normalize_name(nms[[i]])
    if(!is.null(nm)) {
      out[[i]] <- .ft_set_name(out[[i]], nm)
    }
  }
  names(out) <- NULL
  out
}

# Runtime: O(n log n) for sorting + O(n) bulk build.
# Main constructor pipeline: validate lengths, normalize keys/type, build
# canonical entries, stable-sort, merge required monoids, and wrap class state.
# Used by: as_ordered_sequence() and ordered_sequence().
.oms_build_from_items <- function(items, keys = NULL, monoids = NULL) {
  n <- length(items)

  if(n == 0L) {
    if(!is.null(keys) && length(as.list(keys)) > 0L) {
      stop("`keys` must be empty when no elements are supplied.")
    }
    base <- .as_flexseq_build(list(), monoids = .oms_merge_monoids(monoids))
    return(.as_ordered_sequence(base, key_type = NULL))
  }

  if(is.null(keys)) {
    stop("`keys` is required when elements are supplied.")
  }
  key_list <- as.list(keys)
  if(length(key_list) != n) {
    stop("`keys` length must match elements length.")
  }

  entries <- vector("list", n)
  item_names <- names(items)
  key_type <- NULL
  for(i in seq_len(n)) {
    norm <- .oms_normalize_key(key_list[[i]])
    key_type <- .oms_validate_key_type(key_type, norm$key_type)
    entries[[i]] <- .oms_make_entry(items[[i]], norm$key)
  }
  if(!is.null(item_names) && length(item_names) == n) {
    names(entries) <- item_names
  }

  entries <- .oms_order_entries(entries, key_type)

  merged_monoids <- .oms_merge_monoids(monoids)
  base <- .oms_tree_from_ordered_entries(entries, merged_monoids)
  .as_ordered_sequence(base, key_type = key_type)
}

# Runtime: O(n) for ordered entries.
# Build a measured tree from already key-ordered entries (C++ fast path when
# available, otherwise linear R reference path).
# Used by: .oms_build_from_items() and ordered fapply rebuild path.
.oms_tree_from_ordered_entries <- function(entries, monoids) {
  entries <- .oms_prepare_entry_names(entries)
  if(.ft_cpp_can_use(monoids)) {
    return(.as_flexseq(.ft_cpp_tree_from_sorted(entries, monoids)))
  }
  .ft_tree_from_list_linear(entries, monoids)
}

# Runtime: O(n log n) from build and ordering.
#' Build an Ordered Sequence from `x` and `keys`
#'
#' Constructs an `ordered_sequence` by pairing each element of `x` with the
#' corresponding key in `keys`.
#'
#' @param x Elements to add.
#' @param keys Key values with the same length as `x`.
#' @return An `ordered_sequence`.
#' @details
#' Output is always sorted by key.
#'
#' Duplicate keys are allowed; ties preserve input order (stable/FIFO within the
#' same key).
#'
#' Names on `x` are preserved as element names.
#' @examples
#' xs <- as_ordered_sequence(c("d", "a", "b", "a2"), keys = c(4, 1, 2, 1))
#' xs
#' length(elements_between(xs, 1, 1))
#'
#' n <- as_ordered_sequence(setNames(as.list(c("a", "b")), c("ka", "kb")), keys = c(2, 1))
#' n[["kb"]]
#'
#' # Keys can be other comparable types
#' num_by_chr <- as_ordered_sequence(c(20, 10, 30), keys = c("b", "a", "c"))
#' num_by_chr
#' @export
# Public cast/build entry for list/vector-like inputs.
# Used by: users and tests; delegates to .oms_build_from_items().
as_ordered_sequence <- function(x, keys) {
  .oms_build_from_items(as.list(x), keys = keys, monoids = NULL)
}

# Runtime: O(n log n) from build and ordering.
#' Construct an Ordered Sequence
#'
#' Convenience constructor from `...` and matching `keys`.
#'
#' @param ... Elements to add.
#' @param keys Key values with the same length as `...`.
#' @return An `ordered_sequence`.
#' @details
#' Empty construction is supported: `ordered_sequence()` returns an empty
#' ordered sequence.
#'
#' Output is always sorted by key, with stable order across duplicate keys.
#' @examples
#' xs <- ordered_sequence("bb", "a", "ccc", keys = c(2, 1, 3))
#' xs
#' lower_bound(xs, 2)
#'
#' num_by_chr <- ordered_sequence(20, 10, 30, keys = c("b", "a", "c"))
#' num_by_chr
#'
#' ordered_sequence()
#' @export
# Variadic convenience constructor; delegates to .oms_build_from_items().
ordered_sequence <- function(..., keys) {
  if(missing(keys)) {
    keys <- NULL
  }
  .oms_build_from_items(list(...), keys = keys, monoids = NULL)
}

testthat::test_that("coro::loop yields flexseq elements in order", {
  s <- as_flexseq(letters[1:5])
  out <- character(0)
  coro::loop(for (x in s) out <- c(out, x))

  testthat::expect_identical(out, letters[1:5])
})

testthat::test_that("coro::loop over empty flexseq iterates zero times", {
  n <- 0L
  coro::loop(for (x in flexseq()) n <- n + 1L)

  testthat::expect_identical(n, 0L)
})

testthat::test_that("coro::loop strips ft_name from yielded elements of named flexseq", {
  s <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))
  collected <- list()
  coro::loop(for (x in s) collected <- c(collected, list(x)))

  testthat::expect_length(collected, 3L)
  testthat::expect_identical(unlist(collected), 1:3)
  for(el in collected) {
    testthat::expect_null(attr(el, "ft_name", exact = TRUE))
  }
})

testthat::test_that("iteration does not mutate the source flexseq", {
  s <- as_flexseq(letters[1:5])
  before <- as.list(s)
  coro::loop(for (x in s) invisible(NULL))

  testthat::expect_identical(as.list(s), before)
  testthat::expect_identical(length(s), 5L)
})

testthat::test_that("as_iterator.flexseq is exhausted after one full pass", {
  s <- as_flexseq(1:3)
  it <- coro::as_iterator(s)

  testthat::expect_identical(it(), 1L)
  testthat::expect_identical(it(), 2L)
  testthat::expect_identical(it(), 3L)
  testthat::expect_true(coro::is_exhausted(it()))
  testthat::expect_true(coro::is_exhausted(it()))
})

testthat::test_that("iterator streams across deeper tree structure", {
  vals <- as.list(seq_len(200L))
  s <- as_flexseq(vals)
  out <- integer(0)
  coro::loop(for (x in s) out <- c(out, x))

  testthat::expect_identical(out, seq_len(200L))
})

testthat::test_that("as_iterator.flexseq dispatches via class", {
  s <- as_flexseq(letters[1:3])
  it <- coro::as_iterator(s)

  testthat::expect_type(it, "closure")
  testthat::expect_identical(it(), "a")
})

testthat::test_that("loop() is re-exported from coro and works without coro:: prefix", {
  testthat::expect_identical(loop, coro::loop)

  s <- as_flexseq(1:4)
  out <- integer(0)
  loop(for (x in s) out <- c(out, x))

  testthat::expect_identical(out, 1:4)
})

# ---- ordered_sequence iteration ---------------------------------------------

testthat::test_that("loop over ordered_sequence yields values in key-ascending order", {
  os <- as_ordered_sequence(list("two", "one", "three"), keys = c(2, 1, 3))
  out <- character(0)
  loop(for (v in os) out <- c(out, v))

  testthat::expect_identical(out, c("one", "two", "three"))
})

testthat::test_that("loop over ordered_sequence matches as.list()", {
  os <- as_ordered_sequence(list(10, 20, 30, 40), keys = c(4, 1, 3, 2))
  out <- list()
  loop(for (v in os) out <- c(out, list(v)))

  testthat::expect_identical(out, as.list(os))
})

testthat::test_that("loop over empty ordered_sequence iterates zero times", {
  n <- 0L
  loop(for (v in ordered_sequence()) n <- n + 1L)

  testthat::expect_identical(n, 0L)
})

testthat::test_that("iteration does not mutate the source ordered_sequence", {
  os <- as_ordered_sequence(list("a", "b", "c"), keys = c(1, 2, 3))
  before <- as.list(os)
  loop(for (v in os) invisible(NULL))

  testthat::expect_identical(as.list(os), before)
  testthat::expect_identical(length(os), 3L)
})

testthat::test_that("loop over ordered_sequence yields bare values (no entry wrapper)", {
  os <- as_ordered_sequence(list("x", "y"), keys = c(1, 2))
  collected <- list()
  loop(for (v in os) collected <- c(collected, list(v)))

  # values should be bare strings, not list(value=..., key=...)
  testthat::expect_identical(collected[[1]], "x")
  testthat::expect_identical(collected[[2]], "y")
})

# ---- interval_index iteration -----------------------------------------------

testthat::test_that("loop over interval_index yields values in interval tree order", {
  ix <- as_interval_index(list("B", "A", "C"), start = c(1, 0, 5), end = c(3, 2, 8))
  out <- character(0)
  loop(for (v in ix) out <- c(out, v))

  testthat::expect_identical(out, c("A", "B", "C"))
})

testthat::test_that("loop over interval_index matches as.list()", {
  ix <- as_interval_index(
    list("p", "q", "r", "s"),
    start = c(10, 1, 5, 3),
    end   = c(20, 4, 8, 7)
  )
  out <- list()
  loop(for (v in ix) out <- c(out, list(v)))

  testthat::expect_identical(out, as.list(ix))
})

testthat::test_that("loop over empty interval_index iterates zero times", {
  n <- 0L
  loop(for (v in interval_index()) n <- n + 1L)

  testthat::expect_identical(n, 0L)
})

testthat::test_that("iteration does not mutate the source interval_index", {
  ix <- as_interval_index(list("a", "b"), start = c(1, 3), end = c(2, 5))
  before <- as.list(ix)
  loop(for (v in ix) invisible(NULL))

  testthat::expect_identical(as.list(ix), before)
  testthat::expect_identical(length(ix), 2L)
})

testthat::test_that("loop over interval_index yields bare values (no entry wrapper)", {
  ix <- as_interval_index(list("x", "y"), start = c(0, 1), end = c(2, 3))
  collected <- list()
  loop(for (v in ix) collected <- c(collected, list(v)))

  testthat::expect_identical(collected[[1]], "x")
  testthat::expect_identical(collected[[2]], "y")
})

# ---- priority_queue iteration -----------------------------------------------

testthat::test_that("loop over priority_queue yields values in priority-ascending order", {
  pq <- priority_queue("a", "b", "c", priorities = c(3, 1, 2))
  out <- character(0)
  loop(for (v in pq) out <- c(out, v))

  testthat::expect_identical(out, c("b", "c", "a"))
})

testthat::test_that("loop over priority_queue yields FIFO within equal priorities", {
  pq <- priority_queue("a", "b", "c", "d", priorities = c(2, 1, 1, 2))
  out <- character(0)
  loop(for (v in pq) out <- c(out, v))

  # priorities: a=2, b=1, c=1, d=2. ascending = 1,1,2,2.
  # FIFO within ties: b before c (both pri 1); a before d (both pri 2).
  testthat::expect_identical(out, c("b", "c", "a", "d"))
})

testthat::test_that("loop over empty priority_queue iterates zero times", {
  n <- 0L
  loop(for (v in priority_queue()) n <- n + 1L)

  testthat::expect_identical(n, 0L)
})

testthat::test_that("iteration does not mutate the source priority_queue", {
  pq <- priority_queue("a", "b", "c", priorities = c(3, 1, 2))
  before_min <- peek_min(pq)
  loop(for (v in pq) invisible(NULL))

  testthat::expect_identical(length(pq), 3L)
  testthat::expect_identical(peek_min(pq), before_min)
})

testthat::test_that("partial priority_queue iteration leaves source untouched", {
  pq <- priority_queue("a", "b", "c", "d", priorities = c(4, 1, 3, 2))
  seen <- character(0)
  loop(for (v in pq) {
    seen <- c(seen, v)
    if(length(seen) == 2L) break
  })

  testthat::expect_identical(seen, c("b", "d"))
  testthat::expect_identical(length(pq), 4L)
})

testthat::test_that("as_iterator.priority_queue is exhausted after one full pass", {
  pq <- priority_queue("x", "y", priorities = c(2, 1))
  it <- coro::as_iterator(pq)

  testthat::expect_identical(it(), "y")
  testthat::expect_identical(it(), "x")
  testthat::expect_true(coro::is_exhausted(it()))
  testthat::expect_true(coro::is_exhausted(it()))
})

#SO

testthat::test_that("public constructor wrappers align with as_* constructors", {
  fx <- flexseq("a", "b", "c")
  fx_as <- as_flexseq(list("a", "b", "c"))
  testthat::expect_s3_class(fx, "flexseq")
  testthat::expect_equal(as.list(fx), as.list(fx_as))

  pq <- priority_queue(one = "a", two = "b", priorities = c(2, 1))
  pq_as <- as_priority_queue(list(one = "a", two = "b"), priorities = c(2, 1))
  testthat::expect_s3_class(pq, "priority_queue")
  testthat::expect_equal(as.list(pq), as.list(pq_as))
  testthat::expect_equal(peek_min(pq), peek_min(pq_as))
  testthat::expect_equal(peek_max(pq), peek_max(pq_as))

  os <- ordered_sequence(one = "a", two = "b", keys = c(2, 1))
  os_as <- as_ordered_sequence(list(one = "a", two = "b"), keys = c(2, 1))
  testthat::expect_s3_class(os, "ordered_sequence")
  testthat::expect_equal(as.list(os), as.list(os_as))
  testthat::expect_identical(count_key(os, 1), count_key(os_as, 1))
  testthat::expect_identical(count_key(os, 2), count_key(os_as, 2))

  ix <- interval_index(one = "a", two = "b", start = c(2, 1), end = c(4, 3), default_query_bounds = "[]")
  ix_as <- as_interval_index(list(one = "a", two = "b"), start = c(2, 1), end = c(4, 3), default_query_bounds = "[]")
  testthat::expect_s3_class(ix, "interval_index")
  testthat::expect_equal(as.list(ix), as.list(ix_as))
  ix_entries <- lapply(.ivx_entries(ix), function(e) list(start = e$start, end = e$end))
  ix_as_entries <- lapply(.ivx_entries(ix_as), function(e) list(start = e$start, end = e$end))
  testthat::expect_equal(ix_entries, ix_as_entries)
})

testthat::test_that("constructors enforce required metadata args when elements are supplied", {
  testthat::expect_error(priority_queue("a"), "`priorities` is required")
  testthat::expect_error(as_priority_queue(list("a")), "argument \"priorities\" is missing")

  testthat::expect_error(ordered_sequence("a"), "`keys` is required")
  testthat::expect_error(as_ordered_sequence(list("a")), "argument \"keys\" is missing")

  testthat::expect_error(interval_index("a", end = 2), "`start` is required")
  testthat::expect_error(interval_index("a", start = 1), "`end` is required")
  testthat::expect_error(as_interval_index(list("a"), end = 2), "argument \"start\" is missing")
  testthat::expect_error(as_interval_index(list("a"), start = 1), "argument \"end\" is missing")
})

testthat::test_that("as_flexseq on flexseq is identity-safe for named elements", {
  fx <- flexseq(fx = "x", fy = "y")
  out <- as_flexseq(fx)

  testthat::expect_s3_class(out, "flexseq")
  testthat::expect_identical(as.list(out), as.list(fx))
  testthat::expect_identical(names(as.list(out)), c("fx", "fy"))
})

# Snapshot-based tests for the four `print.*` methods.
# Snapshots live in `tests/testthat/_snaps/print/`. On first run testthat
# creates them; subsequent runs diff against them. If you intentionally change
# a print format, accept the new output with `make snapshot-accept` (or
# `testthat::snapshot_accept()` interactively) and commit the updated snapshots.

# --- flexseq -----------------------------------------------------------------

testthat::test_that("print.flexseq handles empty, single, and small named trees", {
  testthat::expect_snapshot(print(flexseq()))
  testthat::expect_snapshot(print(as_flexseq(1)))
  testthat::expect_snapshot(
    print(as_flexseq(setNames(as.list(1:4), letters[1:4])))
  )
})

testthat::test_that("print.flexseq truncates long unnamed trees with head/tail preview", {
  testthat::expect_snapshot(print(as_flexseq(1:12), max_elements = 4))
  # max_elements = 0 shows only the header.
  testthat::expect_snapshot(print(as_flexseq(1:12), max_elements = 0))
})

testthat::test_that("print.flexseq shows custom monoid aggregates when requested", {
  sum_m <- measure_monoid(function(a, b) a + b, 0, function(el) as.numeric(el))
  x <- add_monoids(as_flexseq(1:5), list(sum = sum_m))
  testthat::expect_snapshot(print(x, max_elements = 2, show_custom_monoids = TRUE))
  # With no custom monoids attached, the header still appears.
  testthat::expect_snapshot(
    print(as_flexseq(1:3), max_elements = 0, show_custom_monoids = TRUE)
  )
})

testthat::test_that("print.flexseq rejects invalid arguments with informative errors", {
  testthat::expect_error(print(as_flexseq(1:3), max_elements = -1),
                         "non-negative integer")
  testthat::expect_error(print(as_flexseq(1:3), max_elements = c(1, 2)),
                         "non-negative integer")
  testthat::expect_error(print(as_flexseq(1:3), show_custom_monoids = NA),
                         "TRUE or FALSE")
})

# --- priority_queue ----------------------------------------------------------

testthat::test_that("print.priority_queue handles empty, named, unnamed, and truncated queues", {
  testthat::expect_snapshot(print(priority_queue()))
  testthat::expect_snapshot(
    print(priority_queue(one = 1, two = 2, three = 3, priorities = c(20, 30, 10)))
  )
  testthat::expect_snapshot(
    print(priority_queue(1, 2, 3, priorities = c(2, 1, 3)))
  )
  # Larger queue with truncation.
  big_q <- do.call(priority_queue,
                   c(as.list(letters[1:10]),
                     list(priorities = c(5, 1, 8, 3, 2, 9, 4, 7, 6, 10))))
  testthat::expect_snapshot(print(big_q, max_elements = 4))
})

testthat::test_that("print.priority_queue shows custom monoid aggregates", {
  sum_item <- measure_monoid(function(a, b) a + b, 0,
                             function(entry) as.numeric(entry$value))
  q <- add_monoids(priority_queue(1, 2, 3, priorities = c(2, 1, 3)),
                   list(sum_item = sum_item))
  testthat::expect_snapshot(print(q, max_elements = 0, show_custom_monoids = TRUE))
})

# --- ordered_sequence --------------------------------------------------------

testthat::test_that("print.ordered_sequence handles empty, named, unnamed, and truncated sequences", {
  testthat::expect_snapshot(print(ordered_sequence()))
  testthat::expect_snapshot(
    print(ordered_sequence(one = "a", two = "b", three = "c", keys = c(20, 30, 10)))
  )
  testthat::expect_snapshot(
    print(ordered_sequence("x", "y", "z", keys = c(2, 1, 3)))
  )
  testthat::expect_snapshot(
    print(as_ordered_sequence(letters[1:10], keys = 10:1), max_elements = 4)
  )
})

testthat::test_that("print.ordered_sequence shows custom monoid aggregates", {
  sum_key <- measure_monoid(function(a, b) a + b, 0,
                            function(entry) as.numeric(entry$key))
  ys <- add_monoids(ordered_sequence("a", "b", keys = c(2, 1)),
                    list(sum_key = sum_key))
  testthat::expect_snapshot(print(ys, max_elements = 0, show_custom_monoids = TRUE))
})

# --- interval_index ----------------------------------------------------------

testthat::test_that("print.interval_index handles empty, named, unnamed, and truncated indices", {
  testthat::expect_snapshot(print(interval_index()))
  testthat::expect_snapshot(
    print(interval_index(one = 1, two = 2, three = 3,
                         start = c(20, 30, 10), end = c(25, 37, 24)))
  )
  testthat::expect_snapshot(
    print(interval_index(1, 2, 3, start = c(2, 4, 6), end = c(3, 5, 8),
                         default_query_bounds = "[]"))
  )
  big_ix <- do.call(interval_index,
                    c(as.list(letters[1:10]),
                      list(start = 1:10, end = 2:11)))
  testthat::expect_snapshot(print(big_ix, max_elements = 4))
})

testthat::test_that("print.interval_index shows custom monoid aggregates", {
  width_sum <- measure_monoid(
    function(a, b) a + b, 0,
    function(entry) as.numeric(entry$end - entry$start)
  )
  ix <- add_monoids(
    interval_index(1, 2, start = c(1, 3), end = c(2, 5)),
    list(width_sum = width_sum)
  )
  testthat::expect_snapshot(print(ix, max_elements = 0, show_custom_monoids = TRUE))
})

testthat::test_that("unlist.flexseq mirrors unlist(as.list(.)) for base flexseq", {
  x <- as_flexseq(as.list(1:4))

  testthat::expect_identical(unlist(x), c(1L, 2L, 3L, 4L))
  testthat::expect_identical(
    unlist(x, recursive = FALSE, use.names = FALSE),
    unlist(as.list(x), recursive = FALSE, use.names = FALSE)
  )
})

testthat::test_that("unlist.flexseq forwards use.names behavior", {
  x <- as_flexseq(setNames(as.list(c(10, 20)), c("a", "b")))

  testthat::expect_identical(unlist(x, use.names = TRUE), c(a = 10, b = 20))
  testthat::expect_identical(unlist(x, use.names = FALSE), c(10, 20))
})

testthat::test_that("priority_queue unlist returns payload items", {
  q <- as_priority_queue(
    setNames(as.list(c("x", "y")), c("kx", "ky")),
    priorities = c(2, 1)
  )

  testthat::expect_identical(unlist(q, use.names = TRUE), c(kx = "x", ky = "y"))
  testthat::expect_identical(unlist(q, use.names = FALSE), c("x", "y"))
})

testthat::test_that("as.list.priority_queue supports drop_meta", {
  q <- as_priority_queue(
    setNames(as.list(c("x", "y")), c("kx", "ky")),
    priorities = c(2, 1)
  )

  full <- as.list(q)
  bare <- as.list(q, drop_meta = TRUE)

  testthat::expect_identical(names(full), c("kx", "ky"))
  testthat::expect_identical(full[[1]]$value, "x")
  testthat::expect_identical(full[[1]]$priority, 2)
  testthat::expect_identical(bare, list(kx = "x", ky = "y"))
  testthat::expect_error(as.list(q, drop_meta = NA), "TRUE or FALSE")
})

testthat::test_that("ordered_sequence and interval_index inherit unlist.flexseq", {
  xs <- as_ordered_sequence(
    setNames(list("x1", "x2"), c("a", "b")),
    keys = c(1, 2)
  )
  ix <- as_interval_index(
    setNames(list("i1", "i2"), c("u", "v")),
    start = c(1, 3),
    end = c(2, 5)
  )

  testthat::expect_identical(
    unlist(xs, use.names = TRUE),
    unlist(as.list(xs), use.names = TRUE)
  )
  testthat::expect_identical(
    unlist(ix, recursive = FALSE),
    unlist(as.list(ix), recursive = FALSE)
  )
})

# Smoke tests for `plot_structure()`. These verify the plotting code
# executes end-to-end without error across empty / single / small / larger
# inputs and across each finger-tree-backed structure type — they do NOT
# verify the rendered output (see vdiffr if visual regression testing is
# later desired).

.plot_to_null <- function(expr) {
  grDevices::pdf(tempfile(fileext = ".pdf"))
  on.exit(grDevices::dev.off(), add = TRUE)
  force(expr)
  invisible(NULL)
}

testthat::test_that("plot_structure runs without error on flexseq (empty / single / small / larger / named)", {
  testthat::skip_if_not_installed("igraph")
  testthat::expect_silent(.plot_to_null(plot_structure(flexseq())))
  testthat::expect_silent(.plot_to_null(plot_structure(as_flexseq(1))))
  testthat::expect_silent(.plot_to_null(plot_structure(as_flexseq(1:5))))
  # Larger input to force Deep/Digit/Node branches in graph construction.
  testthat::expect_silent(.plot_to_null(plot_structure(as_flexseq(1:20))))
  # Named input.
  testthat::expect_silent(
    .plot_to_null(plot_structure(as_flexseq(setNames(as.list(1:5), letters[1:5]))))
  )
})

testthat::test_that("plot_structure runs without error on priority_queue", {
  testthat::skip_if_not_installed("igraph")
  testthat::expect_silent(.plot_to_null(plot_structure(priority_queue())))
  testthat::expect_silent(
    .plot_to_null(plot_structure(priority_queue(1, 2, 3, priorities = c(2, 1, 3))))
  )
  big <- do.call(priority_queue,
                 c(as.list(letters[1:15]),
                   list(priorities = seq(15, 1))))
  testthat::expect_silent(.plot_to_null(plot_structure(big)))
})

testthat::test_that("plot_structure runs without error on ordered_sequence", {
  testthat::skip_if_not_installed("igraph")
  testthat::expect_silent(.plot_to_null(plot_structure(ordered_sequence())))
  testthat::expect_silent(
    .plot_to_null(plot_structure(ordered_sequence("a", "b", "c", keys = c(2, 1, 3))))
  )
  testthat::expect_silent(
    .plot_to_null(plot_structure(as_ordered_sequence(letters[1:15], keys = 15:1)))
  )
})

testthat::test_that("plot_structure runs without error on interval_index", {
  testthat::skip_if_not_installed("igraph")
  testthat::expect_silent(.plot_to_null(plot_structure(interval_index())))
  testthat::expect_silent(
    .plot_to_null(plot_structure(interval_index(1, 2, 3, start = c(2, 4, 6), end = c(3, 5, 8))))
  )
  big <- do.call(interval_index,
                 c(as.list(letters[1:15]),
                   list(start = 1:15, end = 2:16)))
  testthat::expect_silent(.plot_to_null(plot_structure(big)))
})

testthat::test_that("plot_structure honors node_label variants and legend = FALSE", {
  testthat::skip_if_not_installed("igraph")
  x <- as_flexseq(1:8)
  testthat::expect_silent(.plot_to_null(plot_structure(x, node_label = "value")))
  testthat::expect_silent(.plot_to_null(plot_structure(x, node_label = "type")))
  testthat::expect_silent(.plot_to_null(plot_structure(x, node_label = "both")))
  testthat::expect_silent(.plot_to_null(plot_structure(x, node_label = "none")))
  testthat::expect_silent(
    .plot_to_null(plot_structure(x, node_label = function(node) node$type))
  )
  testthat::expect_silent(.plot_to_null(plot_structure(x, legend = FALSE)))
  testthat::expect_silent(.plot_to_null(plot_structure(x, title = "A tree", label_edges = TRUE)))
})

#SO

# Runtime: O(n) from traversal + ordered bulk rebuild.
.oms_apply_impl <- function(x, f, ..., preserve_custom_monoids = TRUE) {
  .oms_assert_set(x)
  if(!is.function(f)) {
    stop("`FUN` must be a function.")
  }
  if(!is.logical(preserve_custom_monoids) || length(preserve_custom_monoids) != 1L || is.na(preserve_custom_monoids)) {
    stop("`preserve_custom_monoids` must be TRUE or FALSE.")
  }

  entries <- as.list.flexseq(x)
  n <- length(entries)
  if(n == 0L) {
    return(x)
  }

  out <- vector("list", n)
  out_names <- if(is.null(names(entries))) rep("", n) else names(entries)

  dot_args <- list(...)
  accepts_name <- .ft_fun_accepts_n_positional(f, 3L)
  has_dots <- length(dot_args) > 0L

  for(i in seq_len(n)) {
    e <- entries[[i]]
    cur_name <- out_names[[i]]
    item2 <- if(accepts_name && !has_dots) {
      f(e$value, e$key, cur_name)
    } else if(!accepts_name && !has_dots) {
      f(e$value, e$key)
    } else if(accepts_name) {
      do.call(f, c(list(e$value, e$key, cur_name), dot_args))
    } else {
      do.call(f, c(list(e$value, e$key), dot_args))
    }
    out[[i]] <- .oms_make_entry(value = item2, key_value = e$key)
  }

  if(any(out_names != "")) {
    names(out) <- out_names
  }

  ms <- if(isTRUE(preserve_custom_monoids)) {
    attr(x, "monoids", exact = TRUE)
  } else {
    .oms_merge_monoids(NULL)
  }
  out_tree <- .oms_tree_from_ordered_entries(out, ms)
  .ord_wrap_like(x, out_tree, key_type = .oms_key_type_state(x))
}

#' @method fapply ordered_sequence
#' @export
#' @noRd
fapply.ordered_sequence <- function(X, FUN, ..., preserve_custom_monoids = TRUE) {
  if(!is.function(FUN)) {
    stop("`FUN` must be a function.")
  }
  .oms_apply_impl(X, FUN, ..., preserve_custom_monoids = preserve_custom_monoids)
}

#SO

# theoretically someone could define a monoid that allows splitting a priority
# queue, but since the order of elements doesn't relate to priorities the
# resulting split wouldn't be well defined with respect to priorities.
# so we block splitting operations.

# Runtime: O(1).
#' @method split_by_predicate priority_queue
#' @export
#' @noRd
split_by_predicate.priority_queue <- function(x, predicate, monoid_name) {
  stop("`split_by_predicate()` is not supported for priority_queue. Cast first with `as_flexseq()`.")
}

# Runtime: O(1).
#' @method split_around_by_predicate priority_queue
#' @export
#' @noRd
split_around_by_predicate.priority_queue <- function(t, predicate, monoid_name, accumulator = NULL) {
  stop("`split_around_by_predicate()` is not supported for priority_queue. Cast first with `as_flexseq()`.")
}

# Runtime: O(1).
#' @method split_at priority_queue
#' @export
#' @noRd
split_at.priority_queue <- function(x, at, pull_index = FALSE) {
  stop("`split_at()` is not supported for priority_queue. Cast first with `as_flexseq()`.")
}

#SO

#' Print an Ordered Sequence Summary
#'
#' Prints a compact summary and head/tail preview in key order.
#'
#' @param x An `ordered_sequence`.
#' @param max_elements Maximum number of elements shown in the preview.
#' @param show_custom_monoids Logical; show attached non-default monoids and
#'   their root cached measures.
#' @param ... Passed through to per-element `print()`.
#' @return Invisibly returns `x`.
#' @examples
#' xs <- ordered_sequence(one = "a", two = "b", three = "c", keys = c(20, 30, 10))
#' print(xs, max_elements = 4)
#' sum_key <- measure_monoid(`+`, 0, function(entry) as.numeric(entry$key))
#' ys2 <- add_monoids(ordered_sequence("a", "b", keys = c(2, 1)), list(sum_key = sum_key))
#' print(ys2, max_elements = 0, show_custom_monoids = TRUE)
#'
#' ys <- ordered_sequence("x", "y", "z", keys = c(2, 1, 3))
#' print(ys, max_elements = 3)
#'
#' print(ordered_sequence())
#' @export
#' @method print ordered_sequence
# Runtime: O((k + h) log n), where k = shown elements and h = preview split overhead.
print.ordered_sequence <- function(x, max_elements = 4L, show_custom_monoids = FALSE, ...) {
  .oms_assert_set(x)
  max_elements <- .ft_validate_print_max_elements(max_elements)
  show_custom <- .ft_validate_show_custom_monoids(show_custom_monoids)

  n <- length(x)
  nn <- as.integer(node_measure(x, ".named_count"))
  named <- isTRUE(nn > 0L)
  cat(if(named) "Named" else "Unnamed", " ordered_sequence with ", n, " element", if(n == 1L) "" else "s", ".\n", sep = "")
  if(show_custom) {
    .ft_print_custom_monoids(x, excluded_names = c(".size", ".named_count", ".oms_max_key"))
  }

  if(n == 0L || max_elements == 0L) {
    return(invisible(x))
  }

  cat("\nElements (by key order):\n\n")
  preview <- .pick_preview_sizes(n, max_elements)
  excluded_names <- c(".size", ".named_count", ".oms_max_key")
  for(i in preview$head) {
    entry <- .ft_get_elem_at(x, as.integer(i))
    nm <- .ft_get_name(entry)
    if(named && !is.null(nm)) {
      cat("$", nm, " (key ", .ft_format_scalar(entry$key), ")\n", sep = "")
    } else {
      cat("[[", i, "]] (key ", .ft_format_scalar(entry$key), ")\n", sep = "")
    }
    print(entry$value, ...)
    .ft_print_elem_custom_monoids(x, entry, excluded_names, show_custom)
    cat("\n")
  }

  .ft_print_skipped(preview$skipped)

  for(i in preview$tail) {
    entry <- .ft_get_elem_at(x, as.integer(i))
    nm <- .ft_get_name(entry)
    if(named && !is.null(nm)) {
      cat("$", nm, " (key ", .ft_format_scalar(entry$key), ")\n", sep = "")
    } else {
      cat("[[", i, "]] (key ", .ft_format_scalar(entry$key), ")\n", sep = "")
    }
    print(entry$value, ...)
    .ft_print_elem_custom_monoids(x, entry, excluded_names, show_custom)
    cat("\n")
  }
  invisible(x)
}

#SO
# checks for ensuring measures and monoids are present, consistent, and cached
# also constructors for building fingertree components with measures and monoids attached

# check for MeasureMonoid type
# Runtime: O(1).
if(FALSE) is_measure_monoid <- function(r) NULL
is_measure_monoid(r) %::% . : logical
is_measure_monoid(r) %as% inherits(r, "MeasureMonoid")

# a non-empty named list of MeasureMonoid specs
# Runtime: O(k), where k = length(x).
if(FALSE) is_measure_monoid_list <- function(x) NULL
is_measure_monoid_list(x) %::% . : logical
is_measure_monoid_list(x) %as% {
  is.list(x) && length(x) > 0 && all(vapply(x, is_measure_monoid, logical(1)))
}

# normalize monoid fields to canonical named layout (f, i, measure).
# Runtime: O(1).
if(FALSE) .normalize_measure_monoid_spec <- function(r, monoid_name = NULL) NULL
.normalize_measure_monoid_spec(r, monoid_name = NULL) %::% MeasureMonoid : . : MeasureMonoid
.normalize_measure_monoid_spec(r, monoid_name = NULL) %as% {
  lbl <- if(is.null(monoid_name) || is.na(monoid_name) || monoid_name == "") {
    "measure monoid"
  } else {
    paste0("measure monoid `", monoid_name, "`")
  }

  nms <- names(r)
  read_field <- function(key, idx) {
    if(!is.null(nms) && key %in% nms) {
      return(r[[key]])
    }
    if(length(r) >= idx) {
      return(r[[idx]])
    }
    stop(lbl, " must provide fields `f`, `i`, and `measure`.")
  }

  f <- read_field("f", 1L)
  i <- read_field("i", 2L)
  measure <- read_field("measure", 3L)

  if(!is.function(f)) {
    stop(lbl, " field `f` must be a function.")
  }
  if(!is.function(measure)) {
    stop(lbl, " field `measure` must be a function.")
  }

  out <- list(f = f, i = i, measure = measure)
  class(out) <- class(r)
  out
}

# default size measure used for indexing and structural operations
# Runtime: O(1).
if(FALSE) size_measure_monoid <- function() NULL
size_measure_monoid() %::% MeasureMonoid
size_measure_monoid() %as% {
  measure_monoid(function(a, b) a + b, 0, function(el) 1)
}

# default name-count measure used to enforce named/unnamed tree invariants
# Runtime: O(1).
if(FALSE) named_count_measure_monoid <- function() NULL
named_count_measure_monoid() %::% MeasureMonoid
named_count_measure_monoid() %as% {
  measure_monoid(function(a, b) a + b, 0L, function(el) {
    if(isTRUE(.ft_has_name(el))) 1L else 0L
  })
}

# normalize a candidate element name; NULL/NA/"" are treated as missing
# Runtime: O(1).
if(FALSE) .ft_normalize_name <- function(x) NULL
.ft_normalize_name(x) %::% . : .
.ft_normalize_name(x) %as% {
  if(is.null(x) || length(x) == 0L) {
    return(NULL)
  }
  if(length(x) != 1L) {
    stop("Element names must be scalar.")
  }
  nm <- as.character(x)
  if(is.na(nm) || nm == "") {
    return(NULL)
  }
  nm
}

# read the internal element name payload
# Runtime: O(1).
if(FALSE) .ft_get_name <- function(el) NULL
.ft_get_name(el) %::% . : .
.ft_get_name(el) %as% {
  .ft_normalize_name(attr(el, "ft_name", exact = TRUE))
}

# check whether an element carries an internal name
# Runtime: O(1).
if(FALSE) .ft_has_name <- function(el) NULL
.ft_has_name(el) %::% . : logical
.ft_has_name(el) %as% {
  !is.null(.ft_get_name(el))
}

# attach/remove internal element name payload
# Runtime: O(1).
if(FALSE) .ft_set_name <- function(el, name) NULL
.ft_set_name(el, name) %::% . : . : .
.ft_set_name(el, name) %as% {
  nm <- .ft_normalize_name(name)
  if(is.null(nm)) {
    attr(el, "ft_name") <- NULL
    return(el)
  }
  if(is.null(el)) {
    stop("Cannot attach a name to NULL element.")
  }
  attr(el, "ft_name") <- nm
  el
}

# validate and normalize monoid set; `.size` and `.named_count` are always present.
# Runtime: O(1) under fixed monoid set.
if(FALSE) ensure_size_monoids <- function(monoids) NULL
ensure_size_monoids(monoids) %::% list : list
ensure_size_monoids(monoids) %as% {
  out <- monoids
  if(!is_measure_monoid_list(out)) {
    stop("`monoids` must be a non-empty named list of MeasureMonoid objects.")
  }
  if(is.null(names(out))) {
    stop("Monoid lists must be named.")
  }
  if(any(is.na(names(out))) || any(names(out) == "")) {
    stop("Monoid list names must be non-empty and non-missing.")
  }
  if(any(duplicated(names(out)))) {
    stop("Monoid list names must be unique.")
  }

  for(nm in names(out)) {
    out[[nm]] <- .normalize_measure_monoid_spec(out[[nm]], monoid_name = nm)
  }

  # reserved monoids are canonicalized so all trees share the same invariants.
  out[[".size"]] <- size_measure_monoid()
  out[[".named_count"]] <- named_count_measure_monoid()
  out
}

# combine a list of measure values with the monoid's associative function
# Runtime: O(k), where k = length(measures).
if(FALSE) combine_measures <- function(measures, r) NULL
combine_measures(measures, r) %::% list : MeasureMonoid : .
combine_measures(measures, r) %as% {
  acc <- r$i
  for(m in measures) {
    acc <- r$f(acc, m)
  }
  acc
}

# Hot-path recursive measure helper used by measure recomputation.
# Avoids repeated typed-wrapper dispatch in deep recursion.
# Runtime: best-case O(1) on cached structural children; worst-case O(n_subtree).
.measure_child_named_fast <- function(x, name, rr) {
  if(is_structural_node(x)) {
    cached <- attr(x, "measures", exact = TRUE)
    if(!is.null(cached) && !is.null(cached[[name]])) {
      return(cached[[name]])
    }

    if(inherits(x, "Empty")) {
      return(rr$i)
    }

    if(inherits(x, "Single")) {
      return(.measure_child_named_fast(.subset2(x, 1), name, rr))
    }

    if(inherits(x, "Deep")) {
      acc <- rr$i
      acc <- rr$f(acc, .measure_child_named_fast(.subset2(x, "prefix"), name, rr))
      acc <- rr$f(acc, .measure_child_named_fast(.subset2(x, "middle"), name, rr))
      acc <- rr$f(acc, .measure_child_named_fast(.subset2(x, "suffix"), name, rr))
      return(acc)
    }

    # Digit / Node2 / Node3
    acc <- rr$i
    for(el in x) {
      acc <- rr$f(acc, .measure_child_named_fast(el, name, rr))
    }
    return(acc)
  }

  rr$measure(x)
}

# recursive helper assuming `ms` has already been validated
# Runtime: best-case O(1) when cached measure is present on structural children;
# worst-case O(n_subtree) when recursive fallback is required.
if(FALSE) measure_child_named_impl <- function(x, ms, name, rr) NULL
measure_child_named_impl(x, ms, name, rr) %::% . : list : character : MeasureMonoid : .
measure_child_named_impl(x, ms, name, rr) %as% {
  .measure_child_named_fast(x, name, rr)
}

# compute all cached measures for a structural node across all monoids
# Runtime: typical O(m * k), where m = number of monoids and k = immediate child
# count of `x` (bounded for in-tree nodes); worst-case O(m * n_subtree) when
# recursive fallback is required.
if(FALSE) measure_children <- function(x, monoids) NULL
measure_children(x, monoids) %::% . : list : list
measure_children(x, monoids) %as% {
  ms <- monoids
  nms <- names(ms)
  out <- vector("list", length(ms))
  for(i in seq_along(nms)) {
    out[[i]] <- .measure_child_named_fast(x, nms[[i]], ms[[i]])
  }
  names(out) <- nms
  out
}

# read a structural-node measure by monoid name
# Runtime: O(1).
if(FALSE) node_measure <- function(x, monoid_name) NULL
node_measure(x, monoid_name) %::% . : character : .
node_measure(x, monoid_name) %as% {
  if(!is_structural_node(x)) {
    stop("node_measure expects a structural node.")
  }
  ms <- attr(x, "measures", exact = TRUE)
  if(is.null(ms) || is.null(ms[[monoid_name]])) {
    stop(paste0("Missing cached measure '", monoid_name, "' on node."))
  }
  ms[[monoid_name]]
}

# attach canonical monoids + cached measures to a structural node
# Runtime: same as `measure_children()`; typical O(m * k), worst-case
# O(m * n_subtree) under recursive fallback.
if(FALSE) set_measure <- function(x, monoids) NULL
set_measure(x, monoids) %::% . : list : .
set_measure(x, monoids) %as% {
  if(!is_structural_node(x)) {
    return(x)
  }
  ms <- monoids
  attr(x, "monoids") <- ms
  attr(x, "measures") <- measure_children(x, ms)
  x
}

# attach monoids + measures, reusing unchanged cached entries from `previous`.
# only `recompute_names` are recomputed from children.
# Runtime: O(m) to iterate monoid names plus recompute cost for monoids in
# `recompute_names`; recompute is typical O(k) per monoid on measured children
# and worst-case O(n_subtree) per monoid under recursive fallback.
if(FALSE) set_measure_with_reuse <- function(x, previous, monoids, recompute_names) NULL
set_measure_with_reuse(x, previous, monoids, recompute_names) %::% . : . : list : character : .
set_measure_with_reuse(x, previous, monoids, recompute_names) %as% {
  if(!is_structural_node(x)) {
    return(x)
  }

  ms <- monoids
  nms <- names(ms)
  rec <- unique(intersect(recompute_names, nms))
  old <- if(is_structural_node(previous)) attr(previous, "measures", exact = TRUE) else NULL
  has_old <- !is.null(old)
  rec_flags <- rep(FALSE, length(nms))
  names(rec_flags) <- nms
  if(length(rec) > 0L) {
    rec_flags[rec] <- TRUE
  }

  out <- vector("list", length(ms))
  names(out) <- nms
  for(i in seq_along(nms)) {
    nm <- nms[[i]]
    can_reuse <- !rec_flags[[nm]] && has_old && !is.null(old[[nm]])
    if(can_reuse) {
      out[[i]] <- old[[nm]]
    } else {
      out[[i]] <- .measure_child_named_fast(x, nm, ms[[i]])
    }
  }

  attr(x, "monoids") <- ms
  attr(x, "measures") <- out
  x
}

# only used by add_monoids() to add new monoids to the full tree
# persistent structural copy that updates monoid attrs.
# tree shape is preserved; only requested monoid caches are recomputed.
# Runtime: O(n) in subtree size.
if(FALSE) rebind_tree_monoids <- function(x, monoids, recompute_names) NULL
rebind_tree_monoids(x, monoids, recompute_names) %::% . : list : character : .
rebind_tree_monoids(x, monoids, recompute_names) %as% {
  if(!is_structural_node(x)) {
    return(x)
  }

  if(inherits(x, "Empty")) {
    return(set_measure_with_reuse(Empty(), x, monoids, recompute_names))
  }

  if(inherits(x, "Single")) {
    child <- rebind_tree_monoids(.subset2(x, 1), monoids, recompute_names)
    return(set_measure_with_reuse(Single(child), x, monoids, recompute_names))
  }

  if(inherits(x, "Deep")) {
    pr <- rebind_tree_monoids(.subset2(x,"prefix"), monoids, recompute_names)
    m <- rebind_tree_monoids(.subset2(x,"middle"), monoids, recompute_names)
    sf <- rebind_tree_monoids(.subset2(x,"suffix"), monoids, recompute_names)
    return(set_measure_with_reuse(Deep(pr, m, sf), x, monoids, recompute_names))
  }

  # Digit / Node2 / Node3
  out <- lapply(x, rebind_tree_monoids, monoids = monoids, recompute_names = recompute_names)
  class(out) <- class(x)
  set_measure_with_reuse(out, x, monoids, recompute_names)
}

assert_has_monoids(node) %::% . : .
# Runtime: O(1).
if(FALSE) assert_has_monoids <- function(node) NULL
assert_has_monoids(node) %as% {
  if(!is_structural_node(node)) {
    return(invisible(TRUE))
  }
  ms <- attr(node, "monoids", exact = TRUE)
  if(is.null(ms)) {
    stop("Structural node missing `monoids` attribute.")
  }
  if(!is.list(ms) || is.null(names(ms))) {
    stop("Structural node `monoids` must be a named list.")
  }
  if(!all(c(".size", ".named_count") %in% names(ms))) {
    stop("Structural node `monoids` must include reserved monoids `.size` and `.named_count`.")
  }
  invisible(TRUE)
}

assert_measures_match_monoids(node) %::% . : .
# Runtime: O(1) under fixed monoid set.
if(FALSE) assert_measures_match_monoids <- function(node) NULL
assert_measures_match_monoids(node) %as% {
  if(!is_structural_node(node)) {
    return(invisible(TRUE))
  }
  ms <- attr(node, "monoids", exact = TRUE)
  if(is.null(ms)) {
    stop("Structural node missing `monoids` attribute.")
  }
  cached <- attr(node, "measures", exact = TRUE)
  if(is.null(cached)) {
    stop("Structural node missing `measures` attribute.")
  }
  if(is.null(names(cached)) || !setequal(names(cached), names(ms))) {
    stop("Structural node `measures` names must match `monoids` names.")
  }
  invisible(TRUE)
}

assert_structural_attrs(node) %::% . : .
# Runtime: O(n) in subtree size (recursive full walk).
# Intended use: correctness auditing in tests/debugging, not hot paths.
if(FALSE) assert_structural_attrs <- function(node) NULL
assert_structural_attrs(node) %as% {
  if(!is_structural_node(node)) {
    return(invisible(TRUE))
  }

  assert_has_monoids(node)
  assert_measures_match_monoids(node)

  if(node %isa% Single) {
    assert_structural_attrs(.subset2(node, 1))
  } else if(node %isa% Deep) {
    assert_structural_attrs(.subset2(node,"prefix"))
    assert_structural_attrs(.subset2(node,"middle"))
    assert_structural_attrs(.subset2(node,"suffix"))
  } else {
    for(el in node) {
      assert_structural_attrs(el)
    }
  }

  invisible(TRUE)
}

# construct an Empty with cached measures
# Runtime: O(1).
if(FALSE) measured_empty <- function(monoids) NULL
measured_empty(monoids) %::% list : Empty
measured_empty(monoids) %as% {
  e <- Empty()
  set_measure(e, monoids)
}

# construct a Single with cached measures
# Runtime: O(1).
if(FALSE) measured_single <- function(el, monoids) NULL
measured_single(el, monoids) %::% . : list : Single
measured_single(el, monoids) %as% {
  s <- Single(el)
  set_measure(s, monoids)
}

# construct a Digit with cached measures (size 1..4)
# Runtime: O(1).
if(FALSE) measured_digit <- function(a, monoids) NULL
measured_digit(a, monoids) %::% . : list : Digit
measured_digit(a, monoids) %as% {
  d <- Digit(a)
  set_measure(d, monoids)
}

# Runtime: O(1).
measured_digit(a, b, monoids) %::% . : . : list : Digit
measured_digit(a, b, monoids) %as% {
  d <- Digit(a, b)
  set_measure(d, monoids)
}

# Runtime: O(1).
measured_digit(a, b, c, monoids) %::% . : . : . : list : Digit
measured_digit(a, b, c, monoids) %as% {
  d <- Digit(a, b, c)
  set_measure(d, monoids)
}

# Runtime: O(1).
measured_digit(a, b, c, d1, monoids) %::% . : . : . : . : list : Digit
measured_digit(a, b, c, d1, monoids) %as% {
  d <- Digit(a, b, c, d1)
  set_measure(d, monoids)
}

# construct a Node2 with cached measures
# Runtime: O(1).
if(FALSE) measured_node2 <- function(x, y, monoids) NULL
measured_node2(x, y, monoids) %::% . : . : list : Node
measured_node2(x, y, monoids) %as% {
  n <- Node2(x, y)
  set_measure(n, monoids)
}

# construct a Node3 with cached measures
# Runtime: O(1).
if(FALSE) measured_node3 <- function(x, y, z, monoids) NULL
measured_node3(x, y, z, monoids) %::% . : . : . : list : Node
measured_node3(x, y, z, monoids) %as% {
  n <- Node3(x, y, z)
  set_measure(n, monoids)
}

# construct a Deep with cached measures
# Runtime: O(1) with measured children.
if(FALSE) measured_deep <- function(prefix, middle, suffix, monoids) NULL
measured_deep(prefix, middle, suffix, monoids) %::% Digit : FingerTree : Digit : list : Deep
measured_deep(prefix, middle, suffix, monoids) %as% {
  t <- Deep(prefix, middle, suffix)
  set_measure(t, monoids)
}

#SO

# helpers for converting between digits/nodes/trees and for deconstructing trees

# attach canonical monoid set to a constructed subtree
# Runtime: O(1).
if(FALSE) with_tree_monoids <- function(t, monoids) NULL
with_tree_monoids(t, monoids) %::% FingerTree : list : FingerTree
with_tree_monoids(t, monoids) %as% {
  attr(t, "monoids") <- monoids
  t
}

# construct a Digit from a list of elements.
# input: xs list of 1..4 elements (raw elements or nodes), measure monoid r.
# output: measured Digit(xs...) with preserved order.
# note: empty input returns an empty list() sentinel, used by deepL/deepR rebuild logic.
# Runtime: O(k), where k = length(xs) and k <= 4 for in-tree digits.
if(FALSE) build_digit <- function(xs, monoids) NULL
build_digit(xs, monoids) %::% list : list : .
build_digit(xs, monoids) %as% {
  if(length(xs) == 0) {
    return(list())
  }
  do.call(measured_digit, c(unname(xs), list(monoids = monoids)))
}

# build a Deep node from prefix digit, middle tree, and suffix digit.
# input: pr (Digit), m (FingerTree), sf (Digit), measure monoid r.
# output: measured Deep(pr, m, sf).
# Runtime: O(1) with measured children.
if(FALSE) build_deep <- function(pr, m, sf, monoids) NULL
build_deep(pr, m, sf, monoids) %::% Digit : FingerTree : Digit : list : Deep
build_deep(pr, m, sf, monoids) %as% with_tree_monoids(measured_deep(pr, m, sf, monoids), monoids)

# convert a small list/digit (size 0..4) into a valid measured FingerTree shape.
# Runtime: O(1), since digit size is bounded (0..4).
if(FALSE) digit_to_tree <- function(d, monoids) NULL
digit_to_tree(d, monoids) %::% list : list : FingerTree
digit_to_tree(d, monoids) %as% {
  n <- length(d)
  if(n == 0) { return(with_tree_monoids(measured_empty(monoids), monoids)) }
  if(n == 1) { return(with_tree_monoids(measured_single(d[[1]], monoids), monoids)) }
  if(n == 2) {
    pr <- build_digit(list(d[[1]]), monoids)
    sf <- build_digit(list(d[[2]]), monoids)
    return(build_deep(pr, measured_empty(monoids), sf, monoids))
  }
  if(n == 3) {
    pr <- build_digit(list(d[[1]], d[[2]]), monoids)
    sf <- build_digit(list(d[[3]]), monoids)
    return(build_deep(pr, measured_empty(monoids), sf, monoids))
  }
  if(n == 4) {
    pr <- build_digit(list(d[[1]], d[[2]]), monoids)
    sf <- build_digit(list(d[[3]], d[[4]]), monoids)
    return(build_deep(pr, measured_empty(monoids), sf, monoids))
  }
  stop("digit_to_tree expects a digit of size 0..4")
}

# convert a Node2/Node3 into a measured Digit of its children.
# Runtime: O(1), since Node arity is bounded (2..3).
if(FALSE) node_to_digit <- function(node, monoids) NULL
node_to_digit(node, monoids) %::% Node : list : Digit
node_to_digit(node, monoids) %as% build_digit(as.list(node), monoids)

# viewL: return leftmost element and the remaining tree
# input: t non-empty FingerTree, measure monoid r.
# output: list(value = leftmost element of t, rest = t without that element).
# Runtime: O(log n) worst-case.
if(FALSE) viewL <- function(t, monoids) NULL
viewL(t, monoids) %::% FingerTree : list : list
viewL(t, monoids) %as% {
  if(t %isa% Empty) {
    stop("viewL on Empty")
  }
  if(t %isa% Single) {
    return(list(value = .subset2(t, 1), rest = .as_flexseq(with_tree_monoids(measured_empty(monoids), monoids))))
  }
  pr <- .subset2(t,"prefix")
  if(length(pr) > 1) {
    # Fast path: prefix still has elements after popping its head, so no middle
    # rebalancing is needed.
    head <- pr[[1]]
    tail <- pr[2:length(pr)]
    new_pr <- build_digit(tail, monoids)
    return(list(value = head, rest = .as_flexseq(build_deep(new_pr, .subset2(t,"middle"), .subset2(t,"suffix"), monoids))))
  }
  head <- pr[[1]]
  m <- .subset2(t,"middle")
  if(m %isa% Empty) {
    # Prefix had exactly one element and middle is empty: remaining content is
    # entirely in suffix, so collapse to a minimal tree from that digit.
    return(list(value = head, rest = .as_flexseq(digit_to_tree(.subset2(t,"suffix"), monoids))))
  }
  # Prefix had one element but middle is non-empty: pull the leftmost node from
  # middle, expand it to a digit, and use that as the new prefix.
  res <- viewL(m, monoids)
  node <- res$value
  m_rest <- res$rest
  new_pr <- node_to_digit(node, monoids)
  list(value = head, rest = .as_flexseq(build_deep(new_pr, m_rest, .subset2(t,"suffix"), monoids)))
}

# viewR: return rightmost element and the remaining tree
# input: t non-empty FingerTree, measure monoid r.
# output: list(value = rightmost element of t, rest = t without that element).
# Runtime: O(log n) worst-case.
if(FALSE) viewR <- function(t, monoids) NULL
viewR(t, monoids) %::% FingerTree : list : list
viewR(t, monoids) %as% {
  if(t %isa% Empty) {
    stop("viewR on Empty")
  }
  if(t %isa% Single) {
    return(list(value = .subset2(t, 1), rest = .as_flexseq(with_tree_monoids(measured_empty(monoids), monoids))))
  }
  sf <- .subset2(t,"suffix")
  if(length(sf) > 1) {
    head <- sf[[length(sf)]]
    tail <- sf[1:(length(sf) - 1)]
    new_sf <- build_digit(tail, monoids)
    return(list(value = head, rest = .as_flexseq(build_deep(.subset2(t,"prefix"), .subset2(t,"middle"), new_sf, monoids))))
  }
  head <- sf[[1]]
  m <- .subset2(t,"middle")
  if(m %isa% Empty) {
    return(list(value = head, rest = .as_flexseq(digit_to_tree(.subset2(t,"prefix"), monoids))))
  }
  res <- viewR(m, monoids)
  node <- res$value
  m_rest <- res$rest
  new_sf <- node_to_digit(node, monoids)
  list(value = head, rest = .as_flexseq(build_deep(.subset2(t,"prefix"), m_rest, new_sf, monoids)))
}

# deepL: rebuild Deep, possibly pulling from middle if prefix is empty
# input: pr digit/list for prefix, m middle FingerTree of nodes, sf suffix digit/list.
# output: a valid FingerTree preserving order; if pr is empty it borrows from m or
# collapses to a tree built from sf.
# Runtime: O(log n) worst-case.
if(FALSE) deepL <- function(pr, m, sf, monoids) NULL
deepL(pr, m, sf, monoids) %::% . : FingerTree : . : list : FingerTree
deepL(pr, m, sf, monoids) %as% {
  if(length(pr) > 0) {
    # Normal Deep reconstruction when prefix is non-empty.
    return(.as_flexseq(build_deep(pr, m, sf, monoids)))
  }
  if(m %isa% Empty) {
    # Cannot build Deep with empty prefix and empty middle; collapse to suffix.
    return(.as_flexseq(digit_to_tree(sf, monoids)))
  }
  # Restore a non-empty prefix by borrowing the leftmost node from middle.
  res <- viewL(m, monoids)
  node <- res$value
  m_rest <- res$rest
  new_pr <- node_to_digit(node, monoids)
  .as_flexseq(build_deep(new_pr, m_rest, sf, monoids))
}

# deepR: rebuild Deep, possibly pulling from middle if suffix is empty
# input: pr prefix digit/list, m middle FingerTree of nodes, sf suffix digit/list.
# output: a valid FingerTree preserving order; if sf is empty it borrows from m or
# collapses to a tree built from pr.
# Runtime: O(log n) worst-case.
if(FALSE) deepR <- function(pr, m, sf, monoids) NULL
deepR(pr, m, sf, monoids) %::% . : FingerTree : . : list : FingerTree
deepR(pr, m, sf, monoids) %as% {
  if(length(sf) > 0) {
    return(.as_flexseq(build_deep(pr, m, sf, monoids)))
  }
  if(m %isa% Empty) {
    return(.as_flexseq(digit_to_tree(pr, monoids)))
  }
  res <- viewR(m, monoids)
  node <- res$value
  m_rest <- res$rest
  new_sf <- node_to_digit(node, monoids)
  .as_flexseq(build_deep(pr, m_rest, new_sf, monoids))
}

#SO

#' @method split_around_by_predicate flexseq
#' @export
# Runtime: O(log n) near split point depth.
split_around_by_predicate.flexseq <- function(t, predicate, monoid_name, accumulator = NULL) {
  ctx <- resolve_named_monoid(t, monoid_name)
  ms <- ctx$monoids
  mr <- ctx$monoid

  if(t %isa% Empty) {
    stop("split_around_by_predicate requires a non-empty tree.")
  }

  i <- if(is.null(accumulator)) mr$i else accumulator
  out <- if(.ft_cpp_can_use(ms)) {
    .ft_cpp_split_tree(t, predicate, ms, monoid_name, i)
  } else {
    split_tree_impl_fast(predicate, i, t, ms, mr, monoid_name)
  }
  out$left <- .ft_restore_subclass(.as_flexseq(out$left), t, context = "split_around_by_predicate()")
  out$right <- .ft_restore_subclass(.as_flexseq(out$right), t, context = "split_around_by_predicate()")
  out
}

#' @method split_by_predicate flexseq
#' @export
# Runtime: O(log n) near split point depth.
split_by_predicate.flexseq <- function(x, predicate, monoid_name) {
  ctx <- resolve_named_monoid(x, monoid_name)
  ms <- ctx$monoids
  mr <- ctx$monoid

  if(x %isa% Empty) {
    left_empty <- .ft_restore_subclass(.as_flexseq(measured_empty(ms)), x, context = "split_by_predicate()")
    right_empty <- .ft_restore_subclass(.as_flexseq(measured_empty(ms)), x, context = "split_by_predicate()")
    return(list(left = left_empty, right = right_empty))
  }

  if(predicate(node_measure(x, monoid_name))) {
    s <- if(.ft_cpp_can_use(ms)) {
      .ft_cpp_split_tree(x, predicate, ms, monoid_name, mr$i)
    } else {
      split_tree_impl_fast(predicate, mr$i, x, ms, mr, monoid_name)
    }
    right <- push_front(s$right, s$value)
    left_out <- .ft_restore_subclass(.as_flexseq(s$left), x, context = "split_by_predicate()")
    right_out <- .ft_restore_subclass(.as_flexseq(right), x, context = "split_by_predicate()")
    return(list(left = left_out, right = right_out))
  }

  left_out <- .ft_restore_subclass(.as_flexseq(x), x, context = "split_by_predicate()")
  right_out <- .ft_restore_subclass(.as_flexseq(measured_empty(ms)), x, context = "split_by_predicate()")
  list(left = left_out, right = right_out)
}

#' @method split_at flexseq
#' @export
# Runtime: O(log n) for scalar index splits; O(n) for scalar name lookup + split.
split_at.flexseq <- function(x, at, pull_index = FALSE) {
  if(is.logical(at)) {
    stop("`at` must be a scalar index or name.")
  }

  idx <- NULL
  if(is.numeric(at) && length(at) == 1L && !is.na(at) && at == as.integer(at)) {
    n <- as.integer(node_measure(x, ".size"))
    idx <- .ft_assert_int_indices(as.integer(at), n)
    if(length(idx) != 1L) {
      stop("`at` must be a single valid index.")
    }
    idx <- as.integer(idx[[1L]])
  } else if(is.character(at) && length(at) == 1L && !is.na(at) && at != "") {
    idx <- .ft_match_name_indices(x, at, strict_missing = TRUE)
    idx <- as.integer(idx[[1L]])
  } else {
    stop("`at` must be a positive scalar integer index or a scalar non-empty name.")
  }

  predicate <- function(v) v >= idx
  if(isTRUE(pull_index)) {
    return(split_by_predicate(x, predicate, ".size"))
  }
  split_around_by_predicate(x, predicate, ".size")
}

#SO

#' Concatenate Two Structural Trees
#'
#' Same-name monoids are assumed equivalent; left-tree definitions win.
#' Missing monoids are added to each side before concatenation.
#'
#' @param x A `flexseq` (left side).
#' @param y A `flexseq` (right side).
#' @return Concatenated tree.
#' @keywords internal
# Runtime: O(nx + ny) when monoid harmonization requires add_monoids passes;
# otherwise concat spine work is near-logarithmic in boundary distance for the
# smaller, O(log(min(nx, ny)))
concat_trees <- function(x, y) {
  mx <- resolve_tree_monoids(x, required = TRUE)
  my <- resolve_tree_monoids(y, required = TRUE)

  shared <- intersect(names(mx), names(my))
  shared <- setdiff(shared, c(".size", ".named_count"))

  left_only <- setdiff(names(mx), names(my))
  left_only <- setdiff(left_only, c(".size", ".named_count"))
  right_only <- setdiff(names(my), names(mx))
  right_only <- setdiff(right_only, c(".size", ".named_count"))

  x2 <- if(length(right_only) > 0) add_monoids(x, my[right_only], overwrite = FALSE) else x
  y2 <- if(length(left_only) > 0) add_monoids(y, mx[left_only], overwrite = FALSE) else y

  merged <- c(mx, my[setdiff(names(my), names(mx))])

  if(.ft_cpp_can_use(merged)) {
    return(.as_flexseq(.ft_cpp_concat(x2, y2, merged)))
  }

  .as_flexseq(concat(x2, y2, merged))
}

# Internal: concat two trees that already share identical monoids.
# Skips monoid harmonization. Uses C++ fast path when available.
# Runtime: O(log(min(n1, n2))) via C++; R fallback is slower.
.ft_concat_same_monoids <- function(x, y, monoids) {
  if(.ft_cpp_can_use(monoids)) {
    return(.ft_cpp_concat(x, y, monoids))
  }
  concat(x, y, monoids)
}

#SO

#' Build a Structural Tree from a Vector or List
#'
#' @param x Elements to insert.
#' @param monoids Optional named list of `measure_monoid` objects.
#' @return A finger tree with cached measures for all monoids.
#'   If `x` has names, they are used for name-based indexing and must be
#'   complete (no missing/empty names) and unique.
#' @keywords internal
# Runtime: O(n) for validation + linear bulk build.
tree_from <- function(x, monoids = NULL) {
  # Phase 1: normalize monoid set and decide whether C++ build path is available.
  ms <- if(is.null(monoids)) ensure_size_monoids(list(.size = size_measure_monoid())) else ensure_size_monoids(monoids)
  can_cpp <- .ft_cpp_can_use(ms)

  # Phase 2: normalize container shape and detect caller-supplied outer names.
  x_list <- as.list(x)
  n <- length(x_list)

  resolved_names <- NULL
  in_names <- names(x)
  use_names <- !is.null(in_names) && length(in_names) > 0L
  if(use_names && length(in_names) != n) {
    stop("Input names length must match input length.")
  }

  # Phase 3a: explicit outer names path. Enforce complete/unique naming.
  if(use_names) {
    norm_names <- as.character(in_names)
    missing_name <- is.na(norm_names) | norm_names == ""
    has_any <- any(!missing_name)
    if(has_any) {
      if(any(missing_name)) {
        stop("Mixed named and unnamed elements are not allowed.")
      }
      if(anyDuplicated(norm_names) > 0L) {
        stop("Element names must be unique.")
      }
      resolved_names <- norm_names
    }
  } else if(n > 0L) {
    # Phase 3b: no outer names. Optionally derive names from element payload attrs.
    if(can_cpp) {
      # Common constructor path: if elements have no attrs there are no inline
      # names to preserve, so we can bulk-build directly in C++.
      if(!any(vapply(x_list, function(el) !is.null(attributes(el)), logical(1)))) {
        return(.as_flexseq(.ft_cpp_tree_from(x_list, ms)))
      }
    }

    # Preserve existing behavior: when outer names are absent, derive names
    # from each element payload (ft_name attr or inline scalar names()).
    # Fast local path avoids lambda dispatch in hot construction workloads.
    derived <- vector("list", n)
    has_any <- FALSE
    has_missing <- FALSE
    for(i in seq_len(n)) {
      el <- x_list[[i]]
      nm <- attr(el, "ft_name", exact = TRUE)
      if(!is.null(nm) && length(nm) > 0L) {
        if(length(nm) != 1L) {
          stop("Element names must be scalar.")
        }
        nm <- as.character(nm[[1L]])
        if(is.na(nm) || nm == "") {
          nm <- NULL
        }
      } else {
        nms <- names(el)
        if(is.null(nms) || length(nms) != 1L) {
          nm <- NULL
        } else {
          nm <- nms[[1L]]
          if(is.na(nm) || nm == "") {
            nm <- NULL
          } else {
            nm <- as.character(nm)
          }
        }
      }
      derived[[i]] <- nm
      if(is.null(nm)) {
        has_missing <- TRUE
      } else {
        has_any <- TRUE
      }
    }
    if(has_any) {
      if(has_missing) {
        stop("Mixed named and unnamed elements are not allowed.")
      }
      resolved_names <- unlist(derived, use.names = FALSE)
      if(anyDuplicated(resolved_names) > 0L) {
        stop("Element names must be unique.")
      }
    }
  }

  # Phase 4: dispatch to backend constructor (C++ fast path vs R reference path).
  if(can_cpp) {
    if(is.null(resolved_names)) {
      return(.as_flexseq(.ft_cpp_tree_from(x_list, ms)))
    }
    return(.as_flexseq(.ft_cpp_tree_from_prepared(
      x_list,
      resolved_names,
      ms
    )))
  }

  # R path needs names materialized on each element before structural build.
  if(!is.null(resolved_names)) {
    for(i in seq_along(x_list)) {
      el <- x_list[[i]]
      el <- .ft_set_name(el, resolved_names[[i]])
      # Preserve NULL elements; `[[<- NULL` would drop positions from the list.
      x_list[i] <- list(el)
    }
  }

  .ft_tree_from_list_linear(x_list, ms)
}

# Runtime: O(n) in total elements across recursive levels.
.ft_tree_from_ordered_ref <- function(xs, monoids) {
  n <- length(xs)

  # Small-n base shapes mirror fingertree constructors directly.
  if(n == 0L) {
    return(measured_empty(monoids))
  }
  if(n == 1L) {
    return(measured_single(xs[[1L]], monoids))
  }
  if(n == 2L) {
    return(measured_deep(
      build_digit(list(xs[[1L]]), monoids),
      measured_empty(monoids),
      build_digit(list(xs[[2L]]), monoids),
      monoids
    ))
  }
  if(n == 3L) {
    return(measured_deep(
      build_digit(xs[1:2], monoids),
      measured_empty(monoids),
      build_digit(list(xs[[3L]]), monoids),
      monoids
    ))
  }
  if(n == 4L) {
    return(measured_deep(
      build_digit(xs[1:2], monoids),
      measured_empty(monoids),
      build_digit(xs[3:4], monoids),
      monoids
    ))
  }

  # Reference shape rule: avoid a 1-element middle segment since middle trees
  # are built from Node2/Node3 blocks only.
  prefix_len <- if(n == 5L) 1L else 2L
  suffix_len <- 2L

  prefix <- build_digit(xs[seq_len(prefix_len)], monoids)
  suffix <- build_digit(xs[(n - suffix_len + 1L):n], monoids)
  middle_elems <- xs[(prefix_len + 1L):(n - suffix_len)]
  middle_nodes <- measured_nodes(middle_elems, monoids)
  middle <- .ft_tree_from_ordered_ref(middle_nodes, monoids)

  measured_deep(prefix, middle, suffix, monoids)
}

# Runtime: O(n).
.ft_tree_from_list_linear <- function(x_list, monoids) {
  .as_flexseq(.ft_tree_from_ordered_ref(x_list, monoids))
}

#SO

# Runtime: O(1).
# Formats bounds string into human-readable phrase for headers.
# **Inputs:** scalar `bounds` string.
# **Outputs:** scalar character phrase.
# **Used by:** print.interval_index().
.ivx_bounds_phrase <- function(bounds) {
  paste0(substr(bounds, 1L, 1L), "start, end", substr(bounds, 2L, 2L))
}

#' Print an Interval Index Summary
#'
#' Prints a compact summary with interval bounds and a head/tail preview of
#' payload elements.
#'
#' @param x An `interval_index`.
#' @param max_elements Maximum number of elements shown in the preview.
#' @param show_custom_monoids Logical; show attached non-default monoids and
#'   their root cached measures.
#' @param ... Passed through to per-element `print()`.
#' @return Invisibly returns `x`.
#' @examples
#' ix <- interval_index(
#'   one = 1, two = 2, three = 3,
#'   start = c(20, 30, 10), end = c(25, 37, 24)
#' )
#' print(ix, max_elements = 4)
#' width_sum <- measure_monoid(
#'   `+`, 0, function(entry) as.numeric(entry$end - entry$start)
#' )
#' ix3 <- add_monoids(
#'   interval_index(1, 2, start = c(1, 3), end = c(2, 5)),
#'   list(width_sum = width_sum)
#' )
#' print(ix3, max_elements = 0, show_custom_monoids = TRUE)
#'
#' ix2 <- interval_index(1, 2, 3, start = c(2, 4, 6), end = c(3, 5, 8), default_query_bounds = "[]")
#' print(ix2, max_elements = 3)
#'
#' print(interval_index())
#' @export
#' @method print interval_index
# Runtime: O((k + h) log n), where k = shown elements and h = preview split overhead.
# Pretty-printer with bounded head/tail preview in interval-start order.
# **Inputs:** `x` interval_index; scalar integer `max_elements`; scalar logical `show_custom_monoids`; forwarded `...`.
# **Outputs:** invisibly returns `x`.
# **Used by:** users/tests.
print.interval_index <- function(x, max_elements = 4L, show_custom_monoids = FALSE, ...) {
  .ivx_assert_index(x)
  max_elements <- .ft_validate_print_max_elements(max_elements)
  show_custom <- .ft_validate_show_custom_monoids(show_custom_monoids)

  n <- length(x)
  bounds <- .ivx_resolve_bounds(x, NULL)
  nn <- as.integer(node_measure(x, ".named_count"))
  named <- isTRUE(nn > 0L)

  cat(
    if(named) "Named" else "Unnamed",
    " interval_index with ",
    n,
    " element",
    if(n == 1L) "" else "s",
    ", default query bounds ",
    .ivx_bounds_phrase(bounds),
    ".\n",
    sep = ""
  )
  if(show_custom) {
    .ft_print_custom_monoids(
      x,
      excluded_names = c(".size", ".named_count", ".ivx_max_start", ".ivx_max_end", ".ivx_min_end", ".oms_max_key")
    )
  }

  if(n == 0L || max_elements == 0L) {
    return(invisible(x))
  }

  cat("\nElements (by interval start order):\n\n")
  preview <- .pick_preview_sizes(n, max_elements)
  excluded_names <- c(".size", ".named_count", ".ivx_max_start", ".ivx_max_end", ".ivx_min_end", ".oms_max_key")
  for(i in preview$head) {
    entry <- .ft_get_elem_at(x, as.integer(i))
    nm <- .ft_get_name(entry)
    iv <- paste0(.ft_format_scalar(entry$start), " - ", .ft_format_scalar(entry$end))
    if(named && !is.null(nm)) {
      cat("$", nm, " (interval ", iv, ")\n", sep = "")
    } else {
      cat("[[", i, "]] (interval ", iv, ")\n", sep = "")
    }
    print(entry$value, ...)
    .ft_print_elem_custom_monoids(x, entry, excluded_names, show_custom)
    cat("\n")
  }

  .ft_print_skipped(preview$skipped)

  for(i in preview$tail) {
    entry <- .ft_get_elem_at(x, as.integer(i))
    nm <- .ft_get_name(entry)
    iv <- paste0(.ft_format_scalar(entry$start), " - ", .ft_format_scalar(entry$end))
    if(named && !is.null(nm)) {
      cat("$", nm, " (interval ", iv, ")\n", sep = "")
    } else {
      cat("[[", i, "]] (interval ", iv, ")\n", sep = "")
    }
    print(entry$value, ...)
    .ft_print_elem_custom_monoids(x, entry, excluded_names, show_custom)
    cat("\n")
  }
  invisible(x)
}

#SO

# Runtime: O(log n).
#' Minimum Left Endpoint
#'
#' Returns the smallest left endpoint (`start`) currently present.
#'
#' @param x An `interval_index`.
#' @return Minimum left endpoint, or `NULL` when `x` is empty.
#' @details
#' Because intervals are stored in start order, this reads the first entry's
#' `start`.
#' @examples
#' ix <- interval_index("a", "b", start = c(3, 1), end = c(4, 2))
#' min_endpoint(ix)
#' min_endpoint(interval_index())
#' @export
min_endpoint <- function(x) {
  .ivx_assert_index(x)
  if(length(x) == 0L) {
    return(NULL)
  }
  .ft_get_elem_at(x, 1L)$start
}

# Runtime: O(1).
#' Maximum Right Endpoint
#'
#' Returns the largest right endpoint (`end`) currently present.
#'
#' @param x An `interval_index`.
#' @return Maximum right endpoint, or `NULL` when `x` is empty.
#' @details
#' Uses cached `.ivx_max_end` monoid state.
#' @examples
#' ix <- interval_index("a", "b", start = c(3, 1), end = c(4, 2))
#' max_endpoint(ix)
#' max_endpoint(interval_index())
#' @export
max_endpoint <- function(x) {
  .ivx_assert_index(x)
  m <- node_measure(x, ".ivx_max_end")
  if(!isTRUE(m$has)) {
    return(NULL)
  }
  m$end
}

# Runtime: O(log n + c), where c is candidate count (worst-case O(n)).
#' Peek First Interval Matching a Point
#'
#' @param x An `interval_index`.
#' @param point Query point.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#'   Ignored when `match_at` is not `"interval"`.
#' @param match_at How the query point is matched against each entry. One of
#'   `"interval"` (default; entry interval contains `point`, under `bounds`),
#'   `"start"` (entry's start coordinate equals `point`), `"end"` (entry's end
#'   equals `point`), or `"either"` (start or end equals `point`). The three
#'   coordinate-equality modes are structural and ignore `bounds`.
#' @return The payload value from the first match, or `NULL` on no match.
#' @details
#' Returns the first match in canonical interval order. Use [peek_all_point()] to
#' retrieve all matches as an `interval_index` slice.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(3, 2, 5))
#' peek_point(ix, 2)
#'
#' # Boundary override at an endpoint
#' edge <- interval_index("a", start = 1, end = 3, default_query_bounds = "[)")
#' peek_point(edge, 3)                # default "[)": no match at right endpoint
#' peek_point(edge, 3, bounds = "[]") # closed bounds: endpoint matches
#'
#' # Coordinate-equality modes (bounds irrelevant)
#' peek_point(ix, 2, match_at = "start")  # entry starting at 2
#' peek_point(ix, 3, match_at = "end")    # entry ending at 3
#' @export
peek_point <- function(x, point, bounds = NULL,
                       match_at = c("interval", "start", "end", "either")) {
  .ivx_assert_index(x)
  match_at <- match.arg(match_at)
  qp <- .ivx_normalize_endpoint(point, "point", endpoint_type = .ivx_endpoint_type_state(x))
  if(identical(match_at, "interval")) {
    b <- .ivx_resolve_bounds(x, bounds)
    spec <- .ivx_spec_point(qp, b, .ivx_bounds_flags(b), match_at)
  } else {
    spec <- .ivx_spec_point(qp, NULL, NULL, match_at)
  }
  .ivx_run_relation_query(x, spec, mode = "peek", which = "first")
}

# Runtime: O(log n + c + k), where k is matched count (worst-case O(n)).
#' Peek All Intervals Matching a Point
#'
#' @param x An `interval_index`.
#' @param point Query point.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#'   Ignored when `match_at` is not `"interval"`.
#' @param match_at How the query point is matched against each entry. One of
#'   `"interval"` (default; containment under `bounds`), `"start"`, `"end"`,
#'   or `"either"`. See [peek_point()] for details.
#' @return An `interval_index` slice of all matches (possibly empty).
#' @details
#' The returned `interval_index` can be inspected with [as.list()].
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(3, 2, 5))
#' as.list(peek_all_point(ix, 2))
#'
#' # Entries ending at 3
#' as.list(peek_all_point(ix, 3, match_at = "end"))
#' @export
peek_all_point <- function(x, point, bounds = NULL,
                           match_at = c("interval", "start", "end", "either")) {
  .ivx_assert_index(x)
  match_at <- match.arg(match_at)
  qp <- .ivx_normalize_endpoint(point, "point", endpoint_type = .ivx_endpoint_type_state(x))
  if(identical(match_at, "interval")) {
    b <- .ivx_resolve_bounds(x, bounds)
    spec <- .ivx_spec_point(qp, b, .ivx_bounds_flags(b), match_at)
  } else {
    spec <- .ivx_spec_point(qp, NULL, NULL, match_at)
  }
  .ivx_run_relation_query(x, spec, mode = "peek", which = "all")
}

# Runtime: O(log n + c).
#' Pop First Interval Matching a Point
#'
#' @param x An `interval_index`.
#' @param point Query point.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#'   Ignored when `match_at` is not `"interval"`.
#' @param match_at How the query point is matched against each entry. One of
#'   `"interval"` (default; containment under `bounds`), `"start"`, `"end"`,
#'   or `"either"`. See [peek_point()] for details.
#' @return A list with `element`, `start`, `end`, and `remaining`.
#'   On miss: `element`, `start`, and `end` are `NULL`.
#' @details
#' Removes the first match in canonical interval order. On miss, returns a
#' non-throwing miss object with `remaining = x`.
#' Use [pop_all_point()] to remove all matches.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(3, 2, 5))
#' pop_point(ix, 2)
#' @export
pop_point <- function(x, point, bounds = NULL,
                      match_at = c("interval", "start", "end", "either")) {
  .ivx_assert_index(x)
  match_at <- match.arg(match_at)
  qp <- .ivx_normalize_endpoint(point, "point", endpoint_type = .ivx_endpoint_type_state(x))
  if(identical(match_at, "interval")) {
    b <- .ivx_resolve_bounds(x, bounds)
    spec <- .ivx_spec_point(qp, b, .ivx_bounds_flags(b), match_at)
  } else {
    spec <- .ivx_spec_point(qp, NULL, NULL, match_at)
  }
  .ivx_run_relation_query(x, spec, mode = "pop", which = "first")
}

# Runtime: O(n log n) in worst-case rebuild path.
#' Pop All Intervals Matching a Point
#'
#' @param x An `interval_index`.
#' @param point Query point.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#'   Ignored when `match_at` is not `"interval"`.
#' @param match_at How the query point is matched against each entry. One of
#'   `"interval"` (default; containment under `bounds`), `"start"`, `"end"`,
#'   or `"either"`. See [peek_point()] for details. The `"end"` mode is the
#'   standard primitive for retiring active intervals in a sweep-line.
#' @return A list with `elements` and `remaining`, both `interval_index` objects.
#' @details
#' Use [as.list()] to convert `elements` to a standard R list.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(3, 2, 5))
#' out <- pop_all_point(ix, 2)
#' as.list(out$elements)
#'
#' # Sweep-line retirement: remove everything ending at x = 3
#' retired <- pop_all_point(ix, 3, match_at = "end")
#' as.list(retired$elements)
#' as.list(retired$remaining)
#' @export
pop_all_point <- function(x, point, bounds = NULL,
                          match_at = c("interval", "start", "end", "either")) {
  .ivx_assert_index(x)
  match_at <- match.arg(match_at)
  qp <- .ivx_normalize_endpoint(point, "point", endpoint_type = .ivx_endpoint_type_state(x))
  if(identical(match_at, "interval")) {
    b <- .ivx_resolve_bounds(x, bounds)
    spec <- .ivx_spec_point(qp, b, .ivx_bounds_flags(b), match_at)
  } else {
    spec <- .ivx_spec_point(qp, NULL, NULL, match_at)
  }
  .ivx_run_relation_query(x, spec, mode = "pop", which = "all")
}

# Runtime: O(log n + c).
#' Peek First Interval Overlapping a Query Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return The payload value from the first match, or `NULL` on no match.
#' @details
#' Returns the first match in canonical interval order. Use
#' [peek_all_overlaps()] to retrieve all matches as an `interval_index` slice.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 3, 5), end = c(2, 4, 6))
#' peek_overlaps(ix, 2, 3)
#'
#' # Boundary override at touching endpoints
#' edge <- interval_index("a", start = 1, end = 3, default_query_bounds = "[)")
#' peek_overlaps(edge, 3, 4)                # default "[)": no endpoint overlap
#' peek_overlaps(edge, 3, 4, bounds = "[]") # closed bounds: endpoint overlaps
#' @export
peek_overlaps <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_overlaps(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "peek", which = "first")
}

# Runtime: O(log n + c + k), where k is matched count (worst-case O(n)).
#' Peek All Intervals Overlapping a Query Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return An `interval_index` slice of all matches (possibly empty).
#' @details
#' The returned `interval_index` can be inspected with [as.list()].
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 3, 5), end = c(2, 4, 6))
#' as.list(peek_all_overlaps(ix, 2, 5))
#' @export
peek_all_overlaps <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_overlaps(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "peek", which = "all")
}

# Runtime: O(log n + c).
#' Peek First Interval Containing a Query Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return The payload value from the first match, or `NULL` on no match.
#' @details
#' Returns the first match in canonical interval order. Use
#' [peek_all_containing()] to retrieve all matches as an `interval_index` slice.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(5, 3, 6))
#' peek_containing(ix, 2, 3)
#' @export
peek_containing <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_containing(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "peek", which = "first")
}

# Runtime: O(log n + c + k), where k is matched count (worst-case O(n)).
#' Peek All Intervals Containing a Query Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return An `interval_index` slice of all matches (possibly empty).
#' @details
#' The returned `interval_index` can be inspected with [as.list()].
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 5, 7))
#' as.list(peek_all_containing(ix, 2, 4))
#' @export
peek_all_containing <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_containing(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "peek", which = "all")
}

# Runtime: O(log n + c).
#' Peek First Interval Within a Query Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return The payload value from the first match, or `NULL` on no match.
#' @details
#' Returns the first match in canonical interval order. Use [peek_all_within()]
#' to retrieve all matches as an `interval_index` slice.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 3, 5))
#' peek_within(ix, 1, 4)
#' @export
peek_within <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_within(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "peek", which = "first")
}

# Runtime: O(log n + c + k), where k is matched count (worst-case O(n)).
#' Peek All Intervals Within a Query Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return An `interval_index` slice of all matches (possibly empty).
#' @details
#' The returned `interval_index` can be inspected with [as.list()].
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 3, 5))
#' as.list(peek_all_within(ix, 1, 4))
#' @export
peek_all_within <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_within(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "peek", which = "all")
}

# Runtime: O(log n + c).
#' Pop First Overlapping Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return A list with `element`, `start`, `end`, and `remaining`.
#'   On miss: `element`, `start`, and `end` are `NULL`.
#' @details
#' Removes the first match in canonical interval order. On miss, returns a
#' non-throwing miss object with `remaining = x`.
#' Use [pop_all_overlaps()] to remove all matches.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 3, 5), end = c(2, 4, 6))
#' pop_overlaps(ix, 2, 3)
#' @export
pop_overlaps <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_overlaps(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "pop", which = "first")
}

# Runtime: O(n log n) in worst-case rebuild path.
#' Pop All Overlapping Intervals
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return A list with `elements` and `remaining`, both `interval_index` objects.
#' @details
#' Use [as.list()] to convert `elements` to a standard R list.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 3, 5), end = c(2, 4, 6))
#' out <- pop_all_overlaps(ix, 2, 5)
#' as.list(out$elements)
#' @export
pop_all_overlaps <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_overlaps(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "pop", which = "all")
}

# Runtime: O(log n + c).
#' Pop First Containing Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return A list with `element`, `start`, `end`, and `remaining`.
#'   On miss: `element`, `start`, and `end` are `NULL`.
#' @details
#' Removes the first match in canonical interval order. On miss, returns a
#' non-throwing miss object with `remaining = x`.
#' Use [pop_all_containing()] to remove all matches.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 3, 7))
#' pop_containing(ix, 2, 4)
#' @export
pop_containing <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_containing(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "pop", which = "first")
}

# Runtime: O(n log n) in worst-case rebuild path.
#' Pop All Containing Intervals
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return A list with `elements` and `remaining`, both `interval_index` objects.
#' @details
#' Use [as.list()] to convert `elements` to a standard R list.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 5, 7))
#' out <- pop_all_containing(ix, 2, 4)
#' as.list(out$elements)
#' @export
pop_all_containing <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_containing(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "pop", which = "all")
}

# Runtime: O(log n + c).
#' Pop First Interval Within a Query Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return A list with `element`, `start`, `end`, and `remaining`.
#'   On miss: `element`, `start`, and `end` are `NULL`.
#' @details
#' Removes the first match in canonical interval order. On miss, returns a
#' non-throwing miss object with `remaining = x`.
#' Use [pop_all_within()] to remove all matches.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 3, 5))
#' pop_within(ix, 1, 4)
#' @export
pop_within <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_within(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "pop", which = "first")
}

# Runtime: O(n log n) in worst-case rebuild path.
#' Pop All Intervals Within a Query Interval
#'
#' @param x An `interval_index`.
#' @param start Query interval start.
#' @param end Query interval end.
#' @param bounds Optional boundary override. One of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#' @return A list with `elements` and `remaining`, both `interval_index` objects.
#' @details
#' Use [as.list()] to convert `elements` to a standard R list.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 3, 5))
#' out <- pop_all_within(ix, 1, 4)
#' as.list(out$elements)
#' @export
pop_all_within <- function(x, start, end, bounds = NULL) {
  .ivx_assert_index(x)
  b <- .ivx_resolve_bounds(x, bounds)
  q <- .ivx_normalize_interval(start, end, endpoint_type = .ivx_endpoint_type_state(x))
  spec <- .ivx_spec_within(q, b, .ivx_bounds_flags(b))
  .ivx_run_relation_query(x, spec, mode = "pop", which = "all")
}

with_cpp_mode <- function(flag, expr) {
  old <- getOption("immutables.use_cpp")
  options(immutables.use_cpp = flag)
  on.exit({
    if(is.null(old)) {
      options(immutables.use_cpp = NULL)
    } else {
      options(immutables.use_cpp = old)
    }
  }, add = TRUE)
  force(expr)
}

backend_eval <- function(expr, flag, env = parent.frame()) {
  with_cpp_mode(flag, eval(expr, env))
}

expect_backend_identical <- function(expr, transform = identity, env = parent.frame()) {
  q <- substitute(expr)
  cpp <- transform(backend_eval(q, TRUE, env = env))
  r <- transform(backend_eval(q, FALSE, env = env))
  if(!identical(cpp, r)) {
    diff <- capture.output(all.equal(cpp, r))
    stop(
      paste0(
        "Backend parity mismatch for expression:\n",
        deparse(q),
        "\nDiff:\n",
        paste(diff, collapse = "\n")
      )
    )
  }
  testthat::succeed()
}

expect_wrapper_dispatch <- function(wrapper_name, expr, should_dispatch = TRUE, env = parent.frame()) {
  ns <- asNamespace("Immutables")
  flag <- new.env(parent = emptyenv())
  flag$called <- FALSE
  suppressMessages(base::trace(
    what = wrapper_name,
    tracer = bquote(assign("called", TRUE, envir = .(flag))),
    where = ns,
    print = FALSE
  ))
  on.exit(suppressMessages(base::untrace(what = wrapper_name, where = ns)), add = TRUE)
  eval(substitute(expr), envir = env)
  if(should_dispatch) {
    testthat::expect_true(isTRUE(flag$called), info = paste("Expected wrapper dispatch:", wrapper_name))
  } else {
    testthat::expect_false(isTRUE(flag$called), info = paste("Expected no wrapper dispatch:", wrapper_name))
  }
}

snapshot_tree <- function(t) {
  els <- Immutables:::.ft_to_list(t)
  vals <- lapply(els, Immutables:::.ft_strip_name)
  nms <- vapply(els, function(el) {
    nm <- Immutables:::.ft_get_name(el)
    if(is.null(nm)) "" else nm
  }, character(1))
  list(
    class = class(t),
    values = vals,
    names = nms,
    measures = attr(t, "measures", exact = TRUE),
    monoid_names = names(attr(t, "monoids", exact = TRUE))
  )
}

snapshot_split <- function(s) {
  list(
    left = snapshot_tree(s$left),
    value = Immutables:::.ft_strip_name(s$value),
    right = snapshot_tree(s$right)
  )
}

snapshot_locate <- function(x) {
  list(
    found = x$found,
    value = if(isTRUE(x$found)) Immutables:::.ft_strip_name(x$value) else NULL,
    index = x$index,
    left_measure = x$left_measure,
    hit_measure = x$hit_measure,
    right_measure = x$right_measure
  )
}

parity_scenarios <- c(
  "MeasureMonoid constructor/use",
  "predicate constructor/use",
  "empty_tree",
  "tree_from",
  "tree_from unnamed",
  "add_monoids",
  "append",
  "append named",
  "prepend",
  "prepend named",
  "concat_trees",
  "split_around_by_predicate",
  "split_by_predicate",
  "locate_by_predicate",
  "peek/pop_at helpers",
  "insert_at helper",
  "[ read (integer/logical/name)",
  "[[ read (integer/name)",
  "[<- replacement (integer/logical/name)",
  "[[<- replacement and deletion",
  "$ read",
  "$<- replacement",
  "ordered_sequence insert",
  "interval_index insert and queries",
  "print.FingerTree output",
  "get_graph_df",
  "validate_tree",
  "validate_name_state"
)

cpp_wrapper_coverage <- list(
  .ft_cpp_add_right = c("append"),
  .ft_cpp_add_right_named = c("append named"),
  .ft_cpp_add_left = c("prepend"),
  .ft_cpp_add_left_named = c("prepend named"),
  .ft_cpp_tree_from = c("tree_from unnamed"),
  .ft_cpp_tree_from_prepared = c("tree_from"),
  .ft_cpp_tree_from_sorted = c("ordered_sequence insert"),
  .ft_cpp_concat = c("concat_trees", "interval_index insert and queries"),
  .ft_cpp_oms_insert = c("ordered_sequence insert"),
  .ft_cpp_locate = c("locate_by_predicate"),
  .ft_cpp_split_tree = c("split_around_by_predicate", "split_by_predicate", "interval_index insert and queries"),
  .ft_cpp_split_at_index = c("[[ read (integer/name)", "[[<- replacement and deletion", "peek/pop_at helpers"),
  .ft_cpp_find_name_position = c("$ read", "[ read (integer/logical/name)"),
  .ft_cpp_get_by_index = c("[[ read (integer/name)"),
  .ft_cpp_get_many_by_index = c("[ read (integer/logical/name)"),
  .ft_cpp_name_positions = c("[ read (integer/logical/name)")
)

testthat::test_that("backend parity: coverage map includes all cpp wrappers", {
  ns <- asNamespace("Immutables")
  wrappers <- ls(ns, all.names = TRUE)
  wrappers <- wrappers[startsWith(wrappers, ".ft_cpp_")]
  wrappers <- setdiff(
    wrappers,
    c(".ft_cpp_enabled", ".ft_cpp_eligible_monoids", ".ft_cpp_can_use", ".ft_cpp_can_use_oms_insert")
  )

  testthat::expect_setequal(names(cpp_wrapper_coverage), wrappers)

  for(op in names(cpp_wrapper_coverage)) {
    tags <- cpp_wrapper_coverage[[op]]
    for(tag in tags) {
      testthat::expect_true(tag %in% parity_scenarios, info = paste("Missing parity scenario tag:", tag))
    }
  }
})

testthat::test_that("backend parity: forced backend option gates C++ eligibility", {
  ms <- ensure_size_monoids(list(.size = size_measure_monoid()))
  testthat::expect_true(backend_eval(quote(.ft_cpp_can_use(ms)), TRUE))
  testthat::expect_false(backend_eval(quote(.ft_cpp_can_use(ms)), FALSE))
})

testthat::test_that("backend parity: append dispatch respects forced backend mode", {
  backend_eval(
    quote(expect_wrapper_dispatch(".ft_cpp_add_right", {
      t <- as_flexseq(1:6)
      push_back(t, 7)
    }, should_dispatch = TRUE)),
    TRUE
  )

  backend_eval(
    quote(expect_wrapper_dispatch(".ft_cpp_add_right", {
      t <- as_flexseq(1:6)
      push_back(t, 7)
    }, should_dispatch = FALSE)),
    FALSE
  )
})

testthat::test_that("backend parity: MeasureMonoid constructor/use", {
  expect_backend_identical({
    m <- measure_monoid(function(a, b) a + b, 0, as.numeric)
    t <- add_monoids(as_flexseq(1:5), list(sum = m))
    as.integer(node_measure(t, "sum"))
  })
})

testthat::test_that("backend parity: empty_tree", {
  expect_backend_identical({
    snapshot_tree(flexseq())
  })
})

testthat::test_that("backend parity: tree_from", {
  expect_backend_identical({
    ms <- list(sum = measure_monoid(function(a, b) a + b, 0, as.numeric))
    snapshot_tree(add_monoids(as_flexseq(setNames(as.list(1:8), letters[1:8])), ms))
  })
})

testthat::test_that("backend parity: tree_from unnamed", {
  expect_backend_identical({
    ms <- list(sum = measure_monoid(function(a, b) a + b, 0, as.numeric))
    snapshot_tree(add_monoids(as_flexseq(as.list(1:8)), ms))
  })
})

testthat::test_that("backend parity: tree_from prepared names type check", {
  ms <- ensure_size_monoids(list(.size = size_measure_monoid()))
  testthat::expect_error(
    .ft_cpp_tree_from_prepared(as.list(1:2), list("a", "b"), ms),
    "character vector"
  )
})

testthat::test_that("backend parity: add_monoids", {
  expect_backend_identical({
    t <- as_flexseq(1:8)
    m <- measure_monoid(function(a, b) a + b, 0, as.numeric)
    snapshot_tree(add_monoids(t, list(sum = m)))
  })
})

testthat::test_that("backend parity: append", {
  expect_backend_identical({
    t <- as_flexseq(1:10)
    t <- push_back(t, 11)
    snapshot_tree(t)
  })
})

testthat::test_that("backend parity: append named", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(1:4), LETTERS[1:4]))
    t <- push_back(t, structure(5, names = "E"))
    snapshot_tree(t)
  })
})

testthat::test_that("backend parity: prepend", {
  expect_backend_identical({
    t <- as_flexseq(1:10)
    t <- push_front(t, 0)
    snapshot_tree(t)
  })
})

testthat::test_that("backend parity: prepend named", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(1:4), LETTERS[1:4]))
    t <- push_front(t, structure(0, names = "Z"))
    snapshot_tree(t)
  })
})

testthat::test_that("backend parity: concat_trees", {
  expect_backend_identical({
    t1 <- as_flexseq(1:10)
    t2 <- as_flexseq(11:20)
    snapshot_tree(c(t1, t2))
  })
})

testthat::test_that("backend parity: split_around_by_predicate", {
  expect_backend_identical({
    t <- as_flexseq(letters[1:20])
    snapshot_split(split_around_by_predicate(t, function(v) v >= 8, ".size"))
  })
})

testthat::test_that("backend parity: split_by_predicate", {
  expect_backend_identical({
    t <- as_flexseq(letters[1:20])
    s <- split_by_predicate(t, function(v) v >= 8, ".size")
    list(left = snapshot_tree(s$left), right = snapshot_tree(s$right))
  })
})

testthat::test_that("backend parity: locate_by_predicate", {
  expect_backend_identical({
    t <- as_flexseq(letters[1:20])
    snapshot_locate(locate_by_predicate(t, function(v) v >= 10, ".size", include_metadata = TRUE))
  })
})

testthat::test_that("backend parity: peek/pop_at helpers", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(letters[1:8]), LETTERS[1:8]))
    out <- pop_at(t, 5)
    list(
      peek = peek_at(t, 3),
      popped = out$value,
      remaining = snapshot_tree(out$remaining)
    )
  })
})

testthat::test_that("backend parity: insert_at helper", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(letters[1:8]), LETTERS[1:8]))
    out <- insert_at(t, 4, setNames(as.list(c("x", "y")), c("X1", "X2")))
    snapshot_tree(out)
  })
})

testthat::test_that("backend parity: [ read (integer/logical/name)", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(letters[1:12]), LETTERS[1:12]))
    list(
      int = snapshot_tree(t[c(2, 8, 8, 1)]),
      lgl = snapshot_tree(t[c(TRUE, FALSE, TRUE)]),
      chr = snapshot_tree(t[c("H", "A", "missing", "C")])
    )
  })
})

testthat::test_that("backend parity: [[ read (integer/name)", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(letters[1:12]), LETTERS[1:12]))
    list(int = t[[3]], chr = t[["C"]])
  })
})

testthat::test_that("backend parity: [<- replacement (integer/logical/name)", {
  expect_backend_identical({
    t1 <- as_flexseq(1:12)
    t1[c(2, 5, 5)] <- list(200, 500, 501)
    t1[c(TRUE, FALSE, TRUE)] <- list(9, 8)

    t2 <- as_flexseq(setNames(as.list(letters[1:8]), LETTERS[1:8]))
    t2[c("B", "D")] <- list("bb", "dd")

    list(int_lgl = snapshot_tree(t1), chr = snapshot_tree(t2))
  })
})

testthat::test_that("backend parity: [[<- replacement and deletion", {
  expect_backend_identical({
    t1 <- as_flexseq(1:10)
    t1[[4]] <- 99
    t1[[6]] <- NULL

    t2 <- as_flexseq(setNames(as.list(letters[1:8]), LETTERS[1:8]))
    t2[["C"]] <- structure("cc", names = "Z")
    t2[["E"]] <- NULL

    list(int = snapshot_tree(t1), chr = snapshot_tree(t2))
  })
})

testthat::test_that("backend parity: $ read", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(letters[1:8]), LETTERS[1:8]))
    list(A = t$A, C = t$C)
  })
})

testthat::test_that("backend parity: $<- replacement", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(letters[1:8]), LETTERS[1:8]))
    t$B <- "bbb"
    snapshot_tree(t)
  })
})

testthat::test_that("backend parity: ordered_sequence insert", {
  expect_backend_identical({
    x <- as_ordered_sequence(list("aa", "bb", "c", "ddd"), keys = c(2, 2, 1, 3))
    out <- insert(x, "qq", key = 2)
    as.list(out)
  })
})

testthat::test_that("backend parity: ordered_sequence Date keys use consistent fallback path", {
  expect_backend_identical({
    x <- as_ordered_sequence(
      list("late", "early1", "early2"),
      keys = as.Date(c("2024-01-03", "2024-01-01", "2024-01-01"))
    )
    y <- insert(x, "early3", key = as.Date("2024-01-01"))
    list(
      values = as.list(y),
      lower = lower_bound(y, as.Date("2024-01-01")),
      upper = upper_bound(y, as.Date("2024-01-01"))
    )
  })
})

testthat::test_that("backend parity: interval_index insert and queries", {
  expect_backend_identical({
    x <- as_interval_index(
      list("a", "b", "c"),
      start = c(1, 2, 2),
      end = c(3, 2, 4),
      default_query_bounds = "[]"
    )
    y <- insert(x, "d", start = 2, end = 5)

    p0 <- pop_point(y, 2)
    p1 <- pop_overlaps(y, 2, 3)
    p2 <- pop_all_within(y, 2, 5)
    bounds_tokens <- c("[)", "[]", "()", "(]")
    bounds_matrix <- lapply(bounds_tokens, function(bt) {
      list(
        token = bt,
        point = as.list(peek_all_point(y, 2, bounds = bt)),
        overlaps = as.list(peek_all_overlaps(y, 2, 3, bounds = bt)),
        containing = as.list(peek_all_containing(y, 2, 3, bounds = bt)),
        within = as.list(peek_all_within(y, 2, 3, bounds = bt)),
        pop_point = as.list(pop_all_point(y, 2, bounds = bt)$elements),
        pop_overlaps = as.list(pop_all_overlaps(y, 2, 3, bounds = bt)$elements),
        pop_containing = as.list(pop_all_containing(y, 2, 3, bounds = bt)$elements),
        pop_within = as.list(pop_all_within(y, 2, 3, bounds = bt)$elements)
      )
    })

    match_at_tokens <- c("start", "end", "either")
    match_at_matrix <- lapply(match_at_tokens, function(ma) {
      list(
        mode = ma,
        peek = as.list(peek_all_point(y, 2, match_at = ma)),
        peek_first = peek_point(y, 2, match_at = ma),
        pop = as.list(pop_all_point(y, 2, match_at = ma)$elements),
        pop_remaining = as.list(pop_all_point(y, 2, match_at = ma)$remaining)
      )
    })

    list(
      values = as.list(y),
      intervals = lapply(.ivx_entries(y), function(e) list(start = e$start, end = e$end)),
      bounds_matrix = bounds_matrix,
      match_at_matrix = match_at_matrix,
      point_first = peek_point(y, 2),
      point = as.list(peek_all_point(y, 2)),
      overlaps_first = peek_overlaps(y, 2, 3),
      overlaps = as.list(peek_all_overlaps(y, 2, 3)),
      containing_first = peek_containing(y, 2, 3),
      containing = as.list(peek_all_containing(y, 2, 3)),
      within_first = peek_within(y, 2, 5),
      within = as.list(peek_all_within(y, 2, 5)),
      pop_point_first = list(
        value = p0$value,
        start = p0$start,
        end = p0$end,
        remaining = as.list(p0$remaining)
      ),
      pop_first = list(
        value = p1$value,
        start = p1$start,
        end = p1$end,
        remaining = as.list(p1$remaining)
      ),
      pop_all = list(
        elements = as.list(p2$elements),
        remaining = as.list(p2$remaining)
      )
    )
  })
})

testthat::test_that("backend parity: interval_index user monoid recomputation", {
  expect_backend_identical({
    sum_item <- measure_monoid(function(a, b) a + b, 0, function(entry) as.numeric(entry$value))
    width_sum <- measure_monoid(function(a, b) a + b, 0, function(entry) as.numeric(entry$end - entry$start))

    x <- add_monoids(as_interval_index(
      as.list(c(10, 20, 30)),
      start = c(1, 2, 4),
      end = c(3, 5, 6)
    ), list(sum_item = sum_item, width_sum = width_sum))
    y <- insert(x, 40, start = 3, end = 4)
    z <- fapply(y, function(value, start, end, name) value + 1)
    s <- peek_all_overlaps(z, 2, 3, bounds = "[)")
    p <- pop_all_overlaps(z, 2, 3, bounds = "[)")

    list(
      base = c(sum_item = node_measure(x, "sum_item"), width_sum = node_measure(x, "width_sum")),
      inserted = c(sum_item = node_measure(y, "sum_item"), width_sum = node_measure(y, "width_sum")),
      applied = c(sum_item = node_measure(z, "sum_item"), width_sum = node_measure(z, "width_sum")),
      slice = c(sum_item = node_measure(s, "sum_item"), width_sum = node_measure(s, "width_sum")),
      pop_element = c(
        sum_item = node_measure(p$elements, "sum_item"),
        width_sum = node_measure(p$elements, "width_sum")
      ),
      pop_remaining = c(
        sum_item = node_measure(p$remaining, "sum_item"),
        width_sum = node_measure(p$remaining, "width_sum")
      ),
      slice_values = as.list(s),
      pop_values = as.list(p$elements),
      remaining_values = as.list(p$remaining)
    )
  })
})

testthat::test_that("backend parity: print.FingerTree output", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(1:10), letters[1:10]))
    capture.output(print(t))
  })
})

testthat::test_that("backend parity: get_graph_df", {
  expect_backend_identical({
    t <- as_flexseq(letters[1:12])
    get_graph_df(t)
  })
})

testthat::test_that("backend parity: validate_tree", {
  expect_backend_identical({
    t <- as_flexseq(letters[1:10])
    isTRUE(validate_tree(t))
  })
})

testthat::test_that("backend parity: validate_name_state", {
  expect_backend_identical({
    t <- as_flexseq(setNames(as.list(letters[1:10]), LETTERS[1:10]))
    isTRUE(validate_name_state(t))
  })
})

testthat::test_that("interval_index constructor sorts by start and preserves stable duplicate order", {
  ix <- as_interval_index(
    list("c", "a", "b", "d"),
    start = c(2, 1, 2, 2),
    end = c(5, 3, 4, 6)
  )

  testthat::expect_s3_class(ix, "interval_index")
  testthat::expect_equal(as.list(ix), list("a", "c", "b", "d"))
  testthat::expect_identical(length(ix), 4L)

  entries <- .ivx_entries(ix)
  starts <- lapply(entries, function(e) e$start)
  ends <- lapply(entries, function(e) e$end)
  testthat::expect_equal(starts, as.list(c(1, 2, 2, 2)))
  testthat::expect_equal(ends, as.list(c(3, 5, 4, 6)))
})

testthat::test_that("interval_index validates endpoints and bounds", {
  ok <- as_interval_index("p", start = 2, end = 2)
  testthat::expect_s3_class(ok, "interval_index")

  testthat::expect_error(as_interval_index("x", start = 3, end = 2), "start")
  testthat::expect_error(as_interval_index("x", start = 1, end = 2, default_query_bounds = "bad"), "bounds")
})

testthat::test_that("min_endpoint/max_endpoint expose endpoint extrema", {
  ix <- as_interval_index(list("c", "a", "b"), start = c(2, 1, 2), end = c(5, 3, 4))
  testthat::expect_identical(min_endpoint(ix), 1)
  testthat::expect_identical(max_endpoint(ix), 5)

  empty <- interval_index()
  testthat::expect_null(min_endpoint(empty))
  testthat::expect_null(max_endpoint(empty))
})

testthat::test_that("insert is persistent and appends at right edge of equal-start block", {
  ix <- as_interval_index(list("a", "b", "c"), start = c(1, 2, 2), end = c(2, 3, 4))
  ix2 <- insert(ix, "d", start = 2, end = 5)

  testthat::expect_equal(as.list(ix), list("a", "b", "c"))
  testthat::expect_equal(as.list(ix2), list("a", "b", "c", "d"))

  entries <- .ivx_entries(ix2)
  starts <- lapply(entries, function(e) e$start)
  ends <- lapply(entries, function(e) e$end)
  testthat::expect_equal(starts, as.list(c(1, 2, 2, 2)))
  testthat::expect_equal(ends, as.list(c(2, 3, 4, 5)))
})

testthat::test_that("peek_point honors boundary modes", {
  ix <- as_interval_index(
    list("A", "B", "C", "D"),
    start = c(1, 2, 3, 2),
    end = c(2, 3, 4, 2),
    default_query_bounds = "[)"
  )

  testthat::expect_equal(peek_point(ix, 2, bounds = "[)"), "B")
  testthat::expect_equal(as.list(peek_all_point(ix, 2, bounds = "[]")), list("A", "B", "D"))
  testthat::expect_null(peek_point(ix, 2, bounds = "()"))
  testthat::expect_identical(length(peek_all_point(ix, 2, bounds = "()")), 0L)
  testthat::expect_equal(as.list(peek_all_point(ix, 2, bounds = "(]")), list("A"))
})

testthat::test_that("peek overlap/contain/within queries are deterministic", {
  ix <- as_interval_index(
    list("A", "B", "C", "D"),
    start = c(1, 2, 3, 2),
    end = c(2, 3, 4, 2),
    default_query_bounds = "[)"
  )

  testthat::expect_equal(peek_overlaps(ix, 2, 3, bounds = "[)"), "B")
  testthat::expect_equal(as.list(peek_all_overlaps(ix, 2, 3, bounds = "[]")), list("A", "B", "D", "C"))

  jy <- as_interval_index(
    list("outer", "inner", "tail", "point"),
    start = c(1, 2, 2, 3),
    end = c(5, 3, 5, 3),
    default_query_bounds = "[]"
  )

  testthat::expect_equal(peek_containing(jy, 2, 3), "outer")
  testthat::expect_equal(as.list(peek_all_containing(jy, 2, 3)), list("outer", "inner", "tail"))
  testthat::expect_equal(peek_within(jy, 2, 3), "inner")
  testthat::expect_equal(as.list(peek_all_within(jy, 2, 3)), list("inner", "point"))

  testthat::expect_null(peek_overlaps(ix, 9, 10))
  miss <- peek_all_overlaps(ix, 9, 10)
  testthat::expect_s3_class(miss, "interval_index")
  testthat::expect_identical(length(miss), 0L)
})

testthat::test_that("relation query/pop contracts hold across all bounds tokens", {
  ix <- as_interval_index(
    list("A", "B", "C", "D"),
    start = c(1, 2, 3, 2),
    end = c(2, 3, 4, 2),
    default_query_bounds = "[)"
  )
  vals <- as.list(ix)
  entries <- .ivx_entries(ix)
  starts <- as.numeric(unlist(lapply(entries, function(e) e$start), use.names = FALSE))
  ends <- as.numeric(unlist(lapply(entries, function(e) e$end), use.names = FALSE))

  contains_point <- function(start, end, point, bounds) {
    include_start <- substr(bounds, 1L, 1L) == "["
    include_end <- substr(bounds, 2L, 2L) == "]"
    left_ok <- if(include_start) point >= start else point > start
    right_ok <- if(include_end) point <= end else point < end
    isTRUE(left_ok && right_ok)
  }
  overlaps <- function(a_start, a_end, b_start, b_end, bounds) {
    touching_is_overlap <- identical(bounds, "[]")
    a_before_b <- if(touching_is_overlap) a_end < b_start else a_end <= b_start
    b_before_a <- if(touching_is_overlap) b_end < a_start else b_end <= a_start
    !isTRUE(a_before_b || b_before_a)
  }
  containing <- function(start, end, q_start, q_end, bounds) {
    overlaps(start, end, q_start, q_end, bounds) && start <= q_start && end >= q_end
  }
  within <- function(start, end, q_start, q_end, bounds) {
    overlaps(start, end, q_start, q_end, bounds) && q_start <= start && q_end >= end
  }

  bounds_tokens <- c("[)", "[]", "()", "(]")
  for(bt in bounds_tokens) {
    idx_point <- which(vapply(seq_along(vals), function(i) contains_point(starts[[i]], ends[[i]], 2, bt), logical(1)))
    idx_over <- which(vapply(seq_along(vals), function(i) overlaps(starts[[i]], ends[[i]], 2, 3, bt), logical(1)))
    idx_cont <- which(vapply(seq_along(vals), function(i) containing(starts[[i]], ends[[i]], 2, 3, bt), logical(1)))
    idx_with <- which(vapply(seq_along(vals), function(i) within(starts[[i]], ends[[i]], 2, 3, bt), logical(1)))

    expect_relation <- function(peek_first, peek_all, pop_first, pop_all, idx) {
      expect_all <- vals[idx]
      expect_rest <- vals[setdiff(seq_along(vals), idx)]

      testthat::expect_equal(as.list(peek_all), expect_all)
      if(length(expect_all) == 0L) {
        testthat::expect_null(peek_first)
      } else {
        testthat::expect_equal(peek_first, expect_all[[1L]])
      }

      if(length(expect_all) == 0L) {
        testthat::expect_null(pop_first$value)
        testthat::expect_null(pop_first$start)
        testthat::expect_null(pop_first$end)
        testthat::expect_equal(as.list(pop_first$remaining), vals)
      } else {
        i <- idx[[1L]]
        testthat::expect_equal(pop_first$value, vals[[i]])
        testthat::expect_equal(pop_first$start, starts[[i]])
        testthat::expect_equal(pop_first$end, ends[[i]])
        testthat::expect_equal(as.list(pop_first$remaining), vals[-i])
      }

      testthat::expect_equal(as.list(pop_all$elements), expect_all)
      testthat::expect_equal(as.list(pop_all$remaining), expect_rest)
    }

    expect_relation(
      peek_first = peek_point(ix, 2, bounds = bt),
      peek_all = peek_all_point(ix, 2, bounds = bt),
      pop_first = pop_point(ix, 2, bounds = bt),
      pop_all = pop_all_point(ix, 2, bounds = bt),
      idx = idx_point
    )
    expect_relation(
      peek_first = peek_overlaps(ix, 2, 3, bounds = bt),
      peek_all = peek_all_overlaps(ix, 2, 3, bounds = bt),
      pop_first = pop_overlaps(ix, 2, 3, bounds = bt),
      pop_all = pop_all_overlaps(ix, 2, 3, bounds = bt),
      idx = idx_over
    )
    expect_relation(
      peek_first = peek_containing(ix, 2, 3, bounds = bt),
      peek_all = peek_all_containing(ix, 2, 3, bounds = bt),
      pop_first = pop_containing(ix, 2, 3, bounds = bt),
      pop_all = pop_all_containing(ix, 2, 3, bounds = bt),
      idx = idx_cont
    )
    expect_relation(
      peek_first = peek_within(ix, 2, 3, bounds = bt),
      peek_all = peek_all_within(ix, 2, 3, bounds = bt),
      pop_first = pop_within(ix, 2, 3, bounds = bt),
      pop_all = pop_all_within(ix, 2, 3, bounds = bt),
      idx = idx_with
    )
  }
})

testthat::test_that("point queries honor match_at modes on a fixed fixture", {
  ix <- as_interval_index(
    list("A", "B", "C", "D"),
    start = c(1, 2, 3, 2),
    end = c(3, 4, 5, 2),
    default_query_bounds = "[)"
  )

  # match_at = "start": entries whose start == point, FIFO on ties.
  testthat::expect_equal(peek_point(ix, 2, match_at = "start"), "B")
  testthat::expect_equal(as.list(peek_all_point(ix, 2, match_at = "start")), list("B", "D"))
  testthat::expect_null(peek_point(ix, 99, match_at = "start"))

  # match_at = "end": entries whose end == point.
  testthat::expect_equal(peek_point(ix, 3, match_at = "end"), "A")
  testthat::expect_equal(as.list(peek_all_point(ix, 3, match_at = "end")), list("A"))
  testthat::expect_equal(as.list(peek_all_point(ix, 2, match_at = "end")), list("D"))
  testthat::expect_identical(length(peek_all_point(ix, 99, match_at = "end")), 0L)

  # match_at = "either": union of start/end matches, canonical order by start.
  testthat::expect_equal(as.list(peek_all_point(ix, 2, match_at = "either")), list("B", "D"))
  testthat::expect_equal(as.list(peek_all_point(ix, 3, match_at = "either")), list("A", "C"))

  # pop_* shapes match peek semantics and preserve persistence.
  before <- as.list(ix)
  popped_end <- pop_all_point(ix, 3, match_at = "end")
  testthat::expect_equal(as.list(popped_end$elements), list("A"))
  testthat::expect_equal(as.list(popped_end$remaining), list("B", "D", "C"))
  testthat::expect_equal(as.list(ix), before)

  one_start <- pop_point(ix, 2, match_at = "start")
  testthat::expect_equal(one_start$value, "B")
  testthat::expect_equal(one_start$start, 2)
  testthat::expect_equal(one_start$end, 4)
  testthat::expect_equal(as.list(one_start$remaining), list("A", "D", "C"))
  # (sanity: after popping first of B/D at start==2, D is the only remaining start==2 entry)
})

testthat::test_that("match_at = 'interval' preserves default containment semantics", {
  ix <- as_interval_index(
    list("A", "B", "C"),
    start = c(1, 2, 4),
    end = c(3, 2, 5),
    default_query_bounds = "[)"
  )
  testthat::expect_equal(peek_point(ix, 2), peek_point(ix, 2, match_at = "interval"))
  testthat::expect_equal(
    as.list(peek_all_point(ix, 2)),
    as.list(peek_all_point(ix, 2, match_at = "interval"))
  )
})

testthat::test_that("coordinate-equality match_at modes ignore bounds", {
  ix <- as_interval_index(
    list("A", "B", "C"),
    start = c(1, 2, 3),
    end = c(3, 4, 5),
    default_query_bounds = "[)"
  )
  # With match_at = "end", endpoint is a structural coordinate — bounds override
  # has no effect.
  expected <- as.list(peek_all_point(ix, 3, match_at = "end"))
  for(bt in c("[)", "[]", "()", "(]")) {
    testthat::expect_equal(
      as.list(peek_all_point(ix, 3, match_at = "end", bounds = bt)),
      expected
    )
  }
})

testthat::test_that("match_at on empty indices is non-throwing and empty", {
  empty_ix <- interval_index()
  for(m in c("start", "end", "either")) {
    testthat::expect_null(peek_point(empty_ix, 1, match_at = m))
    testthat::expect_identical(length(peek_all_point(empty_ix, 1, match_at = m)), 0L)
    r <- pop_point(empty_ix, 1, match_at = m)
    testthat::expect_null(r$value)
    testthat::expect_null(r$start)
    testthat::expect_null(r$end)
    testthat::expect_identical(length(r$remaining), 0L)
    ra <- pop_all_point(empty_ix, 1, match_at = m)
    testthat::expect_identical(length(ra$elements), 0L)
    testthat::expect_identical(length(ra$remaining), 0L)
  }
})

testthat::test_that("match_at point queries match a pure-R oracle across bounds tokens", {
  ix <- as_interval_index(
    list("A", "B", "C", "D", "E"),
    start = c(1, 2, 3, 2, 4),
    end = c(3, 4, 5, 2, 4),
    default_query_bounds = "[)"
  )
  vals <- as.list(ix)
  entries <- .ivx_entries(ix)
  starts <- as.numeric(unlist(lapply(entries, function(e) e$start), use.names = FALSE))
  ends <- as.numeric(unlist(lapply(entries, function(e) e$end), use.names = FALSE))

  for(p in c(1, 2, 3, 4, 5, 99)) {
    # start mode
    idx <- which(starts == p)
    testthat::expect_equal(as.list(peek_all_point(ix, p, match_at = "start")), vals[idx])
    # end mode
    idx <- which(ends == p)
    testthat::expect_equal(as.list(peek_all_point(ix, p, match_at = "end")), vals[idx])
    # either mode — canonical start order
    idx <- which(starts == p | ends == p)
    testthat::expect_equal(as.list(peek_all_point(ix, p, match_at = "either")), vals[idx])
  }
})

testthat::test_that("pop helpers follow first/all contracts and preserve persistence", {
  ix <- as_interval_index(
    list("A", "B", "C", "D"),
    start = c(1, 2, 3, 2),
    end = c(2, 3, 4, 2),
    default_query_bounds = "[]"
  )

  first <- pop_overlaps(ix, 2, 3)
  testthat::expect_equal(first$value, "A")
  testthat::expect_equal(first$start, 1)
  testthat::expect_equal(first$end, 2)
  testthat::expect_s3_class(first$remaining, "interval_index")

  all <- pop_all_overlaps(ix, 2, 3)
  testthat::expect_s3_class(all$elements, "interval_index")
  testthat::expect_equal(as.list(all$elements), list("A", "B", "D", "C"))
  testthat::expect_identical(length(all$remaining), 0L)

  point_first <- pop_point(ix, 2)
  testthat::expect_equal(point_first$value, "A")
  testthat::expect_equal(point_first$start, 1)
  testthat::expect_equal(point_first$end, 2)
  testthat::expect_equal(as.list(point_first$remaining), list("B", "D", "C"))

  point_all <- pop_all_point(ix, 2)
  testthat::expect_s3_class(point_all$elements, "interval_index")
  testthat::expect_equal(as.list(point_all$elements), list("A", "B", "D"))
  testthat::expect_equal(as.list(point_all$remaining), list("C"))

  miss_first <- pop_within(ix, 9, 10)
  testthat::expect_null(miss_first$value)
  testthat::expect_null(miss_first$start)
  testthat::expect_null(miss_first$end)
  testthat::expect_equal(as.list(miss_first$remaining), as.list(ix))

  miss_all <- pop_all_containing(ix, 9, 10)
  testthat::expect_s3_class(miss_all$elements, "interval_index")
  testthat::expect_identical(length(miss_all$elements), 0L)
  testthat::expect_equal(as.list(miss_all$remaining), as.list(ix))

  miss_point_first <- pop_point(ix, 9)
  testthat::expect_null(miss_point_first$value)
  testthat::expect_null(miss_point_first$start)
  testthat::expect_null(miss_point_first$end)
  testthat::expect_equal(as.list(miss_point_first$remaining), as.list(ix))

  miss_point_all <- pop_all_point(ix, 9)
  testthat::expect_s3_class(miss_point_all$elements, "interval_index")
  testthat::expect_identical(length(miss_point_all$elements), 0L)
  testthat::expect_equal(as.list(miss_point_all$remaining), as.list(ix))
})

testthat::test_that("ordered key APIs are blocked on interval_index", {
  ix <- as_interval_index(list("a", "b"), start = c(1, 2), end = c(2, 3))

  testthat::expect_error(lower_bound(ix, 2), "not supported for interval_index")
  testthat::expect_error(upper_bound(ix, 2), "not supported for interval_index")
  testthat::expect_error(peek_key(ix, 2), "not supported for interval_index")
  testthat::expect_error(pop_key(ix, 2), "not supported for interval_index")
  testthat::expect_error(elements_between(ix, 1, 2), "not supported for interval_index")
  testthat::expect_error(count_key(ix, 2), "not supported for interval_index")
  testthat::expect_error(count_between(ix, 1, 2), "not supported for interval_index")
})

testthat::test_that("interval_index indexing preserves class and blocks replacement", {
  ix <- as_interval_index(
    setNames(as.list(c("xa", "xb", "xc")), c("a", "b", "c")),
    start = c(1, 2, 3),
    end = c(2, 3, 4)
  )

  sub <- ix[c(1, 3)]
  testthat::expect_s3_class(sub, "interval_index")
  testthat::expect_equal(as.list(sub), list(a = "xa", c = "xc"))
  testthat::expect_identical(names(as.list(sub)), c("a", "c"))

  testthat::expect_equal(ix[[2]], "xb")
  testthat::expect_equal(ix[["c"]], "xc")
  testthat::expect_equal(ix$b, "xb")

  reordered <- NULL
  testthat::expect_warning(
    { reordered <- ix[c(3, 1)] },
    "canonicalizes selector order"
  )
  testthat::expect_equal(as.list(reordered), list(a = "xa", c = "xc"))

  reordered_name <- NULL
  testthat::expect_warning(
    { reordered_name <- ix[c("b", "a")] },
    "canonicalizes selector order"
  )
  testthat::expect_equal(as.list(reordered_name), list(a = "xa", b = "xb"))

  testthat::expect_error(ix[c(2, 2)], "duplicate indices")
  testthat::expect_error(ix[c("a", "a")], "duplicate indices")

  testthat::expect_error({ ix[[1]] <- "qq" }, "not supported")
  testthat::expect_error({ ix[1] <- list("qq") }, "not supported")
  testthat::expect_error({ ix$b <- "qq" }, "not supported")
})

testthat::test_that("front/back/at peek/pop helpers are blocked on interval_index", {
  ix <- as_interval_index("a", start = 1, end = 2)

  testthat::expect_error(peek_front(ix), "not supported for interval_index")
  testthat::expect_error(peek_back(ix), "not supported for interval_index")
  testthat::expect_error(peek_at(ix, 1), "not supported for interval_index")
  testthat::expect_error(pop_front(ix), "not supported for interval_index")
  testthat::expect_error(pop_back(ix), "not supported for interval_index")
  testthat::expect_error(pop_at(ix, 1), "not supported for interval_index")
})

testthat::test_that("fapply for interval_index updates payload only and keeps interval metadata immutable", {
  ix <- as_interval_index(
    setNames(as.list(c("a", "b", "c")), c("ka", "kb", "kc")),
    start = c(3, 1, 2),
    end = c(4, 2, 3),
    default_query_bounds = "[]"
  )
  b0 <- lapply(.ivx_entries(ix), function(e) list(start = e$start, end = e$end))

  ix2 <- fapply(ix, function(value, start, end, name) {
    toupper(value)
  })
  ix2_noname <- fapply(ix, function(value, start, end) {
    paste0(value, "@", start, "-", end)
  })

  testthat::expect_s3_class(ix2, "interval_index")
  testthat::expect_equal(unname(as.list(ix2)), list("B", "C", "A"))
  testthat::expect_equal(ix2[["kb"]], "B")
  testthat::expect_equal(ix2_noname[["kb"]], "b@1-2")

  b2 <- lapply(.ivx_entries(ix2), function(e) list(start = e$start, end = e$end))
  testthat::expect_equal(unname(lapply(b2, function(e) e$start)), unname(lapply(b0, function(e) e$start)))
  testthat::expect_equal(unname(lapply(b2, function(e) e$end)), unname(lapply(b0, function(e) e$end)))

  testthat::expect_error(fapply(ix, 1), "`FUN` must be a function")

  ix3 <- fapply(ix, function(value, start, end, name) {
    list(old = value, at = c(start, end), nm = name)
  })
  testthat::expect_type(ix3[[1]], "list")
  testthat::expect_named(ix3[[1]], c("old", "at", "nm"))
  b3 <- lapply(.ivx_entries(ix3), function(e) list(start = e$start, end = e$end))
  testthat::expect_equal(unname(lapply(b3, function(e) e$start)), unname(lapply(b0, function(e) e$start)))
  testthat::expect_equal(unname(lapply(b3, function(e) e$end)), unname(lapply(b0, function(e) e$end)))
  testthat::expect_error(fapply(ix, identity, preserve_custom_monoids = NA), "`preserve_custom_monoids` must be TRUE or FALSE")
})

testthat::test_that("interval_index recomputes user monoids across insert, fapply, and slices", {
  sum_item <- measure_monoid(function(a, b) a + b, 0, function(entry) as.numeric(entry$value))
  width_sum <- measure_monoid(function(a, b) a + b, 0, function(entry) as.numeric(entry$end - entry$start))

  ix <- add_monoids(as_interval_index(
    as.list(c(10, 20, 30)),
    start = c(1, 2, 4),
    end = c(3, 5, 6)
  ), list(sum_item = sum_item, width_sum = width_sum))
  testthat::expect_equal(node_measure(ix, "sum_item"), 60)
  testthat::expect_equal(node_measure(ix, "width_sum"), 7)

  ix2 <- insert(ix, 40, start = 3, end = 4)
  testthat::expect_equal(node_measure(ix2, "sum_item"), 100)
  testthat::expect_equal(node_measure(ix2, "width_sum"), 8)

  ix3 <- fapply(ix2, function(value, start, end, name) value + 1)
  testthat::expect_equal(node_measure(ix3, "sum_item"), 104)
  testthat::expect_equal(node_measure(ix3, "width_sum"), 8)

  overlaps <- peek_all_overlaps(ix3, 2, 3, bounds = "[)")
  testthat::expect_s3_class(overlaps, "interval_index")
  testthat::expect_equal(as.list(overlaps), as.list(c(11, 21)))
  testthat::expect_equal(node_measure(overlaps, "sum_item"), 32)
  testthat::expect_equal(node_measure(overlaps, "width_sum"), 5)

  popped <- pop_all_overlaps(ix3, 2, 3, bounds = "[)")
  testthat::expect_s3_class(popped$elements, "interval_index")
  testthat::expect_s3_class(popped$remaining, "interval_index")
  testthat::expect_equal(node_measure(popped$elements, "sum_item"), 32)
  testthat::expect_equal(node_measure(popped$remaining, "sum_item"), 72)
  testthat::expect_equal(node_measure(popped$elements, "width_sum"), 5)
  testthat::expect_equal(node_measure(popped$remaining, "width_sum"), 3)
})

testthat::test_that("interval_index custom monoids accept entry measure arguments", {
  ix <- add_monoids(
    as_interval_index(as.list(c(10, 20)), start = c(1, 3), end = c(2, 5)),
    list(
      by_width = measure_monoid(`+`, 0, function(entry) as.numeric(entry$end - entry$start))
    )
  )
  testthat::expect_equal(node_measure(ix, "by_width"), 3)
})

testthat::test_that("fapply can drop custom monoids for interval_index", {
  sum_item <- measure_monoid(`+`, 0, function(entry) as.numeric(entry$value))
  ix <- add_monoids(as_interval_index(as.list(c(10, 20)), start = c(1, 3), end = c(2, 5)), list(sum_item = sum_item))
  ix2 <- fapply(ix, function(value, start, end, name) value + 1, preserve_custom_monoids = FALSE)

  ms <- attr(ix2, "monoids", exact = TRUE)
  testthat::expect_true(!is.null(ms[[".size"]]))
  testthat::expect_true(!is.null(ms[[".named_count"]]))
  testthat::expect_true(!is.null(ms[[".ivx_max_start"]]))
  testthat::expect_true(!is.null(ms[[".ivx_max_end"]]))
  testthat::expect_true(!is.null(ms[[".ivx_min_end"]]))
  testthat::expect_true(is.null(ms[["sum_item"]]))
  testthat::expect_error(node_measure(ix2, "sum_item"), "Missing cached measure")
  testthat::expect_equal(as.list(ix2), as.list(c(11, 21)))
})

testthat::test_that("interval_index casts down to flexseq explicitly", {
  width_sum <- measure_monoid(`+`, 0, function(entry) as.numeric(entry$end - entry$start))
  ix <- add_monoids(as_interval_index(
    setNames(list("x", "y"), c("ix", "iy")),
    start = c(1, 3),
    end = c(2, 5)
  ), list(width_sum = width_sum))

  fx <- as_flexseq(ix)
  testthat::expect_s3_class(fx, "flexseq")
  testthat::expect_false(inherits(fx, "interval_index"))
  testthat::expect_equal(fx[["ix"]], "x")
  testthat::expect_equal(fx[["iy"]], "y")

  ms <- names(attr(fx, "monoids", exact = TRUE))
  testthat::expect_true(all(c(".size", ".named_count") %in% ms))
  testthat::expect_false("width_sum" %in% ms)
  testthat::expect_error(node_measure(fx, "width_sum"), "Missing cached measure")
  testthat::expect_false(".ivx_max_start" %in% ms)
  testthat::expect_false(".ivx_max_end" %in% ms)
  testthat::expect_false(".ivx_min_end" %in% ms)
  testthat::expect_false(".oms_max_key" %in% ms)
})

testthat::test_that("interval_index rejects non-empty start when no elements are supplied", {
  testthat::expect_error(
    interval_index(start = c(1, 2)),
    "`start` must be empty when no elements are supplied"
  )
  testthat::expect_error(
    as_interval_index(list(), start = c(1, 2), end = c(2, 3)),
    "`start` must be empty when no elements are supplied"
  )
})

testthat::test_that("interval_index rejects non-empty end when no elements are supplied", {
  testthat::expect_error(
    interval_index(end = c(1, 2)),
    "`end` must be empty when no elements are supplied"
  )
})

testthat::test_that("interval_index requires start and end when elements are supplied", {
  testthat::expect_error(
    interval_index("a", end = c(2)),
    "`start` is required when elements are supplied"
  )
  testthat::expect_error(
    interval_index("a", start = c(1)),
    "`end` is required when elements are supplied"
  )
})

testthat::test_that("interval_index enforces start/end length matches elements length", {
  testthat::expect_error(
    interval_index("a", "b", start = c(1), end = c(2, 3)),
    "`start` length must match elements length"
  )
  testthat::expect_error(
    interval_index("a", "b", start = c(1, 2), end = c(3)),
    "`end` length must match elements length"
  )
})

testthat::test_that("interval_index falls back to internal merge-sort when `c()` on endpoints fails", {
  # Mirrors the ordered_sequence fallback case: for exotic endpoint classes
  # without a `c()` method, the default `order(do.call(c, starts), idx)` path
  # fails and .ivx_build_from_items delegates to .ivx_merge_sort_indices.
  # `mk_comp` is defined in tests/testthat/helper-mk-comp.R.
  testthat::expect_true(
    inherits(tryCatch(order(do.call(c, list(mk_comp(1), mk_comp(2))), 1:2),
                      error = function(e) e),
             "error")
  )

  vals <- letters[1:7]
  starts <- lapply(c(4, 1, 6, 3, 7, 2, 5), mk_comp)
  ends <- lapply(c(5, 2, 7, 4, 8, 3, 6), mk_comp)
  ix <- as_interval_index(vals, start = starts, end = ends)
  testthat::expect_identical(length(ix), 7L)
  # Sorted by start val ascending: 1,2,3,4,5,6,7 -> b,f,d,a,g,c,e
  testthat::expect_identical(unlist(as.list(ix)), c("b", "f", "d", "a", "g", "c", "e"))
})

testthat::test_that(".ivx_parse_entry validates entry records (direct internal call)", {
  # .ivx_parse_entry is a documented internal validator for serialized entry
  # records during subclass restore, but no current caller path exercises it.
  # These tests pin its behavior via `:::` so the validator cannot silently rot
  # if a future restore hook begins relying on it.
  parse <- Immutables:::.ivx_parse_entry

  # Happy path, with and without explicit `key`.
  out <- parse(list(value = "v", start = 1, end = 2))
  testthat::expect_identical(out$entry$value, "v")
  testthat::expect_identical(out$entry$start, 1)
  testthat::expect_identical(out$entry$end, 2)
  testthat::expect_identical(out$endpoint_type, "numeric")

  out_key <- parse(list(value = "v", start = 1, end = 2, key = 1))
  testthat::expect_identical(out_key$entry$start, 1)

  # Error branches.
  testthat::expect_error(parse(1L),                                     "must be named lists")
  testthat::expect_error(parse(list(1, 2, 3)),                           "must be named lists")
  testthat::expect_error(parse(setNames(list(1, 2, 3), c("a", NA, "c"))),"must be named lists")
  testthat::expect_error(parse(list(value = 1, value = 2, start = 1, end = 2)),
                         "entry fields must be unique")
  testthat::expect_error(parse(list(value = 1, start = 1, end = 2, bogus = 9)),
                         "unsupported field")
  testthat::expect_error(parse(list(value = 1, start = 1)),
                         "must include")
  testthat::expect_error(parse(list(value = "v", start = 1, end = 2, key = 7)),
                         "`key` must equal `start`")
})

testthat::test_that("interval_index supports ordered-factor endpoints (exotic-endpoint slow path)", {
  # Ordered factors aren't in the fast-domain set, so queries route through the
  # pure-R leaf_match predicates .ivx_contains_point / .ivx_overlaps_interval /
  # .ivx_contains_interval instead of the inline fast-path closures.
  lvls <- letters[1:5]
  f <- factor(lvls, levels = lvls, ordered = TRUE)

  # Default bounds are [start, end).
  ix <- interval_index("x", "y", start = c(f[1], f[3]), end = c(f[3], f[5]))
  testthat::expect_identical(length(ix), 2L)

  # Point query: `b` falls inside [a,c) only.
  testthat::expect_identical(peek_point(ix, f[2]), "x")

  # Overlaps query: [c,d) overlaps [c,e) only (not [a,c) because end is exclusive).
  testthat::expect_identical(peek_overlaps(ix, start = f[3], end = f[4]), "y")

  # Containing query: entries whose interval contains [b,c) — only [a,c).
  testthat::expect_identical(peek_containing(ix, start = f[2], end = f[3]), "x")

  # Within query: entries whose interval is within [a,e) — only [a,c).
  testthat::expect_identical(peek_within(ix, start = f[1], end = f[5]), "x")
})

# ---- merge.flexseq ----------------------------------------------------------

testthat::test_that("merge.flexseq concatenates in order", {
  x <- flexseq("a", "b")
  y <- flexseq("c", "d")
  m <- merge(x, y)

  testthat::expect_identical(as.list(m), list("a", "b", "c", "d"))
  testthat::expect_identical(length(m), 4L)
})

testthat::test_that("merge.flexseq handles empty inputs", {
  testthat::expect_identical(as.list(merge(flexseq(), flexseq("x"))), list("x"))
  testthat::expect_identical(as.list(merge(flexseq("x"), flexseq())), list("x"))
  testthat::expect_identical(length(merge(flexseq(), flexseq())), 0L)
})

# ---- merge.priority_queue ---------------------------------------------------

testthat::test_that("merge.priority_queue combines entries and preserves min/max", {
  a <- priority_queue("x", "y", priorities = c(5, 1))
  b <- priority_queue("z", priorities = 3)
  m <- merge(a, b)

  testthat::expect_identical(length(m), 3L)
  testthat::expect_identical(peek_min(m), "y")
  testthat::expect_identical(peek_max(m), "x")
})

testthat::test_that("merge.priority_queue empty short-circuits", {
  a <- priority_queue("p", priorities = 1)
  testthat::expect_identical(length(merge(priority_queue(), a)), 1L)
  testthat::expect_identical(length(merge(a, priority_queue())), 1L)
  testthat::expect_identical(length(merge(priority_queue(), priority_queue())), 0L)
})

testthat::test_that("merge.priority_queue does not mutate sources", {
  a <- priority_queue("x", "y", priorities = c(5, 1))
  b <- priority_queue("z", priorities = 3)
  merge(a, b)

  testthat::expect_identical(length(a), 2L)
  testthat::expect_identical(length(b), 1L)
})

testthat::test_that("merge.priority_queue errors on priority-type mismatch", {
  a <- priority_queue("x", priorities = 1)          # numeric
  b <- priority_queue("y", priorities = "P1")       # character
  testthat::expect_error(merge(a, b), "priority type")
})

testthat::test_that("merge.priority_queue errors on non-priority_queue y", {
  a <- priority_queue("x", priorities = 1)
  testthat::expect_error(merge(a, flexseq("y")), "priority_queue")
})

# ---- merge.ordered_sequence -------------------------------------------------

testthat::test_that("merge.ordered_sequence produces key-ascending merge", {
  a <- ordered_sequence("a1", "a2", "a3", keys = c(1, 3, 5))
  b <- ordered_sequence("b1", "b2", "b3", keys = c(2, 3, 6))
  m <- merge(a, b)

  testthat::expect_identical(
    unlist(as.list(m)),
    c("a1", "b1", "a2", "b2", "a3", "b3")
  )
  testthat::expect_identical(length(m), 6L)
})

testthat::test_that("merge.ordered_sequence uses left-biased FIFO on tied keys", {
  a <- ordered_sequence("x1", "x2", keys = c(1, 1))
  b <- ordered_sequence("y1", "y2", keys = c(1, 1))
  m <- merge(a, b)

  testthat::expect_identical(unlist(as.list(m)), c("x1", "x2", "y1", "y2"))
  testthat::expect_identical(
    unlist(as.list(peek_all_key(m, 1))),
    c("x1", "x2", "y1", "y2")
  )
})

testthat::test_that("merge.ordered_sequence disjoint fast paths match general path", {
  a <- ordered_sequence("a", "b", "c", keys = c(1, 2, 3))
  b <- ordered_sequence("d", "e", "f", keys = c(4, 5, 6))

  testthat::expect_identical(
    unlist(as.list(merge(a, b))),
    c("a", "b", "c", "d", "e", "f")
  )
  testthat::expect_identical(
    unlist(as.list(merge(b, a))),
    c("a", "b", "c", "d", "e", "f")
  )
})

testthat::test_that("merge.ordered_sequence empty short-circuits", {
  a <- ordered_sequence("x", keys = 1)
  testthat::expect_identical(unlist(as.list(merge(ordered_sequence(), a))), "x")
  testthat::expect_identical(unlist(as.list(merge(a, ordered_sequence()))), "x")
  testthat::expect_identical(length(merge(ordered_sequence(), ordered_sequence())), 0L)
})

testthat::test_that("merge.ordered_sequence preserves min_key/max_key", {
  a <- ordered_sequence("a", "b", keys = c(3, 5))
  b <- ordered_sequence("c", "d", keys = c(1, 7))
  m <- merge(a, b)

  testthat::expect_identical(min_key(m), 1)
  testthat::expect_identical(max_key(m), 7)
})

testthat::test_that("merge.ordered_sequence does not mutate sources", {
  a <- ordered_sequence("a", keys = 1)
  b <- ordered_sequence("b", keys = 2)
  merge(a, b)

  testthat::expect_identical(length(a), 1L)
  testthat::expect_identical(length(b), 1L)
})

testthat::test_that("merge.ordered_sequence errors on key-type mismatch", {
  a <- ordered_sequence("a", keys = 1)       # numeric
  b <- ordered_sequence("b", keys = "x")     # character
  testthat::expect_error(merge(a, b), "key type")
})

testthat::test_that("merge.ordered_sequence rejects interval_index input", {
  os <- ordered_sequence("a", keys = 1)
  ix <- interval_index("a", start = 1, end = 2)
  testthat::expect_error(merge(os, ix), "interval_index")
})

# ---- merge.interval_index ---------------------------------------------------

testthat::test_that("merge.interval_index produces start-ascending merge", {
  a <- interval_index("A1", "A2", start = c(1, 5), end = c(4, 8))
  b <- interval_index("B1", "B2", start = c(3, 7), end = c(6, 10))
  m <- merge(a, b)

  testthat::expect_identical(unlist(as.list(m)), c("A1", "B1", "A2", "B2"))
  testthat::expect_identical(length(m), 4L)
})

testthat::test_that("merge.interval_index uses left-biased FIFO on tied starts", {
  a <- interval_index("X1", "X2", start = c(1, 3), end = c(4, 5))
  b <- interval_index("Y1", "Y2", start = c(3, 5), end = c(6, 7))
  m <- merge(a, b)

  testthat::expect_identical(unlist(as.list(m)), c("X1", "X2", "Y1", "Y2"))
})

testthat::test_that("merge.interval_index disjoint fast paths match general path", {
  a <- interval_index("p", "q", start = c(0, 2), end = c(1, 3))
  b <- interval_index("r", "s", start = c(5, 7), end = c(6, 8))

  testthat::expect_identical(
    unlist(as.list(merge(a, b))),
    c("p", "q", "r", "s")
  )
  testthat::expect_identical(
    unlist(as.list(merge(b, a))),
    c("p", "q", "r", "s")
  )
})

testthat::test_that("merge.interval_index preserves min/max endpoints", {
  a <- interval_index("A", "B", start = c(1, 5), end = c(4, 8))
  b <- interval_index("C", "D", start = c(3, 7), end = c(6, 10))
  m <- merge(a, b)

  testthat::expect_identical(min_endpoint(m), 1)
  testthat::expect_identical(max_endpoint(m), 10)
})

testthat::test_that("merge.interval_index point queries work on merged result", {
  a <- interval_index("A", start = 1, end = 5)
  b <- interval_index("B", start = 2, end = 4)
  m <- merge(a, b)

  testthat::expect_identical(
    unlist(as.list(peek_all_point(m, 3))),
    c("A", "B")
  )
})

testthat::test_that("merge.interval_index empty short-circuits", {
  a <- interval_index("A", start = 1, end = 2)
  testthat::expect_identical(unlist(as.list(merge(interval_index(), a))), "A")
  testthat::expect_identical(unlist(as.list(merge(a, interval_index()))), "A")
  testthat::expect_identical(length(merge(interval_index(), interval_index())), 0L)
})

testthat::test_that("merge.interval_index does not mutate sources", {
  a <- interval_index("A", start = 1, end = 2)
  b <- interval_index("B", start = 3, end = 4)
  merge(a, b)

  testthat::expect_identical(length(a), 1L)
  testthat::expect_identical(length(b), 1L)
})

testthat::test_that("merge.interval_index errors on bounds mismatch", {
  ah <- interval_index("x", start = 1, end = 2, default_query_bounds = "[)")
  ac <- interval_index("y", start = 1, end = 2, default_query_bounds = "[]")
  testthat::expect_error(merge(ah, ac), "bounds")
})

testthat::test_that("merge.interval_index errors on endpoint-type mismatch", {
  an <- interval_index("a", start = 1, end = 2)             # numeric
  ac <- interval_index("b", start = "p", end = "q")         # character
  testthat::expect_error(merge(an, ac), "endpoint type")
})

testthat::test_that("merge.interval_index errors on non-interval_index y", {
  a <- interval_index("A", start = 1, end = 2)
  testthat::expect_error(merge(a, flexseq("x")), "interval_ind")
})

library(testthat)
library(Immutables)

# testthat/waldo uses temporary language switching for stable output.
# If LC_ALL is pinned, this emits noisy warnings on every expectation.
Sys.unsetenv("LC_ALL")

test_check("Immutables")

#SO

# recursive fill helper for `.ft_to_list` to avoid repeated list concatenation.
# Runtime: O(n) in number of elements visited.
.ft_to_list_fill <- function(x, st) {
  if(!is_structural_node(x)) {
    st$out[[st$pos]] <- x
    st$pos <- st$pos + 1L
    return(invisible(NULL))
  }
  if(x %isa% Empty) {
    return(invisible(NULL))
  }
  if(x %isa% Single) {
    .ft_to_list_fill(.subset2(x, 1), st)
    return(invisible(NULL))
  }
  if(x %isa% Deep) {
    .ft_to_list_fill(.subset2(x, "prefix"), st)
    .ft_to_list_fill(.subset2(x, "middle"), st)
    .ft_to_list_fill(.subset2(x, "suffix"), st)
    return(invisible(NULL))
  }
  for(el in x) {
    .ft_to_list_fill(el, st)
  }
  invisible(NULL)
}

# flatten a fingertree into its element sequence (left-to-right)
# Runtime: O(n) in number of elements.
.ft_to_list <- function(x) {
  if(!is_structural_node(x)) {
    return(list(x))
  }
  n <- as.integer(node_measure(x, ".size"))
  if(n == 0L) {
    return(list())
  }
  st <- new.env(parent = emptyenv())
  st$out <- vector("list", n)
  st$pos <- 1L
  .ft_to_list_fill(x, st)
  st$out
}

# Runtime: O(k), where k = length(idx).
.ft_assert_int_indices <- function(idx, n) {
  if(is.null(idx)) {
    stop("Index is required.")
  }
  if(!is.numeric(idx) || any(is.na(idx)) || any(idx != as.integer(idx))) {
    stop("Only non-missing integer indices are supported.")
  }
  idx <- as.integer(idx)
  if(length(idx) == 0L) {
    return(idx)
  }
  if(any(idx <= 0L)) {
    stop("Only positive integer indices are supported.")
  }
  if(any(idx > n)) {
    stop("Index out of bounds.")
  }
  idx
}

# Runtime: O(k), where k = length(idx).
.ft_assert_chr_indices <- function(idx) {
  if(is.null(idx)) {
    stop("Index is required.")
  }
  if(!is.character(idx) || any(is.na(idx))) {
    stop("Only non-missing character indices are supported.")
  }
  idx
}

# Runtime: O(n) due `rep_len(..., n)`.
.ft_assert_lgl_indices <- function(idx, n) {
  if(is.null(idx)) {
    stop("Index is required.")
  }
  if(!is.logical(idx)) {
    stop("Only logical indices are supported.")
  }
  if(any(is.na(idx))) {
    stop("Logical indices cannot contain NA.")
  }
  if(length(idx) == 0L) {
    return(logical(0))
  }
  rep_len(idx, n)
}

# Runtime: O(n) worst-case in relevant input/subtree size.
.ft_true_positions <- function(mask) {
  as.integer(which(mask))
}

# minimum character-query width at which full name->position map lookup is used.
# Runtime: O(1).
.ft_name_map_threshold <- function() {
  # Tuned internal constant: switch to one-pass name->position map lookup for
  # medium/large character queries; keep scalar find loop for short vectors.
  8L
}

# Runtime: O(target_len) in recycled output size.
.ft_recycle_values <- function(values, target_len) {
  if(target_len == 0L) {
    return(list())
  }
  if(length(values) == 0L) {
    stop("Replacement has length 0 but selected positions are non-empty.")
  }
  if(length(values) == target_len) {
    return(values)
  }
  rep(values, length.out = target_len)
}

# check whether replacement payloads provide any explicit name hints
# (outer list names, ft_name attrs, or inline scalar names()).
# Runtime: O(k), where k = number of replacement values.
.ft_values_have_name_hints <- function(values) {
  vn <- names(values)
  if(!is.null(vn) && any(!is.na(vn) & vn != "")) {
    return(TRUE)
  }
  if(length(values) == 0L) {
    return(FALSE)
  }
  for(v in values) {
    if(!is.null(.ft_get_name(v)) || !is.null(.ft_name_from_value(v))) {
      return(TRUE)
    }
  }
  FALSE
}

# remove internal naming metadata from user-visible element returns.
# Runtime: O(1).
.ft_strip_name <- function(el) {
  attr(el, "ft_name") <- NULL
  el
}

# prepare selected entries for subset rebuild while preserving name invariants:
# keep names only when all selected entries are named uniquely; otherwise drop all.
# Runtime: O(k), where k = number of selected entries.
.ft_prepare_subset_entries <- function(out) {
  if(length(out) == 0L) {
    return(out)
  }
  nms <- vapply(out, function(el) {
    nm <- .ft_get_name(el)
    if(is.null(nm)) NA_character_ else nm
  }, character(1))
  if(any(is.na(nms)) || any(nms == "") || anyDuplicated(nms) > 0L) {
    return(lapply(out, .ft_strip_name))
  }
  out
}

# collect internal element names in left-to-right order.
# Runtime: O(n).
.ft_collect_names <- function(t) {
  els <- .ft_to_list(t)
  if(length(els) == 0L) {
    return(character(0))
  }
  vapply(els, function(el) {
    nm <- .ft_get_name(el)
    if(is.null(nm)) NA_character_ else nm
  }, character(1))
}

# validate global naming invariant:
# - unnamed tree: .named_count == 0
# - fully named tree: .named_count == .size and names are unique/non-empty
# Runtime: O(n) in the fully-named case (collect/scan all names).
# Intended use: correctness auditing in tests/debugging and explicit validators.
.ft_assert_name_state <- function(t) {
  if(!is_structural_node(t)) {
    return(invisible(TRUE))
  }
  n <- as.integer(node_measure(t, ".size"))
  nn <- as.integer(node_measure(t, ".named_count"))
  if(n == 0L) {
    if(nn != 0L) {
      stop("Invalid name state: empty tree has non-zero named element count.")
    }
    return(invisible(TRUE))
  }
  if(nn == 0L) {
    return(invisible(TRUE))
  }
  if(nn != n) {
    stop("Invalid name state: mixed named and unnamed elements are not allowed.")
  }
  nms <- .ft_collect_names(t)
  if(any(is.na(nms) | nms == "")) {
    stop("Invalid name state: all element names must be non-empty.")
  }
  if(anyDuplicated(nms) > 0L) {
    stop("Invalid name state: element names must be unique.")
  }
  invisible(TRUE)
}

# build a deterministic name -> position map for fully named trees.
# Runtime: O(n).
.ft_name_positions <- function(t) {
  n <- as.integer(node_measure(t, ".size"))
  nn <- as.integer(node_measure(t, ".named_count"))
  if(n == 0L || nn == 0L) {
    stop("Tree has no element names.")
  }
  if(nn != n) {
    stop("Invalid name state: mixed named and unnamed elements are not allowed.")
  }
  nms <- .ft_collect_names(t)
  pos <- seq_len(length(nms))
  names(pos) <- nms
  pos
}

# build name->position map, using C++ when available.
# Runtime: O(n).
.ft_name_positions_fast <- function(t) {
  if(.ft_cpp_enabled()) {
    return(.ft_cpp_name_positions(t))
  }
  .ft_name_positions(t)
}

# measure contribution of one child in terms of leaf-element count.
# Runtime: O(1) using cached `.size` for structural children.
.ft_child_size <- function(el) {
  if(is_structural_node(el)) {
    return(as.integer(node_measure(el, ".size")))
  }
  1L
}

# recursive linear search for one name with early exit.
# Runtime: O(n) worst-case, O(k) until first match.
.ft_find_name_position_impl <- function(x, target, offset = 0L) {
  if(!is_structural_node(x)) {
    nm <- .ft_get_name(x)
    if(!is.null(nm) && identical(nm, target)) {
      return(offset + 1L)
    }
    return(NA_integer_)
  }
  if(x %isa% Empty) {
    return(NA_integer_)
  }
  if(x %isa% Single) {
    return(.ft_find_name_position_impl(.subset2(x, 1), target, offset))
  }
  if(x %isa% Deep) {
    p <- .ft_find_name_position_impl(.subset2(x, "prefix"), target, offset)
    if(!is.na(p)) {
      return(p)
    }
    offset <- offset + as.integer(node_measure(.subset2(x, "prefix"), ".size"))
    p <- .ft_find_name_position_impl(.subset2(x, "middle"), target, offset)
    if(!is.na(p)) {
      return(p)
    }
    offset <- offset + as.integer(node_measure(.subset2(x, "middle"), ".size"))
    return(.ft_find_name_position_impl(.subset2(x, "suffix"), target, offset))
  }
  for(el in x) {
    p <- .ft_find_name_position_impl(el, target, offset)
    if(!is.na(p)) {
      return(p)
    }
    offset <- offset + .ft_child_size(el)
  }
  NA_integer_
}

# find one name position without constructing full name->position map.
# Runtime: O(n) worst-case.
.ft_find_name_position <- function(t, target) {
  .ft_find_name_position_impl(t, target, 0L)
}

# match requested names to positions. If strict_missing is FALSE, missing names
# are represented as NA integer placeholders.
# Runtime: adaptive by query width:
# - scalar query: O(n_lookup) (typically O(n) worst-case linear name scan),
# - short vector queries: O(k * n_lookup),
# - wide queries: O(n + k) via one-pass name->position map.
# where n = tree size and k = length(idx).
.ft_match_name_indices <- function(t, idx, strict_missing = FALSE) {
  n <- as.integer(node_measure(t, ".size"))
  nn <- as.integer(node_measure(t, ".named_count"))
  if(n == 0L || nn == 0L) {
    stop("Tree has no element names.")
  }
  if(nn != n) {
    stop("Invalid name state: mixed named and unnamed elements are not allowed.")
  }
  if(length(idx) == 1L) {
    p <- if(.ft_cpp_enabled()) .ft_cpp_find_name_position(t, idx[[1]]) else .ft_find_name_position(t, idx[[1]])
    if(isTRUE(strict_missing) && is.na(p)) {
      stop("Unknown element name(s): ", idx[[1]])
    }
    return(as.integer(p))
  }

  if(length(idx) >= .ft_name_map_threshold()) {
    name_to_pos <- .ft_name_positions_fast(t)
    pos <- unname(name_to_pos[idx])
    if(isTRUE(strict_missing) && any(is.na(pos))) {
      missing_names <- unique(idx[is.na(pos)])
      stop("Unknown element name(s): ", paste(missing_names, collapse = ", "))
    }
    return(as.integer(pos))
  }

  pos <- integer(length(idx))
  for(k in seq_along(idx)) {
    p <- if(.ft_cpp_enabled()) .ft_cpp_find_name_position(t, idx[[k]]) else .ft_find_name_position(t, idx[[k]])
    if(isTRUE(strict_missing) && is.na(p)) {
      stop("Unknown element name(s): ", idx[[k]])
    }
    pos[[k]] <- p
  }
  as.integer(pos)
}

# internal positional read that preserves ft_name metadata.
# Runtime: O(log n) via indexed locate path (C++ or R fallback).
.ft_get_elem_at <- function(x, idx) {
  if(.ft_cpp_enabled()) {
    return(.ft_cpp_get_by_index(x, idx))
  }
  ctx <- resolve_named_monoid(x, ".size")
  hit <- locate_tree_impl_fast(
    function(v) v >= idx,
    ctx$monoid$i,
    x,
    ctx$monoids,
    ctx$monoid,
    ".size",
    0L
  )
  if(!isTRUE(hit$found)) {
    stop("Index out of bounds.")
  }
  hit$value
}

# bulk positional read helper preserving ft_name metadata.
# Runtime: O(k log n) fallback, where k = length(idx); C++ path is typically faster.
.ft_get_elems_at <- function(x, idx) {
  if(length(idx) == 0L) {
    return(list())
  }
  if(.ft_cpp_enabled()) {
    return(.ft_cpp_get_many_by_index(x, idx))
  }
  lapply(idx, function(i) .ft_get_elem_at(x, i))
}

# extract a name carried directly by a scalar replacement value.
# Runtime: O(1).
.ft_name_from_value <- function(el) {
  nms <- names(el)
  if(is.null(nms) || length(nms) != 1L) {
    return(NULL)
  }
  .ft_normalize_name(nms[[1]])
}

# effective replacement name: explicit outer list name first, then existing
# internal name, then name carried by the replacement value itself.
# Runtime: O(1).
.ft_effective_name <- function(el, explicit_name = NULL) {
  nm <- .ft_normalize_name(explicit_name)
  if(is.null(nm)) {
    nm <- .ft_get_name(el)
  }
  if(is.null(nm)) {
    nm <- .ft_name_from_value(el)
  }
  nm
}

# update name mapping for sequential replacement and enforce uniqueness.
# Runtime: O(k), where k is current map size.
.ft_update_name_map <- function(name_to_pos, name_vec, pos, nm) {
  old_nm <- name_vec[[pos]]
  if(identical(nm, old_nm)) {
    return(list(name_to_pos = name_to_pos, name_vec = name_vec))
  }
  existing <- if(nm %in% names(name_to_pos)) as.integer(name_to_pos[[nm]]) else NULL
  if(!is.null(existing) && existing != pos) {
    stop("Element names must be unique.")
  }
  name_to_pos <- name_to_pos[names(name_to_pos) != old_nm]
  name_to_pos[[nm]] <- pos
  name_vec[[pos]] <- nm
  list(name_to_pos = name_to_pos, name_vec = name_vec)
}

#' Flexseq Indexing
#'
#' Index, replace, and extract elements of a `flexseq` by position or name.
#'
#' @method [ flexseq
#' @param x A `flexseq`.
#' @param i Positive integer indices, character element names, or logical mask.
#'   For `[[`, a single integer or character name.
#' @param ... Unused.
#' @return For `[`: a new `flexseq` containing selected elements in query order.
#'   For character indexing, missing names are represented as `NULL` elements.
#' @details
#' Selector behavior:
#' - Integer indexing (`[`) returns elements in the requested order.
#' - Character indexing (`[`) returns elements in requested name order; unknown
#'   names become `NULL` elements in the output.
#' - Logical indexing (`[`) is positional and follows logical-mask selection.
#'
#' Replacement behavior:
#' - `[<-` is persistent and recycles `value` to the number of selected
#'   positions (with standard R recycling warnings where applicable).
#' - `[[<-` replaces one element; assigning `NULL` removes that element.
#' @examples
#' x <- as_flexseq(letters[1:6])
#' x
#'
#' x2 <- x[c(2, 4, 6)]
#' x2
#'
#' # named lookups return NULL for missing names
#' x3 <- as_flexseq(setNames(as.list(letters[1:4]), c("w", "x", "y", "z")))
#' x4 <- x3[c("y", "missing", "w")]
#' x4
#'
#' # logical indexing supports recycling
#' x[c(TRUE, FALSE)]
#' @export
# Runtime: integer/logical reads O(k log n) in fallback paths (C++ paths are
# typically faster), where k = selected positions. Character reads are adaptive:
# O(k * n_lookup) for short queries, O(n + k) for wider queries via name map.
# In all branches, rebuilding the output tree from selected elements is linear
# in output size.
`[.flexseq` <- function(x, i, ...) {
  if(missing(i)) {
    return(x)
  }
  ms <- resolve_tree_monoids(x, required = TRUE)
  n <- as.integer(node_measure(x, ".size"))

  if(is.logical(i)) {
    mask <- .ft_assert_lgl_indices(i, n)
    idx <- .ft_true_positions(mask)
    if(length(idx) == 0L) {
      return(.ft_restore_subclass(empty_tree(monoids = ms), x, context = "["))
    }
    out <- .ft_prepare_subset_entries(.ft_get_elems_at(x, idx))
    return(.ft_restore_subclass(tree_from(out, monoids = ms), x, context = "["))
  }

  if(is.character(i)) {
    idx <- .ft_assert_chr_indices(i)
    if(length(idx) == 0L) {
      return(.ft_restore_subclass(empty_tree(monoids = ms), x, context = "["))
    }

    strict_missing <- inherits(x, "priority_queue")
    pos <- .ft_match_name_indices(x, idx, strict_missing = strict_missing)
    out <- vector("list", length(pos))
    valid <- which(!is.na(pos))
    if(length(valid) > 0L) {
      vals <- .ft_get_elems_at(x, as.integer(pos[valid]))
      for(j in seq_along(valid)) {
        out[[valid[[j]]]] <- vals[[j]]
      }
      out <- .ft_prepare_subset_entries(out)
    } else {
      out <- .ft_prepare_subset_entries(out)
    }
    return(.ft_restore_subclass(tree_from(out, monoids = ms), x, context = "["))
  }

  idx <- .ft_assert_int_indices(i, n)
  if(length(idx) == 0L) {
    return(.ft_restore_subclass(empty_tree(monoids = ms), x, context = "["))
  }
  out <- .ft_prepare_subset_entries(.ft_get_elems_at(x, idx))
  .ft_restore_subclass(tree_from(out, monoids = ms), x, context = "[")
}

# Runtime: O(k), where k = length(pos).
.ord_normalize_selector_positions <- function(pos) {
  pos <- as.integer(pos)
  if(length(pos) == 0L) {
    return(pos)
  }
  if(any(is.na(pos))) {
    stop("Ordered subsetting does not allow missing indices.")
  }
  if(anyDuplicated(pos) > 0L) {
    stop("Ordered subsetting does not allow duplicate indices.")
  }
  sorted <- sort(pos)
  if(length(sorted) > 1L && !identical(pos, sorted)) {
    warning(
      "Ordered subsetting canonicalizes selector order; pre-sort and unique selectors to silence this warning.",
      call. = FALSE
    )
  }
  sorted
}

#' @rdname sub-.flexseq
#' @method [[ flexseq
#' @return For `[[`: the extracted element (internal name metadata is removed).
#' @examples
#'
#' # [[ extracts one element
#' x <- as_flexseq(letters[1:5])
#' x[[3]]
#'
#' x2 <- as_flexseq(setNames(as.list(letters[1:3]), c("a1", "a2", "a3")))
#' x2[["a2"]]
#' @export
# Runtime: O(log n) for integer lookup; O(n) for name lookup.
`[[.flexseq` <- function(x, i, ...) {
  if(is.character(i) && length(i) == 1L && !is.na(i)) {
    pos <- .ft_match_name_indices(x, i, strict_missing = TRUE)
    return(.ft_strip_name(.ft_get_elem_at(x, pos)))
  }
  n <- .ft_size(x)
  idx <- .ft_assert_int_indices(i, n)
  if(length(idx) != 1L) {
    stop("[[ expects exactly one index.")
  }

  .ft_strip_name(.ft_get_elem_at(x, idx))
}

#' @rdname sub-.flexseq
#' @method [<- flexseq
#' @param value Replacement values; recycled to selected index length.
#' @return For `[<-`: a new `flexseq` with selected elements replaced.
#' @examples
#'
#' # [<- replaces selected elements
#' x <- as_flexseq(1:6)
#' x
#'
#' x2 <- x
#' x2[c(2, 5)] <- list(20, 50)
#' x2
#'
#' # character replacement uses element names
#' x3 <- as_flexseq(setNames(as.list(1:4), c("a", "b", "c", "d")))
#' x3[c("d", "a")] <- list(40, 10)
#' x3
#'
#' # logical replacement supports recycling
#' x4 <- x
#' x4[c(TRUE, FALSE, TRUE, FALSE, TRUE, FALSE)] <- list(1)
#' x4
#' @export
# Runtime: sparse replacement path O(k * n) via repeated point updates; dense
# path O(n + k) via flatten + single rebuild, where n = tree size and
# k = number of replaced positions.
`[<-.flexseq` <- function(x, i, value) {
  ms <- resolve_tree_monoids(x, required = TRUE)
  vals <- as.list(value)
  n <- as.integer(node_measure(x, ".size"))

  if(is.logical(i)) {
    mask <- .ft_assert_lgl_indices(i, n)
    idx <- .ft_true_positions(mask)
    vals2 <- .ft_recycle_values(vals, length(idx))
    return(`[<-.flexseq`(x, idx, vals2))
  }

  if(is.character(i)) {
    idx <- .ft_assert_chr_indices(i)
    vals <- .ft_recycle_values(vals, length(idx))
    if(length(idx) == 0L) {
      return(x)
    }
    xs <- .ft_to_list(x)
    has_name_hints <- .ft_values_have_name_hints(vals)
    if(!has_name_hints) {
      pos <- .ft_match_name_indices(x, idx, strict_missing = TRUE)
      for(k in seq_along(pos)) {
        xs[pos[[k]]] <- list(.ft_set_name(vals[[k]], idx[[k]]))
      }
      return(.ft_restore_subclass(tree_from(xs, monoids = ms), x, context = "[<-"))
    }

    name_to_pos <- .ft_name_positions_fast(x)
    pos <- unname(name_to_pos[idx])
    if(any(is.na(pos))) {
      missing_names <- unique(idx[is.na(pos)])
      stop("Unknown element name(s): ", paste(missing_names, collapse = ", "))
    }
    pos <- as.integer(pos)
    vn <- names(vals)
    name_vec <- names(name_to_pos)
    for(k in seq_along(pos)) {
      name_hint <- if(!is.null(vn)) vn[[k]] else NULL
      nm <- .ft_effective_name(vals[[k]], name_hint)
      if(is.null(nm)) {
        nm <- idx[[k]]
      }
      upd <- .ft_update_name_map(name_to_pos, name_vec, pos[[k]], nm)
      name_to_pos <- upd$name_to_pos
      name_vec <- upd$name_vec
      xs[pos[[k]]] <- list(.ft_set_name(vals[[k]], nm))
    }
    return(.ft_restore_subclass(tree_from(xs, monoids = ms), x, context = "[<-"))
  }

  idx <- .ft_assert_int_indices(i, n)
  vals <- .ft_recycle_values(vals, length(idx))
  if(length(idx) == 0L) {
    return(x)
  }
  vn <- names(vals)
  has_name_hints <- .ft_values_have_name_hints(vals)
  n_named <- as.integer(node_measure(x, ".named_count"))
  use_name_map <- n > 0L && n_named == n

  # Sparse vector replacement is usually faster as repeated point updates.
  # Dense replacement is faster as flatten+single rebuild.
  sparse_cutoff <- max(8L, as.integer(sqrt(max(1L, n))))
  if(length(idx) <= sparse_cutoff) {
    out <- x
    if(use_name_map && has_name_hints) {
      name_to_pos <- .ft_name_positions_fast(x)
      name_vec <- names(name_to_pos)
    }
    for(k in seq_along(idx)) {
      name_hint <- if(!is.null(vn)) vn[[k]] else NULL
      nm <- .ft_effective_name(vals[[k]], name_hint)
      if(is.null(nm)) {
        old <- .ft_get_elem_at(out, idx[[k]])
        nm <- .ft_get_name(old)
      }
      if(use_name_map && has_name_hints && !is.null(nm)) {
        upd <- .ft_update_name_map(name_to_pos, name_vec, idx[[k]], nm)
        name_to_pos <- upd$name_to_pos
        name_vec <- upd$name_vec
      }
      out[[idx[[k]]]] <- .ft_set_name(vals[[k]], nm)
    }
    return(.ft_restore_subclass(out, x, context = "[<-"))
  }

  xs <- .ft_to_list(x)
  if(use_name_map && has_name_hints) {
    name_to_pos <- .ft_name_positions_fast(x)
    name_vec <- names(name_to_pos)
  }
  for(k in seq_along(idx)) {
    name_hint <- if(!is.null(vn)) vn[[k]] else NULL
    nm <- .ft_effective_name(vals[[k]], name_hint)
    if(is.null(nm)) {
      old <- xs[[idx[[k]]]]
      nm <- .ft_get_name(old)
    }
    if(use_name_map && has_name_hints && !is.null(nm)) {
      upd <- .ft_update_name_map(name_to_pos, name_vec, idx[[k]], nm)
      name_to_pos <- upd$name_to_pos
      name_vec <- upd$name_vec
    }
    v <- .ft_set_name(vals[[k]], nm)
    xs[idx[[k]]] <- list(v)
  }
  .ft_restore_subclass(tree_from(xs, monoids = ms), x, context = "[<-")
}

#' @rdname sub-.flexseq
#' @method [[<- flexseq
#' @return For `[[<-`: a new `flexseq` with one element replaced.
#' @examples
#'
#' # [[<- replaces one element
#' x <- as_flexseq(letters[1:4])
#' x2 <- x
#' x2[[2]] <- "ZZ"
#' x2
#'
#' x3 <- as_flexseq(setNames(as.list(1:3), c("x", "y", "z")))
#' x3[["y"]] <- 99
#' x3
#'
#' # assigning NULL removes one element
#' x4 <- as_flexseq(letters[1:4])
#' x4[[2]] <- NULL
#' x4
#' @export
# Runtime: O(log n) via split + push_back + concat.
`[[<-.flexseq` <- function(x, i, value) {
  if(is.character(i) && length(i) == 1L && !is.na(i)) {
    pos <- .ft_match_name_indices(x, i, strict_missing = TRUE)
    if(is.null(value)) {
      return(`[[<-.flexseq`(x, pos, NULL))
    }
    nm <- .ft_effective_name(value)
    if(is.null(nm)) {
      nm <- i
    }
    v <- .ft_set_name(value, nm)
    return(`[[<-.flexseq`(x, pos, v))
  }
  ms <- attr(x, "monoids", exact = TRUE)
  if(is.null(ms)) stop("Tree has no monoids attribute.")
  n <- .ft_size(x)
  idx <- .ft_assert_int_indices(i, n)
  if(length(idx) != 1L) {
    stop("[[<- expects exactly one index.")
  }
  use_cpp <- .ft_cpp_can_use(ms)
  .split_here <- if(use_cpp) {
    function(i) .ft_cpp_split_at_index(x, i, ms)
  } else {
    function(i) split_around_by_predicate(x, function(v) v >= i, ".size")
  }
  if(is.null(value)) {
    s <- .split_here(idx)
    concat_out <- if(use_cpp) .ft_cpp_concat(s$left, s$right, ms) else .ft_concat_same_monoids(s$left, s$right, ms)
    return(.ft_restore_subclass(concat_out, x, context = "[[<-"))
  }
  n_named <- .ft_named_count(x)
  nm <- .ft_get_name_fast(value)
  if(is.null(nm)) nm <- .ft_name_from_value(value)
  old <- NULL
  if(is.null(nm) && n_named > 0L) {
    old <- .ft_get_elem_at(x, idx)
    nm <- .ft_get_name_fast(old)
  }
  if(n > 0L && n_named == n && !is.null(nm)) {
    if(is.null(old)) {
      old <- .ft_get_elem_at(x, idx)
    }
    old_nm <- .ft_get_name_fast(old)
    if(is.null(old_nm)) {
      stop("Invalid name state: mixed named and unnamed elements are not allowed.")
    }
    if(!identical(nm, old_nm)) {
      existing <- if(.ft_cpp_enabled()) .ft_cpp_find_name_position(x, nm) else .ft_find_name_position(x, nm)
      if(!is.na(existing) && as.integer(existing) != as.integer(idx)) {
        stop("Element names must be unique.")
      }
    }
  }
  value <- .ft_set_name_fast(value, nm)

  s <- .split_here(idx)
  left_plus <- .ft_push_back_raw(s$left, value, ms)
  concat_out <- if(use_cpp) .ft_cpp_concat(left_plus, s$right, ms) else .ft_concat_same_monoids(left_plus, s$right, ms)
  .ft_restore_subclass(concat_out, x, context = "[[<-")
}

#SO

# Point containment predicate for fallback/non-fast endpoint types.
# **Inputs:** scalar `start`, `end`, `point`; scalar `bounds` string; scalar `endpoint_type`.
# **Outputs:** scalar logical; TRUE when `point` is inside [start,end] under `bounds`.
# **Used by:** .ivx_spec_point() fallback leaf matcher.
.ivx_contains_point <- function(start, end, point, bounds, endpoint_type) {
  f <- .ivx_bounds_flags(bounds)
  cmp_lo <- .ivx_compare_scalar(point, start, endpoint_type)
  cmp_hi <- .ivx_compare_scalar(point, end, endpoint_type)

  left_ok <- if(isTRUE(f$include_start)) cmp_lo >= 0L else cmp_lo > 0L
  right_ok <- if(isTRUE(f$include_end)) cmp_hi <= 0L else cmp_hi < 0L
  isTRUE(left_ok && right_ok)
}

# Runtime: O(1).
# Interval overlap predicate for fallback/non-fast endpoint types.
# **Inputs:** scalar interval endpoints + bounds + endpoint_type.
# **Outputs:** scalar logical; TRUE when intervals overlap under `bounds`.
# **Used by:** .ivx_spec_overlaps(), .ivx_spec_containing(), .ivx_spec_within() fallback leaf matchers.
.ivx_overlaps_interval <- function(a_start, a_end, b_start, b_end, bounds, endpoint_type) {
  f <- .ivx_bounds_flags(bounds)
  touching_is_overlap <- isTRUE(f$include_start) && isTRUE(f$include_end)

  cmp_aend_bstart <- .ivx_compare_scalar(a_end, b_start, endpoint_type)
  cmp_bend_astart <- .ivx_compare_scalar(b_end, a_start, endpoint_type)

  a_before_b <- if(touching_is_overlap) cmp_aend_bstart < 0L else cmp_aend_bstart <= 0L
  b_before_a <- if(touching_is_overlap) cmp_bend_astart < 0L else cmp_bend_astart <= 0L

  !isTRUE(a_before_b || b_before_a)
}

# Runtime: O(1).
# Interval containment predicate for fallback/non-fast endpoint types.
# **Inputs:** scalar container endpoints, scalar inner endpoints, scalar endpoint_type.
# **Outputs:** scalar logical; TRUE when [inner_start,inner_end] is inside container interval.
# **Used by:** .ivx_spec_containing(), .ivx_spec_within() fallback leaf matchers.
.ivx_contains_interval <- function(container_start, container_end, inner_start, inner_end, endpoint_type) {
  left_ok <- .ivx_compare_scalar(container_start, inner_start, endpoint_type) <= 0L
  right_ok <- .ivx_compare_scalar(container_end, inner_end, endpoint_type) >= 0L
  isTRUE(left_ok && right_ok)
}

# Runtime: O(k) in matched entry count.
# Rebuilds an interval_index from entry payloads while preserving wrapper metadata.
# **Inputs:** `x` interval_index template; `entries` list of interval entry records.
# **Outputs:** interval_index containing `entries` in order (or empty-like if none).
# **Used by:** .ivx_run_relation_query() and all-match pop rebuild path.
.ivx_slice_entries <- function(x, entries) {
  if(length(entries) == 0L) {
    return(.ivx_empty_like(x))
  }

  ms <- resolve_tree_monoids(x, required = TRUE)
  tree <- .ivx_tree_from_ordered_entries(entries, monoids = ms)
  .ivx_wrap_like(x, tree)
}

# Runtime: O(log n) near locate point depth.
# Computes lower/upper bound index from prepared start key and strictness.
# **Inputs:** `x` interval_index; scalar `key`; scalar logical `strict`.
# **Outputs:** 1-based integer insertion/bound index in [1, n+1].
# **Used by:** .ivx_query_candidate_span().
.ivx_bound_index_prepared <- function(x, key, strict = FALSE) {
  n <- length(x)
  if(n == 0L) {
    return(1L)
  }

  pred <- if(!isTRUE(strict)) {
    function(v) {
      isTRUE(v$has) && .ivx_compare_scalar_fast(v$start, key, v$endpoint_type) >= 0L
    }
  } else {
    function(v) {
      isTRUE(v$has) && .ivx_compare_scalar_fast(v$start, key, v$endpoint_type) > 0L
    }
  }

  loc <- locate_by_predicate(x, pred, ".ivx_max_start", include_metadata = TRUE)
  if(!isTRUE(loc$found)) {
    return(as.integer(n + 1L))
  }
  as.integer(loc$metadata$index)
}

# Runtime: O(log n) near bound locate points.
# Computes candidate start-index window from query spec bounds.
# **Inputs:** `x` interval_index; optional scalar `lower`/`upper`; scalar logical strict flags.
# **Outputs:** list(start, end_excl, empty) describing candidate index span.
# **Used by:** .ivx_run_relation_query().
.ivx_query_candidate_span <- function(x, lower = NULL, lower_strict = FALSE, upper = NULL, upper_strict = FALSE) {
  n <- length(x)
  if(n == 0L) {
    return(list(start = 1L, end_excl = 1L, empty = TRUE))
  }

  start <- if(is.null(lower)) {
    1L
  } else {
    # determine start index; if strict == TRUE then first index with start > key,
    # if FALSE then first index with start >= key
    .ivx_bound_index_prepared(x, lower, strict = isTRUE(lower_strict))
  }

  end_excl <- if(is.null(upper)) {
    as.integer(n + 1L)
  } else {
    # upper_strict means use first start >= upper as exclusive bound.
    .ivx_bound_index_prepared(x, upper, strict = !isTRUE(upper_strict))
  }

  if(start > n || end_excl <= start) {
    # empty span
    return(list(start = as.integer(start), end_excl = as.integer(end_excl), empty = TRUE))
  }

  list(start = as.integer(start), end_excl = as.integer(end_excl), empty = FALSE)
}

# Runtime: O(1).
# Returns subtree element count using `.size` measure (or 1 for leaves).
# **Inputs:** structural node or leaf entry.
# **Outputs:** scalar integer subtree size.
# **Used by:** .ivx_collect_query() when pruning whole subtrees.
.ivx_subtree_size <- function(node) {
  if(!.ivx_is_structural_fast(node)) {
    return(1L)
  }
  as.integer(node_measure(node, ".size"))
}

# Runtime: O(log n) near split point depth.
# Splits an interval_index at 1-based position into left/right interval_index trees.
# **Inputs:** `x` interval_index; scalar integer-like `index`.
# **Outputs:** list(left, right) interval_index fragments.
# **Used by:** .ivx_partition_span().
.ivx_split_at_index <- function(x, index) {
  n <- length(x)
  idx <- as.integer(index)
  if(is.na(idx)) {
    stop("`index` must be a non-missing integer.")
  }

  if(n == 0L) {
    empty <- .ivx_empty_like(x)
    return(list(left = empty, right = empty))
  }
  if(idx <= 1L) {
    return(list(left = .ivx_empty_like(x), right = x))
  }
  if(idx > n) {
    return(list(left = x, right = .ivx_empty_like(x)))
  }

  s <- split_by_predicate(x, function(v) v >= idx, ".size")
  list(
    left = .ivx_wrap_like(x, s$left),
    right = .ivx_wrap_like(x, s$right)
  )
}

# Runtime: O(log n) near boundary split points.
# Splits full tree into left/candidate/right partitions for query scanning.
# **Inputs:** `x` interval_index; `span` list with `start` and `end_excl`.
# **Outputs:** list(left, candidate, right) interval_index fragments.
# **Used by:** .ivx_run_relation_query() non-window path.
.ivx_partition_span <- function(x, span) {
  s1 <- .ivx_split_at_index(x, span$start)
  left <- s1$left
  tail <- s1$right
  cand_len <- as.integer(max(0L, span$end_excl - span$start))

  if(cand_len <= 0L) {
    return(list(left = left, candidate = .ivx_empty_like(x), right = tail))
  }

  s2 <- .ivx_split_at_index(tail, cand_len + 1L)
  list(left = left, candidate = s2$left, right = s2$right)
}

# Runtime: O(1).
# Concatenates three interval_index fragments and restores wrapper metadata.
# **Inputs:** interval_index `template`; structural trees `left`, `middle`, `right`.
# **Outputs:** interval_index rebuilt as left + middle + right.
# **Used by:** all-match pop rebuild to construct remaining tree.
.ivx_concat3_like <- function(template, left, middle, right) {
  out <- concat_trees(left, middle)
  out <- concat_trees(out, right)
  .ivx_wrap_like(template, out)
}

# Runtime: O(1).
# Standardized miss return for peek relation endpoints.
# **Inputs:** `x` interval_index; `which` in {"first","all"}.
# **Outputs:** NULL for first; empty interval_index for all.
# **Used by:** .ivx_run_relation_query() miss branches.
.ivx_query_peek_miss <- function(x, which = c("first", "all")) {
  which <- match.arg(which)
  if(identical(which, "all")) {
    return(.ivx_empty_like(x))
  }
  NULL
}

# Runtime: O(1).
# Standardized miss return for pop relation endpoints.
# **Inputs:** `x` interval_index; `which` in {"first","all"}.
# **Outputs:** scalar miss for first; bulk miss for all with empty elements slice.
# **Used by:** .ivx_run_relation_query() miss branches.
.ivx_query_pop_miss <- function(x, which = c("first", "all")) {
  which <- match.arg(which)
  if(identical(which, "all")) {
    return(list(elements = .ivx_empty_like(x), remaining = x))
  }
  list(value = NULL, start = NULL, end = NULL, remaining = x)
}

# Runtime: O(s), where s is traversed subtree size after aggressive pruning 
# based on subtree span and query span.
# Traverses candidate tree with spec predicates and optional unmatched collection.
# **Inputs:**
#
# - `tree`: structural candidate subtree.
# - `no_match_subtree`: function(node) -> logical prune decision.
# - `leaf_match`: function(entry) -> logical exact match decision.
# - scalar logical flags `collect_unmatched`, `stop_after_first`.
#
# **Outputs:**
#
# - list(first_found, first_index, first_entry, matched_entries, unmatched_entries).
#
# **Used by:** .ivx_run_relation_query() traversal path.
.ivx_collect_query <- function(
    tree,
    no_match_subtree,
    leaf_match,
    collect_unmatched = FALSE,
    stop_after_first = FALSE
) {
  # Accumulator state for traversal order, first-hit tracking, and optional
  # matched/unmatched entry capture.
  st <- new.env(parent = emptyenv())
  st$index <- 0L
  st$first_index <- NULL
  st$first_entry <- NULL
  st$m <- 0L
  st$u <- 0L
  st$matched <- list()
  st$unmatched <- list()

  append_many <- function(bucket, entries) {
    if(length(entries) == 0L) {
      return(invisible(NULL))
    }
    if(identical(bucket, "unmatched")) {
      for(el in entries) {
        st$u <- st$u + 1L
        st$unmatched[[st$u]] <- el
      }
      return(invisible(NULL))
    }
    for(el in entries) {
      st$m <- st$m + 1L
      st$matched[[st$m]] <- el
    }
    invisible(NULL)
  }

  walk <- function(node) {
    # Early stop for first-hit queries once a hit has been recorded.
    if(isTRUE(stop_after_first) && !is.null(st$first_index)) {
      return(invisible(NULL))
    }

    # Leaf entry path: evaluate exact predicate and capture into requested
    # output buckets.
    if(!.ivx_is_structural_fast(node)) {
      st$index <- st$index + 1L
      hit <- isTRUE(leaf_match(node))

      if(isTRUE(hit)) {
        if(is.null(st$first_index)) {
          st$first_index <- st$index
          st$first_entry <- node
        }
        st$m <- st$m + 1L
        st$matched[[st$m]] <- node
      } else if(isTRUE(collect_unmatched)) {
        st$u <- st$u + 1L
        st$unmatched[[st$u]] <- node
      }
      return(invisible(NULL))
    }

    # Prune only at Deep nodes where subtree measures summarize meaningful
    # spans; pruning tiny Digit/Node fragments often costs more than it saves.
    can_prune <- inherits(node, "Deep")
    if(isTRUE(can_prune) && isTRUE(no_match_subtree(node))) {
      # Pruned subtree contributes only index offset; optionally capture
      # unmatched entries for all-pop rebuild path.
      sz <- .ivx_subtree_size(node)
      if(isTRUE(collect_unmatched) && sz > 0L) {
        append_many("unmatched", .ivx_entries(node))
      }
      st$index <- st$index + sz
      return(invisible(NULL))
    }

    # Structural descent in left-to-right entry order.
    if(inherits(node, "Empty")) {
      return(invisible(NULL))
    }
    if(inherits(node, "Single")) {
      walk(.subset2(node, 1L))
      return(invisible(NULL))
    }
    if(inherits(node, "Deep")) {
      walk(.subset2(node, "prefix"))
      walk(.subset2(node, "middle"))
      walk(.subset2(node, "suffix"))
      return(invisible(NULL))
    }

    for(el in node) {
      walk(el)
    }
    invisible(NULL)
  }

  walk(tree)

  # Return traversal summary used by peek/pop execution paths.
  list(
    first_found = !is.null(st$first_index),
    first_index = st$first_index,
    first_entry = st$first_entry,
    matched_entries = if(st$m > 0L) st$matched[seq_len(st$m)] else list(),
    unmatched_entries = if(st$u > 0L) st$unmatched[seq_len(st$u)] else list()
  )
}

# Runtime: O(log n + s), where s is traversed candidate subtree size after pruning.
# Main executor for interval relation queries.
# **Inputs:**
#
# - `x`: interval_index.
# - `spec`: query-spec list from .ivx_spec_*.
# - `mode`: "peek" or "pop".
# - `which`: "first" or "all".
#
# **Outputs:**
#
# - peek/first: element value or NULL.
# - peek/all: interval_index slice of matches (possibly empty).
# - pop/first: list(value,start,end,remaining).
# - pop/all: list(elements=<interval_index>, remaining=<interval_index>).
#
# **Used by:** public query API (peek_*/pop_* endpoints).
.ivx_run_relation_query <- function(x, spec, mode = c("peek", "pop"), which = c("first", "all")) {
  mode <- match.arg(mode)
  which <- match.arg(which)
  ms <- resolve_tree_monoids(x, required = TRUE)
  use_window_fast_path <- isTRUE(.ft_cpp_can_use(ms))

  # Phase 1: derive candidate start-index window span from query-spec bounds.
  span <- .ivx_query_candidate_span(
    x,
    lower = spec$lower,
    lower_strict = spec$lower_strict,
    upper = spec$upper,
    upper_strict = spec$upper_strict
  )
  # if the span is empty we return no data (depending on peek/pop/which behavior)
  if(isTRUE(span$empty)) {
    if(identical(mode, "peek")) {
      return(.ivx_query_peek_miss(x, which = which))
    }
    return(.ivx_query_pop_miss(x, which = which))
  }

  # Phase 2: C++-friendly window scan fast path for peek and first-hit pop.
  # When C++ backend is active, candidate-window scanning has lower constants
  # for read-heavy/first-hit queries than split+rebuild traversal.
  if(isTRUE(use_window_fast_path) && (identical(mode, "peek") || identical(which, "first"))) {
    entries <- .ft_get_elems_at(x, seq.int(span$start, span$end_excl - 1L))
    n <- length(entries)
    if(n == 0L) {
      if(identical(mode, "peek")) {
        return(.ivx_query_peek_miss(x, which = which))
      }
      return(.ivx_query_pop_miss(x, which = which))
    }

    first_i <- NULL
    hit <- if(!identical(which, "first")) logical(n) else NULL
    for(i in seq_len(n)) {
      ok <- isTRUE(spec$leaf_match(entries[[i]]))
      if(ok) {
        if(is.null(first_i)) {
          first_i <- i
          if(identical(which, "first")) {
            break
          }
        }
      }
      if(!is.null(hit)) {
        hit[[i]] <- ok
      }
    }

    if(is.null(first_i)) {
      if(identical(mode, "peek")) {
        return(.ivx_query_peek_miss(x, which = which))
      }
      return(.ivx_query_pop_miss(x, which = which))
    }

    # First-hit branches can return value directly (peek) or pop by absolute
    # index without rebuilding candidate partitions.
    if(identical(which, "first")) {
      if(identical(mode, "peek")) {
        return(entries[[first_i]]$value)
      }
      abs_idx <- as.integer(span$start + first_i - 1L)
      return(.ivx_pop_positions(x, abs_idx, which = "first"))
    }

    matched_entries <- entries[hit]
    if(length(matched_entries) == 0L) {
      return(.ivx_query_peek_miss(x, which = which))
    }
    return(.ivx_slice_entries(x, matched_entries))
  }

  # Phase 3: split into left/candidate/right and traverse only candidate tree.
  parts <- .ivx_partition_span(x, span)
  cand <- parts$candidate
  if(length(cand) == 0L) {
    if(identical(mode, "peek")) {
      return(.ivx_query_peek_miss(x, which = which))
    }
    return(.ivx_query_pop_miss(x, which = which))
  }

  # Phase 4A: peek branches (first/all) from collected candidate hits.
  if(identical(mode, "peek")) {
    hit <- .ivx_collect_query(
      cand,
      no_match_subtree = spec$no_match_subtree,
      leaf_match = spec$leaf_match,
      collect_unmatched = FALSE,
      stop_after_first = identical(which, "first")
    )

    if(identical(which, "first")) {
      if(!isTRUE(hit$first_found)) {
        return(.ivx_query_peek_miss(x, which = which))
      }
      return(hit$first_entry$value)
    }

    if(length(hit$matched_entries) == 0L) {
      return(.ivx_query_peek_miss(x, which = which))
    }
    return(.ivx_slice_entries(x, hit$matched_entries))
  }

  # Phase 4B: pop first-hit branch via absolute index.
  if(identical(which, "first")) {
    hit <- .ivx_collect_query(
      cand,
      no_match_subtree = spec$no_match_subtree,
      leaf_match = spec$leaf_match,
      collect_unmatched = FALSE,
      stop_after_first = TRUE
    )

    if(!isTRUE(hit$first_found)) {
      return(.ivx_query_pop_miss(x, which = which))
    }
    abs_idx <- as.integer(span$start + hit$first_index - 1L)
    return(.ivx_pop_positions(x, abs_idx, which = "first"))
  }

  # Phase 4C: pop all branch rebuilds matched slice and remaining tree from
  # matched/unmatched candidate partitions.
  hit <- .ivx_collect_query(
    cand,
    no_match_subtree = spec$no_match_subtree,
    leaf_match = spec$leaf_match,
    collect_unmatched = TRUE,
    stop_after_first = FALSE
  )
  if(length(hit$matched_entries) == 0L) {
    return(.ivx_query_pop_miss(x, which = which))
  }

  # `which = "all"` pop keeps deterministic order by rebuilding matched and
  # unmatched candidate partitions from collected entries.
  matched <- .ivx_slice_entries(x, hit$matched_entries)
  unmatched <- .ivx_slice_entries(x, hit$unmatched_entries)
  remaining <- .ivx_concat3_like(x, parts$left, unmatched, parts$right)

  list(elements = matched, remaining = remaining)
}

# Returns a slice by positions as interval_index, preserving class metadata.
# **Inputs:** `x` interval_index; integer-like vector `positions`.
# **Outputs:** interval_index slice (empty-like when `positions` is empty).
# **Used by:** .ivx_pop_positions() all-branch.
.ivx_slice_positions <- function(x, positions) {
  if(length(positions) == 0L) {
    return(.ivx_empty_like(x))
  }
  x[as.integer(positions)]
}

# Runtime: O(log n) for single index removal; O(n log n) for multi-index rebuild.
# Removes one or more positions from an interval_index.
# **Inputs:** `x` interval_index; integer-like vector `positions`.
# **Outputs:** interval_index with those positions removed.
# **Used by:** .ivx_pop_positions().
.ivx_remove_positions <- function(x, positions) {
  n <- length(x)
  if(length(positions) == 0L) {
    return(x)
  }
  if(length(positions) >= n) {
    return(.ivx_empty_like(x))
  }
  if(length(positions) == 1L) {
    idx <- as.integer(positions[[1]])
    s <- split_around_by_predicate(x, function(v) v >= idx, ".size")
    return(.ivx_wrap_like(x, concat_trees(s$left, s$right)))
  }

  keep <- setdiff(seq_len(n), as.integer(positions))
  x[as.integer(keep)]
}

# Runtime: O(log n + c) for `which = "first"`; O(n log n) for `which = "all"`.
# Pop helper for already-resolved match positions.
# **Inputs:** `x` interval_index; integer-like `positions`; `which` in {"first","all"}.
# **Outputs:**
#
# - which = "first": list(value =<payload value>, start=<scalar>, end=<scalar>, remaining=<interval_index>).
# - which = "all": list(elements=<interval_index slice at `positions`>,
#   remaining=<interval_index with those positions removed>).
#
# **Used by:** .ivx_run_relation_query() fast path and first-hit pop path.
.ivx_pop_positions <- function(x, positions, which = c("first", "all")) {
  which <- match.arg(which)

  if(length(positions) == 0L) {
    if(identical(which, "all")) {
      return(list(elements = .ivx_empty_like(x), remaining = x))
    }
    return(list(value = NULL, start = NULL, end = NULL, remaining = x))
  }

  if(identical(which, "first")) {
    idx <- as.integer(positions[[1]])
    entry <- .ft_get_elem_at(x, idx)
    remaining <- .ivx_remove_positions(x, idx)
    return(list(value = entry$value, start = entry$start, end = entry$end, remaining = remaining))
  }

  matched <- .ivx_slice_positions(x, positions)
  remaining <- .ivx_remove_positions(x, positions)
  list(elements = matched, remaining = remaining)
}

#SO

# Runtime: O(1).
if(FALSE) add_right <- function(e, el, monoids) NULL
add_right(e, el, monoids) %::% Empty : . : list : Single
add_right(e, el, monoids) %as% {
  measured_single(el, monoids)
}

# Runtime: O(1).
add_right(s, el, monoids) %::% Single : . : list : Deep
add_right(s, el, monoids) %as% {
  measured_deep(
    measured_digit(.subset2(s, 1), monoids = monoids),
    measured_empty(monoids),
    measured_digit(el, monoids = monoids),
    monoids
  )
}

# Runtime: O(k), where k is digit length (bounded by 4 in-tree).
add_right(d, el, monoids) %::% Digit : . : list : Digit
add_right(d, el, monoids) %as% {
  oldclasses <- class(d)
  newd <- c(d, list(el))
  class(newd) <- oldclasses
  set_measure(newd, monoids)
}

# symmetric to add_left. Only new nodes get measures.
# Runtime: O(log n) worst-case.
add_right(d, el, monoids) %::% Deep : . : list : Deep
add_right(d, el, monoids) %as% {
  if(length(.subset2(d, "suffix")) == 4) {
    new_suffix <- measured_digit(.subset2(d, "suffix")[[4]], el, monoids = monoids)
    new_middle_node <- measured_node3(.subset2(d, "suffix")[[1]], .subset2(d, "suffix")[[2]], .subset2(d, "suffix")[[3]], monoids)
    new_middle <- add_right(.subset2(d, "middle"), new_middle_node, monoids)
    measured_deep(prefix = .subset2(d, "prefix"), middle = new_middle, suffix = new_suffix, monoids)
  } else {
    new_suffix <- add_right(.subset2(d, "suffix"), el, monoids)
    measured_deep(prefix = .subset2(d, "prefix"), middle = .subset2(d, "middle"), suffix = new_suffix, monoids)
  }
}

# Runtime: O(k log n), where k = length(els).
if(FALSE) add_all_right <- function(t, els, monoids) NULL
add_all_right(t, els, monoids) %::% FingerTree : . : list : FingerTree
add_all_right(t, els, monoids) %as% {
  for(el in els) {
    t <- add_right(t, el, monoids)
  }
  return(t)
}

#SO

# Runtime: O(1).
.ft_validate_print_max_elements <- function(max_elements) {
  out <- as.integer(max_elements)
  if(length(out) != 1L || is.na(out) || out < 0L) {
    stop("`max_elements` must be a single non-negative integer.")
  }
  out
}

# Runtime: O(1).
.ft_validate_show_custom_monoids <- function(show_custom_monoids) {
  if(!is.logical(show_custom_monoids) || length(show_custom_monoids) != 1L || is.na(show_custom_monoids)) {
    stop("`show_custom_monoids` must be TRUE or FALSE.")
  }
  isTRUE(show_custom_monoids)
}

# Runtime: O(1).
.pick_preview_sizes <- function(n, max_elements) {
  if(n <= 0L || max_elements <= 0L) {
    return(list(head = integer(0), tail = integer(0), skipped = max(0L, n)))
  }
  if(n <= max_elements) {
    return(list(head = seq_len(n), tail = integer(0), skipped = 0L))
  }

  head_n <- as.integer(ceiling(max_elements / 2))
  tail_n <- as.integer(max_elements - head_n)

  head <- seq_len(head_n)
  tail <- if(tail_n > 0L) seq.int(n - tail_n + 1L, n) else integer(0)
  skipped <- as.integer(n - length(head) - length(tail))
  list(head = head, tail = tail, skipped = skipped)
}

# Runtime: O(1).
.ft_print_skipped <- function(skipped) {
  if(skipped <= 0L) {
    return(invisible(NULL))
  }
  cat(
    "... (skipping ",
    skipped,
    " element",
    if(skipped == 1L) "" else "s",
    ")\n\n",
    sep = ""
  )
  invisible(NULL)
}

# Runtime: O(1) expected for scalar formatting.
.ft_format_scalar <- function(x) {
  txt <- tryCatch(format(x), error = function(e) NULL)
  if(is.null(txt) || length(txt) == 0L) {
    return("<unprintable>")
  }
  trimws(paste(txt, collapse = " "))
}

# Like `.ft_format_scalar` but preserves structure for non-atomic-scalar
# values. Atomic scalars (e.g. a numeric sum) use `format()`; lists and
# longer vectors use `deparse()` so they render as `list(has = TRUE, ...)`
# instead of a space-joined flat string. Output is clipped to a single line.
# Runtime: O(size of deparsed representation) for structural values.
.ft_format_measure <- function(x) {
  if(is.atomic(x) && length(x) == 1L) {
    return(.ft_format_scalar(x))
  }
  txt <- tryCatch(
    deparse(x, width.cutoff = 60L, nlines = 1L),
    error = function(e) NULL
  )
  if(is.null(txt) || length(txt) == 0L) {
    return("<unprintable>")
  }
  trimws(paste(txt, collapse = " "))
}

# Runtime: O(m), where m = number of monoids on tree root.
.ft_custom_monoid_names <- function(x, excluded_names) {
  monoids <- names(resolve_tree_monoids(x, required = TRUE))
  setdiff(monoids, excluded_names)
}

# Runtime: O(m), where m = number of visible custom monoids.
.ft_print_custom_monoids <- function(x, excluded_names) {
  custom <- .ft_custom_monoid_names(x, excluded_names = excluded_names)
  if(length(custom) == 0L) {
    cat("Custom monoids + measures: <none>\n")
    return(invisible(NULL))
  }

  cat("Custom monoids + measures:\n")
  for(name in custom) {
    cat("  ", name, ": ", .ft_format_measure(node_measure(x, name)), " (aggregate)\n", sep = "")
  }
  invisible(NULL)
}

# Apply each custom (non-excluded) monoid's measure() to a single leaf entry
# and print a "  <name> measure: <value>" line for each. Safe to call when
# `show_custom_monoids` is FALSE — it short-circuits at that point.
# Runtime: O(m), where m = number of custom monoids on the tree.
.ft_print_elem_custom_monoids <- function(x, entry, excluded_names, show_custom_monoids) {
  if(!isTRUE(show_custom_monoids)) {
    return(invisible(NULL))
  }
  monoids <- resolve_tree_monoids(x, required = TRUE)
  custom <- setdiff(names(monoids), excluded_names)
  if(length(custom) == 0L) {
    return(invisible(NULL))
  }
  for(name in custom) {
    val <- monoids[[name]]$measure(entry)
    cat("  ", name, " measure: ", .ft_format_measure(val), "\n", sep = "")
  }
  invisible(NULL)
}

# Runtime: O(log n).
.ft_print_elem_at <- function(x, i, named, show_custom_monoids = FALSE,
                              excluded_names = character(), ...) {
  el <- .ft_get_elem_at(x, as.integer(i))
  nm <- .ft_get_name(el)
  if(isTRUE(named) && !is.null(nm)) {
    cat("$", nm, "\n", sep = "")
  } else {
    cat("[[", i, "]]\n", sep = "")
  }
  print(.ft_strip_name(el), ...)
  .ft_print_elem_custom_monoids(x, el, excluded_names, show_custom_monoids)
  cat("\n")
  invisible(NULL)
}

#' Print a compact summary of a finger tree
#'
#' @method print FingerTree
#' @param x FingerTree.
#' @param max_elements Maximum number of elements shown in preview (`head + tail`).
#'   Default `4`.
#' @param show_custom_monoids Logical; show attached non-default monoids and
#'   their root cached measures. Default `FALSE`.
#' @param ... Passed through to `print()` for preview elements.
#' @return `x`, invisibly.
#' @examples
#' x <- as_flexseq(setNames(as.list(1:6), letters[1:6]))
#' print(x, max_elements = 4)
#' sum_m <- measure_monoid(`+`, 0, function(el) as.numeric(el))
#' print(add_monoids(as_flexseq(1:3), list(sum = sum_m)), max_elements = 0, show_custom_monoids = TRUE)
#'
#' y <- as_flexseq(as.list(1:6))
#' print(y, max_elements = 3)
#' @keywords internal
# Runtime: O((k + h) log n), where k = shown elements and h = preview split overhead.
print.FingerTree <- function(x, max_elements = 4L, show_custom_monoids = FALSE, ...) {
  n <- as.integer(node_measure(x, ".size"))
  nn <- as.integer(node_measure(x, ".named_count"))
  named <- isTRUE(nn > 0L)
  max_elements <- .ft_validate_print_max_elements(max_elements)
  show_custom <- .ft_validate_show_custom_monoids(show_custom_monoids)
  preview <- .pick_preview_sizes(n, max_elements)
  cat(if(named) "Named" else "Unnamed", " flexseq with ", n, " element", if(n == 1L) "" else "s", ".\n", sep = "")
  if(show_custom) {
    .ft_print_custom_monoids(x, excluded_names = c(".size", ".named_count"))
  }

  if(n == 0L || max_elements == 0L) {
    return(invisible(x))
  }

  cat("\nElements:\n\n")

  excluded_names <- c(".size", ".named_count")
  for(i in preview$head) {
    .ft_print_elem_at(x, i, named, show_custom_monoids = show_custom,
                      excluded_names = excluded_names, ...)
  }

  .ft_print_skipped(preview$skipped)

  for(i in preview$tail) {
    .ft_print_elem_at(x, i, named, show_custom_monoids = show_custom,
                      excluded_names = excluded_names, ...)
  }
  invisible(x)
}

#' Print a flexseq
#'
#' @name print.flexseq
#' @method print flexseq
#' @param x A `flexseq`.
#' @param max_elements Maximum number of elements shown in preview (`head + tail`).
#' @param show_custom_monoids Logical; show attached non-default monoids and
#'   their root cached measures.
#' @param ... Passed through to per-element `print()`.
#' @return The input `x`, returned invisibly. Called for its side effect
#'   of printing a formatted preview of the sequence to the console.
#' @examples
#' x <- as_flexseq(setNames(as.list(1:6), letters[1:6]))
#' print(x, max_elements = 4)
#'
#' y <- as_flexseq(as.list(1:6))
#' print(y, max_elements = 3)
#' @export
# Runtime: Delegates to `print.FingerTree`.
print.flexseq <- function(x, max_elements = 4L, show_custom_monoids = FALSE, ...) {
  print.FingerTree(x, max_elements = max_elements, show_custom_monoids = show_custom_monoids, ...)
}

#' @rdname print.FingerTree
#' @method print Deep
#' @keywords internal
# Runtime: Delegates to `print.FingerTree`.
print.Deep <- function(x, ...) {
  print.FingerTree(x, ...)
}

#' @rdname print.FingerTree
#' @method print Single
#' @keywords internal
# Runtime: Delegates to `print.FingerTree`.
print.Single <- function(x, ...) {
  print.FingerTree(x, ...)
}

#' @rdname print.FingerTree
#' @method print Empty
#' @keywords internal
# Runtime: Delegates to `print.FingerTree`.
print.Empty <- function(x, ...) {
  print.FingerTree(x, ...)
}

#SO

# parse `$` name argument into a scalar character key.
# Runtime: O(1).
.ft_dollar_name <- function(name_expr) {
  nm <- as.character(name_expr)
  if(length(nm) != 1L || is.na(nm) || nm == "") {
    stop("`$` expects a single valid name.")
  }
  nm
}

#' @rdname sub-.flexseq
#' @method $ flexseq
#' @param name Element name (for `$` and `$<-`).
#' @return For `$`: the matched element.
#' @examples
#'
#' # $ extracts by name
#' x <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))
#' x$b
#' @export
# Runtime: O(1) for structural-field fallback on unnamed trees; O(n) worst-case
# for name lookup (single-name locate with early exit).
`$.flexseq` <- function(x, name) {
  nm <- .ft_dollar_name(substitute(name))
  # Preserve structural-field `$` access on unnamed trees for internal traversal
  # and developer ergonomics. Use actual list-field names instead of hard-coding.
  ms <- attr(x, "measures", exact = TRUE)
  nn <- if(!is.null(ms) && !is.null(ms[[".named_count"]])) as.integer(ms[[".named_count"]]) else NA_integer_
  if(!is.na(nn) && nn == 0L) {
    # if a name is requested but the tree is unnamed, 
    # we treat the tree as an R list (allowing e.g. x$prefix even if x is a 
    # flexseq, for internal use)
    if(nm %in% names(unclass(x))) {
      return(.subset2(x, nm))
    }
  }
  `[[.flexseq`(x, nm)
}

#' @rdname sub-.flexseq
#' @method $<- flexseq
#' @return For `$<-`: updated tree with one named element replaced.
#' @examples
#'
#' # $<- replaces by name
#' x <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))
#' x$b <- 20
#' x$b
#' @export
# Runtime: O(n), delegated to `[[<-.flexseq`.
`$<-.flexseq` <- function(x, name, value) {
  nm <- .ft_dollar_name(substitute(name))
  `[[<-.flexseq`(x, nm, value)
}

#SO

# Runtime: O(1).
if(FALSE) add_left <- function(e, el, monoids) NULL
add_left(e, el, monoids) %::% Empty : . : list : Single
add_left(e, el, monoids) %as% {
  measured_single(el, monoids)
}

# Runtime: O(1).
add_left(s, el, monoids) %::% Single : . : list : Deep
add_left(s, el, monoids) %as% {
  measured_deep(
    measured_digit(el, monoids = monoids),
    measured_empty(monoids),
    measured_digit(.subset2(s, 1), monoids = monoids),
    monoids
  )
}

# this is actually a pain, because we implement digits as lists with various attributes
# we can't just c() the element to the list, as we'd get a list(el, old_list), rather than list(el, old_list[[1]], old_list[[2]]), etc.
# We prepend via c(list(el), d) and then restore class attributes.
# Runtime: O(k), where k is digit length (bounded by 4 in-tree),
# so O(1), *assuming* cached measures are available (which should be the case
# by invariant)
add_left(d, el, monoids) %::% Digit : . : list : Digit
add_left(d, el, monoids) %as% {
  oldclasses <- class(d)
  newd <- c(list(el), d)
  class(newd) <- oldclasses
  set_measure(newd, monoids)
}

# if the prefix has 4 elements [p1, p2, p3, p4]: keep el and p1 as the
# new 2-element prefix, pack p2, p3, p4 into a Node3 pushed into the
# middle tree.
# otherwise: simple add to the prefix digit.
# Runtime: O(log n) worst-case.
add_left(d, el, monoids) %::% Deep : . : list : Deep
add_left(d, el, monoids) %as% {
  if(length(.subset2(d, "prefix")) == 4) {
    new_prefix <- measured_digit(el, .subset2(d, "prefix")[[1]], monoids = monoids)
    new_middle_node <- measured_node3(.subset2(d, "prefix")[[2]], .subset2(d, "prefix")[[3]], .subset2(d, "prefix")[[4]], monoids)
    new_middle <- add_left(.subset2(d, "middle"), new_middle_node, monoids)
    measured_deep(prefix = new_prefix, middle = new_middle, suffix = .subset2(d, "suffix"), monoids)
  } else {
    new_prefix <- add_left(.subset2(d, "prefix"), el, monoids)
    measured_deep(prefix = new_prefix, middle = .subset2(d, "middle"), suffix = .subset2(d, "suffix"), monoids)
  }
}

# Runtime: O(k log n), where k = length(els).
if(FALSE) add_all_left <- function(t, els, monoids) NULL
add_all_left(t, els, monoids) %::% FingerTree : . : list : FingerTree
add_all_left(t, els, monoids) %as% {
  for(el in rev(els)) {
    t <- add_left(t, el, monoids)
  }
  return(t)
}

#SO

# Applies a payload transform across interval entries while preserving interval
# coordinates and names.
# **Inputs:**
#
# - `x`: interval_index.
# - `f`: function(value, start, end, name, ...) -> new value.
# - `...`: forwarded to `f`.
# - `preserve_custom_monoids`: scalar logical.
# **Outputs:** interval_index with transformed payload items.
# **Used by:** fapply.interval_index().
.ivx_apply_impl <- function(x, f, ..., preserve_custom_monoids = TRUE) {
  .ivx_assert_index(x)
  if(!is.function(f)) {
    stop("`FUN` must be a function.")
  }
  if(!is.logical(preserve_custom_monoids) || length(preserve_custom_monoids) != 1L || is.na(preserve_custom_monoids)) {
    stop("`preserve_custom_monoids` must be TRUE or FALSE.")
  }

  entries <- .ivx_entries(x)
  n <- length(entries)
  if(n == 0L) {
    return(x)
  }

  out_entries <- vector("list", n)

  dot_args <- list(...)
  accepts_name <- .ft_fun_accepts_n_positional(f, 4L)
  has_dots <- length(dot_args) > 0L

  for(i in seq_len(n)) {
    e <- entries[[i]]
    # inline .ft_get_name / .ft_normalize_name to avoid lambda.r dispatch per element
    nm <- attr(e, "ft_name", exact = TRUE)
    if(!is.null(nm) && (!is.character(nm) || length(nm) != 1L || is.na(nm) || !nzchar(nm))) {
      nm <- NULL
    }
    cur_name <- if(is.null(nm)) "" else nm
    item2 <- if(accepts_name && !has_dots) {
      f(e$value, e$start, e$end, cur_name)
    } else if(!accepts_name && !has_dots) {
      f(e$value, e$start, e$end)
    } else if(accepts_name) {
      do.call(f, c(list(e$value, e$start, e$end, cur_name), dot_args))
    } else {
      do.call(f, c(list(e$value, e$start, e$end), dot_args))
    }
    entry2 <- .ivx_make_entry(item2, e$start, e$end)
    # inline .ft_set_name to avoid lambda.r dispatch per element
    if(!is.null(nm)) attr(entry2, "ft_name") <- nm
    out_entries[[i]] <- entry2
  }

  ms <- if(isTRUE(preserve_custom_monoids)) {
    resolve_tree_monoids(x, required = TRUE)
  } else {
    .ivx_merge_monoids(NULL, endpoint_type = .ivx_endpoint_type_state(x))
  }
  out_tree <- .ivx_tree_from_ordered_entries(out_entries, ms)
  .ivx_wrap_like(x, out_tree)
}

# Runtime: O(n) from traversal + ordered bulk rebuild.
# S3 fapply method for interval_index.
# **Inputs:** `X` interval_index; `FUN` transform function; `...`; scalar logical
# `preserve_custom_monoids`.
# **Outputs:** interval_index with transformed payload items.
# **Used by:** public `fapply()` generic dispatch.
#' @method fapply interval_index
#' @export
#' @noRd
fapply.interval_index <- function(X, FUN, ..., preserve_custom_monoids = TRUE) {
  if(!is.function(FUN)) {
    stop("`FUN` must be a function.")
  }
  .ivx_apply_impl(X, FUN, ..., preserve_custom_monoids = preserve_custom_monoids)
}

#SO

# Runtime: O(n) over tree size for any non-trivial update (rebind/recompute pass).
#' @method add_monoids priority_queue
#' @export
#' @noRd
add_monoids.priority_queue <- function(t, monoids, overwrite = FALSE) {
  if(length(monoids) > 0L) {
    bad <- intersect(names(monoids), c(".size", ".named_count", ".pq_min", ".pq_max"))
    if(length(bad) > 0L) {
      stop("Reserved monoid names cannot be supplied for priority_queue: ", paste(bad, collapse = ", "))
    }
  }
  add_monoids.flexseq(t, monoids, overwrite = overwrite)
}

# Runtime: O(log n) near right edge, with O(1) local name-state checks.
.pq_append_entry <- function(q, entry) {
  ms <- attr(q, "monoids", exact = TRUE)
  if(is.null(ms)) {
    stop("Tree has no monoids attribute.")
  }
  m <- attr(q, "measures", exact = TRUE)
  if(is.null(m)) {
    stop("Tree has no measures attribute.")
  }

  n <- as.integer(m[[".size"]])
  nn <- as.integer(m[[".named_count"]])
  if(n > 0L && nn != 0L && nn != n) {
    stop("Invalid tree name state: mixed named/unnamed elements.")
  }

  nm <- .ft_get_name(entry)
  if(nn == 0L) {
    if(!is.null(nm) && n > 0L) {
      stop("Cannot mix named and unnamed elements (insert would create mixed named and unnamed tree).")
    }
    if(.ft_cpp_can_use(ms)) {
      out <- if(is.null(nm)) .ft_cpp_add_right(q, entry, ms) else .ft_cpp_add_right_named(q, entry, nm, ms)
      return(.pq_wrap_like(q, out))
    }
    entry2 <- if(is.null(nm)) entry else .ft_set_name(entry, nm)
    return(.pq_wrap_like(q, add_right(q, entry2, ms)))
  }

  if(is.null(nm)) {
    stop("Cannot mix named and unnamed elements (insert would create mixed named and unnamed tree).")
  }
  if(.ft_cpp_can_use(ms)) {
    return(.pq_wrap_like(q, .ft_cpp_add_right_named(q, entry, nm, ms)))
  }
  .pq_wrap_like(q, add_right(q, .ft_set_name(entry, nm), ms))
}

# Runtime: O(log n) near right edge.
#' @method insert priority_queue
#' @export
#' @noRd
insert.priority_queue <- function(x, element, priority, name = NULL, ...) {
  q <- x
  .pq_assert_queue(q)
  parsed <- .pq_make_entry(element, priority, priority_type = .pq_priority_type_state(q))
  entry <- parsed$entry

  if(!is.null(name)) {
    entry <- .ft_set_name(entry, name)
  }
  .pq_append_entry(q, entry)
}

# Runtime: O(1).
.pq_empty_like <- function(x) {
  ms <- resolve_tree_monoids(x, required = TRUE)
  .pq_wrap_like(x, empty_tree(monoids = ms))
}

# Runtime: O(k), where k = number of requested positions.
.pq_slice_positions <- function(x, positions) {
  if(length(positions) == 0L) {
    return(.pq_empty_like(x))
  }
  .pq_wrap_like(x, `[.flexseq`(x, as.integer(positions)))
}

# Runtime: O(log n) for single deletion; O(n log n) for multi-position rebuild.
.pq_remove_positions <- function(x, positions) {
  n <- length(x)
  if(length(positions) == 0L) {
    return(x)
  }
  pos <- sort(unique(as.integer(positions)))
  if(length(pos) >= n) {
    return(.pq_empty_like(x))
  }
  if(length(pos) == 1L) {
    idx <- pos[[1L]]
    s <- split_around_by_predicate.flexseq(x, function(v) v >= idx, ".size")
    return(.pq_wrap_like(x, concat_trees(s$left, s$right)))
  }
  keep <- setdiff(seq_len(n), pos)
  .pq_wrap_like(x, `[.flexseq`(x, as.integer(keep)))
}

# Runtime: O(n) scan to collect all tie positions for one extrema value.
.pq_extreme_positions <- function(x, monoid_name) {
  .pq_assert_queue(x)
  if(length(x) == 0L) {
    return(integer(0))
  }
  target <- node_measure(x, monoid_name)
  if(!isTRUE(target$has)) {
    return(integer(0))
  }
  target_priority <- target$priority
  domain <- .ft_scalar_domain(target_priority)
  entries <- .ft_to_list(x)
  idx <- which(vapply(
    entries,
    function(e) {
      .ft_scalar_equal_fast(
        e$priority,
        target_priority,
        domain = domain,
        error_message = "Priority values must support scalar ordering with `<` and `>`."
      )
    },
    logical(1)
  ))
  as.integer(idx)
}

# Runtime: O(log n) near locate point depth.
.pq_peek <- function(x, monoid_name) {
  .pq_assert_queue(x)
  if(length(x) == 0L) {
    return(NULL)
  }

  target <- node_measure(x, monoid_name)
  pred <- function(v) .pq_measure_equal(v, target)
  ctx <- resolve_named_monoid(x, monoid_name)
  ms <- ctx$monoids
  mr <- ctx$monoid
  loc <- if(.ft_cpp_can_use(ms)) {
    .ft_cpp_locate(x, pred, ms, monoid_name, mr$i)
  } else {
    locate_tree_impl_fast(pred, mr$i, x, ms, mr, monoid_name, 0L)
  }
  loc$value[["value"]]
}

# Runtime: O(log n) near split point depth.
.pq_extract <- function(x, monoid_name) {
  .pq_assert_queue(x)
  if(length(x) == 0L) {
    return(list(value = NULL, priority = NULL, remaining = x))
  }

  target <- node_measure(x, monoid_name)
  pred <- function(v) .pq_measure_equal(v, target)
  ctx <- resolve_named_monoid(x, monoid_name)
  ms <- ctx$monoids
  mr <- ctx$monoid
  s <- if(.ft_cpp_can_use(ms)) {
    .ft_cpp_split_tree(x, pred, ms, monoid_name, mr$i)
  } else {
    split_tree_impl_fast(pred, mr$i, x, ms, mr, monoid_name)
  }

  rest <- concat_trees(s$left, s$right)
  rest <- .pq_wrap_like(x, rest)

  list(
    value = s$value[["value"]],
    priority = s$value[["priority"]],
    remaining = rest
  )
}

# Runtime: O(n) due tie-run position scan.
.pq_peek_all <- function(x, monoid_name) {
  .pq_assert_queue(x)
  pos <- .pq_extreme_positions(x, monoid_name)
  .pq_slice_positions(x, pos)
}

# Runtime: O(n log n) worst-case from tie-run extraction + remainder rebuild.
.pq_extract_all <- function(x, monoid_name) {
  .pq_assert_queue(x)
  pos <- .pq_extreme_positions(x, monoid_name)
  if(length(pos) == 0L) {
    return(list(elements = .pq_empty_like(x), remaining = x))
  }
  list(
    elements = .pq_slice_positions(x, pos),
    remaining = .pq_remove_positions(x, pos)
  )
}

# Runtime: O(log n) near locate point depth.
#' Peek Minimum-Priority Element
#'
#' Returns the element at the minimum priority without modifying the queue.
#'
#' @param x A `priority_queue`.
#' @return Element at minimum priority, or `NULL` when `x` is empty.
#' @details
#' Ties are stable: when multiple elements share minimum priority, this returns
#' the earliest element in queue order.
#' @examples
#' x <- priority_queue("a", "b", "c", priorities = c(2, 1, 1))
#' peek_min(x)
#' peek_min(priority_queue())
#' @export
peek_min <- function(x) {
  .pq_peek(x, ".pq_min")
}

# Runtime: O(log n) near locate point depth.
#' Peek Maximum-Priority Element
#'
#' Returns the element at the maximum priority without modifying the queue.
#'
#' @param x A `priority_queue`.
#' @return Element at maximum priority, or `NULL` when `x` is empty.
#' @details
#' Ties are stable: when multiple elements share maximum priority, this returns
#' the earliest element in queue order.
#' @examples
#' x <- priority_queue("a", "b", "c", priorities = c(2, 3, 3))
#' peek_max(x)
#' peek_max(priority_queue())
#' @export
peek_max <- function(x) {
  .pq_peek(x, ".pq_max")
}

# Runtime: O(1).
#' Minimum Priority Value
#'
#' Returns the current minimum priority scalar in the queue.
#'
#' @param x A `priority_queue`.
#' @return Minimum priority value, or `NULL` when `x` is empty.
#' @details
#' Uses cached `.pq_min` monoid state.
#' @examples
#' q <- priority_queue("a", "b", priorities = c(2, 1))
#' min_priority(q)
#' min_priority(priority_queue())
#' @export
min_priority <- function(x) {
  .pq_assert_queue(x)
  m <- node_measure(x, ".pq_min")
  if(!isTRUE(m$has)) {
    return(NULL)
  }
  m$priority
}

# Runtime: O(1).
#' Maximum Priority Value
#'
#' Returns the current maximum priority scalar in the queue.
#'
#' @param x A `priority_queue`.
#' @return Maximum priority value, or `NULL` when `x` is empty.
#' @details
#' Uses cached `.pq_max` monoid state.
#' @examples
#' q <- priority_queue("a", "b", priorities = c(2, 1))
#' max_priority(q)
#' max_priority(priority_queue())
#' @export
max_priority <- function(x) {
  .pq_assert_queue(x)
  m <- node_measure(x, ".pq_max")
  if(!isTRUE(m$has)) {
    return(NULL)
  }
  m$priority
}

# Runtime: O(n) due tie-run scan.
#' Peek All Minimum-Priority Elements
#'
#' Returns the full minimum-priority tie run as a `priority_queue`.
#'
#' @param x A `priority_queue`.
#' @return A `priority_queue` containing all minimum-priority elements in stable
#'   queue order. Returns an empty queue when `x` is empty.
#' @details
#' The return is another `priority_queue()`, use `as.list()` to convert 
#' the result to a standard R list.
#' @examples
#' x <- priority_queue("a", "b", "c", priorities = c(2, 1, 1))
#' peek_all_min(x)
#' @export
peek_all_min <- function(x) {
  .pq_peek_all(x, ".pq_min")
}

# Runtime: O(n) due tie-run scan.
#' Peek All Maximum-Priority Elements
#'
#' Returns the full maximum-priority tie run as a `priority_queue`.
#'
#' @param x A `priority_queue`.
#' @return A `priority_queue` containing all maximum-priority elements in stable
#'   queue order. Returns an empty queue when `x` is empty.
#' @details
#' The return is another `priority_queue()`, use `as.list()` to convert 
#' the result to a standard R list.
#' @examples
#' x <- priority_queue("a", "b", "c", priorities = c(2, 3, 3))
#' peek_all_max(x)
#' @export
peek_all_max <- function(x) {
  .pq_peek_all(x, ".pq_max")
}

# Runtime: O(log n) near split point depth.
#' Pop Minimum-Priority Element
#'
#' Removes one minimum-priority element and returns it with the remaining queue.
#'
#' @param x A `priority_queue`.
#' @return A list with fields:
#' - `value`: removed element, or `NULL` when `x` is empty.
#' - `priority`: removed priority, or `NULL` when `x` is empty.
#' - `remaining`: queue after removal.
#' @details
#' Ties are stable: when multiple elements share minimum priority, the earliest
#' element in queue order is removed.
#' @examples
#' x <- priority_queue("a", "b", "c", priorities = c(2, 1, 1))
#' out <- pop_min(x)
#' out$value
#' out$priority
#' out$remaining
#' pop_min(priority_queue())
#' @export
pop_min <- function(x) {
  .pq_extract(x, ".pq_min")
}

# Runtime: O(log n) near split point depth.
#' Pop Maximum-Priority Element
#'
#' Removes one maximum-priority element and returns it with the remaining queue.
#'
#' @param x A `priority_queue`.
#' @return A list with fields:
#' - `value`: removed element, or `NULL` when `x` is empty.
#' - `priority`: removed priority, or `NULL` when `x` is empty.
#' - `remaining`: queue after removal.
#' @details
#' Ties are stable: when multiple elements share maximum priority, the earliest
#' element in queue order is removed.
#' @examples
#' x <- priority_queue("a", "b", "c", priorities = c(2, 3, 3))
#' out <- pop_max(x)
#' out$value
#' out$priority
#' out$remaining
#' pop_max(priority_queue())
#' @export
pop_max <- function(x) {
  .pq_extract(x, ".pq_max")
}

# Runtime: O(n log n) worst-case.
#' Pop All Minimum-Priority Elements
#'
#' Removes the full minimum-priority tie run.
#'
#' @param x A `priority_queue`.
#' @return A list with fields:
#' - `elements`: `priority_queue` of removed minimum-priority elements.
#' - `remaining`: queue after removal.
#' @details
#' The return `elements` is another `priority_queue()`, use `as.list()` to 
#' convert the result to a standard R list.
#' @examples
#' x <- priority_queue("a", "b", "c", priorities = c(2, 1, 1))
#' out <- pop_all_min(x)
#' out$elements
#' out$remaining
#' @export
pop_all_min <- function(x) {
  .pq_extract_all(x, ".pq_min")
}

# Runtime: O(n log n) worst-case.
#' Pop All Maximum-Priority Elements
#'
#' Removes the full maximum-priority tie run.
#'
#' @param x A `priority_queue`.
#' @return A list with fields:
#' - `elements`: `priority_queue` of removed maximum-priority elements.
#' - `remaining`: queue after removal.
#' @details
#' The return `elements` is another `priority_queue()`, use `as.list()` to 
#' convert the result to a standard R list.
#' @examples
#' x <- priority_queue("a", "b", "c", priorities = c(2, 3, 3))
#' out <- pop_all_max(x)
#' out$elements
#' out$remaining
#' @export
pop_all_max <- function(x) {
  .pq_extract_all(x, ".pq_max")
}

#' Merge Two Priority Queues
#'
#' Returns a new `priority_queue` containing every entry from both inputs,
#' preserving each queue's internal insertion order (entries of `x` come
#' first, then entries of `y`).
#'
#' @method merge priority_queue
#' @param x A `priority_queue`.
#' @param y A `priority_queue`.
#' @param ... Unused.
#' @return A new `priority_queue` of size `length(x) + length(y)`.
#' @details
#' The cached `.pq_min` / `.pq_max` monoids recompute automatically on the
#' merged tree, so `peek_min()` / `peek_max()` reflect the combined extremum
#' immediately.
#'
#' Both queues must share the same priority type and the same monoid set;
#' mismatches error rather than being silently harmonized. Merging an
#' empty queue with a non-empty queue returns the non-empty queue unchanged.
#'
#' Both inputs are left unmodified.
#' @examples
#' a <- priority_queue("x", "y", priorities = c(5, 1))
#' b <- priority_queue("z", priorities = 3)
#' m <- merge(a, b)
#' peek_min(m)
#' length(m)
#' @export
# Runtime: O(log(min(m, n))) via concat_trees.
merge.priority_queue <- function(x, y, ...) {
  if(!inherits(y, "priority_queue")) {
    stop("Both arguments to `merge()` must be priority_queues.")
  }

  if(length(x) == 0L) return(y)
  if(length(y) == 0L) return(x)

  if(!identical(attr(x, "pq_priority_type", exact = TRUE),
                attr(y, "pq_priority_type", exact = TRUE))) {
    stop("Cannot merge priority_queues with different priority types.")
  }
  if(!identical(sort(names(attr(x, "monoids", exact = TRUE))),
                sort(names(attr(y, "monoids", exact = TRUE))))) {
    stop("Cannot merge priority_queues with different monoid sets.")
  }

  ms <- attr(x, "monoids", exact = TRUE)
  out <- if(.ft_cpp_can_use(ms)) {
    .ft_cpp_concat(x, y, ms)
  } else {
    concat_trees(x, y)
  }
  .pq_wrap_like(x, out)
}

testthat::test_that("priority_queue constructor and length", {
  q0 <- priority_queue()
  testthat::expect_true(inherits(q0, "priority_queue"))
  testthat::expect_true(inherits(q0, "flexseq"))
  testthat::expect_identical(length(q0), 0L)

  q <- priority_queue("a", "b", "c", priorities = c(3, 1, 2))
  testthat::expect_identical(length(q), 3L)
  testthat::expect_equal(peek_min(q), "b")
  testthat::expect_equal(peek_max(q), "a")
})

testthat::test_that("empty priority_queue scalar peek/pop are non-throwing misses", {
  x <- priority_queue()

  testthat::expect_null(peek_min(x))
  testthat::expect_null(peek_max(x))
  testthat::expect_null(min_priority(x))
  testthat::expect_null(max_priority(x))

  pmin <- pop_min(x)
  pmax <- pop_max(x)
  testthat::expect_null(pmin$value)
  testthat::expect_null(pmin$priority)
  testthat::expect_identical(length(pmin$remaining), 0L)
  testthat::expect_null(pmax$value)
  testthat::expect_null(pmax$priority)
  testthat::expect_identical(length(pmax$remaining), 0L)
})

testthat::test_that("min_priority/max_priority read cached extrema", {
  q <- priority_queue("a", "b", "c", priorities = c(3, 1, 2))
  testthat::expect_identical(min_priority(q), 1)
  testthat::expect_identical(max_priority(q), 3)
})

testthat::test_that("min/max pop uses sequence order on ties", {
  q <- priority_queue("a", "b", "c", "d", priorities = c(2, 1, 1, 2))

  e1 <- pop_min(q)
  testthat::expect_equal(e1$value, "b")
  testthat::expect_equal(e1$priority, 1)

  e2 <- pop_min(e1$remaining)
  testthat::expect_equal(e2$value, "c")
  testthat::expect_equal(e2$priority, 1)

  x1 <- pop_max(q)
  testthat::expect_equal(x1$value, "a")
  testthat::expect_equal(x1$priority, 2)

  x2 <- pop_max(x1$remaining)
  testthat::expect_equal(x2$value, "d")
  testthat::expect_equal(x2$priority, 2)
})

testthat::test_that("bulk extrema helpers return full tie runs", {
  q <- priority_queue("a", "b", "c", "d", priorities = c(2, 1, 1, 2))

  pmin <- peek_all_min(q)
  pmax <- peek_all_max(q)
  testthat::expect_s3_class(pmin, "priority_queue")
  testthat::expect_s3_class(pmax, "priority_queue")
  testthat::expect_equal(lapply(as.list(pmin), function(e) e$value), list("b", "c"))
  testthat::expect_equal(lapply(as.list(pmax), function(e) e$value), list("a", "d"))

  out_min <- pop_all_min(q)
  testthat::expect_s3_class(out_min$elements, "priority_queue")
  testthat::expect_s3_class(out_min$remaining, "priority_queue")
  testthat::expect_equal(lapply(as.list(out_min$elements), function(e) e$value), list("b", "c"))
  testthat::expect_equal(lapply(as.list(out_min$remaining), function(e) e$value), list("a", "d"))

  out_max <- pop_all_max(q)
  testthat::expect_s3_class(out_max$elements, "priority_queue")
  testthat::expect_s3_class(out_max$remaining, "priority_queue")
  testthat::expect_equal(lapply(as.list(out_max$elements), function(e) e$value), list("a", "d"))
  testthat::expect_equal(lapply(as.list(out_max$remaining), function(e) e$value), list("b", "c"))
})

testthat::test_that("bulk extrema single-element removal preserves pq measures", {
  q <- priority_queue("a", "b", "c", priorities = c(3, 1, 2))

  out_min <- pop_all_min(q)
  testthat::expect_equal(lapply(as.list(out_min$elements), function(e) e$value), list("b"))
  testthat::expect_equal(peek_min(out_min$remaining), "c")
  testthat::expect_equal(peek_max(out_min$remaining), "a")
  testthat::expect_no_error(capture.output(print(out_min$remaining)))

  out_max <- pop_all_max(q)
  testthat::expect_equal(lapply(as.list(out_max$elements), function(e) e$value), list("a"))
  testthat::expect_equal(peek_min(out_max$remaining), "b")
  testthat::expect_equal(peek_max(out_max$remaining), "c")
  testthat::expect_no_error(capture.output(print(out_max$remaining)))
})

testthat::test_that("as.list.priority_queue returns entry records in queue sequence order", {
  q <- as_priority_queue(
    setNames(as.list(c("x", "y", "z")), c("kx", "ky", "kz")),
    priorities = c(2, 1, 3)
  )

  out <- as.list(q)
  testthat::expect_type(out, "list")
  testthat::expect_identical(names(out), c("kx", "ky", "kz"))
  testthat::expect_equal(unname(lapply(out, function(e) e$value)), list("x", "y", "z"))
  testthat::expect_equal(unname(lapply(out, function(e) e$priority)), as.list(c(2, 1, 3)))
})

testthat::test_that("insert is persistent and supports names", {
  q <- as_priority_queue(setNames(as.list(c("x", "y")), c("kx", "ky")), priorities = c(5, 1))
  q2 <- insert(q, "z", priority = 1, name = "kz")

  testthat::expect_equal(peek_min(q), "y")
  testthat::expect_equal(peek_min(q2), "y")
  testthat::expect_equal(q2[["kz"]], "z")
  testthat::expect_equal(q2$kz, "z")
  testthat::expect_equal(length(q2), 3L)
})

testthat::test_that("priority_queue disallows sequence-style mutation helpers", {
  q <- priority_queue("b", "c", priorities = c(2, 3))
  testthat::expect_error(push_front(q, list(value = "a", priority = 1)), "Cast first")
  testthat::expect_error(push_back(q, list(value = "d", priority = 4)), "Cast first")
  testthat::expect_error(peek_front(q), "not supported for priority_queue")
  testthat::expect_error(peek_back(q), "not supported for priority_queue")
  testthat::expect_error(peek_at(q, 1), "not supported for priority_queue")
  testthat::expect_error(pop_front(q), "not supported for priority_queue")
  testthat::expect_error(pop_back(q), "not supported for priority_queue")
  testthat::expect_error(pop_at(q, 1), "not supported for priority_queue")
  testthat::expect_error(insert_at(q, 1, "x"), "not supported for priority_queue")
  testthat::expect_error(c(q, q), "Cast first")
})

testthat::test_that("priority_queue character subset is strict on missing names", {
  q <- as_priority_queue(
    setNames(as.list(c("a", "b", "c")), c("ka", "kb", "kc")),
    priorities = c(3, 1, 2)
  )
  s <- q[c("kb", "ka")]

  testthat::expect_s3_class(s, "priority_queue")
  testthat::expect_equal(peek_min(s), "b")
  testthat::expect_error(q[c("kb", "missing")], "Unknown element name")
})

testthat::test_that("priority_queue disallows positional/logical indexing and replacements", {
  q <- priority_queue("a", "b", priorities = c(2, 1))

  testthat::expect_error(q[[1]], "scalar character names only")
  testthat::expect_error(q[1:2], "character name indexing only")
  testthat::expect_error(q[c(TRUE, FALSE)], "character name indexing only")
  testthat::expect_error({ q[[1]] <- list(value = "x", priority = 5) }, "Cast first")
  testthat::expect_error({ q[c(1, 2)] <- list("u", "v") }, "Cast first")
})

testthat::test_that("priority queue carries required monoids", {
  q <- priority_queue(1, 2, priorities = c(10, 20))
  ms <- names(attr(q, "monoids", exact = TRUE))
  testthat::expect_true(all(c(".size", ".named_count", ".pq_min", ".pq_max") %in% ms))
})

testthat::test_that("priority_queue casts down to flexseq explicitly", {
  sum_priority <- measure_monoid(`+`, 0, function(entry) as.numeric(entry$priority))
  q_named <- add_monoids(
    as_priority_queue(setNames(as.list(c("x", "y")), c("kx", "ky")), priorities = c(2, 1)),
    list(sum_priority = sum_priority)
  )
  x <- as_flexseq(q_named)

  testthat::expect_s3_class(x, "flexseq")
  testthat::expect_false(inherits(x, "priority_queue"))
  testthat::expect_equal(x[["kx"]], "x")
  testthat::expect_equal(x[["ky"]], "y")

  ms <- names(attr(x, "monoids", exact = TRUE))
  testthat::expect_true(all(c(".size", ".named_count") %in% ms))
  testthat::expect_false(".pq_min" %in% ms)
  testthat::expect_false(".pq_max" %in% ms)
  testthat::expect_false("sum_priority" %in% ms)
  testthat::expect_error(node_measure(x, "sum_priority"), "Missing cached measure")

  x_unnamed <- as_flexseq(priority_queue("x", "y", priorities = c(2, 1)))
  x2 <- push_back(x_unnamed, "z")
  testthat::expect_s3_class(x2, "flexseq")
  testthat::expect_equal(length(x2), 3L)
})

testthat::test_that("fapply maps priority queue items with read-only metadata", {
  q <- priority_queue("a", "bb", "ccc", priorities = c(1, 3, 2))
  q2 <- fapply(
    q,
    function(value, priority, name) {
      toupper(value)
    }
  )

  testthat::expect_equal(peek_min(q2), "A")
  testthat::expect_equal(peek_max(q2), "BB")
  testthat::expect_identical(length(q2), 3L)

  pmin <- pop_min(q2)
  pmax <- pop_max(q2)
  testthat::expect_identical(pmin$priority, 1)
  testthat::expect_identical(pmax$priority, 3)
})

testthat::test_that("fapply respects sequence order for ties", {
  q <- as_priority_queue(
    c("a", "b", "c"),
    priorities = c(1, 1, 1)
  )
  q2 <- fapply(q, function(value, priority, name) toupper(value))

  # All priorities tie; left-most in sequence wins.
  testthat::expect_equal(peek_min(q2), "A")
  p1 <- pop_min(q2)
  p2 <- pop_min(p1$remaining)
  testthat::expect_equal(p1$value, "A")
  testthat::expect_equal(p2$value, "B")
})

testthat::test_that("fapply preserves priority queue entry names", {
  q <- as_priority_queue(setNames(as.list(c("x", "y")), c("kx", "ky")), priorities = c(2, 1))
  q2 <- fapply(q, function(value, priority, name) {
    paste(name, value, sep = ":")
  })

  testthat::expect_equal(q2[["kx"]], "kx:x")
  testthat::expect_equal(q2[["ky"]], "ky:y")
  testthat::expect_equal(q2$kx, "kx:x")
})

testthat::test_that("fapply can drop custom monoids for priority_queue", {
  sum_item <- measure_monoid(`+`, 0, function(entry) as.numeric(entry$value))
  q <- add_monoids(priority_queue(1, 2, priorities = c(2, 1)), list(sum_item = sum_item))
  q2 <- fapply(q, function(value, priority, name) value * 10, preserve_custom_monoids = FALSE)

  ms <- attr(q2, "monoids", exact = TRUE)
  testthat::expect_true(!is.null(ms[[".size"]]))
  testthat::expect_true(!is.null(ms[[".named_count"]]))
  testthat::expect_true(!is.null(ms[[".pq_min"]]))
  testthat::expect_true(!is.null(ms[[".pq_max"]]))
  testthat::expect_true(is.null(ms[["sum_item"]]))
  testthat::expect_error(node_measure(q2, "sum_item"), "Missing cached measure")
  testthat::expect_equal(peek_min(q2), 20)
})

testthat::test_that("priority_queue custom monoids accept entry measure arguments", {
  q <- add_monoids(
    priority_queue(10, 20, priorities = c(2, 1)),
    list(
      by_priority = measure_monoid(`+`, 0, function(entry) as.numeric(entry$priority))
    )
  )
  testthat::expect_equal(node_measure(q, "by_priority"), 3)
})

testthat::test_that("fapply validates priority queue inputs", {
  q <- priority_queue("a", priorities = 1)
  testthat::expect_error(fapply.priority_queue(as_flexseq(1:3), FUN = identity), "`q` must be a priority_queue")
  testthat::expect_error(fapply(q, 1), "`FUN` must be a function")
  testthat::expect_error(fapply(q, identity, preserve_custom_monoids = NA), "`preserve_custom_monoids` must be TRUE or FALSE")
})

testthat::test_that("priority_queue blocks split helpers and allows locate helper", {
  q <- priority_queue("a", "b", priorities = c(2, 1))
  testthat::expect_error(split_by_predicate(q, function(v) v >= 1, ".size"), "Cast first")
  testthat::expect_error(split_around_by_predicate(q, function(v) v >= 1, ".size"), "Cast first")
  testthat::expect_error(split_at(q, 1), "Cast first")

  loc <- locate_by_predicate(q, function(v) v >= 1, ".size", include_metadata = TRUE)
  testthat::expect_true(loc$found)
  testthat::expect_identical(loc$value$value, "a")
  testthat::expect_identical(loc$metadata$index, 1L)
})

testthat::test_that("priority_queue supports Date priorities with FIFO ties", {
  p <- as.Date(c("2024-01-03", "2024-01-01", "2024-01-01"))
  q <- priority_queue("late", "early1", "early2", priorities = p)

  testthat::expect_equal(peek_min(q), "early1")
  testthat::expect_equal(peek_max(q), "late")

  out1 <- pop_min(q)
  out2 <- pop_min(out1$remaining)
  testthat::expect_equal(out1$value, "early1")
  testthat::expect_equal(out2$value, "early2")
  testthat::expect_s3_class(out1$priority, "Date")
  testthat::expect_true(inherits(out2$priority, "Date"))

  q2 <- insert(q, "early3", priority = as.Date("2024-01-01"))
  testthat::expect_equal(pop_min(pop_min(q2)$remaining)$value, "early2")
})

testthat::test_that("priority_queue supports POSIXct priorities with type-preserving pops", {
  p <- as.POSIXct(
    c("2024-01-01 12:00:00", "2024-01-01 10:00:00", "2024-01-01 10:00:00"),
    tz = "UTC"
  )
  q <- priority_queue("late", "early1", "early2", priorities = p)

  testthat::expect_equal(peek_min(q), "early1")
  out <- pop_min(q)
  testthat::expect_equal(out$value, "early1")
  testthat::expect_s3_class(out$priority, "POSIXct")
  testthat::expect_s3_class(out$priority, "POSIXt")
})

testthat::test_that("priority_queue rejects mixed priority domains and missing priorities", {
  testthat::expect_error(
    as_priority_queue(list("a", "b"), priorities = list(1, "1")),
    "Incompatible priority type"
  )
  testthat::expect_error(
    as_priority_queue(list("a", "b"), priorities = list(1, as.Date("2024-01-01"))),
    "Incompatible priority type"
  )
  testthat::expect_error(
    as_priority_queue(
      list("a", "b"),
      priorities = list(as.Date("2024-01-01"), as.POSIXct("2024-01-01 00:00:00", tz = "UTC"))
    ),
    "Incompatible priority type"
  )

  q_num <- priority_queue("a", priorities = 1)
  q_date <- priority_queue("a", priorities = as.Date("2024-01-01"))
  testthat::expect_error(insert(q_num, "b", priority = "1"), "Incompatible priority type")
  testthat::expect_error(insert(q_num, "b", priority = as.Date("2024-01-01")), "Incompatible priority type")
  testthat::expect_error(
    insert(q_date, "b", priority = as.POSIXct("2024-01-01 00:00:00", tz = "UTC")),
    "Incompatible priority type"
  )

  testthat::expect_error(
    priority_queue("a", priorities = as.Date(NA)),
    "`priority` must be non-missing"
  )
  testthat::expect_error(
    insert(q_date, "b", priority = as.Date(NA)),
    "`priority` must be non-missing"
  )
})

testthat::test_that("fapply dispatches for flexseq", {
  x <- as_flexseq(1:4)
  y <- fapply(x, function(v) v * 3)

  testthat::expect_s3_class(y, "flexseq")
  testthat::expect_identical(lapply(.ft_to_list(y), .ft_strip_name), as.list(c(3, 6, 9, 12)))
  testthat::expect_error(fapply(x, 1), "`FUN` must be a function")
})

testthat::test_that("fapply dispatches for priority_queue", {
  q <- priority_queue("a", "bb", "ccc", priorities = c(1, 3, 2))
  q2 <- fapply(q, function(value, priority, name) {
    toupper(value)
  })
  q3 <- fapply(q, function(value, priority) {
    paste0(value, ":", priority)
  })

  testthat::expect_s3_class(q2, "priority_queue")
  testthat::expect_equal(peek_min(q2), "A")
  testthat::expect_equal(peek_max(q2), "BB")
  testthat::expect_equal(peek_min(q3), "a:1")
  testthat::expect_equal(peek_max(q3), "bb:3")
  testthat::expect_error(fapply(q, 1), "`FUN` must be a function")
})

testthat::test_that("fapply dispatches for ordered_sequence", {
  xs <- as_ordered_sequence(setNames(list("x1", "x2", "x3"), c("a", "b", "c")), keys = c(1, 1, 2))

  xs_item <- fapply(xs, function(value, key, name) {
    toupper(value)
  })
  testthat::expect_s3_class(xs_item, "ordered_sequence")
  testthat::expect_equal(unname(as.list(xs_item)), list("X1", "X2", "X3"))
  testthat::expect_identical(names(as.list(xs_item)), c("a", "b", "c"))

  xs_tagged <- fapply(xs, function(value, key, name) {
    paste(value, key, name, sep = "|")
  })
  xs2 <- fapply(xs, function(value, key) {
    paste(value, key, sep = "|")
  })
  testthat::expect_equal(unname(as.list(xs_tagged)), list("x1|1|a", "x2|1|b", "x3|2|c"))
  testthat::expect_identical(names(as.list(xs_tagged)), c("a", "b", "c"))
  testthat::expect_equal(count_key(xs_tagged, 1), 2L)
  testthat::expect_equal(count_key(xs_tagged, 2), 1L)
  testthat::expect_equal(unname(as.list(xs2)), list("x1|1", "x2|1", "x3|2"))

  testthat::expect_error(fapply(xs, 1), "`FUN` must be a function")
})

#SO

#' Fapply with S3 dispatch
#'
#' `fapply()` is an S3 generic for applying functions over immutable
#' structures with type-specific dispatch.
#'
#' @param X Object to apply over.
#' @param FUN Function to apply.
#' @param ... Method-specific arguments.
#' @return Method-dependent result.
#' @details
#' `fapply()` returns a new object and preserves class semantics.
#'
#' Method signatures:
#' - `fapply.flexseq(X, FUN, ..., preserve_custom_monoids = TRUE)`
#' - `fapply.priority_queue(X, FUN, ..., preserve_custom_monoids = TRUE)`
#' - `fapply.ordered_sequence(X, FUN, ..., preserve_custom_monoids = TRUE)`
#' - `fapply.interval_index(X, FUN, ..., preserve_custom_monoids = TRUE)`
#'
#' For `priority_queue`, `ordered_sequence`, and `interval_index`,
#' `FUN` receives structured fields (`value` plus metadata) and should return the
#' new payload value only; ordering metadata is preserved. The name argument is
#' optional: callbacks that only take value/metadata also work.
#'
#' If supported by the method, `preserve_custom_monoids = TRUE` keeps added user
#' monoids; `FALSE` rebuilds with required structural monoids only.
#' @examples
#' x <- flexseq("a", "b", "c")
#' fapply(x, toupper)
#'
#' q <- priority_queue(one = "a", two = "b", priorities = c(2, 1))
#' fapply(q, function(value, priority, name) paste0(value, priority))
#'
#' o <- ordered_sequence(one = "a", two = "b", keys = c(2, 1))
#' fapply(o, function(value, key, name) paste0(value, "_", key))
#'
#' ix <- interval_index(one = "a", two = "b", start = c(1, 3), end = c(2, 4))
#' fapply(ix, function(value, start, end, name) paste0(value, "[", start, ",", end, "]"))
#' @seealso [flexseq()], [priority_queue()], [ordered_sequence()], [interval_index()]
#' @export
fapply <- function(X, FUN, ...) {
  UseMethod("fapply")
}

#' Add or Merge Measure Monoids
#'
#' Attaches one or more named [measure_monoid()] definitions to an existing
#' immutable structure.
#'
#' @param t Immutable structure (`flexseq` and subclasses).
#' @param monoids Named list of [measure_monoid()] objects.
#' @param overwrite Logical; if `TRUE`, replace existing monoids with the same
#'   names. If `FALSE`, existing names are kept.
#' @return A persistent copy with updated monoid definitions and cached measures.
#' @details
#' Mechanics:
#' - Each monoid name defines an independent accumulated measure over elements in the tree.
#' - New monoids are computed for all elements and cached in the returned object.
#' - Existing monoids are unchanged unless `overwrite = TRUE`.
#' - Structural/reserved monoid names cannot be replaced.
#'
#' Measure-function signatures:
#' - `flexseq`: `measure(entry)` where `entry` is the stored element.
#' - `ordered_sequence`: `measure(entry)` where `entry` is `list(value, key)`.
#' - `priority_queue`: `measure(entry)` where `entry` is `list(value, priority)`.
#' - `interval_index`: `measure(entry)` where `entry` is `list(value, start, end)`.
#'
#' This operation is persistent: `t` is not modified.
#'
#' Use this when you want fast predicate scans (for example with
#' [locate_by_predicate()], [split_by_predicate()], [split_around_by_predicate()])
#' driven by domain-specific accumulated values.
#' @examples
#' x <- flexseq(10, 20, 30)
#'
#' running_sum <- measure_monoid(`+`, 0, as.numeric)
#' x2 <- add_monoids(x, list(sum = running_sum))
#' attr(x2, "measures")$sum
#'
#' # Use the monoid in a split query
#' split_around_by_predicate(x2, function(v) v >= 30, "sum")
#'
#' # Overwrite an existing monoid definition
#' running_count <- measure_monoid(`+`, 0L, function(e) 1L)
#' x3 <- add_monoids(x2, list(sum = running_count), overwrite = TRUE)
#' attr(x3, "measures")$sum
#' @export
#' @seealso [measure_monoid()], [get_measure()], [get_measures()]
# Runtime: O(1) dispatch.
add_monoids <- function(t, monoids, overwrite = FALSE) {
  UseMethod("add_monoids")
}

# Runtime: O(1).
#' @export
#' @noRd
add_monoids.default <- function(t, monoids, overwrite = FALSE) {
  cls <- class(t)
  cls_txt <- if(length(cls) == 0L) "unknown" else paste(cls, collapse = "/")
  stop(sprintf("No `add_monoids()` method for class '%s'.", cls_txt))
}

#' Coerce Objects to `flexseq`
#'
#' `as_flexseq()` is the canonical way to obtain a plain `flexseq` for full
#' sequence-style operations.
#'
#' For base vectors/lists, this builds a new `flexseq` preserving element order
#' and names.
#'
#' For specialized immutable subclasses (`priority_queue`,
#' `ordered_sequence`, `interval_index`), this intentionally drops subclass
#' semantics and returns a plain `flexseq`.
#'
#' @param x Input object.
#' @return A plain `flexseq`.
#' @details
#' This is an S3 generic. Notable method behavior:
#' - `as_flexseq.flexseq(x)` returns `x` unchanged.
#' - `as_flexseq.priority_queue(x)` returns payload items.
#' - `as_flexseq.ordered_sequence(x)` returns payload items.
#' - `as_flexseq.interval_index(x)` returns payload items.
#'
#' For advanced types, custom monoids are dropped and the rebuilt `flexseq`
#' keeps only structural monoids (`.size`, `.named_count`). For
#' `priority_queue`, a flexseq of full entry records can be obtained by
#' composing with [as.list()]: `as_flexseq(as.list(x))`. For
#' `ordered_sequence` and `interval_index`, `as.list()` also returns
#' payload-only lists, so no direct record-preserving cast is provided.
#' @examples
#' x <- as_flexseq(1:3)
#' x
#'
#' q <- priority_queue("a", "b", priorities = c(2, 1))
#' as_flexseq(q)
#'
#' o <- ordered_sequence("a", "b", keys = c(2, 1))
#' as_flexseq(o)
#' @seealso [flexseq()], [priority_queue()], [ordered_sequence()], [interval_index()]
#' @export
# Runtime: O(1) generic dispatch.
as_flexseq <- function(x) {
  UseMethod("as_flexseq")
}

#' Locate First Predicate Match
#'
#' Scans accumulated monoid values and returns the first element where
#' `predicate` becomes `TRUE`, without rebuilding split trees.
#'
#' @param t A `flexseq` (or subclass).
#' @param predicate Function applied to accumulated monoid values.
#' @param monoid_name Name of the monoid used for scanning.
#' @param accumulator Optional starting accumulator value.
#' @param include_metadata Logical; include scan metadata.
#' @return If `include_metadata = FALSE`, a list with:
#' - `found`: logical flag.
#' - `value`: matched element when found, otherwise `NULL`.
#'
#' If `include_metadata = TRUE`, adds `metadata` with:
#' - `left_measure`
#' - `hit_measure`
#' - `right_measure`
#' - `index`
#' @details
#' This is the read-only analogue of [split_around_by_predicate()].
#'
#' As with split helpers, a common setup is a custom monoid created with
#' [measure_monoid()] and attached via [add_monoids()].
#'
#' `value` is the matched leaf entry for the input structure:
#' - `flexseq`: stored user element.
#' - `ordered_sequence`: `list(value, key)`.
#' - `priority_queue`: `list(value, priority)`.
#' - `interval_index`: `list(value, start, end)`.
#' @examples
#' x <- flexseq("a", "b", "c", "d")
#' size_monoid <- measure_monoid(`+`, 0L, function(e) 1L)
#' x2 <- add_monoids(x, list(size = size_monoid))
#'
#' locate_by_predicate(x2, function(v) v >= 3L, "size")
#' locate_by_predicate(x2, function(v) v >= 3L, "size", include_metadata = TRUE)
#' @export
locate_by_predicate <- function(t, predicate, monoid_name, accumulator = NULL, include_metadata = FALSE) {
  UseMethod("locate_by_predicate")
}

#' Split Around First Predicate Match
#'
#' Splits at the first point where a `predicate` function becomes `TRUE` 
#' while scanning the sequence.
#'
#' @param t A `flexseq` (or subclass).
#' @param predicate Function applied to accumulated monoid values.
#' @param monoid_name Name of the monoid used for scanning.
#' @param accumulator Optional starting accumulator value.
#' @return A list with fields:
#' - `left`: elements before the split point.
#' - `value`: the matched element at the split point.
#' - `right`: elements after the split point.
#' @details
#' This function generally requires the sequence be annotated with
#' a `measure_monoid()`; see the examples and `measure_monoid()`
#' for more information.
#'
#' `value` is the matched leaf entry for the input structure:
#' - `flexseq`: stored user element.
#' - `ordered_sequence`: `list(value, key)`.
#' - `priority_queue`: `list(value, priority)`.
#' - `interval_index`: `list(value, start, end)`.
#'
#' `left` and `right` preserve subclass when the input is a subclass of
#' `flexseq`.
#' @examples
#' x <- flexseq("a", "b", "c", "d")
#' 
#' # Each element e has measure 1; the accumulated measure for
#' # a set of elements is computed by the associative function `+`
#' # along with the identity value 0.
#' size_monoid <- measure_monoid(`+`, 0L, function(e) 1L)
#' x2 <- add_monoids(x, list(size = size_monoid))
#' 
#' # the first time the measure stored in the size monoid
#' # accumulates to greater than or equal to 3 is the 3rd
#' # element in sequence.
#' split_around_by_predicate(x2, function(v) v >= 3L, "size")
#'
#' # Split at the first element
#' split_around_by_predicate(x2, function(v) v >= 1L, "size")
#'
#' # Split at the last element
#' split_around_by_predicate(x2, function(v) v >= 4L, "size")
#' @export
split_around_by_predicate <- function(t, predicate, monoid_name, accumulator = NULL) {
  UseMethod("split_around_by_predicate")
}

#' Split into Left and Right Parts
#'
#' Splits at the first point where `predicate` becomes `TRUE` while scanning
#' the sequence.
#'
#' @param x A `flexseq` (or subclass).
#' @param predicate Function applied to accumulated monoid values.
#' @param monoid_name Name of the monoid used for scanning.
#' @return A list with fields:
#' - `left`: elements before the split point.
#' - `right`: elements from the split point onward.
#' @details
#' This is the two-way variant of [split_around_by_predicate()].
#'
#' As with [split_around_by_predicate()], a common setup is a custom monoid
#' created with [measure_monoid()] and attached via [add_monoids()].
#'
#' `left` and `right` preserve subclass when the input is a subclass of
#' `flexseq`.
#' @examples
#' x <- flexseq("a", "b", "c", "d")
#' size_monoid <- measure_monoid(`+`, 0L, function(e) 1L)
#' x2 <- add_monoids(x, list(size = size_monoid))
#' split_by_predicate(x2, function(v) v >= 3L, "size")
#'
#' # Split at the first element
#' split_by_predicate(x2, function(v) v >= 1L, "size")
#' @export
split_by_predicate <- function(x, predicate, monoid_name) {
  UseMethod("split_by_predicate")
}

#' Split at a Position or Name
#'
#' Splits by a single one-based position or a single element name.
#'
#' @param x A `flexseq`.
#' @param at A single positive integer position or a single character name.
#' @param pull_index Controls output shape:
#' - `FALSE` (default): returns `list(left, value, right)`.
#' - `TRUE`: returns `list(left, right)`.
#' @return A split result with shape controlled by `pull_index`.
#' @details
#' `split_at(x, at, pull_index = FALSE)` is a convenience wrapper around
#' [split_around_by_predicate()] using positional scanning.
#'
#' `split_at(x, at, pull_index = TRUE)` is the two-way variant using
#' [split_by_predicate()].
#' @examples
#' x <- flexseq("a", "b", "c", "d")
#' split_at(x, 3)
#' split_at(x, 3, pull_index = TRUE)
#'
#' n <- flexseq(a = 1, b = 2, c = 3)
#' split_at(n, "b")
#' @export
split_at <- function(x, at, pull_index = FALSE) {
  UseMethod("split_at")
}

#' Insert an Element
#'
#' Inserts an element into a structure-specific position according to class
#' semantics.
#'
#' @param x Object to insert into.
#' @param ... Method-specific arguments.
#' @return Updated object of the same class as `x`.
#' @details
#' `insert()` is an S3 generic. Required arguments in `...` depend on `x`:
#' - `priority_queue`: `element`, `priority` (optional `name`)
#' - `ordered_sequence`: `element`, `key` (optional `name`)
#' - `interval_index`: `element`, `start`, `end` (optional `name`)
#'
#' This operation is persistent: `x` is not modified.
#' @examples
#' q <- priority_queue("a", "b", priorities = c(2, 1))
#' insert(q, "c", priority = 3)
#'
#' o <- ordered_sequence("a", "c", keys = c(1, 3))
#' insert(o, "b", key = 2)
#'
#' iv <- interval_index("A", "B", start = c(1, 5), end = c(3, 8))
#' insert(iv, "C", start = 2, end = 6)
#' @seealso [priority_queue()], [ordered_sequence()], [interval_index()],
#'   [insert_at()]
#' @export
insert <- function(x, ...) {
  UseMethod("insert")
}

#' @export
#' @noRd
insert.default <- function(x, ...) {
  cls <- class(x)
  cls_txt <- if(length(cls) == 0L) "unknown" else paste(cls, collapse = "/")
  stop(sprintf("No `insert()` method for class '%s'.", cls_txt))
}

testthat::test_that("monoid caches are stored in measures attr", {
  m_count <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  t <- add_monoids(as_flexseq(letters[1:5]), list(count = m_count))

  ms <- attr(t, "measures")
  testthat::expect_identical(ms$count, 5)
  testthat::expect_identical(ms$.size, 5)
})

testthat::test_that("measures propagate through prepend/append", {
  m_count <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  t <- add_monoids(flexseq(), list(count = m_count))
  t <- push_back(t, "a")
  t <- push_back(t, "b")
  t <- push_front(t, "z")

  ms <- attr(t, "measures")
  testthat::expect_identical(ms$count, 3)
  testthat::expect_identical(ms$.size, 3)
})

testthat::test_that("custom measure works with list elements", {
  m_len <- measure_monoid(function(a, b) a + b, 0, function(el) length(el))
  t <- add_monoids(flexseq(), list(len = m_len))
  t <- push_back(t, list(1, 2, 3))
  t <- push_back(t, list("a"))
  t <- push_back(t, list(TRUE, FALSE))

  testthat::expect_identical(attr(t, "measures")$len, 6)
})

testthat::test_that("multiple measures propagate through concat", {
  m_sum <- measure_monoid(function(a, b) a + b, 0, function(el) el)
  m_count <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  ms <- list(sum = m_sum, count = m_count)

  t1 <- add_monoids(as_flexseq(1:3), ms)
  t1 <- push_back(t1, 4)
  t2 <- add_monoids(as_flexseq(10:11), ms)
  t <- c(t1, t2)

  root <- attr(t, "measures")
  testthat::expect_identical(root$sum, 31)
  testthat::expect_identical(root$count, 6)
  testthat::expect_identical(root$.size, 6)
})

testthat::test_that("str() works for flexseq-like classes with restricted indexing", {
  x <- as_flexseq(list(a = 1, b = 2, c = 3))
  q <- priority_queue(one = 1, two = 2, three = 3, priorities = 1:3)
  os <- ordered_sequence(one = 1, two = 2, three = 3, keys = 1:3)
  ix <- interval_index(
    one = 1, two = 2, three = 3,
    start = c(1, 5, 10),
    end = c(3, 7, 12)
  )

  out_x <- testthat::expect_no_error(utils::capture.output(utils::str(x)))
  out_q <- testthat::expect_no_error(utils::capture.output(utils::str(q)))
  out_os <- testthat::expect_no_error(utils::capture.output(utils::str(os)))
  out_ix <- testthat::expect_no_error(utils::capture.output(utils::str(ix)))

  testthat::expect_true(any(grepl("^List of ", out_x)))
  testthat::expect_true(any(grepl("^List of ", out_q)))
  testthat::expect_true(any(grepl("^List of ", out_os)))
  testthat::expect_true(any(grepl("^List of ", out_ix)))

  blocked_msg <- "supports scalar character names only"
  testthat::expect_false(any(grepl(blocked_msg, out_q, fixed = TRUE)))
  testthat::expect_false(any(grepl(blocked_msg, out_os, fixed = TRUE)))
  testthat::expect_false(any(grepl(blocked_msg, out_ix, fixed = TRUE)))
})

testthat::test_that("print show_custom_monoids displays custom names and root measures only", {
  fx <- add_monoids(
    as_flexseq(as.list(c(1, 2))),
    list(sum_item = measure_monoid(`+`, 0, function(el) as.numeric(el)))
  )
  pq <- add_monoids(
    priority_queue(one = 1, two = 2, priorities = c(2, 1)),
    list(sum_item = measure_monoid(`+`, 0, function(entry) as.numeric(entry$value)))
  )
  os <- add_monoids(
    ordered_sequence(one = 1, two = 2, keys = c(2, 1)),
    list(sum_key = measure_monoid(`+`, 0, function(entry) as.numeric(entry$key)))
  )
  ix <- add_monoids(
    interval_index(one = 1, two = 2, start = c(1, 3), end = c(2, 5)),
    list(width_sum = measure_monoid(`+`, 0, function(entry) as.numeric(entry$end - entry$start)))
  )

  out_fx <- utils::capture.output(print(fx, max_elements = 0, show_custom_monoids = TRUE))
  out_pq <- utils::capture.output(print(pq, max_elements = 0, show_custom_monoids = TRUE))
  out_os <- utils::capture.output(print(os, max_elements = 0, show_custom_monoids = TRUE))
  out_ix <- utils::capture.output(print(ix, max_elements = 0, show_custom_monoids = TRUE))

  testthat::expect_true(any(grepl("Custom monoids + measures:", out_fx, fixed = TRUE)))
  testthat::expect_true(any(grepl("sum_item: 3", out_fx, fixed = TRUE)))
  testthat::expect_false(any(grepl(".size", out_fx, fixed = TRUE)))

  testthat::expect_true(any(grepl("Custom monoids + measures:", out_pq, fixed = TRUE)))
  testthat::expect_true(any(grepl("sum_item: 3", out_pq, fixed = TRUE)))
  testthat::expect_false(any(grepl(".pq_min", out_pq, fixed = TRUE)))
  testthat::expect_false(any(grepl(".pq_max", out_pq, fixed = TRUE)))

  testthat::expect_true(any(grepl("Custom monoids + measures:", out_os, fixed = TRUE)))
  testthat::expect_true(any(grepl("sum_key: 3", out_os, fixed = TRUE)))
  testthat::expect_false(any(grepl(".oms_max_key", out_os, fixed = TRUE)))

  testthat::expect_true(any(grepl("Custom monoids + measures:", out_ix, fixed = TRUE)))
  testthat::expect_true(any(grepl("width_sum: 3", out_ix, fixed = TRUE)))
  testthat::expect_false(any(grepl(".ivx_max_start", out_ix, fixed = TRUE)))
  testthat::expect_false(any(grepl(".ivx_max_end", out_ix, fixed = TRUE)))
  testthat::expect_false(any(grepl(".ivx_min_end", out_ix, fixed = TRUE)))

  out_default <- utils::capture.output(print(fx, max_elements = 0))
  testthat::expect_false(any(grepl("Custom monoids + measures:", out_default, fixed = TRUE)))
})

testthat::test_that("print show_custom_monoids prints <none> with no custom monoids and validates input", {
  plain_fx <- as_flexseq(as.list(c(1, 2)))
  plain_pq <- priority_queue(1, 2, priorities = c(2, 1))
  plain_os <- ordered_sequence(1, 2, keys = c(2, 1))
  plain_ix <- interval_index(1, 2, start = c(1, 3), end = c(2, 5))

  out_fx <- utils::capture.output(print(plain_fx, max_elements = 0, show_custom_monoids = TRUE))
  out_pq <- utils::capture.output(print(plain_pq, max_elements = 0, show_custom_monoids = TRUE))
  out_os <- utils::capture.output(print(plain_os, max_elements = 0, show_custom_monoids = TRUE))
  out_ix <- utils::capture.output(print(plain_ix, max_elements = 0, show_custom_monoids = TRUE))

  testthat::expect_true(any(grepl("Custom monoids + measures: <none>", out_fx, fixed = TRUE)))
  testthat::expect_true(any(grepl("Custom monoids + measures: <none>", out_pq, fixed = TRUE)))
  testthat::expect_true(any(grepl("Custom monoids + measures: <none>", out_os, fixed = TRUE)))
  testthat::expect_true(any(grepl("Custom monoids + measures: <none>", out_ix, fixed = TRUE)))

  testthat::expect_error(print(plain_fx, max_elements = 0, show_custom_monoids = NA), "TRUE or FALSE")
  testthat::expect_error(print(plain_pq, max_elements = 0, show_custom_monoids = 1), "TRUE or FALSE")
})

testthat::test_that("get_measure returns cached aggregate for a user-attached monoid", {
  sum_m <- measure_monoid(function(a, b) a + b, 0, function(el) el)
  x <- add_monoids(as_flexseq(c(3, 1, 4, 1, 5, 9, 2, 6)), list(sum = sum_m))
  testthat::expect_identical(get_measure(x, "sum"), 31)
})

testthat::test_that("get_measure returns built-in .size as element count", {
  x <- as_flexseq(letters[1:5])
  testthat::expect_identical(get_measure(x, ".size"), 5)
  testthat::expect_identical(get_measure(flexseq(), ".size"), 0)
})

testthat::test_that("get_measure reads list-valued built-in monoids on priority_queue", {
  q <- priority_queue("a", "b", "c", priorities = c(5, 1, 3))
  m_min <- get_measure(q, ".pq_min")
  testthat::expect_true(isTRUE(m_min$has))
  testthat::expect_identical(m_min$priority, 1)

  m_max <- get_measure(q, ".pq_max")
  testthat::expect_true(isTRUE(m_max$has))
  testthat::expect_identical(m_max$priority, 5)
})

testthat::test_that("get_measure aggregate reflects splits via split_by_predicate", {
  sum_m <- measure_monoid(function(a, b) a + b, 0, function(el) el)
  x <- add_monoids(as_flexseq(c(1, 2, 3, 4, 5)), list(sum = sum_m))
  parts <- split_by_predicate(x, function(m) m >= 6, "sum")
  testthat::expect_identical(get_measure(parts$left, "sum") + get_measure(parts$right, "sum"), 15)
})

testthat::test_that("get_measures returns a per-element flexseq", {
  sum_m <- measure_monoid(function(a, b) a + b, 0, function(el) el)
  x <- add_monoids(as_flexseq(c(3, 1, 4, 1, 5, 9)), list(sum = sum_m))
  per <- get_measures(x, "sum")
  testthat::expect_true("flexseq" %in% class(per))
  testthat::expect_identical(length(per), 6L)
  testthat::expect_identical(unlist(as.list(per)), c(3, 1, 4, 1, 5, 9))
})

testthat::test_that("get_measures on an empty flexseq returns an empty flexseq", {
  sum_m <- measure_monoid(function(a, b) a + b, 0, function(el) el)
  x <- add_monoids(flexseq(), list(sum = sum_m))
  per <- get_measures(x, "sum")
  testthat::expect_true("flexseq" %in% class(per))
  testthat::expect_identical(length(per), 0L)
})

testthat::test_that("get_measures on priority_queue yields per-element .pq_min records", {
  q <- priority_queue("a", "b", "c", priorities = c(5, 1, 3))
  per <- get_measures(q, ".pq_min")
  testthat::expect_identical(length(per), 3L)
  rec1 <- as.list(per)[[1]]
  testthat::expect_true(isTRUE(rec1$has))
  testthat::expect_identical(rec1$priority, 5)
})

testthat::test_that("get_measure errors when x is not a finger-tree structure", {
  testthat::expect_error(get_measure(list(), "sum"), "finger-tree structure")
  testthat::expect_error(get_measure(1:3, ".size"), "finger-tree structure")
  testthat::expect_error(get_measures(list(), "sum"), "finger-tree structure")
})

testthat::test_that("get_measure errors when monoid_name is not a single non-empty string", {
  x <- as_flexseq(1:3)
  testthat::expect_error(get_measure(x, 1),              "single non-empty string")
  testthat::expect_error(get_measure(x, c("a", "b")),    "single non-empty string")
  testthat::expect_error(get_measure(x, NA_character_),  "single non-empty string")
  testthat::expect_error(get_measure(x, ""),             "single non-empty string")
  testthat::expect_error(get_measures(x, NA_character_), "single non-empty string")
})

testthat::test_that("get_measure errors when the monoid is not attached", {
  x <- as_flexseq(1:3)
  testthat::expect_error(get_measure(x, "not_attached"), "not attached")
  testthat::expect_error(get_measures(x, "not_attached"), "not attached")
})

testthat::test_that("fapply maps flexseq values in order", {
  x <- as_flexseq(1:5)
  y <- fapply(x, function(v) v * 10)

  testthat::expect_s3_class(y, "flexseq")
  testthat::expect_identical(as.integer(node_measure(y, ".size")), 5L)
  testthat::expect_identical(
    lapply(.ft_to_list(y), .ft_strip_name),
    as.list(c(10, 20, 30, 40, 50))
  )
})

testthat::test_that("fapply preserves element names for named flexseq", {
  x <- as_flexseq(setNames(as.list(1:4), c("a", "b", "c", "d")))
  y <- fapply(x, function(v) v + 1L)

  nms <- .ft_collect_names(y)
  testthat::expect_identical(nms, c("a", "b", "c", "d"))
  testthat::expect_identical(y[["c"]], 4L)
})

testthat::test_that("fapply preserves user monoids for flexseq", {
  sum_m <- measure_monoid(`+`, 0, as.numeric)
  x <- add_monoids(as_flexseq(1:4), list(sum = sum_m))
  y <- fapply(x, function(v) v * 2)

  ms <- attr(y, "monoids", exact = TRUE)
  testthat::expect_true(!is.null(ms[[".size"]]))
  testthat::expect_true(!is.null(ms[[".named_count"]]))
  testthat::expect_true(!is.null(ms[["sum"]]))
  testthat::expect_identical(node_measure(y, "sum"), 20)
  testthat::expect_identical(as.integer(node_measure(y, ".size")), 4L)
  testthat::expect_false(identical(node_measure(x, "sum"), node_measure(y, "sum")))
  testthat::expect_identical(node_measure(x, "sum"), 10)
})

testthat::test_that("fapply can drop user monoids for flexseq", {
  sum_m <- measure_monoid(`+`, 0, as.numeric)
  x <- add_monoids(as_flexseq(1:4), list(sum = sum_m))
  y <- fapply(x, function(v) v * 2, preserve_custom_monoids = FALSE)

  ms <- attr(y, "monoids", exact = TRUE)
  testthat::expect_true(!is.null(ms[[".size"]]))
  testthat::expect_true(!is.null(ms[[".named_count"]]))
  testthat::expect_true(is.null(ms[["sum"]]))
  testthat::expect_error(node_measure(y, "sum"), "Missing cached measure")
  testthat::expect_identical(as.integer(node_measure(y, ".size")), 4L)
})

testthat::test_that("fapply validates flexseq inputs", {
  x <- as_flexseq(1:3)
  testthat::expect_error(fapply.flexseq(list(1, 2, 3), FUN = identity), "`x` must be a flexseq")
  testthat::expect_error(fapply(x, 1), "`FUN` must be a function")
  testthat::expect_error(fapply(x, identity, preserve_custom_monoids = NA), "`preserve_custom_monoids` must be TRUE or FALSE")
})

testthat::test_that("split_around_by_predicate returns distinguished element and context", {
  mr <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  t <- add_monoids(as_flexseq(letters[1:6]), list(count = mr))

  s <- split_around_by_predicate(t, function(v) v >= 4, ".size")
  testthat::expect_identical(s$value, "d")
  testthat::expect_identical(attr(s$left, "measures")$.size, 3)
  testthat::expect_identical(attr(s$right, "measures")$.size, 2)
})

testthat::test_that("split_by_predicate returns left/right trees with split element prepended to right", {
  mr <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  t <- add_monoids(as_flexseq(letters[1:6]), list(count = mr))

  s <- split_by_predicate(t, function(v) v >= 4, ".size")
  testthat::expect_identical(attr(s$left, "measures")$.size, 3)
  testthat::expect_identical(attr(s$right, "measures")$.size, 3)
  testthat::expect_identical(tree_chars(s$left), "abc")
  testthat::expect_identical(tree_chars(s$right), "def")
})

testthat::test_that("split_by_predicate handles empty and predicate-never-true cases", {
  e <- flexseq()
  s0 <- split_by_predicate(e, function(v) v >= 1, ".size")
  testthat::expect_true(s0$left %isa% Empty)
  testthat::expect_true(s0$right %isa% Empty)

  t <- as_flexseq(letters[1:4])
  s <- split_by_predicate(t, function(v) v >= 10, ".size")
  testthat::expect_identical(attr(s$left, "measures")$.size, 4)
  testthat::expect_true(s$right %isa% Empty)
})

testthat::test_that("split_by_predicate boundary at first and last element", {
  t <- as_flexseq(letters[1:5])

  s1 <- split_by_predicate(t, function(v) v >= 1, ".size")
  testthat::expect_true(s1$left %isa% Empty)
  testthat::expect_identical(attr(s1$right, "measures")$.size, 5)
  testthat::expect_identical(tree_chars(s1$right), "abcde")

  s2 <- split_by_predicate(t, function(v) v >= 5, ".size")
  testthat::expect_identical(attr(s2$left, "measures")$.size, 4)
  testthat::expect_identical(attr(s2$right, "measures")$.size, 1)
  testthat::expect_identical(tree_chars(s2$left), "abcd")
  testthat::expect_identical(tree_chars(s2$right), "e")
})

testthat::test_that("split_around_by_predicate respects custom accumulator offsets", {
  t <- as_flexseq(letters[1:5])

  s <- split_around_by_predicate(t, function(v) v >= 8, ".size", accumulator = 5)
  testthat::expect_identical(s$value, "c")
  testthat::expect_identical(attr(s$left, "measures")$.size, 2)
  testthat::expect_identical(attr(s$right, "measures")$.size, 2)
})

testthat::test_that("split_around_by_predicate errors when monoid name is missing", {
  t <- as_flexseq(letters[1:4])
  testthat::expect_error(split_around_by_predicate(t, function(v) v >= 2, "unknown"), "not found")
})

testthat::test_that("split_at supports scalar integer index", {
  t <- as_flexseq(letters[1:6])

  s <- split_at(t, 4)
  testthat::expect_identical(tree_chars(s$left), "abc")
  testthat::expect_identical(s$value, "d")
  testthat::expect_identical(tree_chars(s$right), "ef")

  s2 <- split_at(t, 4, pull_index = TRUE)
  testthat::expect_identical(tree_chars(s2$left), "abc")
  testthat::expect_identical(tree_chars(s2$right), "def")
})

testthat::test_that("split_at supports scalar name", {
  t <- as_flexseq(setNames(as.list(letters[1:6]), LETTERS[1:6]))

  s <- split_at(t, "D")
  testthat::expect_identical(.ft_strip_name(s$value), "d")
  testthat::expect_identical(attr(s$left, "measures")$.size, 3)
  testthat::expect_identical(attr(s$right, "measures")$.size, 2)

  s2 <- split_at(t, "D", pull_index = TRUE)
  testthat::expect_identical(attr(s2$left, "measures")$.size, 3)
  testthat::expect_identical(attr(s2$right, "measures")$.size, 3)
})

#SO

validate_fingertree_invariants <- function(t) {
  is_homogeneous <- function(xs) {
    if(length(xs) <= 1) {
      return(TRUE)
    }
    kinds <- vapply(xs, is_structural_node, logical(1))
    all(kinds == kinds[[1]])
  }

  walk <- function(x) {
    if(x %isa% Empty) {
      testthat::expect_true(!is.null(attr(x, "monoids")))
      testthat::expect_true(!is.null(attr(x, "measures")))
      return(invisible(TRUE))
    }

    if(x %isa% Single) {
      testthat::expect_identical(length(x), 1L)
      testthat::expect_true(!is.null(attr(x, "monoids")))
      testthat::expect_true(!is.null(attr(x, "measures")))
      return(invisible(TRUE))
    }

    if(x %isa% Deep) {
      testthat::expect_true(length(x$prefix) >= 1 && length(x$prefix) <= 4)
      testthat::expect_true(length(x$suffix) >= 1 && length(x$suffix) <= 4)
      testthat::expect_true(is_homogeneous(x$prefix))
      testthat::expect_true(is_homogeneous(x$suffix))
      testthat::expect_true(!is.null(attr(x, "monoids")))
      testthat::expect_true(!is.null(attr(x, "measures")))
      walk(x$middle)
      return(invisible(TRUE))
    }

    if(x %isa% Node) {
      testthat::expect_true(length(x) %in% c(2L, 3L))
      testthat::expect_true(is_homogeneous(x))
      testthat::expect_true(!is.null(attr(x, "monoids")))
      testthat::expect_true(!is.null(attr(x, "measures")))
      for(el in x) {
        if(is_structural_node(el)) {
          walk(el)
        }
      }
      return(invisible(TRUE))
    }

    if(x %isa% Digit) {
      testthat::expect_true(length(x) >= 1 && length(x) <= 4)
      testthat::expect_true(is_homogeneous(x))
      testthat::expect_true(!is.null(attr(x, "monoids")))
      testthat::expect_true(!is.null(attr(x, "measures")))
      for(el in x) {
        if(is_structural_node(el)) {
          walk(el)
        }
      }
      return(invisible(TRUE))
    }
  }

  walk(t)
  invisible(TRUE)
}

testthat::test_that("structural invariants hold after mixed updates", {
  mr <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  t <- add_monoids(flexseq(), list(count = mr))
  for(i in 1:30) {
    if(i %% 2 == 0) {
      t <- push_back(t, letters[i %% 26 + 1])
    } else {
      t <- push_front(t, letters[i %% 26 + 1])
    }
  }
  validate_fingertree_invariants(t)
})

testthat::test_that("structural invariants hold after concat and split", {
  mr <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  t1 <- add_monoids(as_flexseq(letters[1:12]), list(count = mr))
  t2 <- add_monoids(as_flexseq(letters[13:20]), list(count = mr))
  t <- c(t1, t2)
  validate_fingertree_invariants(t)

  s <- split_by_predicate(t, function(v) v >= 10, ".size")
  validate_fingertree_invariants(s$left)
  validate_fingertree_invariants(s$right)
})

testthat::test_that("measured deep trees retain root measure through updates", {
  mr <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  t <- add_monoids(as_flexseq(letters[1:10]), list(count = mr))
  t <- push_front(t, "z")
  t <- push_back(t, "y")
  testthat::expect_identical(attr(t, "measures")$.size, 12)
  validate_fingertree_invariants(t)
})

testthat::test_that("read indexing supports [] and [[ with positive integer indices", {
  t <- as_flexseq(1:6)

  testthat::expect_equal(t[[3]], 3)

  s <- t[c(2, 4, 6)]
  testthat::expect_equal(sum(unlist(as.list(s), use.names = FALSE)), 12)
  testthat::expect_identical(attr(s, "measures")$.size, 3)
})

testthat::test_that("read indexing supports duplicates and arbitrary order", {
  t <- as_flexseq(letters[1:5])
  s <- t[c(5, 2, 2, 4)]
  testthat::expect_identical(tree_chars(s), "ebbd")
  testthat::expect_identical(attr(s, "measures")$.size, 4)
})

testthat::test_that("indexing enforces bounds and integer-only indices", {
  r <- measure_monoid(function(a, b) a + b, 0, function(el) el)
  t <- add_monoids(as_flexseq(1:4), list(sum = r))

  testthat::expect_error(t[[0]], "positive integer")
  testthat::expect_error(t[[5]], "out of bounds")
  testthat::expect_error(t[[c(1, 2)]], "\\[\\[ expects exactly one index")
  testthat::expect_error(t[c(1, -2)], "positive integer")
  testthat::expect_error(t[c(1, 2.5)], "integer indices")
})

testthat::test_that("replacement indexing supports [[<- and [<- with recycling semantics", {
  t <- as_flexseq(1:5)

  t2 <- t
  t2[[2]] <- 20
  testthat::expect_identical(sum(unlist(as.list(t2), use.names = FALSE)), 33)
  testthat::expect_identical(t2[[2]], 20)

  t3 <- t
  t3[c(1, 5)] <- list(10, 50)
  testthat::expect_identical(sum(unlist(as.list(t3), use.names = FALSE)), 69)
  testthat::expect_identical(t3[[1]], 10)
  testthat::expect_identical(t3[[5]], 50)

  t4 <- t
  t4[c(1, 2, 3)] <- list(99, 100)
  testthat::expect_identical(t4[[1]], 99)
  testthat::expect_identical(t4[[2]], 100)
  testthat::expect_identical(t4[[3]], 99)
})

testthat::test_that("[[<- keeps monoid attrs and updates one element via split path", {
  sum_m <- measure_monoid(`+`, 0, as.numeric)
  t <- add_monoids(as_flexseq(1:6), list(sum = sum_m))
  t2 <- t
  t2[[4]] <- 99

  testthat::expect_identical(attr(t2, "measures")$.size, 6)
  testthat::expect_identical(attr(t2, "measures")$sum, 116)
  testthat::expect_identical(t2[[4]], 99)
})

testthat::test_that("[[<- with NULL removes one element by index or name", {
  sum_m <- measure_monoid(`+`, 0, as.numeric)
  t <- add_monoids(as_flexseq(1:5), list(sum = sum_m))
  t[[3]] <- NULL
  testthat::expect_identical(attr(t, "measures")$.size, 4)
  testthat::expect_identical(attr(t, "measures")$sum, 12)
  testthat::expect_identical(t[[3]], 4L)

  tn <- as_flexseq(setNames(as.list(1:4), c("a", "b", "c", "d")))
  tn[["c"]] <- NULL
  testthat::expect_identical(attr(tn, "measures")$.size, 3)
  testthat::expect_identical(attr(tn, "measures")$.named_count, 3L)
  testthat::expect_error(tn[["c"]], "Unknown element name")
  testthat::expect_identical(tn[["d"]], 4L)
})

testthat::test_that("[<- applies duplicate indices sequentially (last write wins)", {
  t <- as_flexseq(letters[1:5])
  t[c(2, 2)] <- list("X", "Y")
  testthat::expect_identical(t[[2]], "Y")
})

testthat::test_that("[<- supports arbitrary replacement order", {
  t <- as_flexseq(letters[1:5])
  t[c(5, 1, 4)] <- list("u", "v", "w")
  testthat::expect_identical(t[[1]], "v")
  testthat::expect_identical(t[[4]], "w")
  testthat::expect_identical(t[[5]], "u")
})

testthat::test_that("[<- keeps structural attrs and custom monoid measures coherent", {
  sum_m <- measure_monoid(`+`, 0, as.numeric)
  t <- add_monoids(as_flexseq(1:5), list(sum = sum_m))
  t[c(2, 4)] <- list(20, 40)

  testthat::expect_identical(attr(t, "measures")$.size, 5)
  testthat::expect_identical(attr(t, "measures")$sum, 69)
  testthat::expect_identical(sum(unlist(as.list(t), use.names = FALSE)), 69)
})

testthat::test_that("[<- zero-length index is a no-op when replacement is empty", {
  t <- as_flexseq(letters[1:5])
  before <- tree_chars(t)
  t2 <- t
  t2[integer(0)] <- list()

  testthat::expect_identical(tree_chars(t2), before)
  testthat::expect_identical(attr(t2, "measures")$.size, attr(t, "measures")$.size)
})

testthat::test_that("size measure is available by default for indexing even if not provided", {
  sum_only <- measure_monoid(function(a, b) a + b, 0, function(el) el)
  t <- add_monoids(as_flexseq(1:3), list(sum = sum_only))
  ms <- attr(t, "measures")
  testthat::expect_identical(ms$.size, 3)
  testthat::expect_equal(t[[2]], 2)
})

testthat::test_that("single-element read indexing follows locate_by_predicate(.size) semantics", {
  t <- as_flexseq(letters[1:8])
  for(i in 1:8) {
    hit <- locate_by_predicate(t, function(v) v >= i, ".size")
    testthat::expect_true(hit$found)
    testthat::expect_identical(t[[i]], hit$value)
  }
})

testthat::test_that("name invariants enforce complete unique naming when names are present", {
  t <- as_flexseq(setNames(as.list(1:4), c("a", "b", "c", "d")))
  testthat::expect_identical(attr(t, "measures")$.size, 4)
  testthat::expect_identical(attr(t, "measures")$.named_count, 4L)

  mixed <- as.list(1:3)
  names(mixed) <- c("x", "", "z")
  testthat::expect_error(as_flexseq(mixed), "Mixed named and unnamed")

  dup <- as.list(1:3)
  names(dup) <- c("x", "x", "z")
  testthat::expect_error(as_flexseq(dup), "must be unique")
})

testthat::test_that("character read indexing returns ordered results and NULL placeholders", {
  t <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))
  s <- t[c("c", "missing", "a")]

  testthat::expect_identical(attr(s, "measures")$.size, 3)
  testthat::expect_identical(s[[1]], 3L)
  testthat::expect_null(s[[2]])
  testthat::expect_identical(s[[3]], 1L)
})

testthat::test_that("character lookup paths are equivalent for short and wide queries", {
  vals <- as.list(1:20)
  names(vals) <- paste0("k", seq_len(20))
  t <- as_flexseq(vals)

  q_short <- c("k3", "missing", "k1")
  s_short <- t[q_short]
  exp_short <- lapply(q_short, function(nm) if(nm %in% names(vals)) vals[[nm]] else NULL)
  got_short <- lapply(seq_along(q_short), function(i) s_short[[i]])
  testthat::expect_identical(got_short, exp_short)

  q_wide <- c("k3", "missing", "k1", "k10", "k10", "k20", "missing", "k2", "k19", "k4")
  s_wide <- t[q_wide]
  exp_wide <- lapply(q_wide, function(nm) if(nm %in% names(vals)) vals[[nm]] else NULL)
  got_wide <- lapply(seq_along(q_wide), function(i) s_wide[[i]])
  testthat::expect_identical(got_wide, exp_wide)
})

testthat::test_that("character [[ is strict single-name lookup", {
  t <- as_flexseq(setNames(as.list(letters[1:3]), c("x", "y", "z")))
  testthat::expect_identical(t[["y"]], "b")
  testthat::expect_error(t[["missing"]], "Unknown element name")
})

testthat::test_that("character replacement indexing updates by existing names only", {
  t <- as_flexseq(setNames(as.list(1:4), c("a", "b", "c", "d")))
  t[c("b", "b")] <- list(20, 30)
  testthat::expect_identical(t[["b"]], 30)
  testthat::expect_identical(attr(t, "measures")$.named_count, 4L)

  t[c("a", "c", "d")] <- list(7, 8)
  testthat::expect_identical(t[["a"]], 7)
  testthat::expect_identical(t[["c"]], 8)
  testthat::expect_identical(t[["d"]], 7)

  testthat::expect_error({
    t2 <- t
    t2[c("a", "missing")] <- list(1, 2)
  }, "Unknown element name")
})

testthat::test_that("replacement names win and must remain unique", {
  t <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))
  repl <- list(10, 30)
  names(repl) <- c("z", "y")
  t[c("a", "c")] <- repl

  testthat::expect_identical(t[["z"]], 10)
  testthat::expect_identical(t[["y"]], 30)
  testthat::expect_error(t[["a"]], "Unknown element name")

  bad <- list(100, 200)
  names(bad) <- c("k", "k")
  testthat::expect_error({
    t2 <- as_flexseq(setNames(as.list(1:3), c("u", "v", "w")))
    t2[c("u", "v")] <- bad
  }, "must be unique")
})

testthat::test_that("single-name replacement enforces uniqueness without map rebuild semantics change", {
  t <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))

  testthat::expect_error({
    t2 <- t
    t2[[1]] <- stats::setNames(9, "b")
  }, "must be unique")

  t3 <- t
  t3[[1]] <- stats::setNames(9, "a")
  testthat::expect_identical(unname(t3[["a"]]), 9)
  testthat::expect_identical(attr(t3, "measures")$.named_count, 3L)
})

testthat::test_that("append and prepend preserve global name-state invariant", {
  unnamed <- as_flexseq(1:3)
  testthat::expect_error(
    push_back(unnamed, stats::setNames(4, "x")),
    "mixed named and unnamed"
  )

  named <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))
  testthat::expect_error(push_front(named, 0), "mixed named and unnamed")
})

testthat::test_that("$ provides strict exact-name access and replacement", {
  t <- as_flexseq(setNames(as.list(1:4), c("a", "b", "c", "d")))
  testthat::expect_identical(t$b, 2L)
  testthat::expect_identical(t$b, t[["b"]])
  testthat::expect_error(t$missing, "Unknown element name")

  t$b <- 99
  testthat::expect_identical(t$b, 99)
  testthat::expect_identical(attr(t, "measures")$.named_count, 4L)

  t$b <- NULL
  testthat::expect_identical(attr(t, "measures")$.size, 3)
  testthat::expect_identical(attr(t, "measures")$.named_count, 3L)
  testthat::expect_error(t$b, "Unknown element name")
})

testthat::test_that("logical read indexing supports recycling and rejects NA", {
  t <- as_flexseq(1:5)
  s <- t[c(TRUE, FALSE)]
  testthat::expect_identical(attr(s, "measures")$.size, 3)
  testthat::expect_identical(s[[1]], 1L)
  testthat::expect_identical(s[[2]], 3L)
  testthat::expect_identical(s[[3]], 5L)

  z <- t[logical(0)]
  testthat::expect_identical(attr(z, "measures")$.size, 0)
  testthat::expect_error(t[c(TRUE, NA)], "cannot contain NA")
})

testthat::test_that("logical replacement supports recycling and preserves size", {
  t <- as_flexseq(1:6)
  t[c(TRUE, FALSE, TRUE)] <- list(99)
  testthat::expect_identical(attr(t, "measures")$.size, 6)
  testthat::expect_identical(t[[1]], 99)
  testthat::expect_identical(t[[3]], 99)
  testthat::expect_identical(t[[4]], 99)
  testthat::expect_identical(t[[6]], 99)

  testthat::expect_error({
    t2 <- as_flexseq(1:4)
    t2[c(TRUE, FALSE)] <- list()
  }, "length 0")
})

testthat::test_that("as.list.flexseq returns elements in sequence order", {
  x <- as_flexseq(letters[1:5])
  out <- as.list(x)

  testthat::expect_type(out, "list")
  testthat::expect_identical(out, as.list(letters[1:5]))
})

testthat::test_that("as.list.flexseq preserves element names", {
  x <- as_flexseq(setNames(as.list(1:4), c("a", "b", "c", "d")))
  out <- as.list(x)

  testthat::expect_identical(names(out), c("a", "b", "c", "d"))
  testthat::expect_identical(unname(out), as.list(1:4))
})

testthat::test_that("as.list.flexseq strips internal ft_name metadata", {
  x <- as_flexseq(setNames(as.list(1:3), c("x", "y", "z")))
  out <- as.list(x)

  testthat::expect_null(attr(out[[1]], "ft_name", exact = TRUE))
  testthat::expect_null(attr(out[[2]], "ft_name", exact = TRUE))
  testthat::expect_null(attr(out[[3]], "ft_name", exact = TRUE))
})

testthat::test_that("viewL and viewR return boundary element and remainder", {
  m_count <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  t <- add_monoids(as_flexseq(letters[1:5]), list(count = m_count))
  ms <- attr(t, "monoids")

  vl <- Immutables:::viewL(t, ms)
  testthat::expect_identical(vl$value, "a")
  testthat::expect_identical(tree_chars(vl$rest), "bcde")

  vr <- Immutables:::viewR(t, ms)
  testthat::expect_identical(vr$value, "e")
  testthat::expect_identical(tree_chars(vr$rest), "abcd")
})

testthat::test_that("viewL/viewR on Single produce Empty remainder", {
  t <- as_flexseq("x")
  ms <- attr(t, "monoids")

  vl <- Immutables:::viewL(t, ms)
  vr <- Immutables:::viewR(t, ms)

  testthat::expect_identical(vl$value, "x")
  testthat::expect_identical(vr$value, "x")
  testthat::expect_true(vl$rest %isa% Empty)
  testthat::expect_true(vr$rest %isa% Empty)
})

testthat::test_that("deepL/deepR collapse correctly when middle is Empty", {
  ms <- ensure_size_monoids(list(.size = size_measure_monoid()))
  pr <- list()
  sf <- Digit("c", "d")
  t1 <- Immutables:::deepL(pr, measured_empty(ms), sf, ms)
  testthat::expect_identical(tree_chars(t1), "cd")

  pr2 <- Digit("a", "b")
  sf2 <- list()
  t2 <- Immutables:::deepR(pr2, measured_empty(ms), sf2, ms)
  testthat::expect_identical(tree_chars(t2), "ab")
})

testthat::test_that("measured view helpers preserve measures attrs", {
  t <- as_flexseq(letters[1:4])
  ms <- attr(t, "monoids")

  vl <- Immutables:::viewL(t, ms)
  vr <- Immutables:::viewR(t, ms)

  testthat::expect_identical(attr(vl$rest, "measures")$.size, 3)
  testthat::expect_identical(attr(vr$rest, "measures")$.size, 3)
})

testthat::test_that("locate_by_predicate finds same distinguished element as split_around_by_predicate", {
  t <- as_flexseq(letters[1:8])

  l <- locate_by_predicate(t, function(v) v >= 5, ".size")
  s <- split_around_by_predicate(t, function(v) v >= 5, ".size")

  testthat::expect_true(l$found)
  testthat::expect_identical(l$value, s$value)
})

testthat::test_that("locate_by_predicate metadata for .size is symmetric", {
  t <- as_flexseq(letters[1:6])

  l <- locate_by_predicate(t, function(v) v >= 4, ".size", include_metadata = TRUE)

  testthat::expect_true(l$found)
  testthat::expect_identical(l$value, "d")
  testthat::expect_identical(l$metadata$left_measure, 3)
  testthat::expect_identical(l$metadata$hit_measure, 4)
  testthat::expect_identical(l$metadata$right_measure, 2)
  testthat::expect_identical(l$metadata$index, 4L)
})

testthat::test_that("locate_by_predicate metadata works with non-size monoid and .size index", {
  sum_m <- measure_monoid(`+`, 0, as.numeric)
  t <- add_monoids(as_flexseq(1:6), list(sum = sum_m))

  l <- locate_by_predicate(t, function(v) v >= 10, "sum", include_metadata = TRUE)

  testthat::expect_true(l$found)
  testthat::expect_identical(l$value, 4L)
  testthat::expect_identical(l$metadata$left_measure, 6)
  testthat::expect_identical(l$metadata$hit_measure, 10)
  testthat::expect_identical(l$metadata$right_measure, 11)
  testthat::expect_identical(l$metadata$index, 4L)
})

testthat::test_that("locate_by_predicate returns not-found without reconstruction", {
  t <- as_flexseq(letters[1:5])

  l <- locate_by_predicate(t, function(v) v >= 99, ".size")
  testthat::expect_false(l$found)
  testthat::expect_null(l$value)

  lm <- locate_by_predicate(t, function(v) v >= 99, ".size", include_metadata = TRUE)
  testthat::expect_false(lm$found)
  testthat::expect_null(lm$value)
  testthat::expect_identical(lm$metadata$left_measure, 5)
  testthat::expect_null(lm$metadata$hit_measure)
  testthat::expect_null(lm$metadata$right_measure)
  testthat::expect_null(lm$metadata$index)
})

#SO

# Runtime: O(1).
# Determines endpoint scalar domain tag.
# **Inputs:** scalar endpoint `v`.
# **Outputs:** scalar character domain tag.
# **Used by:** endpoint normalization and monoid measure functions.
.ivx_endpoint_type <- function(v) {
  .ft_scalar_domain(v)
}

# Runtime: O(1).
# Safe scalar comparator with defensive error messaging.
# **Inputs:** scalar `a`,`b`; optional `endpoint_type`.
# **Outputs:** integer in {-1,0,1}.
# **Used by:** normalization and fallback relation predicates.
.ivx_compare_scalar <- function(a, b, endpoint_type = NULL) {
  .ft_scalar_compare(
    a,
    b,
    error_message = "Interval endpoints must support scalar ordering with `<` and `>`."
  )
}

# Runtime: O(1).
# Checks whether endpoint domain can use fast comparator branches.
# **Inputs:** scalar character `endpoint_type`.
# **Outputs:** scalar logical.
# **Used by:** query spec leaf matcher selection.
.ivx_is_fast_endpoint_type <- function(endpoint_type) {
  .ft_scalar_is_fast_domain(endpoint_type)
}

# Runtime: O(1).
# Fast scalar comparator for known fast scalar domains.
# **Inputs:** scalar `a`,`b`; optional `endpoint_type`.
# **Outputs:** integer in {-1,0,1}.
# **Used by:** monoid reducers and bound queries.
.ivx_compare_scalar_fast <- function(a, b, endpoint_type = NULL) {
  .ft_scalar_compare_fast(
    a,
    b,
    domain = endpoint_type,
    error_message = "Interval endpoints must support scalar ordering with `<` and `>`."
  )
}

# Runtime: O(1).
# Monoid reducer selecting max `start` summary.
# **Inputs:** monoid state lists `a`,`b`.
# **Outputs:** monoid state list with greatest start.
# **Used by:** .ivx_max_start_monoid().
.ivx_choose_max_start <- function(a, b) {
  if(!isTRUE(a$has)) {
    return(b)
  }
  if(!isTRUE(b$has)) {
    return(a)
  }
  if(!identical(a$endpoint_type, b$endpoint_type)) {
    stop("Incompatible endpoint types encountered in interval_index measures.")
  }
  cmp <- .ivx_compare_scalar_fast(a$start, b$start, endpoint_type = a$endpoint_type)
  if(cmp >= 0L) a else b
}

# Runtime: O(1).
# Measure monoid for subtree max start endpoint.
# **Inputs:** none.
# **Outputs:** MeasureMonoid object.
# **Used by:** .ivx_merge_monoids() required set.
.ivx_max_start_monoid <- function() {
  measure_monoid(
    .ivx_choose_max_start,
    list(has = FALSE, start = NULL, endpoint_type = NULL),
    function(el) {
      list(
        has = TRUE,
        start = el$start,
        endpoint_type = .ivx_endpoint_type(el$start)
      )
    }
  )
}

# Runtime: O(1).
# Monoid reducer selecting max `end` summary.
# **Inputs:** monoid state lists `a`,`b`.
# **Outputs:** monoid state list with greatest end.
# **Used by:** .ivx_max_end_monoid().
.ivx_choose_max_end <- function(a, b) {
  if(!isTRUE(a$has)) {
    return(b)
  }
  if(!isTRUE(b$has)) {
    return(a)
  }
  if(!identical(a$endpoint_type, b$endpoint_type)) {
    stop("Incompatible endpoint types encountered in interval_index measures.")
  }
  cmp <- .ivx_compare_scalar_fast(a$end, b$end, endpoint_type = a$endpoint_type)
  if(cmp >= 0L) a else b
}

# Runtime: O(1).
# Measure monoid for subtree max end endpoint.
# **Inputs:** none.
# **Outputs:** MeasureMonoid object.
# **Used by:** .ivx_merge_monoids() required set.
.ivx_max_end_monoid <- function() {
  measure_monoid(
    .ivx_choose_max_end,
    list(has = FALSE, end = NULL, endpoint_type = NULL),
    function(el) {
      list(
        has = TRUE,
        end = el$end,
        endpoint_type = .ivx_endpoint_type(el$end)
      )
    }
  )
}

# Runtime: O(1).
# Monoid reducer selecting min `end` summary.
# **Inputs:** monoid state lists `a`,`b`.
# **Outputs:** monoid state list with smallest end.
# **Used by:** .ivx_min_end_monoid().
.ivx_choose_min_end <- function(a, b) {
  if(!isTRUE(a$has)) {
    return(b)
  }
  if(!isTRUE(b$has)) {
    return(a)
  }
  if(!identical(a$endpoint_type, b$endpoint_type)) {
    stop("Incompatible endpoint types encountered in interval_index measures.")
  }
  cmp <- .ivx_compare_scalar_fast(a$end, b$end, endpoint_type = a$endpoint_type)
  if(cmp <= 0L) a else b
}

# Runtime: O(1).
# Measure monoid for subtree min end endpoint.
# **Inputs:** none.
# **Outputs:** MeasureMonoid object.
# **Used by:** .ivx_merge_monoids() required set.
.ivx_min_end_monoid <- function() {
  measure_monoid(
    .ivx_choose_min_end,
    list(has = FALSE, end = NULL, endpoint_type = NULL),
    function(el) {
      list(
        has = TRUE,
        end = el$end,
        endpoint_type = .ivx_endpoint_type(el$end)
      )
    }
  )
}

# Runtime: O(1).
# Reports whether interval endpoint type can reuse ordered key monoids.
# **Inputs:** scalar character `endpoint_type`.
# **Outputs:** scalar logical.
# **Used by:** insert/build monoid assembly.
.ivx_supports_oms_key_type <- function(endpoint_type) {
  isTRUE(endpoint_type %in% c("numeric", "character", "logical"))
}

# Runtime: O(m), where m = number of user-supplied monoids.
# Merges user monoids with required interval (and optional ordered-key) monoids.
# **Inputs:** optional user `monoids`; optional scalar `endpoint_type`.
# **Outputs:** normalized monoid list including reserved required monoids.
# **Used by:** constructors, insert on empty, apply when dropping custom monoids.
.ivx_merge_monoids <- function(monoids = NULL, endpoint_type = NULL) {
  user <- if(is.null(monoids)) list() else monoids
  if(length(user) > 0L) {
    bad <- intersect(names(user), c(".size", ".named_count", ".ivx_max_start", ".ivx_max_end", ".ivx_min_end", ".oms_max_key"))
    if(length(bad) > 0L) {
      stop(paste0("Reserved monoid names cannot be supplied for interval_index: ", paste(bad, collapse = ", ")))
    }
  }

  req <- list(
    .ivx_max_start = .ivx_max_start_monoid(),
    .ivx_max_end = .ivx_max_end_monoid(),
    .ivx_min_end = .ivx_min_end_monoid()
  )
  if(.ivx_supports_oms_key_type(endpoint_type)) {
    req <- c(req, .oms_required_monoids())
  }
  ensure_size_monoids(c(user, req))
}

#SO
# note: .size and .count are priveledged and ensured for all trees

# Runtime: O(1) fast-path.
if(FALSE) resolve_tree_monoids <- function(t, required = FALSE) NULL
resolve_tree_monoids(t, required) %::% . : logical : .
resolve_tree_monoids(t, required = FALSE) %as% {
  ms <- attr(t, "monoids", exact = TRUE)
  if(required && is.null(ms)) {
    stop("Tree has no monoids attribute.")
  }
  if(is.null(ms)) {
    return(NULL)
  }
  ms
}

# Runtime: O(1) expected (attribute read + list lookup).
if(FALSE) resolve_named_monoid <- function(t, monoid_name) NULL
resolve_named_monoid(t, monoid_name) %::% . : character : list
resolve_named_monoid(t, monoid_name) %as% {
  if(!is.character(monoid_name) || length(monoid_name) != 1L || is.na(monoid_name) || monoid_name == "") {
    stop("`monoid_name` must be a single non-empty string.")
  }
  ms <- resolve_tree_monoids(t, required = TRUE)
  mr <- ms[[monoid_name]]
  if(is.null(mr)) {
    stop(paste0("Monoid name `", monoid_name, "` not found in tree monoids."))
  }
  list(monoids = ms, monoid = mr)
}

# Runtime: O(1) under fixed monoid set.
if(FALSE) merge_monoid_sets <- function(base, add, overwrite = FALSE) NULL
merge_monoid_sets(base, add, overwrite) %::% list : list : logical : list
merge_monoid_sets(base, add, overwrite = FALSE) %as% {
  b <- base
  a <- add

  overlap <- intersect(names(b), names(a))
  overlap <- setdiff(overlap, c(".size", ".named_count"))
  if(length(overlap) > 0 && !isTRUE(overwrite)) {
    stop("Monoid names already exist: ", paste(overlap, collapse = ", "), ". Use overwrite = TRUE to replace.")
  }

  if(length(overlap) > 0 && isTRUE(overwrite)) {
    for(nm in overlap) {
      b[[nm]] <- a[[nm]]
    }
  }

  add_only <- setdiff(names(a), names(b))
  if(length(add_only) > 0) {
    b <- c(b, a[add_only])
  }
  b
}

#SO

# Internal: push a pre-prepared element (ft_name already set) without re-checking
# name state. Used by [[<- after all name validation has already been done.
# Runtime: O(log n) tree update.
.ft_push_back_raw <- function(x, element, monoids) {
  out <- if(.ft_cpp_can_use(monoids)) {
    .ft_cpp_add_right(x, element, monoids)
  } else {
    add_right(x, element, monoids)
  }
  .ft_restore_subclass(out, x, context = "push")
}

# dispatch add-right through C++/R backend and restore class stack.
# Runtime: O(log n) tree update + subclass restoration.
.ft_push_back_dispatch <- function(x, element, monoids, context = "push_back()") {
  out <- if(.ft_cpp_can_use(monoids)) {
    .ft_cpp_add_right(x, element, monoids)
  } else {
    add_right(x, element, monoids)
  }
  .ft_restore_subclass(out, x, context = context)
}

# dispatch named add-right through C++/R backend and restore class stack.
# Runtime: O(log n) tree update + subclass restoration.
.ft_push_back_named_dispatch <- function(x, element, name, monoids, context = "push_back()") {
  out <- if(.ft_cpp_can_use(monoids)) {
    .ft_cpp_add_right_named(x, element, name, monoids)
  } else {
    add_right(x, .ft_set_name(element, name), monoids)
  }
  .ft_restore_subclass(out, x, context = context)
}

# dispatch add-left through C++/R backend and restore class stack.
# Runtime: O(log n) tree update + subclass restoration.
.ft_push_front_dispatch <- function(x, element, monoids, context = "push_front()") {
  out <- if(.ft_cpp_can_use(monoids)) {
    .ft_cpp_add_left(x, element, monoids)
  } else {
    add_left(x, element, monoids)
  }
  .ft_restore_subclass(out, x, context = context)
}

# dispatch named add-left through C++/R backend and restore class stack.
# Runtime: O(log n) tree update + subclass restoration.
.ft_push_front_named_dispatch <- function(x, element, name, monoids, context = "push_front()") {
  out <- if(.ft_cpp_can_use(monoids)) {
    .ft_cpp_add_left_named(x, element, name, monoids)
  } else {
    add_left(x, .ft_set_name(element, name), monoids)
  }
  .ft_restore_subclass(out, x, context = context)
}

# Runtime: O(log n) tree update, with O(1) local name-state checks.
.ft_push_back_impl <- function(x, value, context = "push_back()") {
  monoids <- attr(x, "monoids", exact = TRUE)
  if(is.null(monoids)) {
    stop("Tree has no monoids attribute.")
  }
  measures <- attr(x, "measures", exact = TRUE)
  if(is.null(measures)) {
    stop("Tree has no measures attribute.")
  }
  size <- as.integer(measures[[".size"]])
  named_count <- as.integer(measures[[".named_count"]])
  if(size > 0L && named_count != 0L && named_count != size) {
    stop("Invalid tree name state: mixed named/unnamed elements.")
  }

  element <- value
  if(named_count == 0L) {
    element_name <- .ft_get_name_fast(element)
    if(is.null(element_name)) {
      element_name <- .ft_name_from_value(element)
    }

    # Unnamed tree mode: unnamed element is always fine.
    if(is.null(element_name)) {
      return(.ft_push_back_dispatch(x, element, monoids, context = context))
    }

    # Named element may only seed an empty tree; otherwise it would mix states.
    if(size > 0L) {
      stop("Cannot mix named and unnamed elements (push_back would create mixed named and unnamed tree).")
    }
    return(.ft_push_back_named_dispatch(x, element, element_name, monoids, context = context))
  }

  # Named tree mode: inserted element must carry a usable name.
  element_name <- .ft_get_name_fast(element)
  if(is.null(element_name)) {
    element_name <- .ft_name_from_value(element)
  }
  if(is.null(element_name)) {
    stop("Cannot mix named and unnamed elements (push_back would create mixed named and unnamed tree).")
  }
  .ft_push_back_named_dispatch(x, element, element_name, monoids, context = context)
}

#' Push an Element to the Back
#'
#' Returns a new sequence with `value` appended at the right end.
#'
#' @param x A `flexseq`.
#' @param value Element to append.
#' @return Updated `flexseq`.
#' @details
#' This operation is persistent: `x` is not modified.
#'
#' Elements can be named, but only if all are uniquely named (no missing
#' names). 
#' @examples
#' s <- as_flexseq(letters[1:3])
#' s2 <- push_back(s, "d")
#' s2
#' s  # unchanged
#'
#' n <- as_flexseq(list(two = 2, three = 3))
#' new_el <- 4
#' names(new_el) <- "four"
#' push_back(n, new_el)
#'
#' # Named/unnamed mixes are rejected
#' try(push_back(n, 5))
#' @export
# Runtime: O(log n) tree update, with O(1) local name-state checks.
push_back <- function(x, value) {
  if(inherits(x, "ordered_sequence")) {
    .ft_stop_ordered_like(x, "push_back", "Use `insert()`.")
  }
  if(inherits(x, "priority_queue")) {
    stop("`push_back()` is not supported for priority_queue. Cast first with `as_flexseq()`.")
  }
  .ft_push_back_impl(x, value, context = "push_back()")
}

#' Push an Element to the Front
#'
#' Returns a new sequence with `value` prepended at the left end.
#'
#' @param x A `flexseq`.
#' @param value Element to prepend.
#' @return Updated `flexseq`.
#' @details
#' This operation is persistent: `x` is not modified.
#'
#' Elements can be named, but only if all are uniquely named (no missing
#' names). 
#' @examples
#' s <- as_flexseq(letters[2:4])
#' s2 <- push_front(s, "a")
#' s2
#' s  # unchanged
#'
#' n <- as_flexseq(list(two = 2, three = 3))
#' new_el <- 1
#' names(new_el) <- "one"
#' push_front(n, new_el)
#'
#' # Named/unnamed mixes are rejected
#' try(push_front(n, 0))
#' @export
# Runtime: O(log n) tree update, with O(1) local name-state checks.
push_front <- function(x, value) {
  if(inherits(x, "ordered_sequence")) {
    .ft_stop_ordered_like(x, "push_front", "Use `insert()`.")
  }
  if(inherits(x, "priority_queue")) {
    stop("`push_front()` is not supported for priority_queue. Cast first with `as_flexseq()`.")
  }

  monoids <- attr(x, "monoids", exact = TRUE)
  if(is.null(monoids)) {
    stop("Tree has no monoids attribute.")
  }
  measures <- attr(x, "measures", exact = TRUE)
  if(is.null(measures)) {
    stop("Tree has no measures attribute.")
  }
  size <- as.integer(measures[[".size"]])
  named_count <- as.integer(measures[[".named_count"]])
  if(size > 0L && named_count != 0L && named_count != size) {
    stop("Invalid tree name state: mixed named/unnamed elements.")
  }

  element <- value
  if(named_count == 0L) {
    element_name <- .ft_get_name_fast(element)
    if(is.null(element_name)) {
      element_name <- .ft_name_from_value(element)
    }

    # Unnamed tree mode: unnamed element is always fine.
    if(is.null(element_name)) {
      return(.ft_push_front_dispatch(x, element, monoids, context = "push_front()"))
    }

    # Named element may only seed an empty tree; otherwise it would mix states.
    if(size > 0L) {
      stop("Cannot mix named and unnamed elements (push_front would create mixed named and unnamed tree).")
    }
    return(.ft_push_front_named_dispatch(x, element, element_name, monoids, context = "push_front()"))
  }

  # Named tree mode: inserted element must carry a usable name.
  element_name <- .ft_get_name_fast(element)
  if(is.null(element_name)) {
    element_name <- .ft_name_from_value(element)
  }
  if(is.null(element_name)) {
    stop("Cannot mix named and unnamed elements (push_front would create mixed named and unnamed tree).")
  }
  .ft_push_front_named_dispatch(x, element, element_name, monoids, context = "push_front()")
}

# Runtime: O(1) for empty rebuild metadata, class restoration depends on type.
# pop_* on size-1 should return an empty value of the same semantic type, not just raw flexseq.
.ft_empty_same_type <- function(x, context = "pop") {
  monoids <- resolve_tree_monoids(x, required = TRUE)
  empty <- empty_tree(monoids = monoids)
  if(inherits(x, "interval_index")) {
    return(.ivx_wrap_like(x, empty))
  }
  if(inherits(x, "ordered_sequence")) {
    return(.ord_wrap_like(x, empty))
  }
  .ft_restore_subclass(empty, x, context = context)
}

# Runtime: O(1).
.ft_unwrap_public_value <- function(x, element) {
  # Ordered/interval subclasses store internal entry records; public end-ops
  # return payload items to match user-facing sequence semantics.
  if(inherits(x, "interval_index")) {
    if(is.list(element) && ("value" %in% names(element))) {
      return(element$value)
    }
    return(element)
  }
  if(inherits(x, "ordered_sequence")) {
    if(is.list(element) && ("value" %in% names(element))) {
      return(element$value)
    }
    return(element)
  }
  element
}

# validate one positional index for public peek/pop-at helpers.
# Runtime: O(1).
.ft_validate_scalar_position <- function(index, n) {
  resolved <- .ft_assert_int_indices(index, n)
  if(length(resolved) != 1L) {
    stop("`index` must be a single positive integer.")
  }
  resolved
}

# validate one positional index for miss-aware public peek/pop-at helpers.
# Returns NULL when index is a valid positive integer but out of bounds.
# Runtime: O(1).
.ft_validate_scalar_position_missable <- function(index, n) {
  if(is.null(index)) {
    stop("Index is required.")
  }
  if(!is.numeric(index) || any(is.na(index)) || any(index != as.integer(index))) {
    stop("Only non-missing integer indices are supported.")
  }
  idx <- as.integer(index)
  if(length(idx) != 1L) {
    stop("`index` must be a single positive integer.")
  }
  if(idx <= 0L) {
    stop("Only positive integer indices are supported.")
  }
  if(idx > n) {
    return(NULL)
  }
  idx
}

# validate one insertion index in [1, n + 1].
# Runtime: O(1).
.ft_validate_scalar_insert_position <- function(index, n) {
  resolved <- .ft_assert_int_indices(index, n + 1L)
  if(length(resolved) != 1L) {
    stop("`index` must be a single positive integer.")
  }
  resolved
}

# normalize insert_at payload into a plain list of elements.
# Runtime: O(k), where k = number of inserted elements.
.ft_coerce_insert_values <- function(values) {
  if(inherits(values, "priority_queue")) {
    stop("`values` cannot be a priority_queue. Cast first with `as_flexseq()`.")
  }
  if(inherits(values, "ordered_sequence")) {
    stop(sprintf(
      "`values` cannot be an ordered sequence subclass (%s). Cast first with `as_flexseq()`.",
      .ft_ordered_owner_class(values)
    ))
  }
  if(inherits(values, "flexseq")) {
    return(as.list(values))
  }
  if(is.list(values)) {
    return(values)
  }
  as.list(values)
}

# enforce global named/unnamed consistency for insert_at.
# Runtime: O(1) from cached size/name counts.
.ft_validate_insert_name_state <- function(x, insert_tree, context = "insert_at()") {
  measures_x <- attr(x, "measures", exact = TRUE)
  measures_insert <- attr(insert_tree, "measures", exact = TRUE)
  if(is.null(measures_x) || is.null(measures_insert)) {
    stop("Tree has no measures attribute.")
  }
  n_x <- as.integer(measures_x[[".size"]])
  nn_x <- as.integer(measures_x[[".named_count"]])
  n_i <- as.integer(measures_insert[[".size"]])
  nn_i <- as.integer(measures_insert[[".named_count"]])

  if((n_x > 0L && nn_x != 0L && nn_x != n_x) || (n_i > 0L && nn_i != 0L && nn_i != n_i)) {
    stop("Invalid tree name state: mixed named/unnamed elements.")
  }
  if(n_x == 0L || n_i == 0L) {
    return(invisible(TRUE))
  }
  x_named <- (nn_x == n_x)
  i_named <- (nn_i == n_i)
  if(x_named != i_named) {
    stop("Cannot mix named and unnamed elements (", context, " would create mixed named and unnamed tree).")
  }
  invisible(TRUE)
}

#' Peek at the Front Element
#'
#' Returns the first element without modifying the sequence.
#'
#' @param x A `flexseq`.
#' @return First element, or `NULL` when `x` is empty.
#' @details
#' Returns the payload element without modifying `x`.
#' @examples
#' x <- flexseq("a", "b", "c")
#' peek_front(x)
#'
#' peek_front(flexseq())
#' @export
# Runtime: O(log n) lookup by scalar index.
peek_front <- function(x) {
  if(inherits(x, "interval_index")) {
    stop("`peek_front()` is not supported for interval_index. Use interval query helpers (`peek_*`, `pop_*`, and `peek_point()` for point lookup).")
  }
  if(inherits(x, "priority_queue")) {
    stop("`peek_front()` is not supported for priority_queue. Use `peek_min()`/`peek_max()`, or cast with `as_flexseq()`.")
  }
  if(!inherits(x, "flexseq")) {
    stop("`x` must be a flexseq.")
  }
  if(length(x) == 0L) {
    return(NULL)
  }
  .ft_unwrap_public_value(x, x[[1L]])
}

#' Peek at the Back Element
#'
#' Returns the last element without modifying the sequence.
#'
#' @param x A `flexseq`.
#' @return Last element, or `NULL` when `x` is empty.
#' @details
#' Returns the payload element without modifying `x`.
#' @examples
#' x <- flexseq("a", "b", "c")
#' peek_back(x)
#'
#' peek_back(flexseq())
#' @export
# Runtime: O(log n) lookup by scalar index.
peek_back <- function(x) {
  if(inherits(x, "interval_index")) {
    stop("`peek_back()` is not supported for interval_index. Use interval query helpers (`peek_*`, `pop_*`, and `peek_point()` for point lookup).")
  }
  if(inherits(x, "priority_queue")) {
    stop("`peek_back()` is not supported for priority_queue. Use `peek_min()`/`peek_max()`, or cast with `as_flexseq()`.")
  }
  if(!inherits(x, "flexseq")) {
    stop("`x` must be a flexseq.")
  }
  n <- length(x)
  if(n == 0L) {
    return(NULL)
  }
  .ft_unwrap_public_value(x, x[[n]])
}

#' Peek at an Element by Position
#'
#' Returns the element at a one-based index without modifying the sequence.
#'
#' @param x A `flexseq`.
#' @param index One-based position to read.
#' @return Element at `index`, or `NULL` when `index` is out of bounds.
#' @details
#' Positive integer indices beyond `length(x)` return `NULL`.
#' Invalid indices (`NA`, non-integer, `<= 0`, or length not equal to 1) error.
#' @examples
#' x <- flexseq("a", "b", "c")
#' peek_at(x, 2)
#' peek_at(x, 10)
#'
#' try(peek_at(x, 0))
#' @export
# Runtime: O(log n) lookup by scalar index.
peek_at <- function(x, index) {
  if(inherits(x, "interval_index")) {
    stop("`peek_at()` is not supported for interval_index. Use interval query helpers (`peek_*`, `pop_*`, and `peek_point()` for point lookup).")
  }
  if(inherits(x, "priority_queue")) {
    stop("`peek_at()` is not supported for priority_queue. Use `peek_min()`/`peek_max()`, or cast with `as_flexseq()`.")
  }
  if(!inherits(x, "flexseq")) {
    stop("`x` must be a flexseq.")
  }
  n <- length(x)
  idx <- .ft_validate_scalar_position_missable(index, n)
  if(is.null(idx)) {
    return(NULL)
  }
  element <- .ft_strip_name(.ft_get_elem_at(x, idx))
  .ft_unwrap_public_value(x, element)
}

#' Pop the Front Element
#'
#' Returns the first element and the remaining sequence.
#'
#' @param x A `flexseq`.
#' @return A list with fields:
#' - `value`: the first element, or `NULL` when `x` is empty.
#' - `remaining`: the sequence after removing the first element.
#' @details
#' This operation is persistent: `x` is not modified.
#'
#' On empty input, returns a non-throwing miss object with
#' `value = NULL` and `remaining = x`.
#' @examples
#' s <- flexseq("a", "b", "c")
#' out <- pop_front(s)
#' out$value
#' out$remaining
#' s  # unchanged
#'
#' pop_front(flexseq())
#' @export
# Runtime: O(log n) for one read plus one subset.
pop_front <- function(x) {
  if(inherits(x, "interval_index")) {
    stop("`pop_front()` is not supported for interval_index. Use interval query helpers (`peek_*`, `pop_*`, and `peek_point()` for point lookup).")
  }
  if(inherits(x, "priority_queue")) {
    stop("`pop_front()` is not supported for priority_queue. Use `pop_min()`/`pop_max()`, or cast with `as_flexseq()`.")
  }
  if(!inherits(x, "flexseq")) {
    stop("`x` must be a flexseq.")
  }
  n <- .ft_size(x)
  if(n == 0L) {
    return(list(value = NULL, remaining = x))
  }
  ms <- attr(x, "monoids", exact = TRUE)
  if(.ft_cpp_can_use(ms)) {
    s <- .ft_cpp_split_at_index(x, 1L, ms)
    element <- .ft_unwrap_public_value(x, .ft_strip_name(s$value))
    remaining <- if(n == 1L) .ft_empty_same_type(x, context = "pop_front()") else .ft_restore_subclass(s$right, x, context = "pop_front()")
    return(list(value = element, remaining = remaining))
  }
  s <- split_around_by_predicate(x, function(v) v >= 1L, ".size")
  element <- .ft_unwrap_public_value(x, .ft_strip_name(s$value))
  remaining <- if(n == 1L) .ft_empty_same_type(x, context = "pop_front()") else s$right
  list(value = element, remaining = remaining)
}

#' Pop the Back Element
#'
#' Returns the last element and the remaining sequence.
#'
#' @param x A `flexseq`.
#' @return A list with fields:
#' - `value`: the last element, or `NULL` when `x` is empty.
#' - `remaining`: the sequence after removing the last element.
#' @details
#' This operation is persistent: `x` is not modified.
#'
#' On empty input, returns a non-throwing miss object with
#' `value = NULL` and `remaining = x`.
#' @examples
#' s <- flexseq("a", "b", "c")
#' out <- pop_back(s)
#' out$value
#' out$remaining
#' s  # unchanged
#'
#' pop_back(flexseq())
#' @export
# Runtime: O(log n) for one read plus one subset.
pop_back <- function(x) {
  if(inherits(x, "interval_index")) {
    stop("`pop_back()` is not supported for interval_index. Use interval query helpers (`peek_*`, `pop_*`, and `peek_point()` for point lookup).")
  }
  if(inherits(x, "priority_queue")) {
    stop("`pop_back()` is not supported for priority_queue. Use `pop_min()`/`pop_max()`, or cast with `as_flexseq()`.")
  }
  if(!inherits(x, "flexseq")) {
    stop("`x` must be a flexseq.")
  }
  n <- .ft_size(x)
  if(n == 0L) {
    return(list(value = NULL, remaining = x))
  }
  ms <- attr(x, "monoids", exact = TRUE)
  if(.ft_cpp_can_use(ms)) {
    s <- .ft_cpp_split_at_index(x, n, ms)
    element <- .ft_unwrap_public_value(x, .ft_strip_name(s$value))
    remaining <- if(n == 1L) .ft_empty_same_type(x, context = "pop_back()") else .ft_restore_subclass(s$left, x, context = "pop_back()")
    return(list(value = element, remaining = remaining))
  }
  s <- split_around_by_predicate(x, function(v) v >= n, ".size")
  element <- .ft_unwrap_public_value(x, .ft_strip_name(s$value))
  remaining <- if(n == 1L) .ft_empty_same_type(x, context = "pop_back()") else s$left
  list(value = element, remaining = remaining)
}

#' Pop an Element by Position
#'
#' Returns the selected element and the remaining sequence.
#'
#' @param x A `flexseq`.
#' @param index One-based position to remove.
#' @return A list with fields:
#' - `value`: the element at `index`, or `NULL` when `index` is out of bounds.
#' - `remaining`: the sequence after removing the selected element.
#' @details
#' This operation is persistent: `x` is not modified.
#'
#' Positive integer indices beyond `length(x)` return a non-throwing miss object
#' with `value = NULL` and `remaining = x`.
#' Invalid indices (`NA`, non-integer, `<= 0`, or length not equal to 1) error.
#' @examples
#' x <- flexseq("a", "b", "c", "d")
#' out <- pop_at(x, 3)
#' out$value
#' out$remaining
#' x  # unchanged
#'
#' pop_at(x, 10)
#' try(pop_at(x, 0))
#' @export
# Runtime: O(log n) for split plus O(log n) concat.
pop_at <- function(x, index) {
  if(inherits(x, "interval_index")) {
    stop("`pop_at()` is not supported for interval_index. Use interval query helpers (`peek_*`, `pop_*`, and `peek_point()` for point lookup).")
  }
  if(inherits(x, "priority_queue")) {
    stop("`pop_at()` is not supported for priority_queue. Use `pop_min()`/`pop_max()`, or cast with `as_flexseq()`.")
  }
  if(!inherits(x, "flexseq")) {
    stop("`x` must be a flexseq.")
  }
  n <- length(x)
  idx <- .ft_validate_scalar_position_missable(index, n)
  if(is.null(idx)) {
    return(list(value = NULL, remaining = x))
  }
  ms <- attr(x, "monoids", exact = TRUE)
  if(.ft_cpp_can_use(ms)) {
    s <- .ft_cpp_split_at_index(x, idx, ms)
    element <- .ft_unwrap_public_value(x, .ft_strip_name(s$value))
    remaining <- if(n == 1L) {
      .ft_empty_same_type(x, context = "pop_at()")
    } else {
      .ft_restore_subclass(.ft_cpp_concat(s$left, s$right, ms), x, context = "pop_at()")
    }
    return(list(value = element, remaining = remaining))
  }
  s <- split_around_by_predicate(x, function(v) v >= idx, ".size")
  element <- .ft_unwrap_public_value(x, .ft_strip_name(s$value))
  remaining <- if(n == 1L) {
    .ft_empty_same_type(x, context = "pop_at()")
  } else {
    .ft_restore_subclass(.ft_concat_same_monoids(s$left, s$right, ms), x, context = "pop_at()")
  }
  list(value = element, remaining = remaining)
}

#' Insert Elements at a Position
#'
#' Inserts `values` before the current element at `index`.
#'
#' @param x A `flexseq`.
#' @param index One-based insertion position in `[1, length(x) + 1]`.
#' @param values Values to insert.
#' @return Updated sequence with inserted values.
#' @details
#' `values` is interpreted as a collection of elements to splice in.
#'
#' Common cases:
#' - Atomic vector (`c("x", "y")`): inserts one element per vector entry.
#' - List (`list("x", "y")`): inserts one element per list entry.
#' - `flexseq`: inserts all of its elements.
#' - Empty input (`list()` or `flexseq()`): no change.
#'
#' To insert one composite object (for example, a vector or a list) as a single
#' element, wrap it in `list(...)`.
#'
#' This operation is persistent: `x` is not modified.
#' @examples
#' x <- flexseq("a", "b", "c", "d")
#' insert_at(x, 3, c("x", "y"))
#' insert_at(x, 1, "start")
#' insert_at(x, length(x) + 1, "end")
#'
#' # Insert one vector as a single element
#' insert_at(x, 3, list(c("u", "v")))
#' @export
# Runtime: O(k log k) to build inserted payload + O(log n) split + concat work.
insert_at <- function(x, index, values) {
  if(inherits(x, "ordered_sequence")) {
    .ft_stop_ordered_like(x, "insert_at", "Use `insert()`.")
  }
  if(inherits(x, "priority_queue")) {
    stop("`insert_at()` is not supported for priority_queue. Cast first with `as_flexseq()`.")
  }
  if(!inherits(x, "flexseq")) {
    stop("`x` must be a flexseq.")
  }

  n <- length(x)
  insert_index <- .ft_validate_scalar_insert_position(index, n)
  insert_values <- .ft_coerce_insert_values(values)
  if(length(insert_values) == 0L) {
    return(x)
  }

  monoids <- resolve_tree_monoids(x, required = TRUE)
  insert_tree <- tree_from(insert_values, monoids = monoids)
  .ft_validate_insert_name_state(x, insert_tree, context = "insert_at()")

  out <- if(n == 0L) {
    insert_tree
  } else if(insert_index == 1L) {
    concat_trees(insert_tree, x)
  } else if(insert_index == (n + 1L)) {
    concat_trees(x, insert_tree)
  } else {
    # Split once at insertion boundary, then stitch left + inserted + right.
    split_ctx <- split_by_predicate(x, function(v) v >= insert_index, ".size")
    concat_trees(concat_trees(split_ctx$left, insert_tree), split_ctx$right)
  }
  .ft_restore_subclass(out, x, context = "insert_at()")
}

testthat::test_that("Elements can be lists and data.frames", {
  df <- data.frame(x = 1:2, y = c("a", "b"), stringsAsFactors = FALSE)
  payload <- list(a = 1, b = 2)
  t <- flexseq()
  t <- push_back(t, payload)
  t <- push_back(t, df)

  reduced <- as.list(t)
  testthat::expect_identical(length(reduced), 2L)
  testthat::expect_true(is.list(reduced[[1]]))
  testthat::expect_true(is.data.frame(reduced[[2]]))
  testthat::expect_identical(reduced[[1]], payload)
  testthat::expect_identical(reduced[[2]], df)
})

testthat::test_that("tree construction defaults to .size monoid", {
  t <- as_flexseq(1:3)
  ms <- attr(t, "measures")
  testthat::expect_identical(ms$.size, 3)
  testthat::expect_true(!is.null(attr(t, "monoids")[[".size"]]))

  e <- flexseq()
  testthat::expect_true(e %isa% Empty)
  testthat::expect_identical(attr(e, "measures")$.size, 0)
})

testthat::test_that("flexseq class does not inherit from list and length is element count", {
  t <- as_flexseq(letters[1:5])
  testthat::expect_true("flexseq" %in% class(t))
  testthat::expect_false("list" %in% class(t))
  testthat::expect_identical(length(t), 5L)
})

testthat::test_that("as_flexseq returns payload-only flexseq and rejects extra args", {
  from_default <- as_flexseq(1:3)
  from_flex <- as_flexseq(as_flexseq(1:3))
  testthat::expect_identical(as.list(from_default), as.list(1:3))
  testthat::expect_identical(as.list(from_flex), as.list(1:3))

  # drop_meta was removed; passing it should error (no `...` fallthrough).
  testthat::expect_error(as_flexseq(1:3, drop_meta = TRUE))
  testthat::expect_error(as_flexseq(as_flexseq(1:3), drop_meta = FALSE))
})

testthat::test_that("add_monoids merges and supports overwrite flag", {
  t <- as_flexseq(1:5)
  sum_m <- measure_monoid(function(a, b) a + b, 0, function(el) el)
  t2 <- add_monoids(t, list(sum = sum_m))

  testthat::expect_identical(attr(t2, "measures")$sum, 15)
  testthat::expect_identical(attr(t2, "measures")$.size, 5)

  sum2 <- measure_monoid(function(a, b) a + b + 1, 0, function(el) el)
  testthat::expect_error(
    add_monoids(t2, list(sum = sum2), overwrite = FALSE),
    "already exist"
  )

  t3 <- add_monoids(t2, list(sum = sum2), overwrite = TRUE)
  testthat::expect_true(!identical(attr(t2, "measures")$sum, attr(t3, "measures")$sum))
})

testthat::test_that("monoid specs are normalized to canonical named layout", {
  raw <- list(
    function(a, b) a + b,
    0L,
    function(el) as.integer(el)
  )
  class(raw) <- c("measure_monoid", "MeasureMonoid", "list")

  x <- add_monoids(as_flexseq(1:5), list(sum = raw))
  ms <- attr(x, "monoids", exact = TRUE)
  sum_m <- ms$sum

  testthat::expect_identical(names(sum_m), c("f", "i", "measure"))
  testthat::expect_true(is.function(sum_m$f))
  testthat::expect_true(is.function(sum_m$measure))
  testthat::expect_identical(sum_m$measure(7L), 7L)
})

testthat::test_that("add_monoids blocks reserved monoid names by type", {
  sum_m <- measure_monoid(function(a, b) a + b, 0, function(el) 1L)

  fx <- as_flexseq(1:3)
  testthat::expect_error(
    add_monoids(fx, list(.size = sum_m), overwrite = TRUE),
    "Reserved monoid names cannot be supplied for flexseq"
  )

  pq <- priority_queue("a", "b", priorities = c(2, 1))
  testthat::expect_error(
    add_monoids(pq, list(.pq_min = sum_m), overwrite = TRUE),
    "Reserved monoid names cannot be supplied for priority_queue"
  )

  os <- as_ordered_sequence(list("a", "b"), keys = c(1, 2))
  testthat::expect_error(
    add_monoids(os, list(.oms_max_key = sum_m), overwrite = TRUE),
    "Reserved monoid names cannot be supplied for ordered_sequence"
  )

  ix <- interval_index("a", start = 1, end = 2)
  testthat::expect_error(
    add_monoids(ix, list(.ivx_max_end = sum_m), overwrite = TRUE),
    "Reserved monoid names cannot be supplied for interval_index"
  )
})

testthat::test_that("advanced structures require entry-style measure functions", {
  pq <- priority_queue("a", priorities = 1)
  pq_tuple <- measure_monoid(`+`, 0, function(value, priority) as.numeric(priority))
  testthat::expect_error(add_monoids(pq, list(m = pq_tuple)))

  os <- as_ordered_sequence(list("a"), keys = 1)
  os_tuple <- measure_monoid(`+`, 0, function(value, key) as.numeric(key))
  testthat::expect_error(add_monoids(os, list(m = os_tuple)))

  ix <- as_interval_index(list("a"), start = 1, end = 2)
  ix_tuple <- measure_monoid(`+`, 0, function(value, start, end) as.numeric(end - start))
  testthat::expect_error(add_monoids(ix, list(m = ix_tuple)))
})

testthat::test_that("concat_trees unions monoids on shared names", {
  a <- measure_monoid(function(x, y) x + y, 0, function(el) el)
  b <- measure_monoid(function(x, y) x + y, 0, function(el) 1)

  t1 <- add_monoids(as_flexseq(1:2), list(sum = a))
  t2 <- add_monoids(as_flexseq(3:4), list(cnt = b))
  t <- c(t1, t2)
  testthat::expect_true(all(c(".size", "sum", "cnt") %in% names(attr(t, "monoids"))))

  # shared name path: left definition is assumed authoritative
  t3 <- add_monoids(as_flexseq(1:2), list(sum = a))
  t4 <- add_monoids(as_flexseq(3:4), list(sum = measure_monoid(function(x, y) x + y, 0, function(el) 1)))
  t_merged <- c(t3, t4)
  testthat::expect_true("sum" %in% names(attr(t_merged, "monoids")))
})

testthat::test_that("peek/pop helpers work and are persistent", {
  x <- as_flexseq(letters[1:4])

  testthat::expect_identical(peek_front(x), "a")
  testthat::expect_identical(peek_back(x), "d")
  testthat::expect_identical(peek_at(x, 2), "b")

  pf <- pop_front(x)
  testthat::expect_identical(pf$value, "a")
  testthat::expect_identical(as.list(pf$remaining), as.list(letters[2:4]))

  pb <- pop_back(x)
  testthat::expect_identical(pb$value, "d")
  testthat::expect_identical(as.list(pb$remaining), as.list(letters[1:3]))

  pm <- pop_at(x, 3)
  testthat::expect_identical(pm$value, "c")
  testthat::expect_identical(as.list(pm$remaining), as.list(c("a", "b", "d")))

  testthat::expect_identical(as.list(x), as.list(letters[1:4]))
})

testthat::test_that("legacy payload field names are absent after value rename", {
  x <- as_flexseq(letters[1:2])
  out <- pop_front(x)
  testthat::expect_identical(out$value, "a")
  testthat::expect_null(out$element)
})

testthat::test_that("peek/pop helpers return non-throwing miss results on empty input", {
  x <- flexseq()
  testthat::expect_null(peek_front(x))
  testthat::expect_null(peek_back(x))
  testthat::expect_null(peek_at(x, 1))

  pf <- pop_front(x)
  pb <- pop_back(x)
  pa <- pop_at(x, 1)
  testthat::expect_null(pf$value)
  testthat::expect_identical(length(pf$remaining), 0L)
  testthat::expect_null(pb$value)
  testthat::expect_identical(length(pb$remaining), 0L)
  testthat::expect_null(pa$value)
  testthat::expect_identical(length(pa$remaining), 0L)
})

testthat::test_that("peek_at/pop_at validate positional index shape and treat out-of-range as miss", {
  x <- as_flexseq(letters[1:4])
  testthat::expect_error(peek_at(x, 0), "positive integer")
  testthat::expect_null(peek_at(x, 5))
  testthat::expect_error(peek_at(x, c(1, 2)), "single positive integer")

  testthat::expect_error(pop_at(x, 0), "positive integer")
  miss <- pop_at(x, 5)
  testthat::expect_null(miss$value)
  testthat::expect_identical(as.list(miss$remaining), as.list(x))
  testthat::expect_error(pop_at(x, c(1, 2)), "single positive integer")
})

testthat::test_that("insert_at inserts before index and preserves source persistence", {
  x <- as_flexseq(letters[1:4])

  a <- insert_at(x, 1, "z")
  testthat::expect_identical(as.list(a), as.list(c("z", "a", "b", "c", "d")))

  b <- insert_at(x, 3, c("x", "y"))
  testthat::expect_identical(as.list(b), as.list(c("a", "b", "x", "y", "c", "d")))

  c <- insert_at(x, 5, list("q"))
  testthat::expect_identical(as.list(c), as.list(c("a", "b", "c", "d", "q")))

  ins <- as_flexseq(c("m", "n"))
  d <- insert_at(x, 2, ins)
  testthat::expect_identical(as.list(d), as.list(c("a", "m", "n", "b", "c", "d")))

  testthat::expect_identical(as.list(x), as.list(letters[1:4]))
})

testthat::test_that("insert_at validates bounds and supports empty insert payload", {
  x <- as_flexseq(letters[1:4])
  testthat::expect_error(insert_at(x, 0, "z"), "positive integer")
  testthat::expect_error(insert_at(x, 6, "z"), "out of bounds")
  testthat::expect_error(insert_at(x, c(1, 2), "z"), "single positive integer")

  x2 <- insert_at(x, 2, list())
  testthat::expect_identical(as.list(x2), as.list(x))
})

testthat::test_that("insert_at enforces global named/unnamed consistency", {
  unnamed <- as_flexseq(1:3)
  named <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))

  testthat::expect_error(
    insert_at(unnamed, 2, setNames(as.list(9), "k")),
    "mixed named and unnamed"
  )
  testthat::expect_error(
    insert_at(named, 2, list(9)),
    "mixed named and unnamed"
  )

  out <- insert_at(named, 2, setNames(as.list(c(9, 8)), c("k1", "k2")))
  testthat::expect_identical(out[["k1"]], 9)
  testthat::expect_identical(out[["k2"]], 8)
  testthat::expect_identical(length(out), 5L)
})

testthat::test_that("add_monoids() errors on an unsupported class with an informative message", {
  testthat::expect_error(
    add_monoids(list(), list()),
    "No `add_monoids\\(\\)` method for class"
  )
})

testthat::test_that("insert() errors on an unsupported class with an informative message", {
  testthat::expect_error(
    insert(list()),
    "No `insert\\(\\)` method for class"
  )
})

#SO

# Runtime: O(1).
.oms_assert_set <- function(x) {
  if(!inherits(x, "ordered_sequence") || !is_structural_node(x)) {
    stop("`x` must be an ordered_sequence.")
  }
  invisible(TRUE)
}

# Runtime: O(1).
.oms_key_type_state <- function(x) {
  attr(x, "oms_key_type", exact = TRUE)
}

# Runtime: O(1).
.as_ordered_sequence <- function(x, key_type) {
  if(!is_structural_node(x)) {
    stop("Expected a structural tree node.")
  }
  class(x) <- unique(c("ordered_sequence", "flexseq", setdiff(class(x), "list")))
  attr(x, "oms_key_type") <- key_type
  x
}

# Runtime: O(1).
.ord_wrap_like <- function(template, tree, key_type = NULL) {
  resolved_key_type <- if(is.null(key_type)) .oms_key_type_state(template) else key_type
  .as_ordered_sequence(tree, key_type = resolved_key_type)
}

# Runtime: O(1).
.oms_validate_key_type <- function(oms_key_type, new_key_type) {
  if(is.null(oms_key_type)) {
    return(new_key_type)
  }
  if(!identical(oms_key_type, new_key_type)) {
    stop("Incompatible key type for this ordered_sequence.")
  }
  oms_key_type
}

# Runtime: O(1).
.oms_empty_tree_like <- function(template) {
  ms <- attr(template, "monoids", exact = TRUE)
  .as_flexseq(measured_empty(ms))
}

# Runtime: O(1).
.oms_coerce_lgl_scalar <- function(v, arg_name) {
  if(!is.logical(v) || length(v) != 1L || is.na(v)) {
    stop(sprintf("`%s` must be TRUE or FALSE.", arg_name))
  }
  isTRUE(v)
}

# Runtime: O(n) in number of selected entries.
.oms_extract_items <- function(entries) {
  lapply(entries, function(e) e$value)
}

#SO

# choose the left-most (a) entry on ties
# Runtime: O(1).
.pq_choose_min <- function(a, b) {
  if(!isTRUE(a$has)) {
    return(b)
  }
  if(!isTRUE(b$has)) {
    return(a)
  }
  domain <- .ft_scalar_domain(a$priority)
  cmp <- .ft_scalar_compare_fast(
    a$priority,
    b$priority,
    domain = domain,
    error_message = "Priority values must support scalar ordering with `<` and `>`."
  )
  if(cmp < 0L) {
    return(a)
  }
  if(cmp > 0L) {
    return(b)
  }
  a
}

# choose the left-most (a) entry on ties
# Runtime: O(1).
.pq_choose_max <- function(a, b) {
  if(!isTRUE(a$has)) {
    return(b)
  }
  if(!isTRUE(b$has)) {
    return(a)
  }
  domain <- .ft_scalar_domain(a$priority)
  cmp <- .ft_scalar_compare_fast(
    a$priority,
    b$priority,
    domain = domain,
    error_message = "Priority values must support scalar ordering with `<` and `>`."
  )
  if(cmp > 0L) {
    return(a)
  }
  if(cmp < 0L) {
    return(b)
  }
  a
}

# Runtime: O(1).
pq_min_measure_monoid <- function() {
  measure_monoid(
    .pq_choose_min,
    # `has` distinguishes true identity (empty aggregate) from measured values.
    # We avoid overloading `priority = NULL` as state.
    list(has = FALSE, priority = NULL),
    function(el) {
      list(has = TRUE, priority = el$priority)
    }
  )
}

# Runtime: O(1).
pq_max_measure_monoid <- function() {
  measure_monoid(
    .pq_choose_max,
    # Same sentinel contract as min monoid identity.
    list(has = FALSE, priority = NULL),
    function(el) {
      list(has = TRUE, priority = el$priority)
    }
  )
}

# Runtime: O(1).
.pq_measure_equal <- function(a, b) {
  if(!isTRUE(a$has) && !isTRUE(b$has)) {
    return(TRUE)
  }
  if(!isTRUE(a$has) || !isTRUE(b$has)) {
    return(FALSE)
  }
  domain <- .ft_scalar_domain(a$priority)
  .ft_scalar_equal_fast(
    a$priority,
    b$priority,
    domain = domain,
    error_message = "Priority values must support scalar ordering with `<` and `>`."
  )
}

# Runtime: O(1).
.pq_required_monoids <- function() {
  list(
    .pq_min = pq_min_measure_monoid(),
    .pq_max = pq_max_measure_monoid()
  )
}

#SO

# Runtime: O(n) total (O(n) traversal + O(n) linear rebuild).
.seq_apply_impl <- function(x, f, ..., preserve_custom_monoids = TRUE) {
  if(!inherits(x, "flexseq")) {
    stop("`x` must be a flexseq.")
  }
  if(!is.function(f)) {
    stop("`f` must be a function.")
  }
  if(!is.logical(preserve_custom_monoids) || length(preserve_custom_monoids) != 1L || is.na(preserve_custom_monoids)) {
    stop("`preserve_custom_monoids` must be TRUE or FALSE.")
  }

  els <- .ft_to_list(x)
  n <- length(els)
  out <- vector("list", n)

  for(i in seq_len(n)) {
    out[[i]] <- f(.ft_strip_name(els[[i]]), ...)
  }

  if(n > 0L) {
    el_names <- vapply(els, function(el) {
      nm <- .ft_get_name(el)
      if(is.null(nm)) NA_character_ else nm
    }, character(1))
    has_named <- any(!is.na(el_names))
    if(has_named) {
      if(any(is.na(el_names))) {
        stop("Invalid name state: mixed named and unnamed elements are not allowed.")
      }
      names(out) <- el_names
    }
  }

  out_monoids <- if(isTRUE(preserve_custom_monoids)) {
    resolve_tree_monoids(x, required = TRUE)
  } else {
    ensure_size_monoids(list(.size = size_measure_monoid()))
  }

  .as_flexseq_build(out, monoids = out_monoids)
}

# Runtime: O(1) per call-site introspection.
# Returns TRUE when `f` can accept at least `n` positional args.
.ft_fun_accepts_n_positional <- function(f, n) {
  if(is.primitive(f)) {
    return(TRUE)
  }
  fm <- formals(f)
  if(is.null(fm)) {
    return(TRUE)
  }
  fml_names <- names(fm)
  has_dots <- any(fml_names == "...")
  n_fixed <- sum(fml_names != "...")
  has_dots || (n_fixed >= n)
}

#' @method fapply flexseq
#' @export
#' @noRd
fapply.flexseq <- function(X, FUN, ..., preserve_custom_monoids = TRUE) {
  if(!is.function(FUN)) {
    stop("`FUN` must be a function.")
  }
  .seq_apply_impl(X, FUN, ..., preserve_custom_monoids = preserve_custom_monoids)
}

#SO

# Runtime: O(1).
.ft_scalar_domain <- function(v) {
  if(inherits(v, "Date")) {
    return("Date")
  }
  if(inherits(v, "POSIXct")) {
    return("POSIXct")
  }
  if(is.numeric(v)) {
    return("numeric")
  }
  if(is.character(v)) {
    return("character")
  }
  if(is.logical(v)) {
    return("logical")
  }
  cls <- class(v)
  if(length(cls) > 0L) {
    return(paste0("class:", paste(cls, collapse = "/")))
  }
  paste0("typeof:", typeof(v))
}

# Runtime: O(1).
.ft_scalar_is_fast_domain <- function(domain) {
  isTRUE(domain %in% c("numeric", "character", "logical", "Date", "POSIXct"))
}

# Runtime: O(1).
.ft_scalar_compare <- function(a, b, error_message = "Values must support scalar ordering with `<` and `>`.") {
  # Strict comparator for validation-facing paths.
  # Always guards base Ops with tryCatch and shape checks.
  lt <- suppressWarnings(tryCatch(a < b, error = function(e) e))
  gt <- suppressWarnings(tryCatch(a > b, error = function(e) e))

  if(
    inherits(lt, "error") ||
    inherits(gt, "error") ||
    !is.logical(lt) ||
    !is.logical(gt) ||
    length(lt) != 1L ||
    length(gt) != 1L ||
    is.na(lt) ||
    is.na(gt)
  ) {
    stop(error_message)
  }

  if(isTRUE(lt)) {
    return(-1L)
  }
  if(isTRUE(gt)) {
    return(1L)
  }
  0L
}

# Runtime: O(1).
.ft_scalar_compare_fast <- function(a, b, domain = NULL, error_message = "Values must support scalar ordering with `<` and `>`.") {
  # Hot-path comparator used by monoid aggregation and tight scans.
  # For known scalar domains, use direct Ops first; otherwise fallback to
  # the strict comparator above to keep behavior safe for custom classes.
  if(.ft_scalar_is_fast_domain(domain)) {
    lt <- a < b
    gt <- a > b
    if(
      is.logical(lt) &&
      is.logical(gt) &&
      length(lt) == 1L &&
      length(gt) == 1L &&
      !is.na(lt) &&
      !is.na(gt)
    ) {
      if(isTRUE(lt)) {
        return(-1L)
      }
      if(isTRUE(gt)) {
        return(1L)
      }
      return(0L)
    }
  }
  .ft_scalar_compare(a, b, error_message = error_message)
}

# Runtime: O(1).
.ft_scalar_equal_fast <- function(a, b, domain = NULL, error_message = "Values must support scalar ordering with `<` and `>`.") {
  # Equality helper mirrors compare strategy: fast primitive check first, then
  # fallback through comparison semantics.
  if(.ft_scalar_is_fast_domain(domain)) {
    eq <- a == b
    if(is.logical(eq) && length(eq) == 1L && !is.na(eq)) {
      return(isTRUE(eq))
    }
  }
  .ft_scalar_compare_fast(a, b, domain = domain, error_message = error_message) == 0L
}

# Runtime: O(1).
.ft_normalize_scalar_orderable <- function(value, arg_name = "value") {
  # 1) Scalar shape + non-missing checks.
  if(length(value) != 1L) {
    stop(sprintf("`%s` must be a scalar value.", arg_name))
  }
  domain <- .ft_scalar_domain(value)
  if(.ft_scalar_is_fast_domain(domain)) {
    na_flag <- is.na(value)
  } else {
    na_flag <- suppressWarnings(tryCatch(is.na(value), error = function(e) FALSE))
  }
  if(is.logical(na_flag) && length(na_flag) == 1L && isTRUE(na_flag)) {
    stop(sprintf("`%s` must be non-missing.", arg_name))
  }

  # 2) Domain-specific canonical coercion.
  norm_value <- if(identical(domain, "numeric")) {
    as.numeric(value)
  } else if(identical(domain, "character")) {
    as.character(value)
  } else if(identical(domain, "logical")) {
    as.logical(value)
  } else if(identical(domain, "Date")) {
    as.Date(value)
  } else if(identical(domain, "POSIXct")) {
    as.POSIXct(value)
  } else {
    value
  }

  # 3) Confirm the normalized value supports scalar ordering.
  .ft_scalar_compare_fast(
    norm_value,
    norm_value,
    domain = domain,
    error_message = sprintf("`%s` must support scalar ordering with `<` and `>`.", arg_name)
  )
  list(value = norm_value, value_type = domain)
}

# Runtime: O(1).
.oms_key_type <- function(key) {
  .ft_scalar_domain(key)
}

# Runtime: O(1).
.oms_is_cpp_key_type <- function(key_type) {
  isTRUE(key_type %in% c("numeric", "character", "logical"))
}

# Runtime: O(1).
.oms_compare_key <- function(a, b, key_type) {
  .ft_scalar_compare_fast(
    a,
    b,
    domain = key_type,
    error_message = "Ordered keys must support scalar ordering with `<` and `>`."
  )
}

# Runtime: O(1).
.oms_normalize_key <- function(key_value) {
  norm <- .ft_normalize_scalar_orderable(key_value, arg_name = "key")
  list(key = norm$value, key_type = norm$value_type)
}

# Runtime: O(1).
.oms_choose_max_key <- function(a, b) {
  if(!isTRUE(a$has)) {
    return(b)
  }
  if(!isTRUE(b$has)) {
    return(a)
  }
  if(!identical(a$key_type, b$key_type)) {
    stop("Incompatible key types encountered in ordered_sequence measures.")
  }
  cmp <- .oms_compare_key(a$key, b$key, a$key_type)
  if(cmp >= 0L) a else b
}

# Runtime: O(1).
oms_max_key_monoid <- function() {
  measure_monoid(
    .oms_choose_max_key,
    list(has = FALSE, key = NULL, key_type = NULL),
    function(el) {
      list(has = TRUE, key = el$key, key_type = .oms_key_type(el$key))
    }
  )
}

# Runtime: O(1).
.oms_required_monoids <- function() {
  list(
    .oms_max_key = oms_max_key_monoid()
  )
}

# Runtime: O(1).
.oms_merge_monoids <- function(monoids = NULL) {
  user <- if(is.null(monoids)) list() else monoids
  if(length(user) > 0L) {
    bad <- intersect(names(user), c(".size", ".named_count", ".oms_max_key"))
    if(length(bad) > 0L) {
      stop(paste0("Reserved monoid names cannot be supplied for ordered_sequence: ", paste(bad, collapse = ", ")))
    }
  }
  ensure_size_monoids(c(user, .oms_required_monoids()))
}

#SO

#' Construct a Measure Monoid Specification
#'
#' Defines how element-level values are measured and combined as accumulated
#' tree metadata.
#'
#' @param f Associative binary function over measure values.
#' @param i Identity value for `f`.
#' @param measure Function mapping one element to its measure value.
#' @return An object of class `measure_monoid`.
#' @details
#' A measure monoid has three parts:
#' - `measure(entry)`: maps each stored leaf entry to a measure value.
#' - `f(left, right)`: combines two measure values.
#' - `i`: identity value for `f`.
#'
#' Requirements:
#' - `f` should be associative.
#' - `i` should satisfy `f(i, x) == x` and `f(x, i) == x`.
#' - `measure()` outputs must be compatible with `f` and `i`.
#'
#' Developer APIs are leaf-entry oriented:
#' - `flexseq`: entry is the stored user element.
#' - `ordered_sequence`: entry is `list(value, key)`.
#' - `priority_queue`: entry is `list(value, priority)`.
#' - `interval_index`: entry is `list(value, start, end)`.
#'
#' `measure_monoid()` only constructs the specification; it becomes active after
#' being attached to a structure via [add_monoids()].
#' @examples
#' sum_m <- measure_monoid(`+`, 0, as.numeric)
#' x <- as_flexseq(1:5)
#' x2 <- add_monoids(x, list(sum = sum_m))
#' attr(x2, "measures")$sum
#' split_around_by_predicate(x2, function(v) v >= 6, "sum")
#'
#' # Count elements
#' count_m <- measure_monoid(`+`, 0L, function(el) 1L)
#' x3 <- add_monoids(x, list(count = count_m))
#' attr(x3, "measures")$count
#'
#' # Character-width accumulation
#' width_m <- measure_monoid(`+`, 0L, function(el) nchar(as.character(el)))
#' s <- as_flexseq(c("aa", "b", "cccc"))
#' s2 <- add_monoids(s, list(width = width_m))
#' attr(s2, "measures")$width
#' @export
# Runtime: O(1).
measure_monoid <- function(f, i, measure) {
  res <- list(f = f, i = i, measure = measure)
  class(res) <- c("measure_monoid", "MeasureMonoid", class(res))
  res
}

testthat::test_that("split_digit handles first/middle/last boundaries", {
  mr <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  ms <- ensure_size_monoids(list(count = mr))
  d <- Digit("a", "b", "c")

  s1 <- Immutables:::split_digit(function(v) v >= 1, 0, d, ms, "count")
  testthat::expect_identical(s1$value, "a")
  testthat::expect_identical(unclass(s1$left), list())
  testthat::expect_identical(unclass(s1$right), list("b", "c"))

  s2 <- Immutables:::split_digit(function(v) v >= 2, 0, d, ms, "count")
  testthat::expect_identical(s2$value, "b")
  testthat::expect_identical(unclass(s2$left), list("a"))
  testthat::expect_identical(unclass(s2$right), list("c"))

  s3 <- Immutables:::split_digit(function(v) v >= 3, 0, d, ms, "count")
  testthat::expect_identical(s3$value, "c")
  testthat::expect_identical(unclass(s3$left), list("a", "b"))
  testthat::expect_identical(unclass(s3$right), list())
})

testthat::test_that("split_digit works when digit holds nodes", {
  mr <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  ms <- ensure_size_monoids(list(count = mr))
  n2 <- Node2("a", "b")
  n3 <- Node3("c", "d", "e")
  d <- Digit(n2, n3)

  s <- Immutables:::split_digit(function(v) v >= 3, 0, d, ms, "count")
  testthat::expect_true(s$value %isa% Node)
  testthat::expect_identical(length(s$left), 1L)
  testthat::expect_identical(length(s$right), 0L)
})

testthat::test_that("split_digit errors for invalid preconditions", {
  mr <- measure_monoid(function(a, b) a + b, 0, function(el) 1)
  ms <- ensure_size_monoids(list(count = mr))
  testthat::expect_error(
    Immutables:::split_digit(function(v) v >= 1, 0, list(), ms, "count"),
    "empty digit"
  )

  d <- Digit("a", "b")
  testthat::expect_error(
    Immutables:::split_digit(function(v) v >= 10, 0, d, ms, "count"),
    "predicate never became true"
  )
})

#SO

#' Validate full tree invariants (debug/test utility)
#'
#' Performs expensive full-tree auditing of:
#' - structural attributes (`monoids`/`measures`) consistency
#' - global name-state invariants
#'
#' Intended for debugging and tests, not hot runtime paths.
#'
#' @param t FingerTree.
#' @return `TRUE` invisibly; errors if invariant violations are found.
#' @examples
#' x <- as_flexseq(letters[1:10])
#' validate_tree(x)
#' @export
# Runtime: O(n) full traversal. Intended for debugging/tests.
validate_tree <- function(t) {
  assert_structural_attrs(t)
  .ft_assert_name_state(t)
  invisible(TRUE)
}

#' Validate name-state invariants only (debug/test utility)
#'
#' Checks that trees are either fully unnamed or fully named with unique,
#' non-empty names.
#'
#' Intended for debugging and tests, not hot runtime paths.
#'
#' @param t FingerTree.
#' @return `TRUE` invisibly; errors if name invariants are violated.
#' @examples
#' x <- as_flexseq(setNames(as.list(letters[1:4]), letters[1:4]))
#' validate_name_state(x)
#' @export
# Runtime: O(n) when tree is named (name collection and uniqueness checks).
validate_name_state <- function(t) {
  .ft_assert_name_state(t)
  invisible(TRUE)
}

testthat::test_that("validate utilities pass on valid trees", {
  t <- as_flexseq(letters[1:12])
  tree_check <- withVisible(validate_tree(t))
  name_check <- withVisible(validate_name_state(t))

  testthat::expect_false(tree_check$visible)
  testthat::expect_false(name_check$visible)
  testthat::expect_true(isTRUE(tree_check$value))
  testthat::expect_true(isTRUE(name_check$value))
})

testthat::test_that("validate_name_state catches mixed named/unnamed trees", {
  t <- as_flexseq(letters[1:3])
  # Force invalid named-count metadata for validation-path coverage only.
  bad <- t
  ms <- attr(bad, "measures", exact = TRUE)
  ms[[".named_count"]] <- as.integer(ms[[".size"]])
  attr(bad, "measures") <- ms
  testthat::expect_error(validate_name_state(bad), "non-empty")
})

#SO

#' Indexing for Interval Indexes
#'
#' Read indexing returns `interval_index` subsets while preserving interval/key
#' order; out-of-order selectors are canonicalized with a warning. Replacement
#' indexing is blocked.
#'
#' @name sub-.interval_index
#' @param x An `interval_index`.
#' @param i Index input.
#' @param value Replacement value (unsupported).
#' @param ... Unused.
#' @return Read methods return interval payload values/subsets; replacement forms
#'   always error.
#' @details
#' Read indexing preserves canonical interval order in returned subsets.
#'
#' - Integer/character vectors are treated as selectors and canonicalized to
#'   interval-order output.
#' - Out-of-order selector vectors trigger a warning and are canonicalized.
#' - Duplicate selectors are rejected.
#' - `[[` and `$` return payload values.
#' - Replacement indexing (`[<-`, `[[<-`, `$<-`) is unsupported.
#' @examples
#' ix <- interval_index(a = "A", b = "B", c = "C", start = c(1, 3, 5), end = c(2, 4, 6))
#'
#' ix[c(3, 1)]         # warning; result returned in interval order
#' ix[c("c", "a")]     # warning; result returned in interval order
#' ix[c(TRUE, FALSE, TRUE)]
#' ix[["b"]]
#' ix$b
#'
#' try(ix[c(2, 2)])
#' try(ix$b <- "updated")
NULL

# Runtime: O(k log n) for reads + O(k log k) selector normalization.
# Subset method preserving interval-index wrapper semantics.
# **Inputs:** `x` interval_index; `i` integer/logical/character index.
# **Outputs:** interval_index subset (empty-like when no positions selected).
# **Used by:** user indexing and internal position-based slicing.
#' @rdname sub-.interval_index
#' @method [ interval_index
#' @export
`[.interval_index` <- function(x, i, ...) {
  if(missing(i)) {
    return(x)
  }
  .ivx_assert_index(x)

  ms <- resolve_tree_monoids(x, required = TRUE)
  n <- as.integer(node_measure(x, ".size"))

  if(is.logical(i)) {
    mask <- .ft_assert_lgl_indices(i, n)
    idx <- .ft_true_positions(mask)
    if(length(idx) == 0L) {
      return(.ivx_wrap_like(x, empty_tree(monoids = ms)))
    }
    out <- .ft_prepare_subset_entries(.ft_get_elems_at(x, idx))
    return(.ivx_wrap_like(x, tree_from(out, monoids = ms)))
  }

  if(is.character(i)) {
    idx <- .ft_assert_chr_indices(i)
    if(length(idx) == 0L) {
      return(.ivx_wrap_like(x, empty_tree(monoids = ms)))
    }
    pos <- .ft_match_name_indices(x, idx, strict_missing = TRUE)
    pos <- .ord_normalize_selector_positions(pos)
    out <- .ft_prepare_subset_entries(.ft_get_elems_at(x, pos))
    return(.ivx_wrap_like(x, tree_from(out, monoids = ms)))
  }

  idx <- .ft_assert_int_indices(i, n)
  if(length(idx) == 0L) {
    return(.ivx_wrap_like(x, empty_tree(monoids = ms)))
  }
  idx <- .ord_normalize_selector_positions(idx)
  out <- .ft_prepare_subset_entries(.ft_get_elems_at(x, idx))
  .ivx_wrap_like(x, tree_from(out, monoids = ms))
}

# Runtime: O(log n) by index, O(n_lookup) by name.
# Scalar extract method returning payload value (not full entry record).
# **Inputs:** `x` interval_index; scalar index/name `i`.
# **Outputs:** payload value.
# **Used by:** `$.interval_index()`, users.
#' @rdname sub-.interval_index
#' @method [[ interval_index
#' @export
`[[.interval_index` <- function(x, i, ...) {
  .ivx_assert_index(x)
  entry <- `[[.flexseq`(x, i, ...)
  if(!is.list(entry) || !("value" %in% names(entry))) {
    stop("Malformed interval_index entry.")
  }
  entry$value
}

# Runtime: O(1).
# Replacement subset is intentionally blocked for ordered interval semantics.
# **Inputs:** `x` interval_index; `i`; `value`.
# **Outputs:** none (always errors).
# **Used by:** users.
#' @rdname sub-.interval_index
#' @method [<- interval_index
#' @export
`[<-.interval_index` <- function(x, i, value) {
  stop("`[<-` is not supported for interval_index.")
}

# Runtime: O(1).
# Replacement scalar extract is intentionally blocked.
# **Inputs:** `x` interval_index; `i`; `value`.
# **Outputs:** none (always errors).
# **Used by:** users.
#' @rdname sub-.interval_index
#' @method [[<- interval_index
#' @export
`[[<-.interval_index` <- function(x, i, value) {
  stop("`[[<-` is not supported for interval_index.")
}

# Name-based scalar read forwarding to `[[.interval_index`.
# **Inputs:** `x` interval_index; symbol/character `name`.
# **Outputs:** payload value.
# **Used by:** users.
#' @rdname sub-.interval_index
#' @method $ interval_index
#' @param name Element name (for `$` and `$<-`).
#' @return For `$`: the matched payload element.
#' @export
`$.interval_index` <- function(x, name) {
  nm <- .ft_dollar_name(substitute(name))
  `[[.interval_index`(x, nm)
}

# Name-based replacement is intentionally blocked.
# **Inputs:** `x` interval_index; `name`; `value`.
# **Outputs:** none (always errors).
# **Used by:** users.
#' @rdname sub-.interval_index
#' @method $<- interval_index
#' @param value Replacement value (unsupported).
#' @return No return value; always errors because replacement indexing is unsupported.
#' @export
`$<-.interval_index` <- function(x, name, value) {
  stop("`$<-` is not supported for interval_index.")
}

#SO

# Runtime: O(n) total from entry traversal + linear queue rebuild.
.pq_apply_impl <- function(q, f, ..., preserve_custom_monoids = TRUE) {
  .pq_assert_queue(q)
  if(!is.function(f)) {
    stop("`f` must be a function.")
  }
  if(!is.logical(preserve_custom_monoids) || length(preserve_custom_monoids) != 1L || is.na(preserve_custom_monoids)) {
    stop("`preserve_custom_monoids` must be TRUE or FALSE.")
  }

  entries <- as.list(q)
  n <- length(entries)
  out <- vector("list", n)
  out_names <- if(is.null(names(entries))) rep("", n) else names(entries)

  dot_args <- list(...)
  accepts_name <- .ft_fun_accepts_n_positional(f, 3L)
  has_dots <- length(dot_args) > 0L

  for(i in seq_len(n)) {
    e <- entries[[i]]
    cur_name <- out_names[[i]]
    item2 <- if(accepts_name && !has_dots) {
      f(e$value, e$priority, cur_name)
    } else if(!accepts_name && !has_dots) {
      f(e$value, e$priority)
    } else if(accepts_name) {
      do.call(f, c(list(e$value, e$priority, cur_name), dot_args))
    } else {
      do.call(f, c(list(e$value, e$priority), dot_args))
    }

    out[[i]] <- list(
      value = item2,
      priority = e$priority
    )
  }

  if(any(out_names != "")) {
    names(out) <- out_names
  }

  user_monoids <- if(isTRUE(preserve_custom_monoids)) {
    all_monoids <- attr(q, "monoids", exact = TRUE)
    custom <- all_monoids[setdiff(names(all_monoids), c(".size", ".named_count", ".pq_min", ".pq_max"))]
    if(length(custom) == 0L) NULL else custom
  } else {
    NULL
  }

  q2 <- .as_flexseq_build(out, monoids = .pq_merge_monoids(user_monoids))
  .pq_wrap_like(q, q2)
}

#' @method fapply priority_queue
#' @export
#' @noRd
fapply.priority_queue <- function(X, FUN, ..., preserve_custom_monoids = TRUE) {
  if(!is.function(FUN)) {
    stop("`FUN` must be a function.")
  }
  .pq_apply_impl(X, FUN, ..., preserve_custom_monoids = preserve_custom_monoids)
}

#SO

#' Display Internal Structure of a flexseq
#'
#' @method str flexseq
#' @param object A `flexseq`.
#' @param ... Passed to [utils::str()].
#' @return `NULL`, invisibly.
#' @examples
#' x <- flexseq(a = 1, b = list(k = 2))
#' str(x)
#' @keywords internal
#' @export
str.flexseq <- function(object, ...) {
  utils::str(unclass(object), ...)
  invisible(NULL)
}

#SO

# identify internal structural nodes (vs user elements)
# Runtime: O(1).
if(FALSE) is_structural_node <- function(x) NULL
is_structural_node(x) %::% . : logical
is_structural_node(x) %as% {
  x %isa% FingerTree || x %isa% Deep || x %isa% Digit || x %isa% Node || x %isa% Single || x %isa% Empty
}

#SO

#' @method as_flexseq priority_queue
#' @export
# Runtime: O(n) from list materialization + linear rebuild.
as_flexseq.priority_queue <- function(x) {
  entries <- as.list(x, drop_meta = TRUE)
  .as_flexseq_build.default(entries, monoids = NULL)
}

#' Coerce Priority Queue to List
#'
#' Returns queue entries as a plain list of records with fields `value` and
#' `priority`, in queue sequence order.
#'
#' @method as.list priority_queue
#' @param x A `priority_queue`.
#' @param ... Unused.
#' @param drop_meta Logical scalar. When `FALSE` (default), returns full queue
#'   entry records (`value` + `priority`). When `TRUE`, returns payload values
#'   only.
#' @return A plain list of queue entry records (`drop_meta = FALSE`) or payload
#'   values (`drop_meta = TRUE`).
#' @details
#' Each returned entry is a record with fields `value` and `priority`.
#' Entry names (when present) are preserved on the returned list.
#' @examples
#' q <- priority_queue("a", "b", priorities = c(2, 1))
#' as.list(q)
#' as.list(q, drop_meta = TRUE)
#' @export
# Runtime: O(n).
as.list.priority_queue <- function(x, ..., drop_meta = FALSE) {
  .pq_assert_queue(x)
  drop_flag <- .ft_validate_drop_meta(drop_meta)
  entries <- as.list.flexseq(x, ...)
  if(!drop_flag) {
    return(entries)
  }

  values <- lapply(entries, function(e) e$value)
  names(values) <- names(entries)
  values
}

#' Priority Queue Length
#'
#' @method length priority_queue
#' @param x A `priority_queue`.
#' @return Integer length.
#' @details
#' Uses cached size metadata and runs in O(1).
#' @examples
#' q <- priority_queue("a", "b", priorities = c(2, 1))
#' length(q)
#'
#' length(priority_queue())
#' @export
length.priority_queue <- function(x) {
  as.integer(node_measure(x, ".size"))
}

#SO

# Fast plain-R helpers that bypass lambda.r dispatch overhead (~100-190us/call).
# These are hot-path equivalents of lambda.r-defined functions.

# Runtime: O(1). Fast replacement for node_measure(x, ".size").
.ft_size <- function(x) {
  as.integer(attr(x, "measures", exact = TRUE)[[".size"]])
}

# Runtime: O(1). Fast replacement for node_measure(x, ".named_count").
.ft_named_count <- function(x) {
  as.integer(attr(x, "measures", exact = TRUE)[[".named_count"]])
}

# Runtime: O(1). Fast replacement for .ft_get_name(el) (skips lambda.r dispatch).
.ft_get_name_fast <- function(el) {
  nm <- attr(el, "ft_name", exact = TRUE)
  if(is.null(nm)) return(NULL)
  if(!is.character(nm) || length(nm) != 1L || is.na(nm) || !nzchar(nm)) return(NULL)
  nm
}

# Runtime: O(1). Fast replacement for .ft_set_name(el, name) (skips lambda.r dispatch).
# Caller must pass NULL or a validated non-empty string.
.ft_set_name_fast <- function(el, name) {
  attr(el, "ft_name") <- name
  el
}

# mark a structural tree as a user-facing flexseq object.
# Runtime: O(1).
.as_flexseq <- function(x) {
  if(!inherits(x, "FingerTree")) {
    stop("Expected a structural tree node.")
  }
  class(x) <- unique(c("flexseq", setdiff(class(x), "list")))
  x
}

# restore subtype after shared flexseq operations.
# Runtime: O(1) for flexseq/ordered_sequence/priority_queue, O(n) when
# interval_index restore validation is required.
.ft_restore_subclass <- function(out, source, context = "flexseq operation") {
  if(!inherits(out, "FingerTree")) {
    stop("Expected a structural tree node.")
  }
  if(inherits(source, "interval_index")) {
    return(.ivx_restore_tree(out, template = source, context = context))
  }
  if(inherits(source, "ordered_sequence")) {
    key_type <- attr(source, "oms_key_type", exact = TRUE)
    return(.as_ordered_sequence(out, key_type = key_type))
  }
  if(inherits(source, "priority_queue")) {
    return(.pq_wrap_like(source, out))
  }
  .as_flexseq(out)
}

# identify the concrete ordered-like class for user-facing error messages.
# Runtime: O(1).
.ft_ordered_owner_class <- function(x) {
  cls <- class(x)
  if(length(cls) == 0L) {
    return("ordered_sequence")
  }
  keep <- setdiff(
    cls,
    c(
      "ordered_sequence", "flexseq", "list",
      "FingerTree", "Deep", "Single", "Empty", "Digit", "Node"
    )
  )
  if(length(keep) > 0L) {
    return(keep[[1L]])
  }
  if(inherits(x, "interval_index")) {
    return("interval_index")
  }
  "ordered_sequence"
}

# Runtime: O(1).
.ft_stop_ordered_like <- function(x, fn_name, advice = "Use `insert()`.") {
  target <- .ft_ordered_owner_class(x)
  stop(sprintf("`%s()` is not supported for %s. %s", fn_name, target, advice))
}

# normalize constructor input to list without dropping names.
# Runtime: O(n), where n is number of constructor inputs.
.flexseq_input_list <- function(x) {
  if(is.list(x)) {
    return(x)
  }
  as.list(x)
}

# Runtime: O(1).
.ft_validate_drop_meta <- function(drop_meta) {
  if(!is.logical(drop_meta) || length(drop_meta) != 1L || is.na(drop_meta)) {
    stop("`drop_meta` must be TRUE or FALSE.")
  }
  isTRUE(drop_meta)
}

#' Construct a Persistent Flexible Sequence
#'
#' `flexseq(...)` creates an immutable sequence from `...`, preserving element
#' order and optional names, with efficient persistent updates.
#'
#' It is list-like in payload flexibility (any R object per element), but
#' sequence-oriented in API (`push_*`, `peek_*`, `pop_*`, indexing, split/concat).
#'
#' @param ... Sequence elements.
#' @return A `flexseq` object.
#' @details
#' `flexseq` is the base general-purpose structure in this package.
#' Specialized structures such as `priority_queue`, `ordered_sequence`, and
#' `interval_index` build on related internals but expose narrower semantics.
#'
#' `flexseq` operations are persistent: updates return new objects and do not
#' mutate prior versions.
#' @examples
#' x <- flexseq(1, 2, 3)
#' x
#'
#' y <- push_front(x, 0)
#' y
#' x  # unchanged
#'
#' named <- flexseq(a = 1, b = 2)
#' named
#' named[["a"]]
#' @seealso [as_flexseq()], [priority_queue()], [ordered_sequence()], [interval_index()]
#' @export
flexseq <- function(...) {
  .as_flexseq_build(list(...), monoids = NULL)
}

#' Internal builder dispatched by `as_flexseq()` methods.
#'
# Runtime: O(1) generic dispatch.
#' @noRd
.as_flexseq_build <- function(x, monoids = NULL) {
  UseMethod(".as_flexseq_build")
}

# Runtime: O(n) over element count via linear bulk tree construction.
.as_flexseq_build.default <- function(x, monoids = NULL) {
  t <- tree_from(x, monoids = monoids)
  .as_flexseq(t)
}

#' @method as_flexseq default
#' @export
# Runtime: O(n) over element count via linear bulk tree construction.
as_flexseq.default <- function(x) {
  .as_flexseq_build.default(x, monoids = NULL)
}

#' @method as_flexseq flexseq
#' @export
# Runtime: O(1).
as_flexseq.flexseq <- function(x) {
  x
}

#' Concatenate Sequences
#'
#' @method c flexseq
#' @param ... `flexseq` objects.
#' @param recursive Unused; must be `FALSE`.
#' @return A concatenated `flexseq`.
#' @details
#' `c()` is supported for `flexseq` and returns a new concatenated `flexseq`.
#'
#' For `priority_queue`, `ordered_sequence`, and `interval_index`, `c()` is not
#' supported because concatenation can violate structure-specific invariants.
#' Cast first with [as_flexseq()] when sequence-style concatenation is intended,
#' noting that this drops ordering and priority metadata.
#' @examples
#' x <- flexseq("a", "b")
#' y <- flexseq("c", "d")
#' c(x, y)
#'
#' q1 <- priority_queue("a", priorities = 2)
#' q2 <- priority_queue("b", priorities = 1)
#' try(c(q1, q2))
#' c(as_flexseq(q1), as_flexseq(q2))
#'
#' o1 <- ordered_sequence("a", keys = 1)
#' o2 <- ordered_sequence("b", keys = 2)
#' try(c(o1, o2))
#' @export
# Runtime: O(sum(n_i)) worst-case with monoid harmonization.
c.flexseq <- function(..., recursive = FALSE) {
  if(isTRUE(recursive)) {
    stop("`recursive = TRUE` is not supported for flexseq.")
  }
  xs <- list(...)
  if(length(xs) == 0L) {
    return(flexseq())
  }
  out <- xs[[1]]
  for(i in 2:length(xs)) {
    out <- concat_trees(out, xs[[i]])
  }
  .ft_restore_subclass(out, xs[[1]], context = "c()")
}

#' @export
#' @noRd
# Runtime: O(1).
c.priority_queue <- function(..., recursive = FALSE) {
  stop("`c()` is not supported for priority_queue. Cast first with `as_flexseq()`.")
}

#' Merge Two Sequences
#'
#' Returns a new `flexseq` containing all elements of `x` followed by all
#' elements of `y`. Thin wrapper over [c()] for API uniformity across the
#' package's merge methods; `c(x, y)` and `merge(x, y)` are equivalent for
#' `flexseq`.
#'
#' @method merge flexseq
#' @param x A `flexseq`.
#' @param y A `flexseq`.
#' @param ... Unused.
#' @return A new `flexseq`.
#' @details
#' For ordered types (`ordered_sequence`, `interval_index`), `merge()` performs
#' a proper sorted merge respecting keys/intervals — see [merge.ordered_sequence()]
#' and [merge.interval_index()]. For `priority_queue`, see
#' [merge.priority_queue()].
#' @examples
#' x <- flexseq("a", "b")
#' y <- flexseq("c", "d")
#' merge(x, y)
#' @export
# Runtime: O(log(min(m, n))) via concat_trees.
merge.flexseq <- function(x, y, ...) {
  c(x, y)
}

#' Sequence Length
#'
#' @method length flexseq
#' @param x A `flexseq`.
#' @return Number of elements in the sequence.
#' @details
#' Uses cached size metadata and runs in O(1).
#' @examples
#' x <- flexseq("a", "b")
#' length(x)
#'
#' length(flexseq())
#' @export
# Runtime: O(1) using cached `.size` measure.
length.flexseq <- function(x) {
  .ft_size(x)
}

#' Coerce a Sequence to Base List
#'
#' Returns elements in left-to-right sequence order.
#'
#' @method as.list flexseq
#' @param x A `flexseq`.
#' @param ... Unused.
#' @return A base R list of sequence elements.
#' @details
#' Returns payload elements in sequence order. If the sequence is fully named,
#' those names are preserved on the returned list.
#' @examples
#' x <- flexseq("a", "b", "c")
#' as.list(x)
#'
#' n <- flexseq(a = 1, b = 2)
#' as.list(n)
#' @export
# Runtime: O(n) over number of elements.
as.list.flexseq <- function(x, ...) {
  els <- .ft_to_list(x)
  n <- length(els)
  if(n == 0L) {
    return(list())
  }

  out <- vector("list", n)
  nms <- character(n)
  has_names <- FALSE
  for(i in seq_len(n)) {
    el <- els[[i]]
    out[[i]] <- .ft_strip_name(el)
    nm <- .ft_get_name(el)
    if(is.null(nm)) {
      nms[[i]] <- ""
    } else {
      nms[[i]] <- nm
      has_names <- TRUE
    }
  }
  if(has_names) {
    names(out) <- nms
  }
  out
}

#' Coerce a Sequence to an Atomic Vector
#'
#' Convenience wrapper around [base::unlist()] over [as.list()].
#'
#' @method unlist flexseq
#' @param x A `flexseq`.
#' @param recursive Passed through to [base::unlist()].
#' @param use.names Passed through to [base::unlist()].
#' @return An atomic vector built from [as.list.flexseq()].
#' @details
#' For `priority_queue`, this unwraps queue entries to payload items before
#' unlisting (equivalent to `unlist(as.list(x, drop_meta = TRUE), ...)`).
#'
#' Inherited by `ordered_sequence` and `interval_index` through the shared
#' class stack.
#' @examples
#' x <- flexseq(1, 2, 3)
#' unlist(x)
#'
#' q <- priority_queue("a", "b", priorities = c(2, 1))
#' unlist(q)
#' @export
# Runtime: O(n) over number of elements.
unlist.flexseq <- function(x, recursive = TRUE, use.names = TRUE) {
  if(inherits(x, "priority_queue")) {
    return(unlist(as.list(x, drop_meta = TRUE), recursive = recursive, use.names = use.names))
  }
  unlist(as.list(x), recursive = recursive, use.names = use.names)
}

#SO

# read-only locate helpers: no reconstruction, only traversal + metadata.

# Runtime: O(1) with cached measures; may recurse for unannotated structural children.
if(FALSE) locate_child_measure <- function(x, ms, monoid_name, mr) NULL
locate_child_measure(x, ms, monoid_name, mr) %::% . : list : character : MeasureMonoid : .
locate_child_measure(x, ms, monoid_name, mr) %as% {
  if(is_structural_node(x)) {
    cached <- attr(x, "measures", exact = TRUE)
    if(!is.null(cached) && !is.null(cached[[monoid_name]])) {
      return(cached[[monoid_name]])
    }
    return(measure_child_named_impl(x, ms, monoid_name, mr))
  }
  mr$measure(x)
}

# Runtime: O(1).
if(FALSE) locate_child_size <- function(x, ms) NULL
locate_child_size(x, ms) %::% . : list : integer
locate_child_size(x, ms) %as% {
  sr <- ms[[".size"]]
  if(is.null(sr)) {
    return(1L)
  }
  out <- if(is_structural_node(x)) {
    as.integer(node_measure(x, ".size"))
  } else {
    as.integer(sr$measure(x))
  }
  if(is.na(out) || out < 0L) {
    stop(".size monoid must produce non-negative integer sizes.")
  }
  out
}

# Runtime: O(k), where k = length(xs).
if(FALSE) locate_sequence_measure <- function(xs, ms, monoid_name, mr) NULL
locate_sequence_measure(xs, ms, monoid_name, mr) %::% list : list : character : MeasureMonoid : .
locate_sequence_measure(xs, ms, monoid_name, mr) %as% {
  if(length(xs) == 0L) {
    return(mr$i)
  }
  acc <- mr$i
  for(el in xs) {
    acc <- mr$f(acc, locate_child_measure(el, ms, monoid_name, mr))
  }
  acc
}

# Runtime: O(k), where k = digit length (<= 4 in normal tree structure).
if(FALSE) locate_digit <- function(p, i, digit, monoids, monoid_name, i_size = 0L) NULL
locate_digit(p, i, digit, monoids, monoid_name, i_size) %::% Function : . : list : list : character : integer : list
locate_digit(p, i, digit, monoids, monoid_name, i_size = 0L) %as% {
  if(length(digit) == 0) {
    stop("locate_digit called with empty digit")
  }

  ms <- monoids
  mr <- ms[[monoid_name]]
  if(is.null(mr)) {
    stop(paste0("Unknown measure monoid '", monoid_name, "'."))
  }
  locate_digit_impl(p, i, digit, ms, mr, monoid_name, i_size)
}

# Runtime: O(k), where k = digit length (<= 4 in normal tree structure).
locate_digit_impl <- function(p, i, digit, ms, mr, monoid_name, i_size = 0L) {
  # `acc` tracks measure strictly before current candidate; `size_before`
  # tracks positional offset for index metadata.
  acc <- i
  size_before <- as.integer(i_size)

  for(idx in seq_along(digit)) {
    el <- digit[[idx]]
    m_el <- locate_child_measure(el, ms, monoid_name, mr)
    n_el <- locate_child_size(el, ms)
    acc_after <- mr$f(acc, m_el)

    if(p(acc_after)) {
      if(is_structural_node(el)) {
        # If the hit is a structural child, recurse into it so `value` is a leaf
        # value and metadata stays consistent with leaf-level locate semantics.
        return(locate_tree_impl_fast(p, acc, el, ms, mr, monoid_name, size_before))
      }

      right <- if(idx == length(digit)) list() else digit[(idx + 1):length(digit)]
      right_measure <- locate_sequence_measure(right, ms, monoid_name, mr)
      return(list(
        found = TRUE,
        value = el,
        left_measure = acc,
        hit_measure = acc_after,
        right_measure = right_measure,
        index = size_before + 1L
      ))
    }

    acc <- acc_after
    size_before <- size_before + n_el
  }

  list(found = FALSE, value = NULL, left_measure = acc, hit_measure = NULL, right_measure = NULL, index = NULL)
}

# Runtime: O(log n) near locate point depth.
if(FALSE) locate_tree_impl <- function(p, i, t, monoids, monoid_name, i_size = 0L) NULL
locate_tree_impl(p, i, t, monoids, monoid_name, i_size) %::% Function : . : . : list : character : integer : list
locate_tree_impl(p, i, t, monoids, monoid_name, i_size = 0L) %as% {
  ms <- monoids
  mr <- ms[[monoid_name]]
  if(is.null(mr)) {
    stop(paste0("Unknown measure monoid '", monoid_name, "'."))
  }
  locate_tree_impl_fast(p, i, t, ms, mr, monoid_name, i_size)
}

# Runtime: O(log n) near locate point depth.
locate_tree_impl_fast <- function(p, i, t, ms, mr, monoid_name, i_size = 0L) {
  if(!is_structural_node(t)) {
    return(locate_digit_impl(p, i, list(t), ms, mr, monoid_name, as.integer(i_size)))
  }

  if(t %isa% Empty) {
    return(list(found = FALSE, value = NULL, left_measure = i, hit_measure = NULL, right_measure = NULL, index = NULL))
  }

  if(t %isa% Single) {
    return(locate_digit_impl(p, i, list(.subset2(t, 1)), ms, mr, monoid_name, as.integer(i_size)))
  }

  if(t %isa% Deep) {
    # Deep(prefix, middle, suffix): use cached subtree measures to determine
    # which branch contains the first predicate flip without scanning all leaves.
    mpr <- node_measure(.subset2(t,"prefix"), monoid_name)
    mm <- node_measure(.subset2(t,"middle"), monoid_name)
    msf <- node_measure(.subset2(t,"suffix"), monoid_name)

    # `vpr` = i <> measure(prefix), `vm` = i <> measure(prefix) <> measure(middle).
    vpr <- mr$f(i, mpr)
    vm <- mr$f(vpr, mm)

    # Positional offsets for index metadata when descending middle/suffix.
    npr <- as.integer(node_measure(.subset2(t,"prefix"), ".size"))
    nm <- as.integer(node_measure(.subset2(t,"middle"), ".size"))

    if(p(vpr)) {
      # Hit is in prefix. Any right-side measure returned by recursion must be
      # extended by full middle and suffix measures from this Deep node.
      res <- locate_tree_impl_fast(p, i, .subset2(t,"prefix"), ms, mr, monoid_name, as.integer(i_size))
      if(res$found) {
        res$right_measure <- mr$f(mr$f(res$right_measure, mm), msf)
      }
      return(res)
    }

    if(p(vm)) {
      # Hit is in middle. Start accumulator at `vpr` and offset index by prefix size.
      res <- locate_tree_impl_fast(p, vpr, .subset2(t,"middle"), ms, mr, monoid_name, as.integer(i_size + npr))
      if(res$found) {
        # Suffix remains entirely to the right of the located element.
        res$right_measure <- mr$f(res$right_measure, msf)
      }
      return(res)
    }

    # Otherwise hit is in suffix: carry full prefix+middle accumulators and
    # positional offset into the suffix recursion.
    return(locate_tree_impl_fast(p, vm, .subset2(t,"suffix"), ms, mr, monoid_name, as.integer(i_size + npr + nm)))
  }

  # Node / Digit structural lists
  locate_digit_impl(p, i, as.list(t), ms, mr, monoid_name, as.integer(i_size))
}

#SO

# Canonical interval entry constructor.
# **Inputs:** payload `value`; scalar `start`/`end`.
# **Outputs:** entry list(value,start,end,key) with key==start.
# **Used by:** constructor/insert/apply/parse rebuild paths.
.ivx_make_entry <- function(value, start, end) {
  # `key` mirrors `start` so interval entries remain compatible with ordered
  # key monoid paths (e.g. `.oms_max_key`) when those monoids are present.
  list(value = value, start = start, end = end, key = start)
}


# Applies validated element names to entry records as `ft_name` attrs.
# **Inputs:** `entries` list of entry records (optionally named).
# **Outputs:** entries list with normalized per-entry names and stripped list names.
# **Used by:** .ivx_tree_from_ordered_entries().
.ivx_prepare_entry_names <- function(entries) {
  if(length(entries) == 0L) {
    return(entries)
  }
  nms <- names(entries)
  if(is.null(nms)) {
    return(entries)
  }

  out <- entries
  for(i in seq_along(out)) {
    nm <- .ft_normalize_name(nms[[i]])
    if(!is.null(nm)) {
      out[[i]] <- .ft_set_name(out[[i]], nm)
    }
  }
  names(out) <- NULL
  out
}

# Runtime: O(n) for ordered entries.
# Bulk-builds a measured tree from already ordered entries.
# **Inputs:** `entries` ordered entry list; `monoids` normalized monoid list.
# **Outputs:** structural tree/flexseq node carrying the entries.
# **Used by:** constructors and rebuild helpers.
.ivx_tree_from_ordered_entries <- function(entries, monoids) {
  entries <- .ivx_prepare_entry_names(entries)
  if(.ft_cpp_can_use(monoids)) {
    return(.as_flexseq(.ft_cpp_tree_from_sorted(entries, monoids)))
  }
  .ft_tree_from_list_linear(entries, monoids)
}

# Runtime: O(n log n) for stable merge sort by start.
# Stable merge-sort over entry indices by `start`.
# **Inputs:** integer index vector `idx`; `entries` list; scalar `endpoint_type`.
# **Outputs:** reordered integer index vector.
# **Used by:** .ivx_order_entries() fallback path.
.ivx_merge_sort_indices <- function(idx, entries, endpoint_type) {
  n <- length(idx)
  if(n <= 1L) {
    return(idx)
  }

  mid <- as.integer(n %/% 2L)
  left <- .ivx_merge_sort_indices(idx[seq_len(mid)], entries, endpoint_type)
  right <- .ivx_merge_sort_indices(idx[(mid + 1L):n], entries, endpoint_type)

  out <- integer(n)
  i <- 1L
  j <- 1L
  k <- 1L
  while(i <= length(left) && j <= length(right)) {
    cmp <- .ivx_compare_scalar(entries[[left[[i]]]]$start, entries[[right[[j]]]]$start, endpoint_type)
    if(cmp <= 0L) {
      out[[k]] <- left[[i]]
      i <- i + 1L
    } else {
      out[[k]] <- right[[j]]
      j <- j + 1L
    }
    k <- k + 1L
  }

  while(i <= length(left)) {
    out[[k]] <- left[[i]]
    i <- i + 1L
    k <- k + 1L
  }
  while(j <= length(right)) {
    out[[k]] <- right[[j]]
    j <- j + 1L
    k <- k + 1L
  }
  out
}

# Runtime: O(n log n) stable by start and FIFO on ties.
# Produces start-ordered entries with stable tie handling.
# **Inputs:** `entries` list; scalar `endpoint_type`.
# **Outputs:** ordered entry list.
# **Used by:** .ivx_build_from_items().
.ivx_order_entries <- function(entries, endpoint_type) {
  if(length(entries) <= 1L) {
    return(entries)
  }

  idx <- seq_along(entries)
  starts <- lapply(entries, function(e) e$start)
  ord <- tryCatch(
    order(do.call(c, starts), idx),
    error = function(e) NULL
  )
  if(is.null(ord) || length(ord) != length(entries)) {
    ord <- .ivx_merge_sort_indices(idx, entries, endpoint_type)
  }
  entries[ord]
}

# Runtime: O(n log n) from sort + bulk build.
# Core builder from user value/start/end vectors.
# **Inputs:**
#
# - `items`: list payloads.
# - `start`,`end`: vectors/lists of scalar endpoints or NULL.
# - `bounds`: scalar bounds string.
# - `monoids`: optional user monoid list.
# **Outputs:** interval_index.
# **Used by:** as_interval_index(), interval_index(), .as_interval_index_build().
.ivx_build_from_items <- function(items, start = NULL, end = NULL, bounds = "[)", monoids = NULL) {
  n <- length(items)

  if(n == 0L) {
    if(!is.null(start) && length(as.list(start)) > 0L) {
      stop("`start` must be empty when no elements are supplied.")
    }
    if(!is.null(end) && length(as.list(end)) > 0L) {
      stop("`end` must be empty when no elements are supplied.")
    }

    base <- .as_flexseq_build(list(), monoids = .ivx_merge_monoids(monoids))
    return(.as_interval_index(base, endpoint_type = NULL, bounds = bounds))
  }

  if(is.null(start)) {
    stop("`start` is required when elements are supplied.")
  }
  if(is.null(end)) {
    stop("`end` is required when elements are supplied.")
  }

  starts <- as.list(start)
  ends <- as.list(end)

  if(length(starts) != n) {
    stop("`start` length must match elements length.")
  }
  if(length(ends) != n) {
    stop("`end` length must match elements length.")
  }

  entries <- vector("list", n)
  item_names <- names(items)
  endpoint_type <- NULL

  for(i in seq_len(n)) {
    norm <- .ivx_normalize_interval(starts[[i]], ends[[i]], endpoint_type = endpoint_type)
    endpoint_type <- norm$endpoint_type

    entries[[i]] <- .ivx_make_entry(items[[i]], norm$start, norm$end)
  }
  if(!is.null(item_names) && length(item_names) == n) {
    names(entries) <- item_names
  }

  entries <- .ivx_order_entries(entries, endpoint_type)
  merged_monoids <- .ivx_merge_monoids(monoids, endpoint_type = endpoint_type)
  base <- .ivx_tree_from_ordered_entries(entries, merged_monoids)
  .as_interval_index(base, endpoint_type = endpoint_type, bounds = bounds)
}

# Public coercion constructor.
# **Inputs:** `x` coercible to list; `start`/`end` endpoint vectors; `bounds`.
# **Outputs:** interval_index.
# **Used by:** users/tests.
#' Build an Interval Index from `x`, `start`, and `end`
#'
#' Constructs an `interval_index` by pairing each element of `x` with
#' corresponding `start` and `end` endpoints.
#'
#' @param x Elements to add.
#' @param start Start endpoints with the same length as `x`.
#' @param end End endpoints with the same length as `x`.
#' @param default_query_bounds Boundary convention used as the default for
#'   query operations on this index: one of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#'   Per-query `peek_*` / `pop_*` calls may override via their own `bounds`
#'   argument.
#' @return An `interval_index`.
#' @details
#' Output is ordered by interval `start`.
#'
#' Names on `x` are preserved as element names.
#' @examples
#' ix <- as_interval_index(c("a", "b", "c"), start = c(1, 2, 2), end = c(3, 2, 4))
#' ix
#' as.list(peek_all_point(ix, 2))
#'
#' # Endpoints can be other comparable types
#' ix_date <- as_interval_index(
#'   c("phase1", "phase2"),
#'   start = as.Date(c("2024-01-01", "2024-01-10")),
#'   end = as.Date(c("2024-01-05", "2024-01-15"))
#' )
#' ix_date
#' @export
as_interval_index <- function(x, start, end, default_query_bounds = "[)") {
  .as_interval_index_build(x, start = start, end = end, bounds = default_query_bounds, monoids = NULL)
}

# Runtime: O(n log n) from sort + bulk build.
# Internal constructor wrapper accepting optional custom monoids.
# **Inputs:** `x`, `start`, `end`, `bounds`, optional `monoids`.
# **Outputs:** interval_index.
# **Used by:** as_interval_index(), interval_index(), internal rebuild paths.
.as_interval_index_build <- function(x, start, end, bounds = "[)", monoids = NULL) {
  .ivx_build_from_items(as.list(x), start = start, end = end, bounds = bounds, monoids = monoids)
}

# Runtime: O(n log n) from sort + bulk build.
# Variadic convenience constructor.
# **Inputs:** variadic payload args; `start`,`end`; `bounds`.
# **Outputs:** interval_index.
# **Used by:** users/tests.
#' Construct an Interval Index
#'
#' Convenience constructor from `...`, `start`, and `end`.
#'
#' @param ... Elements to add.
#' @param start Start endpoints matching `...`.
#' @param end End endpoints matching `...`.
#' @param default_query_bounds Boundary convention used as the default for
#'   query operations on this index: one of `"[)"`, `"[]"`, `"()"`, `"(]"`.
#'   Per-query `peek_*` / `pop_*` calls may override via their own `bounds`
#'   argument.
#' @return An `interval_index`.
#' @details
#' Empty construction is supported: `interval_index()` returns an empty index.
#'
#' Output is ordered by interval `start`.
#' @examples
#' ix <- interval_index("a", "b", "c", start = c(1, 2, 2), end = c(3, 2, 4))
#' ix
#'
#' interval_index()
#' @export
interval_index <- function(..., start, end, default_query_bounds = "[)") {
  if(missing(start)) {
    start <- NULL
  }
  if(missing(end)) {
    end <- NULL
  }
  .as_interval_index_build(list(...), start = start, end = end, bounds = default_query_bounds, monoids = NULL)
}

#SO

#' Iterate over a `flexseq` (coro iterator)
#'
#' Returns a lazy iterator that yields payload elements left-to-right.
#' Use with [loop()] as the canonical iteration form:
#'
#' ```
#' loop(for (x in s) print(x))
#' ```
#'
#' @param x A `flexseq`.
#' @return A `coro` iterator function.
#' @details
#' Iteration uses repeated left-view (`viewL`) and is O(n) total, O(1) amortized
#' per step. The original `x` is not modified; the iterator holds a private
#' cursor over progressively-smaller tails.
#'
#' For named sequences, internal name metadata is stripped from yielded values
#' to match `peek_front(s)` semantics. Access names via [as.list()] when needed.
#'
#' Inherited by `ordered_sequence` and `interval_index`: for those subclasses
#' the yielded value is the unwrapped payload (keys / interval endpoints
#' dropped), in key-ascending / start-position order respectively. See
#' [as_iterator.priority_queue()] for the priority-order override.
#'
#' # Do not use plain `for` directly
#' Writing `for (x in s) ...` (without [loop()]) will not dispatch this method.
#' R's `for` walks the object's underlying list storage at the C level and
#' bypasses S3 `length`/`[[`, so it silently yields raw finger-tree internals
#' (Digit/Empty/Deep nodes) rather than sequence elements. Always wrap with
#' [loop()], or call [as.list()] first for an eager copy.
#' @examples
#' s <- flexseq("a", "b", "c")
#' loop(for (x in s) print(x))
#' @exportS3Method coro::as_iterator
as_iterator.flexseq <- function(x) {
  monoids <- attr(x, "monoids", exact = TRUE)
  if(is.null(monoids)) {
    stop("Tree has no monoids attribute.")
  }
  state <- x
  function() {
    if(length(state) == 0L) {
      return(coro::exhausted())
    }
    step <- viewL(state, monoids)
    state <<- step$rest
    .ft_unwrap_public_value(x, .ft_strip_name(step$value))
  }
}

#' Iterate over a `priority_queue` (coro iterator)
#'
#' Returns a lazy iterator that yields payload elements in priority-ascending
#' order. Use with [loop()]:
#'
#' ```
#' loop(for (x in pq) print(x))
#' ```
#'
#' @param x A `priority_queue`.
#' @return A `coro` iterator function.
#' @details
#' Traversal is driven by repeated [pop_min()]: each step is O(log n), so full
#' traversal is O(n log n). Ties within equal priorities are yielded in FIFO
#' insertion order (inherited from [pop_min()]).
#'
#' The original `x` is not modified; the iterator holds a private cursor and
#' partial iteration (e.g. via `break`) leaves the source intact.
#'
#' Each yielded value is the bare payload (matching [peek_min()]). Use
#' [fapply()] if your callback needs the priority alongside the value, or cast
#' with [as_flexseq()] for insertion-order iteration.
#' @examples
#' pq <- priority_queue("a", "b", "c", priorities = c(3, 1, 2))
#' loop(for (x in pq) print(x))  # "b", "c", "a"
#' @exportS3Method coro::as_iterator
as_iterator.priority_queue <- function(x) {
  state <- x
  function() {
    if(length(state) == 0L) {
      return(coro::exhausted())
    }
    step <- pop_min(state)
    state <<- step$remaining
    step$value
  }
}

#' Iterate over an iterator (re-exported from coro)
#'
#' Re-exported [coro::loop()]. Enables `for`-loop-style iteration over
#' `immutables` structures without needing to load `coro` separately.
#'
#' @param loop A `for` loop expression.
#' @return `NULL`, invisibly. Called for side effects.
#' @seealso [coro::loop()] for full documentation.
#' @examples
#' s <- flexseq(1, 2, 3)
#' loop(for (x in s) print(x))
#' @importFrom coro loop
#' @export
loop <- coro::loop

#SO

# validate common inputs for get_measure / get_measures
# Runtime: O(1).
.get_measure_validate <- function(x, monoid_name, context) {
  if(!is_structural_node(x)) {
    stop("`x` must be a finger-tree structure (flexseq or subclass).", call. = FALSE)
  }
  if(!is.character(monoid_name) || length(monoid_name) != 1L || is.na(monoid_name) || monoid_name == "") {
    stop("`monoid_name` must be a single non-empty string.", call. = FALSE)
  }
  monoids <- attr(x, "monoids", exact = TRUE)
  if(is.null(monoids) || is.null(monoids[[monoid_name]])) {
    stop("Monoid '", monoid_name, "' is not attached to `x`. ",
         "Attach it with `add_monoids()` first.", call. = FALSE)
  }
  monoids[[monoid_name]]
}

#' Read a Cached Subtree Measure
#'
#' Returns the monoid's accumulated measure across the whole (sub)tree at
#' the root of `x`. This is the cached value that internal operations use
#' for O(log n) locate and split.
#'
#' @param x A flexseq (or any immutables subclass).
#' @param monoid_name Name of an attached monoid (e.g. `.size`, a custom
#'   name from [add_monoids()], or a built-in like `.pq_min` /
#'   `.oms_max_key`).
#' @return The cached monoid value at the root of `x`. Shape depends on
#'   the monoid — typically atomic (e.g. a numeric sum) but may be a list
#'   for built-ins that carry auxiliary state.
#' @details
#' On the full tree `x`, this is the aggregate over every element. After a
#' split — e.g. `s <- split_by_predicate(x, p, "sum")` — call
#' `get_measure(s$left, "sum")` to read the aggregate of just the left
#' side in O(1).
#'
#' For per-element measures, see [get_measures()].
#' @seealso [get_measures()], [measure_monoid()], [add_monoids()]
#' @examples
#' sum_m <- measure_monoid(`+`, 0, function(el) el)
#' x <- add_monoids(as_flexseq(c(3, 1, 4, 1, 5, 9, 2, 6)),
#'                  list(sum = sum_m))
#' get_measure(x, "sum")         # total across the whole tree
#' get_measure(x, ".size")       # built-in: element count
#'
#' q <- priority_queue("a", "b", "c", priorities = c(5, 1, 3))
#' get_measure(q, ".pq_min")     # built-in, list-valued
#' @export
# Runtime: O(1), reads cached attribute.
get_measure <- function(x, monoid_name) {
  .get_measure_validate(x, monoid_name, "get_measure")
  node_measure(x, monoid_name)
}

#' Read Per-Element Monoid Measures
#'
#' Applies the named monoid's `measure()` function to each leaf entry in
#' sequence order, returning the results as a new `flexseq`.
#'
#' @param x A flexseq (or any immutables subclass).
#' @param monoid_name Name of an attached monoid.
#' @return A plain unnamed `flexseq` of length `length(x)`. Entry `i` is
#'   the monoid's `measure()` applied to the `i`-th leaf entry.
#' @details
#' Each leaf in a finger tree stores a structure-specific *entry*
#' (flexseq: the raw element; ordered_sequence: `list(value, key)`;
#' priority_queue: `list(value, priority)`; interval_index:
#' `list(value, start, end)`). A monoid's `measure()` function is defined
#' over that entry shape, and this accessor exposes the per-element
#' values that would be combined to produce the cached aggregate.
#'
#' Returns a `flexseq` rather than a base list so results can be piped
#' back into other immutables operations; use [as.list()] or [unlist()]
#' to convert.
#' @seealso [get_measure()], [measure_monoid()], [fapply()]
#' @examples
#' sum_m <- measure_monoid(`+`, 0, function(el) el)
#' x <- add_monoids(as_flexseq(c(3, 1, 4, 1, 5, 9)), list(sum = sum_m))
#' get_measures(x, "sum") |> unlist()
#'
#' q <- priority_queue("a", "b", "c", priorities = c(5, 1, 3))
#' # Per-element .pq_min measures — each is list(has, priority).
#' get_measures(q, ".pq_min")
#' @export
# Runtime: O(n) over number of elements.
get_measures <- function(x, monoid_name) {
  monoid <- .get_measure_validate(x, monoid_name, "get_measures")
  entries <- .ft_to_list(x)
  if(length(entries) == 0L) {
    return(flexseq())
  }
  as_flexseq(lapply(entries, monoid$measure))
}

#' Build graph data frames for a finger tree
#'
#' @param t FingerTree.
#' @return A list with three elements: `edge_df` (parent/child/label),
#'   `node_df` (node/type/label), and `node_data` — a named list keyed by node
#'   id whose values are lists with fields `type`, `label`, `measures` (cached
#'   monoid values for structural nodes; `NULL` for element leaves), and
#'   `element` (the raw leaf entry for element nodes; `NULL` for structural
#'   nodes).
#' @keywords internal
# Runtime: O(n) over tree nodes/elements.
get_graph_df <- function(t) {
  edge_rows <- list()
  node_rows <- list()
  node_data <- list()

  # Monoids are attached at the tree root; use them to compute the per-leaf
  # measure value exposed on element-node `$measures` so plot/label code can
  # access measures uniformly on structural and leaf nodes.
  root_monoids <- if(is_structural_node(t)) attr(t, "monoids", exact = TRUE) else NULL

  leaf_measures <- function(entry) {
    if(is.null(root_monoids)) {
      return(NULL)
    }
    out <- lapply(root_monoids, function(mo) mo$measure(entry))
    names(out) <- names(root_monoids)
    out
  }

  # The root of a flexseq/subclass carries the outer S3 class first
  # (e.g. c("flexseq", "Deep", ...)); pick the finger-tree structural class
  # for coloring, falling back to class(t)[1] for anything unexpected.
  structural_class <- function(t) {
    hits <- intersect(class(t),
                      c("Deep", "Single", "Empty", "Digit", "Node2", "Node3"))
    if(length(hits) > 0L) hits[[1L]] else class(t)[[1L]]
  }

  add_edge_row <- function(parent, child, label) {
    edge_rows[[length(edge_rows) + 1L]] <<- list(
      parent = as.character(parent),
      child = as.character(child),
      label = as.character(label)
    )
  }

  add_node_row <- function(node, type, label, measures = NULL, element = NULL) {
    key <- as.character(node)
    if(!is.null(node_data[[key]])) {
      return(invisible())
    }
    node_rows[[length(node_rows) + 1L]] <<- list(
      node = key,
      type = as.character(type),
      label = as.character(label)
    )
    node_data[[key]] <<- list(
      id = key,
      type = as.character(type),
      label = as.character(label),
      measures = measures,
      element = element
    )
  }

# Runtime: O(n) worst-case in relevant input/subtree size.
  add_edges <- function(t, path) {

    if(is_structural_node(t) && t %isa% Empty) {
      add_node_row(path, structural_class(t), "",
                   measures = attr(t, "measures", exact = TRUE))
    }
    else if(!is_structural_node(t)) {
      add_node_row(path, "Element",
                   paste0(as.character(unlist(t)), collapse = ", "),
                   element = t,
                   measures = leaf_measures(t))
    } else {
      add_node_row(path, structural_class(t), "",
                   measures = attr(t, "measures", exact = TRUE))

      if(!is.null(names(t))) {
        for(subthing_name in names(t)) {
          subthing <- .subset2(t, subthing_name)
          childid <- paste(path, subthing_name, sep = ":")
          add_edge_row(path, childid, subthing_name)
          add_edges(subthing, childid)
        }
      } else {
        index <- 1
        for(subthing in t) {
          childid <- paste(path, index, sep = ":")
          add_edge_row(path, childid, index)
          add_edges(subthing, childid)
          index <- index + 1
        }
      }
    }
    return(invisible())
  }

  add_edges(t, "root")

  edge_df <- if(length(edge_rows) == 0L) {
    data.frame(parent = character(0), child = character(0), label = character(0), stringsAsFactors = FALSE)
  } else {
    do.call(rbind.data.frame, c(lapply(edge_rows, as.data.frame, stringsAsFactors = FALSE), list(stringsAsFactors = FALSE)))
  }
  node_df <- if(length(node_rows) == 0L) {
    data.frame(node = character(0), type = character(0), label = character(0), stringsAsFactors = FALSE)
  } else {
    do.call(rbind.data.frame, c(lapply(node_rows, as.data.frame, stringsAsFactors = FALSE), list(stringsAsFactors = FALSE)))
  }

  return(list(edge_df, node_df, node_data))
}



#' @importFrom graphics par plot.new polygon strheight strwidth
#' @keywords internal
NULL

# Build a closed polygon path for a rounded rectangle centered at (cx, cy)
# with width w, height h, and corner radius r (clamped to min(w, h) / 2).
# Returns list(x, y) suitable for graphics::polygon().
# Runtime: O(1) in polygon samples; arc sampled at 10 points per corner.
.roundrect_path <- function(cx, cy, w, h, r) {
  r <- min(r, w / 2, h / 2)
  n_arc <- 10L
  ang <- seq(0, pi / 2, length.out = n_arc)

  rx <- cx + w / 2 - r
  lx <- cx - w / 2 + r
  ty <- cy + h / 2 - r
  by <- cy - h / 2 + r

  list(
    x = c(rx + r * cos(ang),           # top-right corner
          lx + r * cos(pi / 2 + ang),  # top-left corner
          lx + r * cos(pi + ang),      # bottom-left corner
          rx + r * cos(3 * pi / 2 + ang)),   # bottom-right corner
    y = c(ty + r * sin(ang),
          ty + r * sin(pi / 2 + ang),
          by + r * sin(pi + ang),
          by + r * sin(3 * pi / 2 + ang))
  )
}

# Register the custom "rounded_rect" igraph vertex shape if not already present.
# The shape sizes each vertex to fit its own label text (so varying-length
# labels all get appropriately sized boxes), and uses `shape_noclip` as the
# clip function — safe here because edges have no arrowheads and the filled
# rectangle is drawn over the edge endpoints.
# Runtime: O(1) per registration; the plot function is O(n) per plot call.
.register_rounded_rect_shape <- function() {
  if("rounded_rect" %in% igraph::shapes()) {
    return(invisible())
  }

  plot_fn <- function(coords, v = NULL, params) {
    vcolor <- params("vertex", "color")
    if(length(vcolor) != 1L && !is.null(v)) vcolor <- vcolor[v]
    fcolor <- params("vertex", "frame.color")
    if(length(fcolor) != 1L && !is.null(v)) fcolor <- fcolor[v]
    vlabel <- params("vertex", "label")
    if(!is.null(v)) vlabel <- vlabel[v]
    lcex <- params("vertex", "label.cex")
    if(length(lcex) != 1L && !is.null(v)) lcex <- lcex[v]

    n <- nrow(coords)
    vcolor <- rep(vcolor, length.out = n)
    fcolor <- rep(fcolor, length.out = n)
    vlabel <- rep(as.character(vlabel), length.out = n)
    lcex <- rep(lcex, length.out = n)

    for(i in seq_len(n)) {
      lbl <- vlabel[i]
      if(is.na(lbl) || nchar(lbl) == 0L) {
        w <- strwidth("oo", cex = lcex[i])
        h <- strheight("M", cex = lcex[i]) * 1.5
      } else {
        lines <- strsplit(lbl, "\n", fixed = TRUE)[[1L]]
        w <- max(strwidth(lines, cex = lcex[i])) + strwidth("oo", cex = lcex[i])
        h <- strheight("Mg", cex = lcex[i]) * length(lines) * 1.35 +
             strheight("o", cex = lcex[i]) * 0.6
      }
      r <- min(w, h) * 0.25
      path <- .roundrect_path(coords[i, 1L], coords[i, 2L], w, h, r)
      polygon(path$x, path$y, col = vcolor[i], border = fcolor[i])
    }
  }

  igraph::add_shape("rounded_rect",
                    clip = igraph::shape_noclip,
                    plot = plot_fn)
  invisible()
}


#' Plot the Internal Finger-Tree Structure
#'
#' Developer-facing visualizer for the finger-tree backing any immutables
#' structure (`flexseq`, `ordered_sequence`, `priority_queue`,
#' `interval_index`). The S3 `plot()` methods on those classes forward to
#' this function, but `plot_structure()` is also callable directly and gives
#' access to the full `node_label` API for custom label formatting. Requires
#' the `igraph` package (listed in `Suggests`).
#'
#' @param t1 A finger-tree-backed immutables structure.
#' @param vertex.size Passed to `igraph::plot.igraph`. Ignored by the default
#'   `"rounded_rect"` shape, which sizes each box to its own label text.
#' @param vertex.shape Vertex shape. Defaults to `"rounded_rect"`, a custom
#'   shape registered on first call that auto-sizes every box to fit its
#'   label (so multi-line labels render cleanly without tuning
#'   `vertex.size`). Any built-in igraph shape (`"circle"`, `"rectangle"`,
#'   `"none"`, ...) also works and is passed through unchanged.
#' @param edge.width Edge width passed to `igraph::plot.igraph`.
#' @param label_edges If `TRUE`, draw the child slot name on each edge
#'   (`"prefix"`/`"middle"`/`"suffix"` for `Deep` nodes, or the numeric
#'   position for `Digit`/`Node2`/`Node3`).
#' @param title Optional plot title.
#' @param node_label Either one of the preset modes `"value"` (default;
#'   payload for elements, blank for structural), `"type"` (node class
#'   name), `"both"` (type + value on separate lines), `"none"`, or a
#'   user-supplied function `function(node)` returning a single character
#'   string per vertex. The `node` argument is a list with fields:
#'   \describe{
#'     \item{`id`}{Internal graph vertex id (string).}
#'     \item{`type`}{Node class: `"Element"`, `"Digit"`, `"Deep"`,
#'       `"Node2"`, `"Node3"`, `"Single"`, or `"Empty"`.}
#'     \item{`label`}{The default label string.}
#'     \item{`measures`}{Named list of accumulated monoid values for the
#'       subtree rooted at this node. For structural nodes these are the
#'       cached values; for element leaves, each entry is the monoid's
#'       `measure()` applied to the leaf entry. Keys include built-ins
#'       (`.size`, `.named_count`, and any structure-specific ones like
#'       `.pq_min`) plus any custom name added via [add_monoids()].}
#'     \item{`element`}{For element nodes, the raw leaf entry. Shape
#'       depends on the structure type; see [measure_monoid()] for the
#'       entry contract. `NULL` for structural nodes.}
#'   }
#'   Measure values are exposed as-is, including list-valued measures
#'   (e.g. the built-in `.pq_min` is `list(has, priority)`).
#' @param label.cex Numeric scalar controlling label text size (passed as
#'   `vertex.label.cex` to igraph). If `NULL` (default), scales
#'   automatically from ~1.0 on small trees down to ~0.55 on dense ones so
#'   the auto-sized `"rounded_rect"` boxes don't collide. Override with a
#'   fixed value for finer control.
#' @param asp Plot aspect ratio (physical y-unit / physical x-unit).
#'   Default `NA` lets the tree fill the device without aspect constraint.
#'   Set to a numeric (e.g. `0.4`) to enforce horizontal stretching,
#'   or `1` for a square plot region.
#' @param legend If `TRUE` (default), draws a horizontal legend below the
#'   tree mapping node-type names to their fill colors, restricted to
#'   types actually present.
#' @param ... Additional arguments passed to `igraph::plot.igraph` (e.g.
#'   `vertex.label.color`, `vertex.frame.color`).
#' @return Invoked for its side effect (draws to the active graphics
#'   device). Returns `NULL` invisibly.
#' @examples
#' if (requireNamespace("igraph", quietly = TRUE)) {
#' \donttest{
#' t <- as_flexseq(letters[1:8])
#' plot_structure(t, title = "Finger tree")
#'
#' # Label every node with its subtree size (leaves contribute 1).
#' plot_structure(as_flexseq(1:10), node_label = function(node) {
#'   paste0(node$type, "\n.size=", node$measures$.size)
#' })
#'
#' # Custom monoid: subtree sum of numeric payloads. Structural nodes show
#' # the accumulated total; leaves show their own contribution.
#' sum_monoid <- measure_monoid(`+`, 0, function(el) el)
#' xs <- add_monoids(as_flexseq(c(3, 1, 4, 1, 5, 9, 2, 6)),
#'                   list(sum = sum_monoid))
#' plot_structure(xs, node_label = function(node) {
#'   if(node$type == "Element") sprintf("%g\nsum=%g", node$element, node$measures$sum)
#'   else sprintf("%s\nsum=%g", node$type, node$measures$sum)
#' })
#'
#' # List-valued built-in measure: priority_queue's .pq_min tracks the min
#' # priority seen in a subtree as list(has, priority). Unpack in the label.
#' pq <- priority_queue("task-a", "task-b", "task-c",
#'                      priorities = c(5, 1, 3))
#' plot_structure(pq, node_label = function(node) {
#'   m <- node$measures$.pq_min
#'   if(node$type == "Element") {
#'     sprintf("%s\np=%g", node$element$value, node$element$priority)
#'   } else if(isTRUE(m$has)) {
#'     sprintf("%s\nmin=%g", node$type, m$priority)
#'   } else {
#'     node$type
#'   }
#' })
#' }
#' }
#' @seealso [measure_monoid()], [add_monoids()]
#' @export
# Runtime: O(n) to build graph structures prior to plotting.
plot_structure <- function(t1, vertex.size = 15, vertex.shape = "rounded_rect",
                           edge.width = 1, label_edges = FALSE, title = NULL,
                           node_label = "value", label.cex = NULL,
                           asp = NA, legend = TRUE, ...) {
  if(!requireNamespace("igraph", quietly = TRUE)) {
    stop("Package 'igraph' is required for plot_structure(). Install it with install.packages('igraph').")
  }

  if(identical(vertex.shape, "rounded_rect")) {
    .register_rounded_rect_shape()
  }

  gdf <- get_graph_df(t1)
  t1_edge_df <- gdf[[1]]
  t1_node_df <- gdf[[2]]
  node_data <- gdf[[3]]

  type_colors <- c(
    Element = "#ffffb3",
    Digit   = "#8dd3c7",
    Deep    = "#bebada",
    Node2   = "#b3de69",
    Node3   = "#fdb462",
    Single  = "#80b1d3",
    Empty   = "#fb8072"
  )
  t1_node_df$color <- unname(type_colors[t1_node_df$type])

  vertices_df <- unique(t1_node_df)
  g <- igraph::graph_from_data_frame(t1_edge_df, vertices = vertices_df, directed = TRUE)

  present_types <- if(isTRUE(legend)) {
    intersect(names(type_colors), unique(vertices_df$type))
  } else {
    character(0L)
  }
  draw_legend <- length(present_types) > 0L

  old_par <- par(no.readonly = TRUE)
  on.exit({
    graphics::layout(1L)
    par(old_par)
  }, add = TRUE)

  if(draw_legend) {
    graphics::layout(matrix(c(1L, 2L), ncol = 1L), heights = c(10, 1))
  }
  # Shrink default margins so the tree fills the panel; leave a little on
  # top for the optional `title`.
  par(mar = c(0.2, 0.2, if(is.null(title)) 0.2 else 2, 0.2))

  if(is.function(node_label)) {
    vertex_ids <- igraph::V(g)$name
    vlabels <- vapply(vertex_ids, function(id) {
      out <- node_label(node_data[[id]])
      if(is.null(out) || length(out) == 0L) "" else as.character(out)[[1L]]
    }, character(1))
  } else {
    node_label <- match.arg(node_label, c("value", "type", "both", "none"))
    if(node_label == "none") {
      vlabels <- rep("", igraph::vcount(g))
    } else if(node_label == "type") {
      vlabels <- igraph::V(g)$type
    } else if(node_label == "both") {
      vlabels <- ifelse(igraph::V(g)$label == "", igraph::V(g)$type, paste0(igraph::V(g)$type, "\n", igraph::V(g)$label))
    } else {
      vlabels <- igraph::V(g)$label
    }
  }

  auto_cex <- if(is.null(label.cex)) {
    # Shrink labels on dense trees so auto-sized rounded_rect boxes don't
    # collide. Tuned so small trees render at cex ~1 and trees approaching
    # ~50+ nodes render around cex ~0.6–0.7.
    n_nodes <- igraph::vcount(g)
    max(0.55, min(1.0, sqrt(25 / n_nodes)))
  } else {
    label.cex
  }

  plot(g,
       layout = igraph::layout_as_tree(g),
       vertex.label = vlabels,
       vertex.label.cex = auto_cex,
       vertex.size = vertex.size,
       vertex.shape = vertex.shape,
       asp = asp,
       edge.arrow.size = 0.4,
       edge.arrow.mode = 0,
       edge.width = edge.width,
       edge.label = ifelse(label_edges, t1_edge_df$label, ""),
       main = title,
       ...
  )

  if(draw_legend) {
    par(mar = c(0, 0, 0, 0))
    plot.new()
    graphics::legend(
      "center",
      legend = present_types,
      fill = unname(type_colors[present_types]),
      border = "black",
      bty = "n",
      cex = 0.8,
      horiz = TRUE
    )
  }

  invisible(NULL)
}

#SO

# Runtime: O(n) overall (entry normalization + linear sequence construction).
#' Build a Priority Queue from `x` and `priorities`
#'
#' Constructs a queue by pairing each element of `x` with the corresponding
#' value in `priorities`.
#'
#' @param x Elements to enqueue.
#' @param priorities Priorities with the same length as `x`.
#' @return A `priority_queue`.
#' @details
#' `x` is interpreted element-wise (via list coercion). Names on `x` are
#' preserved as queue element names.
#'
#' All priorities must be non-missing and mutually comparable.
#' @examples
#' x <- as_priority_queue(letters[1:4], priorities = c(3, 1, 2, 1))
#' x
#' peek_min(x)
#' peek_max(x)
#'
#' # Names are preserved
#' n <- as_priority_queue(setNames(as.list(1:3), c("a", "b", "c")), priorities = c(2, 1, 3))
#' n[["b"]]
#' @export
as_priority_queue <- function(x, priorities) {
  .as_priority_queue_build(x, priorities = priorities, monoids = NULL)
}

# Runtime: O(n) overall (entry normalization + linear sequence construction).
.as_priority_queue_build <- function(x, priorities, monoids = NULL) {
  x_list <- as.list(x)
  n <- length(x_list)

  p_list <- as.list(priorities)
  if(length(p_list) != n) {
    stop("`priorities` length must match elements length.")
  }

  entries <- vector("list", n)
  priority_type <- NULL
  for(i in seq_len(n)) {
    parsed <- .pq_make_entry(x_list[[i]], p_list[[i]], priority_type = priority_type)
    priority_type <- parsed$priority_type
    entries[[i]] <- parsed$entry
  }

  nm <- names(x)
  if(!is.null(nm) && length(nm) > 0L) {
    if(length(nm) != n) {
      stop("`names` length must match elements length.")
    }
    names(entries) <- nm
  }

  q <- .as_flexseq_build(entries, monoids = .pq_merge_monoids(monoids))
  .as_priority_queue(q, priority_type = priority_type)
}

# Runtime: O(n) overall (entry normalization + linear sequence construction).
#' Construct a Priority Queue
#'
#' Creates a `priority_queue` from elements in `...` and matching
#' `priorities`.
#'
#' @param ... Elements to enqueue.
#' @param priorities Priorities with the same length as `...`.
#' @return A `priority_queue`.
#' @details
#' Empty construction is supported: `priority_queue()` returns an empty queue.
#'
#' If elements are named, names are preserved for name-based reads.
#'
#' Queue operations are exposed through `insert()`, `peek_*()`, `pop_*()`,
#' and `fapply()`.
#' @examples
#' x <- priority_queue("a", "b", "c", priorities = c(2, 1, 2))
#' x
#' peek_min(x)
#'
#' empty_q <- priority_queue()
#' peek_min(empty_q)
#' @export
priority_queue <- function(..., priorities) {
  if(missing(priorities)) {
    priorities <- NULL
  }
  .priority_queue_build(..., priorities = priorities, monoids = NULL)
}

# Runtime: O(n) overall (entry normalization + linear sequence construction).
.priority_queue_build <- function(..., priorities = NULL, monoids = NULL) {
  xs <- list(...)
  n <- length(xs)

  if(n == 0L) {
    if(!is.null(priorities) && length(priorities) > 0L) {
      stop("`priorities` must be empty when no elements are supplied.")
    }
    return(.as_priority_queue(.as_flexseq_build(list(), monoids = .pq_merge_monoids(monoids)), priority_type = NULL))
  }

  if(is.null(priorities)) {
    stop("`priorities` is required when elements are supplied.")
  }

  .as_priority_queue_build(xs, priorities = priorities, monoids = monoids)
}

#SO

# Runtime: O(log n) near locate point depth.
# Core bound-search worker on a pre-normalized key; strict=FALSE yields first
# key >= target, strict=TRUE yields first key > target.
# Used by: .oms_bound_index(), .oms_key_span().
.oms_bound_index_prepared <- function(x, key, strict = FALSE) {
  .oms_assert_set(x)
  n <- length(x)
  if(n == 0L) {
    return(1L)
  }

  pred <- if(!isTRUE(strict)) {
    function(v) {
      isTRUE(v$has) && .oms_compare_key(v$key, key, v$key_type) >= 0L
    }
  } else {
    function(v) {
      isTRUE(v$has) && .oms_compare_key(v$key, key, v$key_type) > 0L
    }
  }

  loc <- locate_by_predicate(x, pred, ".oms_max_key", include_metadata = TRUE)
  if(!isTRUE(loc$found)) {
    return(as.integer(n + 1L))
  }
  as.integer(loc$metadata$index)
}

# Runtime: O(n) over tree size for any non-trivial update (rebind/recompute pass).
#' @method add_monoids ordered_sequence
#' @export
#' @noRd
add_monoids.ordered_sequence <- function(t, monoids, overwrite = FALSE) {
  if(length(monoids) > 0L) {
    bad <- intersect(names(monoids), c(".size", ".named_count", ".oms_max_key"))
    if(length(bad) > 0L) {
      target <- .ft_ordered_owner_class(t)
      stop("Reserved monoid names cannot be supplied for ", target, ": ", paste(bad, collapse = ", "))
    }
  }
  add_monoids.flexseq(t, monoids, overwrite = overwrite)
}

# Runtime: O(log n).
#' Minimum Key Value
#'
#' Returns the smallest key currently present in the ordered sequence.
#'
#' @param x An `ordered_sequence`.
#' @return Minimum key, or `NULL` when `x` is empty.
#' @details
#' This follows sequence key order directly.
#' @examples
#' x <- ordered_sequence("a", "b", keys = c(2, 1))
#' min_key(x)
#' min_key(ordered_sequence())
#' @export
min_key <- function(x) {
  .oms_stop_interval_index(x, "min_key")
  .oms_assert_set(x)
  if(length(x) == 0L) {
    return(NULL)
  }
  .ft_get_elem_at(x, 1L)$key
}

# Runtime: O(1).
#' Maximum Key Value
#'
#' Returns the largest key currently present in the ordered sequence.
#'
#' @param x An `ordered_sequence`.
#' @return Maximum key, or `NULL` when `x` is empty.
#' @details
#' Uses cached `.oms_max_key` monoid state.
#' @examples
#' x <- ordered_sequence("a", "b", keys = c(2, 1))
#' max_key(x)
#' max_key(ordered_sequence())
#' @export
max_key <- function(x) {
  .oms_stop_interval_index(x, "max_key")
  .oms_assert_set(x)
  m <- node_measure(x, ".oms_max_key")
  if(!isTRUE(m$has)) {
    return(NULL)
  }
  m$key
}

# Runtime: O(log n) near locate point depth.
# Normalize/validate a key against sequence key_type, then delegate to prepared
# bound search.
# Used by: lower_bound()/upper_bound(), range/count helpers, key span logic.
.oms_bound_index <- function(x, key_value, strict = FALSE) {
  .oms_assert_set(x)
  key_type <- .oms_key_type_state(x)
  norm <- .oms_normalize_key(key_value)
  .oms_validate_key_type(key_type, norm$key_type)
  .oms_bound_index_prepared(x, norm$key, strict = strict)
}

# Runtime: O(1).
# Resolve inclusive/exclusive lower-range boundary into a start index.
# Used by: elements_between(), count_between().
.oms_range_start_index <- function(x, lo, include_lo) {
  if(isTRUE(include_lo)) {
    .oms_bound_index(x, lo, strict = FALSE)
  } else {
    .oms_bound_index(x, lo, strict = TRUE)
  }
}

# Runtime: O(1).
# Resolve inclusive/exclusive upper-range boundary into an exclusive end index.
# Used by: elements_between(), count_between().
.oms_range_end_exclusive_index <- function(x, hi, include_hi) {
  if(isTRUE(include_hi)) {
    .oms_bound_index(x, hi, strict = TRUE)
  } else {
    .oms_bound_index(x, hi, strict = FALSE)
  }
}


# Runtime: O(log n) near insertion/split point depth.
#' @noRd
# Insert one (value,key) while preserving sorted order and FIFO tie stability.
# Used by: insert.ordered_sequence().
.oms_validate_insert_name_state <- function(x, entry, context = "insert()") {
  measures <- attr(x, "measures", exact = TRUE)
  if(is.null(measures)) {
    stop("Tree has no measures attribute.")
  }
  n <- as.integer(measures[[".size"]])
  nn <- as.integer(measures[[".named_count"]])
  if(n > 0L && nn != 0L && nn != n) {
    stop("Invalid tree name state: mixed named and unnamed elements.")
  }
  if(n == 0L) {
    return(invisible(TRUE))
  }
  entry_name <- .ft_get_name(entry)
  if(nn == 0L && !is.null(entry_name)) {
    stop("Cannot mix named and unnamed elements (", context, " would create mixed named and unnamed tree).")
  }
  if(nn == n && is.null(entry_name)) {
    stop("Cannot mix named and unnamed elements (", context, " would create mixed named and unnamed tree).")
  }
  invisible(TRUE)
}

.oms_insert_impl <- function(x, element, key) {
  .oms_assert_set(x)
  norm <- .oms_normalize_key(key)
  key_type <- .oms_validate_key_type(.oms_key_type_state(x), norm$key_type)

  entry <- .oms_make_entry(element, norm$key, name = .ft_effective_name(element))
  .oms_validate_insert_name_state(x, entry, context = "insert()")
  ms <- attr(x, "monoids", exact = TRUE)

  out <- if(.ft_cpp_can_use_oms_insert(ms, key_type)) {
    .as_flexseq(.ft_cpp_oms_insert(x, entry, ms, key_type))
  } else {
    s <- split_by_predicate(
      x,
      function(v) isTRUE(v$has) && .oms_compare_key(v$key, norm$key, v$key_type) > 0L,
      ".oms_max_key"
    )
    # Internal append avoids public ordered_sequence push guards on boundary splits.
    left_plus <- .ft_push_back_impl(s$left, entry, context = "insert()")
    # concat_trees returns a flexseq so need to re-wrap with ordered seq class below
    # the .as_flexseq() above isn't required but matches the return class with this case
    concat_trees(left_plus, s$right)
  }

  .ord_wrap_like(x, out, key_type = key_type)
}

# Runtime: O(log n) near insertion/split point depth.
#' @method insert ordered_sequence
#' @export
#' @noRd
insert.ordered_sequence <- function(x, element, key, ...) {
  .oms_insert_impl(x, element, key)
}

# Runtime: O(log n).
# Compute the contiguous duplicate-key run as [start, end_excl) for one key.
# Used by: peek_key() and pop_key().
.oms_key_span <- function(x, key) {
  .oms_assert_set(x)
  norm <- .oms_normalize_key(key)
  .oms_validate_key_type(.oms_key_type_state(x), norm$key_type)

  start <- .oms_bound_index_prepared(x, norm$key, strict = FALSE)
  end_excl <- .oms_bound_index_prepared(x, norm$key, strict = TRUE)
  list(
    found = isTRUE(end_excl > start),
    key = norm$key,
    start = as.integer(start),
    end_excl = as.integer(end_excl)
  )
}

# Runtime: O(log n) near split points.
# Extract and remove the positional span [start, end_excl) using .size splits.
# Used by: peek_all_key() and pop_all_key().
.oms_slice_key_span <- function(x, start, end_excl) {
  .oms_assert_set(x)
  if(end_excl <= start) {
    empty <- .oms_empty_tree_like(x)
    return(list(matched = empty, rest = .as_flexseq(x)))
  }
  s1 <- split_by_predicate(x, function(v) v >= start, ".size")
  span_len <- as.integer(end_excl - start)
  s2 <- split_by_predicate(s1$right, function(v) v >= (span_len + 1L), ".size")
  list(
    matched = s2$left,
    rest = concat_trees(s1$left, s2$right)
  )
}

# Runtime: O(log n).
#' Find First Element with Key `>=` Query
#'
#' @param x An `ordered_sequence`.
#' @param key Query key.
#' @return A list with fields:
#' - `found`: logical flag.
#' - `index`: one-based position of the first match, or `NULL`.
#' - `value`: matched element, or `NULL`.
#' - `key`: matched key, or `NULL`.
#' @details
#' `lower_bound()` finds the first element with key `>= key`. This includes an
#' exact key match when present, which is useful for starting equality or
#' inclusive range scans.
#' @examples
#' x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
#' lower_bound(x, 2)
#' lower_bound(x, 10)
#' @seealso [upper_bound()]
#' @export
# Public lower-bound query wrapper over .oms_bound_index().
lower_bound <- function(x, key) {
  .oms_stop_interval_index(x, "lower_bound")
  .oms_assert_set(x)
  idx <- .oms_bound_index(x, key, strict = FALSE)
  n <- length(x)

  if(idx > n) {
    return(list(found = FALSE, index = NULL, value = NULL, key = NULL))
  }

  entry <- .ft_get_elem_at(x, as.integer(idx))
  list(found = TRUE, index = idx, value = entry$value, key = entry$key)
}

# Runtime: O(log n).
#' Find First Element with Key `>` Query
#'
#' @param x An `ordered_sequence`.
#' @param key Query key.
#' @return A list with fields:
#' - `found`: logical flag.
#' - `index`: one-based position of the first match, or `NULL`.
#' - `value`: matched element, or `NULL`.
#' - `key`: matched key, or `NULL`.
#' @details
#' `upper_bound()` finds the first element with key `> key`. This skips exact
#' key matches, which is useful for exclusive range endpoints and for finding
#' the position immediately after a duplicate-key run.
#' @examples
#' x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
#' upper_bound(x, 2)
#' upper_bound(x, 10)
#' @seealso [lower_bound()]
#' @export
# Public upper-bound query wrapper over .oms_bound_index().
upper_bound <- function(x, key) {
  .oms_stop_interval_index(x, "upper_bound")
  .oms_assert_set(x)
  idx <- .oms_bound_index(x, key, strict = TRUE)
  n <- length(x)

  if(idx > n) {
    return(list(found = FALSE, index = NULL, value = NULL, key = NULL))
  }

  entry <- .ft_get_elem_at(x, as.integer(idx))
  list(found = TRUE, index = idx, value = entry$value, key = entry$key)
}

# Runtime: O(log n).
#' Peek First Element for One Key
#'
#' Returns the first element whose key equals `key`.
#'
#' @param x An `ordered_sequence`.
#' @param key Query key.
#' @return Matched element, or `NULL` when no matching key exists.
#' @details
#' For duplicate keys, this returns the first element in stable sequence order.
#' @examples
#' x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
#' peek_key(x, 2)
#' peek_key(x, 10)
#' @export
peek_key <- function(x, key) {
  .oms_stop_interval_index(x, "peek_key")
  span <- .oms_key_span(x, key)
  if(!isTRUE(span$found)) {
    return(NULL)
  }
  s <- split_around_by_predicate(x, function(v) v >= span$start, ".size")
  s$value$value
}

# Runtime: O(log n) near split points.
#' Peek All Elements for One Key
#'
#' Returns all elements whose key equals `key`.
#'
#' @param x An `ordered_sequence`.
#' @param key Query key.
#' @return An `ordered_sequence` containing all matches; empty on miss.
#' @details
#' The returned `ordered_sequence` can be inspected with [as.list()].
#' @examples
#' x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
#' out <- peek_all_key(x, 2)
#' as.list(out)
#' @export
peek_all_key <- function(x, key) {
  .oms_stop_interval_index(x, "peek_all_key")
  span <- .oms_key_span(x, key)
  if(!isTRUE(span$found)) {
    return(.ord_wrap_like(x, .oms_empty_tree_like(x)))
  }
  parts <- .oms_slice_key_span(x, span$start, span$end_excl)
  .ord_wrap_like(x, parts$matched)
}

# Runtime: O(log n) near split point depth.
#' Pop First Element for One Key
#'
#' Removes and returns the first element whose key equals `key`.
#'
#' @param x An `ordered_sequence`.
#' @param key Query key.
#' @return A list with fields:
#' - `value`: removed element, or `NULL` on miss.
#' - `key`: removed key, or `NULL` on miss.
#' - `remaining`: ordered sequence after removal.
#' @details
#' For duplicate keys, the first element in stable sequence order is removed.
#' @examples
#' x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
#' out <- pop_key(x, 2)
#' out$value
#' out$remaining
#' pop_key(x, 10)
#' @export
pop_key <- function(x, key) {
  .oms_stop_interval_index(x, "pop_key")
  span <- .oms_key_span(x, key)

  if(!isTRUE(span$found)) {
    return(list(value = NULL, key = NULL, remaining = x))
  }

  s <- split_around_by_predicate(x, function(v) v >= span$start, ".size")
  out <- concat_trees(s$left, s$right)
  seq_out <- .ord_wrap_like(x, out)
  list(value = s$value$value, key = s$value$key, remaining = seq_out)
}

# Runtime: O(log n) near split point depth.
#' Pop All Elements for One Key
#'
#' Removes and returns all elements whose key equals `key`.
#'
#' @param x An `ordered_sequence`.
#' @param key Query key.
#' @return A list with fields:
#' - `elements`: `ordered_sequence` of removed matches.
#' - `remaining`: `ordered_sequence` after removal.
#' @details
#' Use [as.list()] to convert `elements` to a standard R list.
#' @examples
#' x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
#' out <- pop_all_key(x, 2)
#' as.list(out$elements)
#' out$remaining
#' @export
pop_all_key <- function(x, key) {
  .oms_stop_interval_index(x, "pop_all_key")
  span <- .oms_key_span(x, key)

  if(!isTRUE(span$found)) {
    return(list(elements = .ord_wrap_like(x, .oms_empty_tree_like(x)), remaining = x))
  }

  parts <- .oms_slice_key_span(x, span$start, span$end_excl)
  list(
    elements = .ord_wrap_like(x, parts$matched),
    remaining = .ord_wrap_like(x, parts$rest)
  )
}

# Runtime: O(log n) via tree-surgery splits; no per-element gather.
#' Return Elements in a Key Range
#'
#' @param x An `ordered_sequence`.
#' @param from_key Lower bound key.
#' @param to_key Upper bound key.
#' @param include_from Include lower bound when `TRUE`.
#' @param include_to Include upper bound when `TRUE`.
#' @return An `ordered_sequence` of matched elements, in key order. Use
#'   [as.list()] to convert to a plain list.
#' @details
#' Range membership is controlled by `include_from` and `include_to`:
#' - `include_from = TRUE` uses `key >= from_key`; otherwise `key > from_key`.
#' - `include_to = TRUE` uses `key <= to_key`; otherwise `key < to_key`.
#'
#' If no elements fall in the range, returns an empty `ordered_sequence`.
#' @examples
#' x <- ordered_sequence("a", "b", "c", "d", keys = c(1, 2, 2, 3))
#' elements_between(x, 2, 3)
#' as.list(elements_between(x, 2, 3))
#' elements_between(x, 2, 2, include_to = FALSE)
#' @export
# Public range extraction API over bound-index helpers.
elements_between <- function(x, from_key, to_key, include_from = TRUE, include_to = TRUE) {
  .oms_stop_interval_index(x, "elements_between")
  .oms_assert_set(x)
  include_from <- .oms_coerce_lgl_scalar(include_from, "include_from")
  include_to <- .oms_coerce_lgl_scalar(include_to, "include_to")

  start <- .oms_range_start_index(x, from_key, include_from)
  end_excl <- .oms_range_end_exclusive_index(x, to_key, include_to)

  parts <- .oms_slice_key_span(x, start, end_excl)
  .ord_wrap_like(x, parts$matched)
}

# Runtime: O(log n).
#' Count Elements Matching One Key
#'
#' @param x An `ordered_sequence`.
#' @param key Query key.
#' @return Integer count of matches.
#' @details
#' Counts multiplicity for a single key. Returns `0L` when the key is not
#' present.
#' @examples
#' x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
#' count_key(x, 2)
#' count_key(x, 10)
#' @export
# Public multiplicity query for a single key (size of duplicate-key run).
count_key <- function(x, key) {
  .oms_stop_interval_index(x, "count_key")
  .oms_assert_set(x)
  lo <- .oms_bound_index(x, key, strict = FALSE)
  hi <- .oms_bound_index(x, key, strict = TRUE)
  as.integer(max(0L, hi - lo))
}

# Runtime: O(log n).
#' Count Elements in a Key Range
#'
#' @param x An `ordered_sequence`.
#' @param from_key Lower bound key.
#' @param to_key Upper bound key.
#' @param include_from Include lower bound when `TRUE`.
#' @param include_to Include upper bound when `TRUE`.
#' @return Integer count of matches.
#' @details
#' Uses the same range semantics as [elements_between()] but returns only the
#' count:
#' - `include_from = TRUE` uses `key >= from_key`; otherwise `key > from_key`.
#' - `include_to = TRUE` uses `key <= to_key`; otherwise `key < to_key`.
#' @examples
#' x <- ordered_sequence("a", "b", "c", "d", keys = c(1, 2, 2, 3))
#' count_between(x, 2, 3)
#' count_between(x, 2, 2, include_to = FALSE)
#' @export
# Public multiplicity query for a key interval.
count_between <- function(x, from_key, to_key, include_from = TRUE, include_to = TRUE) {
  .oms_stop_interval_index(x, "count_between")
  .oms_assert_set(x)
  include_from <- .oms_coerce_lgl_scalar(include_from, "include_from")
  include_to <- .oms_coerce_lgl_scalar(include_to, "include_to")

  start <- .oms_range_start_index(x, from_key, include_from)
  end_excl <- .oms_range_end_exclusive_index(x, to_key, include_to)
  as.integer(max(0L, end_excl - start))
}

#' Merge Two Ordered Sequences
#'
#' Returns a new `ordered_sequence` containing every entry from both inputs,
#' preserving key order. On duplicate keys, `x`'s entries precede `y`'s
#' (left-biased FIFO).
#'
#' @method merge ordered_sequence
#' @param x An `ordered_sequence`.
#' @param y An `ordered_sequence`.
#' @param ... Unused.
#' @return A new `ordered_sequence` of size `length(x) + length(y)`.
#' @details
#' The merge runs in O(m + n) via a zipper-style traversal, with a fast path
#' to O(log(min(m, n))) when the key ranges are disjoint (all of `x`'s keys
#' `<=` all of `y`'s keys, or vice versa with a strict `<` to keep
#' left-biased FIFO intact on equal boundary keys).
#'
#' Both sequences must share the same key type and the same monoid set;
#' mismatches error rather than being silently harmonized. Merging an
#' empty sequence with a non-empty sequence returns the non-empty one
#' unchanged.
#'
#' Both inputs are left unmodified.
#' @examples
#' a <- ordered_sequence("a1", "a2", "a3", keys = c(1, 3, 5))
#' b <- ordered_sequence("b1", "b2", "b3", keys = c(2, 3, 6))
#' m <- merge(a, b)
#' as.list(m)
#' # At the tied key 3, "a2" precedes "b2".
#' @export
# Runtime: O(m + n); O(log(min(m, n))) on the disjoint fast path.
merge.ordered_sequence <- function(x, y, ...) {
  if(inherits(x, "interval_index") || inherits(y, "interval_index")) {
    stop("Cannot merge `ordered_sequence` with `interval_index`. Merge like-typed structures only.")
  }
  if(!inherits(y, "ordered_sequence")) {
    stop("Both arguments to `merge()` must be ordered_sequences.")
  }

  if(length(x) == 0L) return(y)
  if(length(y) == 0L) return(x)

  kx <- attr(x, "oms_key_type", exact = TRUE)
  ky <- attr(y, "oms_key_type", exact = TRUE)
  if(!identical(kx, ky)) {
    stop("Cannot merge ordered_sequences with different key types.")
  }
  if(!identical(sort(names(attr(x, "monoids", exact = TRUE))),
                sort(names(attr(y, "monoids", exact = TRUE))))) {
    stop("Cannot merge ordered_sequences with different monoid sets.")
  }

  # Disjoint fast paths.
  mx <- max_key(x); my_min <- min_key(y)
  if(.oms_compare_key(mx, my_min, kx) <= 0L) {
    return(.ord_wrap_like(x, concat_trees(x, y)))
  }
  ym <- max_key(y); x_min <- min_key(x)
  if(.oms_compare_key(ym, x_min, kx) < 0L) {
    return(.ord_wrap_like(x, concat_trees(y, x)))
  }

  # General O(m + n) zipper merge on pre-sorted entries.
  xs <- .ft_to_list(x)
  ys <- .ft_to_list(y)
  nxs <- length(xs); nys <- length(ys)
  merged <- vector("list", nxs + nys)
  i <- 1L; j <- 1L; k <- 1L
  while(i <= nxs && j <= nys) {
    if(.oms_compare_key(xs[[i]]$key, ys[[j]]$key, kx) <= 0L) {
      merged[[k]] <- xs[[i]]; i <- i + 1L
    } else {
      merged[[k]] <- ys[[j]]; j <- j + 1L
    }
    k <- k + 1L
  }
  while(i <= nxs) { merged[[k]] <- xs[[i]]; i <- i + 1L; k <- k + 1L }
  while(j <= nys) { merged[[k]] <- ys[[j]]; j <- j + 1L; k <- k + 1L }

  ms <- attr(x, "monoids", exact = TRUE)
  out <- .oms_tree_from_ordered_entries(merged, ms)
  .ord_wrap_like(x, out)
}

testthat::test_that("constructor sorts by key and preserves stable duplicate order", {
  xs <- as_ordered_sequence(
    list("bbb", "a", "cc", "dd", "e"),
    keys = c(3, 1, 2, 2, 1)
  )

  testthat::expect_s3_class(xs, "ordered_sequence")
  testthat::expect_equal(as.list(xs), list("a", "e", "cc", "dd", "bbb"))
  testthat::expect_identical(length(xs), 5L)
})

testthat::test_that("min_key/max_key expose key extrema", {
  xs <- as_ordered_sequence(list("bbb", "a", "cc"), keys = c(3, 1, 2))
  testthat::expect_identical(min_key(xs), 1)
  testthat::expect_identical(max_key(xs), 3)

  empty <- ordered_sequence()
  testthat::expect_null(min_key(empty))
  testthat::expect_null(max_key(empty))
})

testthat::test_that("insert appends at right edge of equal-key block (FIFO ties)", {
  xs <- as_ordered_sequence(list("bb", "aa", "c"), keys = c(2, 2, 1))
  xs2 <- insert(xs, "dd", key = 2)

  testthat::expect_equal(as.list(xs2), list("c", "bb", "aa", "dd"))
  testthat::expect_equal(peek_key(xs2, 2), "bb")

  out1 <- pop_key(xs2, 2)
  out2 <- pop_key(out1$remaining, 2)
  testthat::expect_equal(out1$value, "bb")
  testthat::expect_equal(out2$value, "aa")
})

testthat::test_that("insert works on R backend when insertion is at right boundary", {
  old_cpp <- getOption("immutables.use_cpp")
  options(immutables.use_cpp = FALSE)
  on.exit({
    if(is.null(old_cpp)) {
      options(immutables.use_cpp = NULL)
    } else {
      options(immutables.use_cpp = old_cpp)
    }
  }, add = TRUE)

  xs <- as_ordered_sequence(list("a", "b"), keys = c(1, 2))
  ys <- insert(xs, "c", key = 3)

  testthat::expect_s3_class(ys, "ordered_sequence")
  testthat::expect_equal(as.list(ys), list("a", "b", "c"))
})

testthat::test_that("insert preserves name state for named ordered_sequence", {
  xs_num <- as_ordered_sequence(
    setNames(as.list(c("a", "b")), c("ka", "kb")),
    keys = c(1, 2)
  )
  v_num <- "c"
  names(v_num) <- "kc"
  ys_num <- insert(xs_num, v_num, key = 2)

  testthat::expect_identical(unname(ys_num[["kc"]]), "c")
  testthat::expect_identical(attr(ys_num, "measures")$.named_count, as.integer(length(ys_num)))
  testthat::expect_true(validate_name_state(ys_num))

  xs_date <- as_ordered_sequence(
    setNames(as.list(c("a", "b")), c("da", "db")),
    keys = as.Date(c("2024-01-01", "2024-01-03"))
  )
  v_date <- "c"
  names(v_date) <- "dc"
  ys_date <- insert(xs_date, v_date, key = as.Date("2024-01-02"))

  testthat::expect_identical(unname(ys_date[["dc"]]), "c")
  testthat::expect_identical(attr(ys_date, "measures")$.named_count, as.integer(length(ys_date)))
  testthat::expect_true(validate_name_state(ys_date))
})

testthat::test_that("insert enforces named/unnamed consistency for ordered_sequence", {
  named <- as_ordered_sequence(
    setNames(as.list(c("a", "b")), c("ka", "kb")),
    keys = c(1, 2)
  )
  testthat::expect_error(
    insert(named, "c", key = 3),
    "mixed named and unnamed"
  )

  unnamed <- as_ordered_sequence(list("a", "b"), keys = c(1, 2))
  v_named <- "c"
  names(v_named) <- "kc"
  testthat::expect_error(
    insert(unnamed, v_named, key = 3),
    "mixed named and unnamed"
  )
})

testthat::test_that("lower_bound and upper_bound behave at boundaries", {
  xs <- as_ordered_sequence(list("bbb", "a", "cc", "dddd"), keys = c(3, 1, 2, 4))

  lb1 <- lower_bound(xs, 2)
  ub1 <- upper_bound(xs, 2)
  lb9 <- lower_bound(xs, 9)

  testthat::expect_true(lb1$found)
  testthat::expect_identical(lb1$index, 2L)
  testthat::expect_equal(lb1$value, "cc")

  testthat::expect_true(ub1$found)
  testthat::expect_identical(ub1$index, 3L)
  testthat::expect_equal(ub1$value, "bbb")

  testthat::expect_false(lb9$found)
  testthat::expect_null(lb9$index)
})

testthat::test_that("pop_key/pop_all_key removal by key span", {
  xs <- as_ordered_sequence(list("aa", "bb", "c", "dd", "e"), keys = c(2, 2, 1, 2, 1))

  one <- pop_key(xs, 2)
  testthat::expect_equal(one$value, "aa")
  testthat::expect_equal(as.list(one$remaining), list("c", "e", "bb", "dd"))

  all <- pop_all_key(xs, 2)
  testthat::expect_s3_class(all$elements, "ordered_sequence")
  testthat::expect_equal(as.list(all$elements), list("aa", "bb", "dd"))
  testthat::expect_equal(as.list(all$remaining), list("c", "e"))

  miss_one <- pop_key(xs, 99)
  testthat::expect_null(miss_one$value)
  testthat::expect_equal(as.list(miss_one$remaining), as.list(xs))

  miss_all <- pop_all_key(xs, 99)
  testthat::expect_s3_class(miss_all$elements, "ordered_sequence")
  testthat::expect_identical(length(miss_all$elements), 0L)
  testthat::expect_equal(as.list(miss_all$remaining), as.list(xs))
})

testthat::test_that("elements_between supports inclusivity flags", {
  xs <- as_ordered_sequence(list("a", "bb", "cc", "ddd", "eeee"), keys = c(1, 2, 2, 3, 4))

  e_closed <- elements_between(xs, 2, 3)
  e_open_hi <- elements_between(xs, 2, 3, include_to = FALSE)
  e_open_lo <- elements_between(xs, 2, 3, include_from = FALSE)
  e_miss <- elements_between(xs, 9, 10)

  testthat::expect_s3_class(e_closed, "ordered_sequence")
  testthat::expect_s3_class(e_miss, "ordered_sequence")
  testthat::expect_equal(as.list(e_closed), list("bb", "cc", "ddd"))
  testthat::expect_equal(as.list(e_open_hi), list("bb", "cc"))
  testthat::expect_equal(as.list(e_open_lo), list("ddd"))
  testthat::expect_identical(length(e_miss), 0L)
})

testthat::test_that("peek_key/peek_all_key and pop_key are stable within duplicate key blocks", {
  xs <- as_ordered_sequence(list("a1", "b1", "a2", "a3"), keys = c(1, 2, 1, 1))

  testthat::expect_equal(peek_key(xs, 1), "a1")
  peek_all <- peek_all_key(xs, 1)
  testthat::expect_s3_class(peek_all, "ordered_sequence")
  testthat::expect_equal(as.list(peek_all), list("a1", "a2", "a3"))
  out <- pop_key(xs, 1)
  testthat::expect_equal(out$value, "a1")
  testthat::expect_equal(out$key, 1)
  testthat::expect_equal(as.list(out$remaining), list("a2", "a3", "b1"))

  testthat::expect_null(peek_key(xs, 9))
  testthat::expect_identical(length(peek_all_key(xs, 9)), 0L)
  miss <- pop_key(xs, 9)
  testthat::expect_null(miss$value)
  testthat::expect_null(miss$key)
  testthat::expect_equal(as.list(miss$remaining), as.list(xs))
})

testthat::test_that("count helpers match range and key multiplicities", {
  xs <- as_ordered_sequence(list("a", "bb", "cc", "ddd", "eeee"), keys = c(1, 2, 2, 3, 4))

  testthat::expect_identical(count_key(xs, 2), 2L)
  testthat::expect_identical(count_key(xs, 9), 0L)

  testthat::expect_identical(count_between(xs, 2, 3), 3L)
  testthat::expect_identical(count_between(xs, 2, 3, include_to = FALSE), 2L)
  testthat::expect_identical(count_between(xs, 2, 3, include_from = FALSE), 1L)
  testthat::expect_identical(count_between(xs, 9, 10), 0L)
})

testthat::test_that("split helpers preserve ordered_sequence subclass", {
  xs <- as_ordered_sequence(list("a", "bb", "cc", "ddd", "eeee"), keys = c(1, 2, 2, 3, 4))

  s <- split_by_predicate(xs, function(v) v >= 3L, ".size")
  testthat::expect_s3_class(s$left, "ordered_sequence")
  testthat::expect_s3_class(s$right, "ordered_sequence")
  testthat::expect_equal(as.list(s$left), list("a", "bb"))
  testthat::expect_equal(as.list(s$right), list("cc", "ddd", "eeee"))

  sa <- split_around_by_predicate(xs, function(v) v >= 3L, ".size")
  testthat::expect_s3_class(sa$left, "ordered_sequence")
  testthat::expect_s3_class(sa$right, "ordered_sequence")
  testthat::expect_equal(sa$value$value, "cc")
  testthat::expect_equal(as.list(sa$left), list("a", "bb"))
  testthat::expect_equal(as.list(sa$right), list("ddd", "eeee"))
})

testthat::test_that("order-breaking writes are blocked on ordered types", {
  xs <- as_ordered_sequence(list("a", "b"), keys = c(1, 2))

  testthat::expect_error(c(xs, xs), "not supported")
  testthat::expect_error(push_back(xs, "c"), "not supported")
  testthat::expect_error(push_front(xs, "z"), "not supported")
  testthat::expect_error(insert_at(xs, 1, "z"), "not supported")

  testthat::expect_error({ xs[[1]] <- "z" }, "not supported")
  testthat::expect_error({ xs[1] <- list("z") }, "not supported")
  testthat::expect_error({ xs$a <- "z" }, "not supported")
})

testthat::test_that("ordered replacement blocker messages never leak structural classes", {
  xs <- as_ordered_sequence(list(a = "x", b = "y"), keys = c(1, 2))

  msg1 <- testthat::expect_error({ xs[[1]] <- "z" })$message
  msg2 <- testthat::expect_error({ xs[1] <- list("z") })$message
  msg3 <- testthat::expect_error({ xs$a <- "z" })$message

  testthat::expect_match(msg1, "ordered_sequence")
  testthat::expect_match(msg2, "ordered_sequence")
  testthat::expect_match(msg3, "ordered_sequence")
  testthat::expect_false(grepl("Deep", msg1, fixed = TRUE))
  testthat::expect_false(grepl("Deep", msg2, fixed = TRUE))
  testthat::expect_false(grepl("Deep", msg3, fixed = TRUE))
})

testthat::test_that("ordered subsetting canonicalizes selector order and rejects duplicates", {
  xs <- as_ordered_sequence(
    list(a = "xa", b = "xb", c = "xc", d = "xd"),
    keys = c(1, 2, 3, 4)
  )

  inc <- xs[c(1, 3)]
  testthat::expect_s3_class(inc, "ordered_sequence")
  testthat::expect_equal(as.list(inc), list(a = "xa", c = "xc"))
  testthat::expect_identical(names(as.list(inc)), c("a", "c"))

  by_name <- xs[c("a", "c")]
  testthat::expect_equal(as.list(by_name), list(a = "xa", c = "xc"))
  testthat::expect_identical(names(as.list(by_name)), c("a", "c"))

  testthat::expect_equal(xs[[2]], "xb")
  testthat::expect_equal(xs[["c"]], "xc")
  testthat::expect_equal(xs$c, "xc")

  lgl <- xs[c(TRUE, FALSE)]
  testthat::expect_equal(as.list(lgl), list(a = "xa", c = "xc"))
  testthat::expect_identical(names(as.list(lgl)), c("a", "c"))

  reordered <- NULL
  testthat::expect_warning(
    { reordered <- xs[c(3, 1)] },
    "canonicalizes selector order"
  )
  testthat::expect_equal(as.list(reordered), list(a = "xa", c = "xc"))

  reordered_name <- NULL
  testthat::expect_warning(
    { reordered_name <- xs[c("c", "a")] },
    "canonicalizes selector order"
  )
  testthat::expect_equal(as.list(reordered_name), list(a = "xa", c = "xc"))

  testthat::expect_error(xs[c(2, 2)], "duplicate indices")
  testthat::expect_error(xs[c("a", "a")], "duplicate indices")
  testthat::expect_error(xs[c("a", "missing")], "Unknown element name")
})

testthat::test_that("pop helpers preserve ordered class", {
  xs <- as_ordered_sequence(list("x1", "x2", "x3"), keys = c(1, 2, 3))
  pf <- pop_front(xs)
  pb <- pop_back(xs)
  pm <- pop_at(xs, 2)

  testthat::expect_identical(pf$value, "x1")
  testthat::expect_s3_class(pf$remaining, "ordered_sequence")
  testthat::expect_equal(as.list(pf$remaining), list("x2", "x3"))

  testthat::expect_identical(pb$value, "x3")
  testthat::expect_s3_class(pb$remaining, "ordered_sequence")
  testthat::expect_equal(as.list(pb$remaining), list("x1", "x2"))

  testthat::expect_identical(peek_at(xs, 2), "x2")
  testthat::expect_identical(pm$value, "x2")
  testthat::expect_s3_class(pm$remaining, "ordered_sequence")
  testthat::expect_equal(as.list(pm$remaining), list("x1", "x3"))
})

testthat::test_that("fapply dispatches for ordered_sequence and no reset_ties arg", {
  xs <- as_ordered_sequence(setNames(list("x1", "x2", "x3"), c("a", "b", "c")), keys = c(1, 1, 2))

  xs_item <- fapply(xs, function(value, key, name) {
    toupper(value)
  })
  testthat::expect_s3_class(xs_item, "ordered_sequence")
  testthat::expect_equal(unname(as.list(xs_item)), list("X1", "X2", "X3"))
  testthat::expect_identical(names(as.list(xs_item)), c("a", "b", "c"))

  xs_tagged <- fapply(xs, function(value, key, name) {
    paste(value, key, name, sep = "|")
  })
  testthat::expect_equal(unname(as.list(xs_tagged)), list("x1|1|a", "x2|1|b", "x3|2|c"))
  testthat::expect_identical(names(as.list(xs_tagged)), c("a", "b", "c"))
  testthat::expect_equal(count_key(xs_tagged, 1), 2L)
  testthat::expect_equal(count_key(xs_tagged, 2), 1L)
  testthat::expect_equal(peek_key(xs_tagged, 1), "x1|1|a")

  testthat::expect_error(fapply(xs, 1), "`FUN` must be a function")
  testthat::expect_error(fapply(xs, function(value, key, name) value, reset_ties = TRUE), "unused")
})

testthat::test_that("fapply can drop custom monoids for ordered_sequence", {
  sum_item <- measure_monoid(`+`, 0, function(entry) as.numeric(entry$value))
  xs <- add_monoids(as_ordered_sequence(list(10, 20), keys = c(1, 2)), list(sum_item = sum_item))
  xs2 <- fapply(xs, function(value, key, name) value + 1, preserve_custom_monoids = FALSE)

  ms <- attr(xs2, "monoids", exact = TRUE)
  testthat::expect_true(!is.null(ms[[".size"]]))
  testthat::expect_true(!is.null(ms[[".named_count"]]))
  testthat::expect_true(!is.null(ms[[".oms_max_key"]]))
  testthat::expect_true(is.null(ms[["sum_item"]]))
  testthat::expect_error(node_measure(xs2, "sum_item"), "Missing cached measure")
  testthat::expect_equal(as.list(xs2), list(11, 21))
})

testthat::test_that("ordered_sequence custom monoids accept entry measure arguments", {
  xs <- add_monoids(
    as_ordered_sequence(list(10, 20), keys = c(2, 1)),
    list(
      by_key = measure_monoid(`+`, 0, function(entry) as.numeric(entry$key))
    )
  )
  testthat::expect_equal(node_measure(xs, "by_key"), 3)
})

testthat::test_that("ordered_sequence casts down to flexseq explicitly", {
  sum_key <- measure_monoid(`+`, 0, function(entry) entry$key)
  xs <- add_monoids(as_ordered_sequence(
    setNames(list("x", "y"), c("kx", "ky")),
    keys = c(2, 1)
  ), list(sum_key = sum_key))

  fx <- as_flexseq(xs)
  testthat::expect_s3_class(fx, "flexseq")
  testthat::expect_false(inherits(fx, "ordered_sequence"))
  testthat::expect_equal(fx[["kx"]], "x")
  testthat::expect_equal(fx[["ky"]], "y")

  ms <- names(attr(fx, "monoids", exact = TRUE))
  testthat::expect_true(all(c(".size", ".named_count") %in% ms))
  testthat::expect_false("sum_key" %in% ms)
  testthat::expect_error(node_measure(fx, "sum_key"), "Missing cached measure")
  testthat::expect_false(".oms_max_key" %in% ms)
})

testthat::test_that("ordered_sequence supports Date keys with stable tie handling", {
  keys <- as.Date(c("2024-01-03", "2024-01-01", "2024-01-01", "2024-01-02"))
  xs <- as_ordered_sequence(list("c", "a1", "a2", "b"), keys = keys)

  testthat::expect_equal(as.list(xs), list("a1", "a2", "b", "c"))
  testthat::expect_equal(peek_key(xs, as.Date("2024-01-01")), "a1")

  lb <- lower_bound(xs, as.Date("2024-01-01"))
  ub <- upper_bound(xs, as.Date("2024-01-01"))
  testthat::expect_true(lb$found)
  testthat::expect_identical(lb$index, 1L)
  testthat::expect_equal(lb$value, "a1")
  testthat::expect_s3_class(lb$key, "Date")
  testthat::expect_true(ub$found)
  testthat::expect_identical(ub$index, 3L)
  testthat::expect_equal(ub$value, "b")

  out <- pop_key(xs, as.Date("2024-01-01"))
  testthat::expect_equal(out$value, "a1")
  testthat::expect_s3_class(out$key, "Date")
  testthat::expect_equal(as.list(out$remaining), list("a2", "b", "c"))

  ys <- insert(xs, "a3", key = as.Date("2024-01-01"))
  testthat::expect_equal(as.list(ys), list("a1", "a2", "a3", "b", "c"))
})

testthat::test_that("ordered_sequence supports POSIXct keys with stable tie handling", {
  keys <- as.POSIXct(
    c("2024-01-01 12:00:00", "2024-01-01 10:00:00", "2024-01-01 10:00:00"),
    tz = "UTC"
  )
  xs <- as_ordered_sequence(list("late", "early1", "early2"), keys = keys)

  testthat::expect_equal(as.list(xs), list("early1", "early2", "late"))
  testthat::expect_equal(peek_key(xs, as.POSIXct("2024-01-01 10:00:00", tz = "UTC")), "early1")

  out1 <- pop_key(xs, as.POSIXct("2024-01-01 10:00:00", tz = "UTC"))
  out2 <- pop_key(out1$remaining, as.POSIXct("2024-01-01 10:00:00", tz = "UTC"))
  testthat::expect_equal(out1$value, "early1")
  testthat::expect_equal(out2$value, "early2")
  testthat::expect_s3_class(out1$key, "POSIXct")
})

testthat::test_that("ordered_sequence rejects mixed key domains and missing keys", {
  testthat::expect_error(
    as_ordered_sequence(list("a", "b"), keys = list(1, "1")),
    "Incompatible key type"
  )
  testthat::expect_error(
    as_ordered_sequence(list("a", "b"), keys = list(1, as.Date("2024-01-01"))),
    "Incompatible key type"
  )
  testthat::expect_error(
    as_ordered_sequence(
      list("a", "b"),
      keys = list(as.Date("2024-01-01"), as.POSIXct("2024-01-01 00:00:00", tz = "UTC"))
    ),
    "Incompatible key type"
  )

  xs_num <- as_ordered_sequence(list("a"), keys = 1)
  xs_date <- as_ordered_sequence(list("a"), keys = as.Date("2024-01-01"))
  testthat::expect_error(insert(xs_num, "b", key = "1"), "Incompatible key type")
  testthat::expect_error(insert(xs_num, "b", key = as.Date("2024-01-01")), "Incompatible key type")
  testthat::expect_error(
    insert(xs_date, "b", key = as.POSIXct("2024-01-01 00:00:00", tz = "UTC")),
    "Incompatible key type"
  )

  testthat::expect_error(
    as_ordered_sequence(list("a"), keys = as.Date(NA)),
    "`key` must be non-missing"
  )
  testthat::expect_error(
    insert(xs_date, "b", key = as.Date(NA)),
    "`key` must be non-missing"
  )
})

testthat::test_that("ordered_sequence rejects non-empty keys when no elements are supplied", {
  testthat::expect_error(
    ordered_sequence(keys = c(1, 2)),
    "`keys` must be empty when no elements are supplied"
  )
  testthat::expect_error(
    as_ordered_sequence(list(), keys = c(1, 2)),
    "`keys` must be empty when no elements are supplied"
  )
})

testthat::test_that("ordered_sequence requires keys when elements are supplied", {
  testthat::expect_error(
    ordered_sequence("a", "b"),
    "`keys` is required when elements are supplied"
  )
  testthat::expect_error(
    as_ordered_sequence(list("a", "b"), keys = NULL),
    "`keys` is required when elements are supplied"
  )
})

testthat::test_that("ordered_sequence enforces keys length matches elements length", {
  testthat::expect_error(
    ordered_sequence("a", "b", keys = c(1)),
    "`keys` length must match elements length"
  )
  testthat::expect_error(
    as_ordered_sequence(list("a", "b"), keys = c(1, 2, 3)),
    "`keys` length must match elements length"
  )
})

testthat::test_that("ordered_sequence falls back to internal merge-sort when `c()` on keys fails", {
  # The default sort path in .oms_order_entries uses `order(do.call(c, keys), idx)`.
  # For exotic classes without a `c()` method, `do.call(c, keys)` returns a list
  # which `order()` cannot handle; .oms_build_from_items then delegates to
  # .oms_merge_sort_indices for a stable merge sort using the class's < / >.
  # Class/methods are defined at file top-level (see below) — S3 dispatch for
  # `Ops` needs them in an env the package can see, not inside test_that.

  # Ensure the prerequisite holds (documents the test's premise).
  testthat::expect_true(
    inherits(tryCatch(order(do.call(c, list(mk_comp(1), mk_comp(2))), 1:2),
                      error = function(e) e),
             "error")
  )

  # Large enough to force recursive merge-sort descent.
  vals <- letters[1:7]
  keys <- lapply(c(4, 1, 6, 3, 7, 2, 5), mk_comp)
  xs <- as_ordered_sequence(vals, keys = keys)
  testthat::expect_identical(length(xs), 7L)
  # Sorted ascending by key val: 1,2,3,4,5,6,7 -> b,f,d,a,g,c,e
  testthat::expect_identical(unlist(as.list(xs)), c("b", "f", "d", "a", "g", "c", "e"))
})

testthat::test_that("ordered_sequence supports ordered-factor keys (exotic-key slow path)", {
  # Ordered factors aren't in the fast-domain set (numeric/character/logical/Date/POSIXct)
  # but do support `<`/`>`. This exercises the slow-path key-compare and the
  # non-fast-domain branch of .ft_normalize_scalar_orderable.
  lvls <- c("low", "med", "high")
  f <- factor(lvls, levels = lvls, ordered = TRUE)

  xs <- as_ordered_sequence(c("L", "H", "M"), keys = c(f[1], f[3], f[2]))
  testthat::expect_identical(length(xs), 3L)
  # Sorted by factor level order: low, med, high.
  testthat::expect_identical(unlist(as.list(xs)), c("L", "M", "H"))

  # Key-based range query.
  mid <- elements_between(xs, f[2], f[3])
  testthat::expect_identical(unlist(as.list(mid)), c("M", "H"))
})

#SO

# split a digit at the first point where predicate flips from FALSE to TRUE
# Runtime: O(k), where k = digit length (<= 4 in normal tree structure).
if(FALSE) split_digit <- function(p, i, digit, monoids, monoid_name) NULL
split_digit(p, i, digit, monoids, monoid_name) %::% Function : . : list : list : character : list
split_digit(p, i, digit, monoids, monoid_name) %as% {
  ms <- monoids
  mr <- ms[[monoid_name]]
  if(is.null(mr)) {
    stop(paste0("Unknown measure monoid '", monoid_name, "'."))
  }
  split_digit_impl(p, i, digit, ms, mr, monoid_name)
}

# Runtime: O(k), where k = digit length (<= 4 in normal tree structure).
split_digit_impl <- function(p, i, digit, ms, mr, monoid_name) {
  if(length(digit) == 0) {
    stop("split_digit called with empty digit")
  }

  acc <- i
  for(idx in seq_along(digit)) {
    el <- digit[[idx]]
    # el can be a raw element or a structural node.
    m_el <- if(is_structural_node(el)) {
      cached <- attr(el, "measures", exact = TRUE)
      if(!is.null(cached) && !is.null(cached[[monoid_name]])) {
        cached[[monoid_name]]
      } else {
        # support unannotated Node2/Node3 values in tests and ad-hoc usage.
        measure_child_named_impl(el, ms, monoid_name, mr)
      }
    } else {
      mr$measure(el)
    }
    acc <- mr$f(acc, m_el)
    if(p(acc)) {
      left <- if(idx == 1) list() else digit[1:(idx - 1)]
      right <- if(idx == length(digit)) list() else digit[(idx + 1):length(digit)]
      return(list(left = left, value = el, right = right))
    }
  }

  stop("split_digit: predicate never became true; precondition violated")
}

# Shared test utility: a minimal S3 class that supports `<`, `>`, `==` via
# `Ops` dispatch but has no `c()` method. Used to exercise the merge-sort
# fallback paths in .oms_order_entries / .ivx_order_entries, which fire only
# when `do.call(c, keys)` errors or returns a length-mismatched result.
#
# The methods must be registered via .S3method (not just assigned) so dispatch
# works when the package's internals — in a different namespace than this
# test env — call `<`/`>` on an mk_comp object.

mk_comp <- function(val) {
  structure(list(val = val), class = "mk_comp")
}

Ops.mk_comp <- function(e1, e2) {
  v1 <- if (inherits(e1, "mk_comp")) e1$val else e1
  v2 <- if (inherits(e2, "mk_comp")) e2$val else e2
  get(.Generic)(v1, v2)
}

is.na.mk_comp <- function(x) FALSE

.S3method("Ops", "mk_comp", Ops.mk_comp)
.S3method("is.na", "mk_comp", is.na.mk_comp)

#SO

#' Indexing for Ordered Sequences
#'
#' Read indexing treats vectors as selectors and returns subsets in key order.
#' Out-of-order selectors are canonicalized with a warning.
#' Replacement indexing is blocked to prevent order-breaking writes.
#'
#' @name sub-.ordered_sequence
#' @param x An `ordered_sequence`.
#' @param i Index input.
#' @param value Replacement value (unsupported).
#' @param ... Unused.
#' @return Read methods return ordered payload values/subsets; replacement forms
#'   always error.
#' @details
#' Vector selectors are treated as membership selectors, not output-order
#' instructions.
#'
#' - Integer/character vectors are normalized to unique positions and returned in
#'   canonical sequence order.
#' - Out-of-order selector vectors trigger a warning and are canonicalized.
#' - Duplicate selectors are rejected.
#' - Replacement indexing (`[<-`, `[[<-`, `$<-`) is unsupported.
#' @examples
#' x <- ordered_sequence(a = "A", b = "B", c = "C", keys = c(1, 2, 3))
#'
#' x[c(3, 1)]          # warning; result returned in key order
#' x[c("c", "a")]      # warning; result returned in key order
#' x[c(TRUE, FALSE, TRUE)]
#' x[["b"]]
#' x$b
#'
#' try(x[c(2, 2)])
#' try(x$b <- "updated")
#' @rdname sub-.ordered_sequence
#' @method [ ordered_sequence
#' @export
# Runtime: O(k log n) for reads + O(k log k) selector normalization and rebuild
#          but for name access O(n + k log n) due to how name indexing works
`[.ordered_sequence` <- function(x, i, ...) {
  if(missing(i)) {
    return(x)
  }
  ms <- resolve_tree_monoids(x, required = TRUE)
  n <- as.integer(node_measure(x, ".size"))

  if(is.logical(i)) {
    mask <- .ft_assert_lgl_indices(i, n)
    idx <- .ft_true_positions(mask)
    if(length(idx) == 0L) {
      return(.ord_wrap_like(x, empty_tree(monoids = ms)))
    }
    out <- .ft_prepare_subset_entries(.ft_get_elems_at(x, idx))
    return(.ord_wrap_like(x, tree_from(out, monoids = ms)))
  }

  if(is.character(i)) {
    idx <- .ft_assert_chr_indices(i)
    if(length(idx) == 0L) {
      return(.ord_wrap_like(x, empty_tree(monoids = ms)))
    }
    pos <- .ft_match_name_indices(x, idx, strict_missing = TRUE)
    pos <- .ord_normalize_selector_positions(pos)
    out <- .ft_prepare_subset_entries(.ft_get_elems_at(x, pos))
    return(.ord_wrap_like(x, tree_from(out, monoids = ms)))
  }

  idx <- .ft_assert_int_indices(i, n)
  if(length(idx) == 0L) {
    return(.ord_wrap_like(x, empty_tree(monoids = ms)))
  }
  idx <- .ord_normalize_selector_positions(idx)
  out <- .ft_prepare_subset_entries(.ft_get_elems_at(x, idx))
  .ord_wrap_like(x, tree_from(out, monoids = ms))
}

# Runtime: O(log n) by index, O(n_lookup) by name.
#' @rdname sub-.ordered_sequence
#' @method [[ ordered_sequence
#' @export
`[[.ordered_sequence` <- function(x, i, ...) {
  entry <- `[[.flexseq`(x, i, ...)
  if(!is.list(entry) || !("value" %in% names(entry))) {
    stop("Malformed ordered_sequence entry.")
  }
  entry$value
}

# Runtime: O(1).
#' @rdname sub-.ordered_sequence
#' @method [<- ordered_sequence
#' @export
`[<-.ordered_sequence` <- function(x, i, value) {
  .ft_stop_ordered_like(x, "[<-", "Replacement indexing is not supported for ordered sequences. Consider converting with as_flexseq().")
}

# Runtime: O(1).
#' @rdname sub-.ordered_sequence
#' @method [[<- ordered_sequence
#' @export
`[[<-.ordered_sequence` <- function(x, i, value) {
  .ft_stop_ordered_like(x, "[[<-", "Replacement indexing is not supported. Consider converting with as_flexseq().")
}


# Runtime: O(n_lookup) via strict single-name lookup.
#' @rdname sub-.ordered_sequence
#' @method $ ordered_sequence
#' @param name Element name (for `$` and `$<-`).
#' @export
`$.ordered_sequence` <- function(x, name) {
  nm <- .ft_dollar_name(substitute(name))
  `[[.ordered_sequence`(x, nm)
}

# Runtime: O(1).
#' @rdname sub-.ordered_sequence
#' @method $<- ordered_sequence
#' @export
`$<-.ordered_sequence` <- function(x, name, value) {
  .ft_stop_ordered_like(x, "$<-", "Replacement indexing is not supported. Consider converting with as_flexseq().")
}

#SO

#' Create an Empty Structural Tree
#'
#' @param monoids Optional named list of `measure_monoid` objects.
#' @return An empty finger tree with structural `monoids` and `measures` attrs.
#' @keywords internal
# Runtime: O(1).
empty_tree <- function(monoids = NULL) {
  ms <- if(is.null(monoids)) {
    list(.size = size_measure_monoid())
  } else {
    ensure_size_monoids(monoids)
  }
  measured_empty(ms)
}

#SO

# Runtime: O(1).
# Central type/shape gate for public interval methods.
# **Inputs:** object `x`.
# **Outputs:** invisible TRUE on success; otherwise errors.
# **Used by:** indexing, query API, apply, print, methods.
.ivx_assert_index <- function(x) {
  if(!inherits(x, "interval_index") || !is_structural_node(x)) {
    stop("`x` must be an interval_index.")
  }
  invisible(TRUE)
}

# Runtime: O(1).
# Reads the canonical endpoint domain stored on an interval_index object.
# **Inputs:** `x` interval_index.
# **Outputs:** scalar character endpoint type or NULL.
# **Used by:** insert/query normalization and wrapper restoration.
.ivx_endpoint_type_state <- function(x) {
  attr(x, "ivx_endpoint_type", exact = TRUE)
}

# Runtime: O(1).
# Validates user-facing bound syntax; single source of truth for accepted forms.
# **Inputs:** scalar `bounds` string.
# **Outputs:** normalized bounds string.
# **Used by:** .as_interval_index(), .ivx_resolve_bounds().
.ivx_normalize_bounds <- function(bounds) {
  if(!is.character(bounds) || length(bounds) != 1L || is.na(bounds)) {
    stop("`bounds` must be one of: [), [], (), (].")
  }
  if(!(bounds %in% c("[)", "[]", "()", "(]"))) {
    stop("`bounds` must be one of: [), [], (), (].")
  }
  bounds
}

# Runtime: O(1).
# Decodes bounds string into booleans consumed by query spec builders.
# **Inputs:** scalar `bounds` string.
# **Outputs:** list(include_start=<lgl>, include_end=<lgl>).
# **Used by:** .ivx_spec_* query specification helpers.
.ivx_bounds_flags <- function(bounds) {
  list(
    include_start = identical(substr(bounds, 1L, 1L), "["),
    include_end = identical(substr(bounds, 2L, 2L), "]")
  )
}

# Runtime: O(1).
# Normalizes one endpoint value (scalar, non-missing, compatible type, orderable).
# **Inputs:** endpoint `value`; scalar string `arg_name`; optional scalar `endpoint_type`.
# **Outputs:** list(value=<scalar>, endpoint_type=<scalar character>).
# **Used by:** point-query normalization and interval normalization.
.ivx_normalize_endpoint <- function(value, arg_name, endpoint_type = NULL) {
  if(length(value) != 1L) {
    stop(sprintf("`%s` must be a scalar value.", arg_name))
  }

  na_flag <- suppressWarnings(tryCatch(is.na(value), error = function(e) FALSE))
  if(is.logical(na_flag) && length(na_flag) == 1L && isTRUE(na_flag)) {
    stop(sprintf("`%s` must be non-missing.", arg_name))
  }

  vt <- .ivx_endpoint_type(value)
  if(!is.null(endpoint_type) && !identical(endpoint_type, vt)) {
    stop("Incompatible endpoint type for this interval_index.")
  }

  # Validate orderability for endpoint objects using base Ops.
  .ivx_compare_scalar(value, value, vt)

  list(value = value, endpoint_type = vt)
}

# Runtime: O(1).
# Normalizes start/end pair and enforces start <= end in one place.
# **Inputs:** scalar `start`,`end`; optional scalar `endpoint_type`; arg-name labels.
# **Outputs:** list(start=<scalar>, end=<scalar>, endpoint_type=<scalar character>).
# **Used by:** constructors, insert, relation-query argument validation, entry parse.
.ivx_normalize_interval <- function(start, end, endpoint_type = NULL, start_name = "start", end_name = "end") {
  ns <- .ivx_normalize_endpoint(start, start_name, endpoint_type = endpoint_type)
  ne <- .ivx_normalize_endpoint(end, end_name, endpoint_type = ns$endpoint_type)

  if(.ivx_compare_scalar(ns$value, ne$value, ns$endpoint_type) > 0L) {
    stop("`start` must be <= `end`.")
  }

  list(start = ns$value, end = ne$value, endpoint_type = ns$endpoint_type)
}

# Structural wrapper: attaches interval classes + endpoint/bounds attrs.
# **Inputs:** structural tree `x`; optional scalar `endpoint_type`; scalar `bounds`.
# **Outputs:** interval_index wrapper object.
# **Used by:** constructors and restoration/wrap helpers.
.as_interval_index <- function(x, endpoint_type = NULL, bounds = "[)") {
  if(!is_structural_node(x)) {
    stop("Expected a structural tree node.")
  }

  class(x) <- unique(c("interval_index", "ordered_sequence", "flexseq", setdiff(class(x), "list")))
  attr(x, "ivx_endpoint_type") <- endpoint_type
  attr(x, "ivx_bounds") <- .ivx_normalize_bounds(bounds)

  # Keep ordered_sequence key type available only for primitive key classes.
  attr(x, "oms_key_type") <- if(is.null(endpoint_type) || !(endpoint_type %in% c("numeric", "character", "logical"))) {
    NULL
  } else {
    endpoint_type
  }

  x
}

# Runtime: O(1).
# Re-wraps a structural tree with interval metadata inherited from a template.
# **Inputs:** `template` interval_index; structural `tree`; optional endpoint_type/bounds overrides.
# **Outputs:** interval_index wrapper of `tree`.
# **Used by:** mutation/query rebuild paths and empty-like construction.
.ivx_wrap_like <- function(template, tree, endpoint_type = NULL, bounds = NULL) {
  ep <- if(is.null(endpoint_type)) .ivx_endpoint_type_state(template) else endpoint_type
  b <- .ivx_resolve_bounds(template, bounds)
  .as_interval_index(tree, endpoint_type = ep, bounds = b)
}

# Runtime: O(1).
# Validates/canonicalizes one stored entry payload.
# **Inputs:** entry object `entry`; scalar `context`; optional scalar `endpoint_type`.
# **Outputs:** list(entry=<canonical entry>, endpoint_type=<resolved endpoint type>).
# **Used by:** full-tree validation during subclass restore.
.ivx_parse_entry <- function(entry, context = "interval_index", endpoint_type = NULL) {
  if(!is.list(entry)) {
    stop(context, " entries must be named lists with fields: value, start, end (optional: key).")
  }

  nm <- names(entry)
  if(is.null(nm) || any(is.na(nm)) || any(nm == "")) {
    stop(context, " entries must be named lists with fields: value, start, end (optional: key).")
  }
  if(anyDuplicated(nm) > 0L) {
    stop(context, " entry fields must be unique.")
  }

  bad <- setdiff(nm, c("value", "start", "end", "key"))
  if(length(bad) > 0L) {
    stop(context, " entry contains unsupported field(s): ", paste(bad, collapse = ", "))
  }
  if(!("value" %in% nm) || !("start" %in% nm) || !("end" %in% nm)) {
    stop(context, " entries must include `value`, `start`, and `end`.")
  }

  norm <- .ivx_normalize_interval(entry[["start"]], entry[["end"]], endpoint_type = endpoint_type)

  if("key" %in% nm) {
    # Backward/interop tolerance: accept explicit `key`, but require it to match
    # `start` so interval key semantics stay canonical.
    if(.ivx_compare_scalar(entry[["key"]], norm$start, norm$endpoint_type) != 0L) {
      stop(context, " entry `key` must equal `start`.")
    }
  }

  list(
    entry = .ivx_make_entry(entry[["value"]], norm$start, norm$end),
    endpoint_type = norm$endpoint_type
  )
}

# Runtime: O(1) metadata restoration/wrap.
# Interval subclass restoration for trees produced by shared flexseq ops.
# **Inputs:** structural tree `x`; optional `template` interval_index; scalar `context`.
# **Outputs:** interval_index wrapper preserving endpoint_type/bounds metadata.
# **Used by:** .ft_restore_subclass() in core flexseq API.
.ivx_restore_tree <- function(x, template = NULL, context = "interval_index") {
  endpoint_type <- if(is.null(template)) NULL else .ivx_endpoint_type_state(template)
  bounds <- if(is.null(template)) "[)" else .ivx_resolve_bounds(template, NULL)
  .as_interval_index(x, endpoint_type = endpoint_type, bounds = bounds)
}

# Fast structural-node classifier for internal traversals.
# **Inputs:** `node` object.
# **Outputs:** scalar logical structural classification.
# **Used by:** .ivx_entries() and query-engine subtree size/traversal logic.
.ivx_is_structural_fast <- function(node) {
  cls <- class(node)
  !is.null(cls) && any(cls %in% c("Empty", "Single", "Deep", "Digit", "Node"))
}

# Runtime: O(n).
# Materializes interval entries in left-to-right order with preallocation.
# **Inputs:** `x` interval_index/structural tree.
# **Outputs:** list of interval entry payload records.
# **Used by:** interval bounds/query/apply paths and query-engine prune fallback.
.ivx_entries <- function(x) {
  ms <- attr(x, "measures", exact = TRUE)
  n <- if(!is.null(ms) && !is.null(ms[[".size"]])) {
    as.integer(ms[[".size"]])
  } else {
    as.integer(node_measure(x, ".size"))
  }

  if(is.na(n) || n <= 0L) {
    return(list())
  }

  st <- new.env(parent = emptyenv())
  st$out <- vector("list", n)
  st$pos <- 1L

  fill <- function(node) {
    if(!.ivx_is_structural_fast(node)) {
      st$out[[st$pos]] <- node
      st$pos <- st$pos + 1L
      return(invisible(NULL))
    }

    if(inherits(node, "Empty")) {
      return(invisible(NULL))
    }
    if(inherits(node, "Single")) {
      fill(.subset2(node, 1L))
      return(invisible(NULL))
    }
    if(inherits(node, "Deep")) {
      fill(.subset2(node, "prefix"))
      fill(.subset2(node, "middle"))
      fill(.subset2(node, "suffix"))
      return(invisible(NULL))
    }

    for(el in node) {
      fill(el)
    }
    invisible(NULL)
  }

  fill(x)
  if(!identical(st$pos, as.integer(n + 1L))) {
    return(.ft_to_list(x))
  }
  st$out
}

# Runtime: O(1).
# Builds an empty interval_index preserving class/attrs/monoids from template.
# **Inputs:** `template` interval_index.
# **Outputs:** empty interval_index with template metadata.
# **Used by:** query miss/slice helpers and split partitioning.
.ivx_empty_like <- function(template) {
  ms <- resolve_tree_monoids(template, required = TRUE)
  .ivx_wrap_like(template, empty_tree(monoids = ms))
}

# Runtime: O(1).
# Resolves effective bounds for an operation (override or object default).
# **Inputs:** `x` interval_index; optional scalar `bounds` override.
# **Outputs:** scalar bounds string.
# **Used by:** query API entrypoints, print header, and wrapper restoration.
.ivx_resolve_bounds <- function(x, bounds = NULL) {
  if(!is.null(bounds)) {
    return(.ivx_normalize_bounds(bounds))
  }
  b <- attr(x, "ivx_bounds", exact = TRUE)
  if(is.null(b)) "[)" else b
}

#SO

# Runtime: O(n) over tree size for any non-trivial update (rebind/recompute pass).
#' @method add_monoids flexseq
#' @export
add_monoids.flexseq <- function(t, monoids, overwrite = FALSE) {
  if(length(monoids) > 0L) {
    bad <- intersect(names(monoids), c(".size", ".named_count"))
    if(length(bad) > 0L) {
      stop("Reserved monoid names cannot be supplied for flexseq: ", paste(bad, collapse = ", "))
    }
  }
  add <- ensure_size_monoids(monoids)
  cur <- resolve_tree_monoids(t, required = TRUE)
  merged <- merge_monoid_sets(cur, add, overwrite = overwrite)

  add_only <- setdiff(names(add), names(cur))
  overlap <- setdiff(intersect(names(add), names(cur)), c(".size", ".named_count"))
  recompute_names <- add_only
  if(isTRUE(overwrite) && length(overlap) > 0) {
    recompute_names <- union(recompute_names, overlap)
  }
  recompute_names <- setdiff(recompute_names, c(".size", ".named_count"))

  if(length(recompute_names) == 0) {
    return(t)
  }

  out <- rebind_tree_monoids(t, merged, recompute_names)
  .ft_restore_subclass(out, t, context = "add_monoids()")
}

#SO

#' Indexing for Priority Queues
#'
#' Name-based indexing is supported for reads only. Positional indexing and all
#' replacement indexing are intentionally blocked to preserve queue-first UX.
#'
#' @name sub-.priority_queue
#' @param x A `priority_queue`.
#' @param i Index input. For reads, must be a character name (scalar for `[[`).
#' @param value Replacement value (unsupported).
#' @param ... Unused.
#' @return For `$`/`[[`/`[`: queue payload values or queue subsets by name.
#'   Replacement forms always error.
#' @details
#' `priority_queue` supports name-based read indexing only.
#'
#' - `[`: character vector of names, returns a `priority_queue` subset.
#' - `[[` and `$`: scalar name, return the payload value.
#' - Positional indexing and all replacement indexing forms are unsupported.
#' @examples
#' q <- priority_queue(a = "task-a", b = "task-b", priorities = c(2, 1))
#'
#' q["a"]
#' q[["b"]]
#' q$b
#'
#' try(q[1])
#' try(q$a <- "updated")
NULL

# Runtime: O(k * n_lookup) for short name queries; O(n + k) in map-backed paths.
#' @rdname sub-.priority_queue
#' @method [ priority_queue
#' @export
`[.priority_queue` <- function(x, i, ...) {
  if(missing(i)) {
    return(x)
  }
  if(!is.character(i)) {
    stop("`[.priority_queue` supports character name indexing only. Cast first with `as_flexseq()`.")
  }
  `[.flexseq`(x, i, ...)
}

# Runtime: O(n_lookup) single name lookup + O(log n) element fetch.
#' @rdname sub-.priority_queue
#' @method [[ priority_queue
#' @export
`[[.priority_queue` <- function(x, i, ...) {
  if(!(is.character(i) && length(i) == 1L && !is.na(i))) {
    stop("`[[.priority_queue` supports scalar character names only. Cast first with `as_flexseq()`.")
  }
  entry <- `[[.flexseq`(x, i, ...)
  if(!is.list(entry) || !("value" %in% names(entry))) {
    stop("Malformed priority_queue entry.")
  }
  entry$value
}

# Runtime: O(1).
#' @rdname sub-.priority_queue
#' @method [<- priority_queue
#' @export
`[<-.priority_queue` <- function(x, i, value) {
  stop("`[<-` is not supported for priority_queue. Cast first with `as_flexseq()`.")
}

# Runtime: O(1).
#' @rdname sub-.priority_queue
#' @method [[<- priority_queue
#' @export
`[[<-.priority_queue` <- function(x, i, value) {
  stop("`[[<-` is not supported for priority_queue. Cast first with `as_flexseq()`.")
}

# Runtime: O(n_lookup) single name lookup + O(log n) element fetch.
#' @rdname sub-.priority_queue
#' @method $ priority_queue
#' @param name Element name (for `$` and `$<-`).
#' @export
`$.priority_queue` <- function(x, name) {
  nm <- .ft_dollar_name(substitute(name))
  `[[.priority_queue`(x, nm)
}

# Runtime: O(1).
#' @rdname sub-.priority_queue
#' @method $<- priority_queue
#' @export
`$<-.priority_queue` <- function(x, name, value) {
  stop("`$<-` is not supported for priority_queue. Cast first with `as_flexseq()`.")
}

#SO

##############################
## Node type definitions
##############################

# generic node type
# stores arbitrary child elements in a list; base for Node2/Node3
# Runtime: O(k), where k is number of constructor arguments.
if(FALSE) Node <- function(...) NULL
Node(...) %::% ... : list
Node(...) %as% { 
  res <- list(...)
  return(res)
}

# Node2 and Node3 node types
# represent internal nodes with 2 or 3 children
# Runtime: O(1).
if(FALSE) Node2 <- function(x, y) NULL
Node2(x, y) %::% . : . : list
Node2(x, y) %as% Node(x, y)

# Runtime: O(1).
if(FALSE) Node3 <- function(x, y, z) NULL
Node3(x, y, z) %::% . : . : . : list
Node3(x, y, z) %as% Node(x, y, z)


# Runtime: O(1).
if(FALSE) FingerTree <- function() NULL
FingerTree() %::% FingerTree
FingerTree() %as% Empty()


# basic constructor for inheriting
# internal convenience for creating tree records
# Runtime: O(k), where k is number of constructor arguments.
FingerTree(...) %::% ... : list
FingerTree(...) %as% {
  list(...)
}

# empty node type
# represents an empty tree
# Runtime: O(1).
if(FALSE) Empty <- function() NULL
Empty() %::% FingerTree
Empty() %as% FingerTree(NULL)


# single-element node type
# represents a tree with exactly one element
# Runtime: O(1).
if(FALSE) Single <- function() NULL
Single(x) %::% . : FingerTree
Single(x) %as% FingerTree(x)

# digits are like nodes, but they allow 1 to 4 elements
# used as prefix/suffix containers in Deep
# Runtime: O(k), where k is number of digit elements.
if(FALSE) Digit <- function(...) NULL
Digit(...) %::% ... : list
Digit(...) %as% {
  list(...)
}

# Deep is the main data type, with a prefix (digit), middle (fingertree), and suffix (digit)
# Runtime: O(1).
if(FALSE) Deep <- function(prefix, middle, suffix) NULL
Deep(prefix, middle, suffix) %::% Digit : FingerTree : Digit : FingerTree
Deep(prefix, middle, suffix) %as% {
  FingerTree(prefix = prefix, middle = middle, suffix = suffix)   
}

#SO

#' @method locate_by_predicate flexseq
#' @export
# Runtime: O(log n) near locate point depth.
locate_by_predicate.flexseq <- function(t, predicate, monoid_name, accumulator = NULL, include_metadata = FALSE) {
  ctx <- resolve_named_monoid(t, monoid_name)
  ms <- ctx$monoids
  mr <- ctx$monoid

  if(t %isa% Empty) {
    if(!isTRUE(include_metadata)) {
      return(list(found = FALSE, value = NULL))
    }
    i0 <- if(is.null(accumulator)) mr$i else accumulator
    return(list(
      found = FALSE,
      value = NULL,
      metadata = list(left_measure = i0, hit_measure = NULL, right_measure = NULL, index = NULL)
    ))
  }

  i <- if(is.null(accumulator)) mr$i else accumulator
  total <- mr$f(i, node_measure(t, monoid_name))

  if(!predicate(total)) {
    if(!isTRUE(include_metadata)) {
      return(list(found = FALSE, value = NULL))
    }
    return(list(
      found = FALSE,
      value = NULL,
      metadata = list(left_measure = total, hit_measure = NULL, right_measure = NULL, index = NULL)
    ))
  }

  res <- if(.ft_cpp_can_use(ms)) {
    .ft_cpp_locate(t, predicate, ms, monoid_name, i)
  } else {
    locate_tree_impl_fast(predicate, i, t, ms, mr, monoid_name, 0L)
  }
  if(!isTRUE(include_metadata)) {
    return(list(found = res$found, value = res$value))
  }

  list(
    found = res$found,
    value = res$value,
    metadata = list(
      left_measure = res$left_measure,
      hit_measure = res$hit_measure,
      right_measure = res$right_measure,
      index = res$index
    )
  )
}

# Lets CI force a specific backend for reference / coverage runs via env var
# (options set in the parent process don't propagate through subprocess test
# runners like tools::testInstalledPackage, but env vars do).
if (nzchar(Sys.getenv("IMMUTABLES_USE_CPP"))) {
  options(immutables.use_cpp = as.logical(Sys.getenv("IMMUTABLES_USE_CPP")))
}

#SO

# Runtime: O(1).
.oms_stop_interval_index <- function(x, fn_name) {
  if(inherits(x, "interval_index")) {
    stop(sprintf("`%s()` is not supported for interval_index. Use interval query helpers (`peek_*`, `pop_*`, and `peek_point()`).", fn_name))
  }
  invisible(TRUE)
}

#' Ordered sequences cannot be concatenated with `c()`
#'
#' Rebuild with `ordered_sequence()` / `as_ordered_sequence()` if you need to
#' combine ordered keyed values.
#'
#' @method c ordered_sequence
#' @param ... Ordered objects.
#' @param recursive Unused.
#' @return Never returns; always errors.
#' @export
#' @noRd
# Runtime: O(1).
c.ordered_sequence <- function(..., recursive = FALSE) {
  xs <- list(...)
  target <- if(length(xs) == 0L) "ordered_sequence" else .ft_ordered_owner_class(xs[[1L]])
  stop(sprintf("`c()` is not supported for %s. Cast first with `as_flexseq()`.", target))
}

#SO

# core split implementation following Hinze/Paterson 4.4
#
# split_tree_impl returns a distinguished element with left/right context:
#   list(left = FingerTree, value = element, right = FingerTree)
# It assumes:
# - t is non-empty
# - p(i) is FALSE and p(i <> measure(t)) is TRUE for strict split semantics
# The public wrappers relax these preconditions (matching the paper's split).
# Runtime: O(log n) near split point depth.
if(FALSE) split_tree_impl <- function(p, i, t, monoids, monoid_name) NULL
split_tree_impl(p, i, t, monoids, monoid_name) %::% Function : . : FingerTree : list : character : list
split_tree_impl(p, i, t, monoids, monoid_name) %as% {
  ms <- monoids
  mr <- ms[[monoid_name]]
  if(is.null(mr)) {
    stop(paste0("Unknown measure monoid '", monoid_name, "'."))
  }
  split_tree_impl_fast(p, i, t, ms, mr, monoid_name)
}

# Runtime: O(log n) near split point depth.
split_tree_impl_fast <- function(p, i, t, ms, mr, monoid_name) {
  if(t %isa% Empty) {
    stop("split_tree_impl requires a non-empty tree")
  }

  if(t %isa% Single) {
    return(list(
      left = measured_empty(ms),
      value = .subset2(t, 1),
      right = measured_empty(ms)
    ))
  }

  # Deep(pr, m, sf): test where predicate first flips using cached measures.
  vpr <- mr$f(i, node_measure(.subset2(t,"prefix"), monoid_name))
  vm <- mr$f(vpr, node_measure(.subset2(t,"middle"), monoid_name))

  if(p(vpr)) {
    # split occurs in prefix digit
    s <- split_digit_impl(p, i, .subset2(t,"prefix"), ms, mr, monoid_name)
    # Left context is only the prefix fragment strictly before the hit.
    # `digit_to_tree` handles 0..4 elements and returns the smallest valid tree.
    left_tree <- digit_to_tree(s$left, ms)
    # Right context starts with the suffix of the prefix digit, followed by the
    # original middle and suffix. `deepL` preserves Deep invariants when that
    # rebuilt prefix fragment is empty (borrowing from middle or collapsing).
    right_tree <- deepL(build_digit(s$right, ms), .subset2(t,"middle"), .subset2(t,"suffix"), ms)
    return(list(left = left_tree, value = s$value, right = right_tree))
  }

  if(p(vm)) {
    # split occurs in middle tree, then inside the selected Node2/Node3
    # Stage 1: split middle at node granularity. In a valid Deep shape, middle
    # stores Node2/Node3 blocks (not leaf elements), so `sm$value` is that block.
    sm <- split_tree_impl_fast(p, vpr, .subset2(t,"middle"), ms, mr, monoid_name)
    # Accumulator value immediately before the selected middle node.
    inode <- mr$f(vpr, node_measure(sm$left, monoid_name))
    # Stage 2: split inside the selected node to isolate the leaf element.
    # `as.list(sm$value)` exposes node children as a tiny digit (arity 2 or 3).
    sx <- split_digit_impl(p, inode, as.list(sm$value), ms, mr, monoid_name)
    # Rebuild left and right contexts around the isolated hit. deepR/deepL handle
    # empty-edge cases by borrowing from middle when needed.
    left_tree <- deepR(.subset2(t,"prefix"), sm$left, build_digit(sx$left, ms), ms)
    right_tree <- deepL(build_digit(sx$right, ms), sm$right, .subset2(t,"suffix"), ms)
    return(list(left = left_tree, value = sx$value, right = right_tree))
  }

  # split occurs in suffix digit
  s <- split_digit_impl(p, vm, .subset2(t,"suffix"), ms, mr, monoid_name)
  left_tree <- deepR(.subset2(t,"prefix"), .subset2(t,"middle"), build_digit(s$left, ms), ms)
  right_tree <- digit_to_tree(s$right, ms)
  list(left = left_tree, value = s$value, right = right_tree)
}

---
title: "Developer API"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Developer API}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
if(requireNamespace("pkgload", quietly = TRUE)) {
  pkgload::load_all(".", quiet = TRUE)
} else if(requireNamespace("Immutables", quietly = TRUE)) {
  library(Immutables)
} else {
  stop("Need either installed 'Immutables' or the 'pkgload' package to render this vignette.")
}
```

Every immutables structure is built on finger trees annotated with *monoids* —
cached summaries that support O(log n) search and split. This vignette covers
the developer API for defining custom monoids, using predicate-based locate and
split operations, and validating tree invariants. These operations work
uniformly across `flexseq`, `ordered_sequence`, `priority_queue`, and
`interval_index`.


## Measures and monoids

A monoid is defined by three pieces: a binary combining function `f`, an
identity value `i`, and a `measure` function that maps a single leaf entry to
its measure value. The combining function *must* be associative 
(`f(a, b) == f(b, a)`), and the identity must satisfy `f(i, x) == x` and `f(x, i) == x`. 

Measure values typically represent something you would want to aggregate over a 
subset of elements, with different measures and aggregations resulting in unique
properties.

When operating on a `flexseq`,
you can think of a monoid's `f` as being applied left-to-right (a left fold), 
accumulating measure values at each element. The calculation for the first 
element also uses `f`, which is where `i` is used, as the other input. 


```
elements:  x1,  x2,  x3
measures:  m1,  m2,  m3,  ...
monoid-accumulated measures (infix notation):
  v1 = i %f% m1
  v2 = (i %f% m1) %f% m2            = v1 %f% m2
  v3 = ((i %f% m1) %f% m2) %f% m3   = v2 %f% m3
  ...
```

This left-to-rightness is only well defined for `flexseq` (using element order 
defined by the user in construction), `ordered_sequence` (always ordered by key), 
and `interval_index` (always ordered by interval start). Priority queue element
order is maintained internally with new elements being added on the right, not
in priority order. There may still be uses for monoids that aggregate priority values. 
This is because we can compute an aggregate 
over any *subsequence*, not just prefixes. Internally elements
are stored as leaf nodes in a tree structure, and aggregating measures upward
can be done purely recursively, and cached for fast access at any level.

This monoid keeps track of the sum of values:

```{r}
sum_monoid <- measure_monoid(
  f = `+`,
  i = 0,
  measure = function(el) el
)
```

Next we attach it to
a structure with `add_monoids()`, which accepts a named list of monoids. 
We can pass `show_custom_monoids = TRUE` to `print()` to see this new information:

```{r}
set.seed(42)
task_times <- runif(20, min = 1, max = 10) |>
  round(digits = 1) |>
  as_flexseq()

x <- add_monoids(task_times, list(sum = sum_monoid))

print(x, show_custom_monoids = TRUE)
```

The `plot_structure()` functions visualizes the internal tree layout. Passing a function to
`node_label` lets us render both leaf values and cached structural measures produced by monoids:

```{r fig.height=4, fig.width=10}
plot_structure(x, node_label = function(node) {
  if(node$type == "Element") sprintf("%.1f\nΣ=%.1f", node$element, node$measures$sum)
  else sprintf("Σ=%.1f", node$measures$sum)
})
```

The tree now caches cumulative sum annotations at every internal node,
enabling $O(log n)$ search by accumulated sum. Setting `overwrite = TRUE`
replaces an existing monoid with the same name.

Each structure type reserves certain monoid names for internal use (e.g. `.size`,
`.named_count`, `.pq_min`). Attempting to add a monoid with a reserved name will
error.


## Leaf entry contract

The provided `measure` function receives a structure-specific *entry*, not just the
stored value. The entry shapes are:

- **`flexseq`**: the stored element directly.
- **`ordered_sequence`**: `list(value, key)`.
- **`priority_queue`**: `list(value, priority)`.
- **`interval_index`**: `list(value, start, end)`.

This means a monoid that works on one structure type may not work on another.
Here is a monoid that sums keys on an `ordered_sequence`:

```{r}
key_sum <- measure_monoid(`+`, 0L, function(entry) as.integer(entry$key))

os <- as_ordered_sequence(c("alice", "bob", "carol"), keys = c(10, 20, 30))
os <- add_monoids(os, list(key_sum = key_sum))

print(os, show_custom_monoids = TRUE)
```


## Locating and splitting by predicate

Predicate-based operations scan accumulated monoid values left-to-right and find
the first point where a predicate transitions from `FALSE` to `TRUE`. The
predicate must be *monotonic*: once it returns `TRUE`, it must stay `TRUE` for
all subsequent accumulated values. This is what enables the $O(log n)$ search for a
single split point.

`locate_by_predicate()` finds the *first* element where the predicate fires,
without splitting the tree. With `include_metadata = TRUE` it returns additional
scan information. The predicate function is applied to accumulated measure
values for a given monoid name; here we find the first element where the accumulated
sum exceeds $25$.

```{r}
loc <- locate_by_predicate(x, function(v) v > 25, "sum", include_metadata = TRUE)
str(loc)
```

The `$found` entry indicates that a node where the predicate fires was found; if
this is false the predicate is `FALSE` for all elements through the end. The `$value`
entry gives the measure of the node where the predicate turned `TRUE`.

The optional metadata includes `left_measure` (accumulated value before the match),
`hit_measure` (the matched element's own accumulated measure), `right_measure` (accumulated
value of everything after), and `index` (1-based position). In this example index
4 is where the first time where the accumulated sum is greater than 25, which we can
verify in the plot above ($9.2 + 9.4 + 3.6 = 22.2$, adding in $8.5$ brings the sum to
$30.7$).

`split_by_predicate()` splits the structure at the predicate point, returning
`$left` and `$right`. The matched element is the first element of `$right`:

```{r}
s <- split_by_predicate(x, function(v) v > 25, "sum")
s$left
s$right
```

`split_around_by_predicate()` is similar but extracts the matched element into a
separate `$value` field:

```{r}
sa <- split_around_by_predicate(x, function(v) v > 25, "sum")
sa$left
sa$value
sa$right
```

Both split variants accept an optional `accumulator` argument to start scanning
from a value other than the monoid identity. `split_at()` is a convenience
wrapper that splits by position using the built-in `.size` monoid (which counts nodes;
each leaf node's measure is `1` and it uses the `sum` operator):

```{r}
split_at(x, index = 5)
```


## Built-in monoids in action

As described by [Hinze and Paterson](https://www.cs.ox.ac.uk/ralf.hinze/publications/FingerTrees.pdf),
this monoid + predicate pattern is remarkably flexible. 
The `priority_queue` and `ordered_sequence` types for example are not actually separate
data structures, but monoid-annotated `flexseq` structures plus thin wrappers that
invoke `locate_by_predicate()` and friends with the right predicate. The
same primitives introduced above power every `peek_*`, `pop_*`, and bound
query in the package. (Interval indices also follow these patterns, but
are more complex than would be useful to review here.)

### Priority queue: `.pq_min` and `.pq_max`

Priority queues store elements in insertion order, not priority order, so a
linear scan would be the only way to find the min-priority element without
some extra bookkeeping. That bookkeeping is `.pq_min` (and its twin
`.pq_max`): a monoid whose cached value at every internal node is the
smallest priority anywhere in that subtree, combined by picking the side
with the smaller priority (left-most on ties, which preserves first-in
first-out for equal priorities).

`pop_min(q)` then works in three lines of logic: read the root aggregate
as the *target* priority, form the predicate "has the accumulated subtree
absorbed a slot with this priority yet?", and hand it to
`split_around_by_predicate()` over `.pq_min`. The descent uses cached
subtree aggregates to prune branches whose min is a different value, so
the whole thing runs in $O(\log n)$ even though the leaves themselves are
not in priority order.

### Ordered sequence: `.oms_max_key`

An ordered sequence is a `flexseq` whose elements are kept sorted by key.
Its built-in `.oms_max_key` monoid caches the largest key in each subtree,
combined by picking the larger of its two inputs.

Because the sequence is already key-ordered, the accumulated max-key is
monotonically non-decreasing left-to-right, so a predicate like "is the
running max at least my query?" transitions from `FALSE` to `TRUE` exactly
once at the correct location.
The `lower_bound()` function wires that predicate into `locate_by_predicate()` on
`.oms_max_key` and returns the first element with key at or above the
target; `peek_key()`, `pop_key()`, `count_between()`, and the rest all
follow the same template with different predicate shapes and 
splitting/concatenating.

In practice, `.pq_min`, `.oms_max_key` and other measures are stored as lists,
containing the measure of interest (min priority value, max key value) as well
as other helper information such as the key type to ensure consistent 
comparisons. The identity `i` and associative combining function `f` operate
on list-based measures of the right shape.


## Example: DNA prefix search

Custom monoids and predicate splits combine naturally for index-style queries.
Here we build a persistent subsequence index over a DNA string and use it to find
all positions where a query pattern occurs as a prefix.

```{r}
sequence <- "ACGCGCTCGCGCATAGTCGCGCCTG"
query <- "CGCGC" # goal: find indices where query occurs
```

The idea is to store each position in the string as a value, keyed on a subsequence long
enough to contain the query (we'll keep 10-letter subsequences):

```
key: ACGCGC..., value = 1
key: CGCGCT..., value = 2
key: GCGCTC..., value = 3
...
```

However, ordered sequences actually store their values in key-order:

```
key: ACGCGC..., value = 1
key: AGTCGC..., value = 15
key: ATAGTC..., value = 13
...
```

Since they are in key-order, matches will be found in a run: all subsequences
starting with the query will be next to each other. To extract them we just
need to find the first one (a splitting operation over a monoid), and the last
(another splitting operation over a monoid).

First we define a monoid that tracks the maximum key (subsequence string) seen so far. The
built-in `.oms_max_key` monoid already tracks this, but we define our own to
illustrate the pattern. Recall from above that the measuring function takes a list
with shape `list(value = ..., key = ...)` when used with an `ordered_sequence`;
in this case we just need it to return the key as the measure.

We can add the monoid to an empty tree, and values for new entries will be automatically
computed on addition/removal etc.

```{r}
max_seq <- measure_monoid(max, i = -Inf, measure = function(entry) entry$key)

subseqs <- ordered_sequence()
subseqs <- add_monoids(subseqs, list(max_seq = max_seq))
```

Next we index every position in the DNA string by its length-10 subsequence,
inserting them into the structure:

```{r}
for(i in 1:nchar(sequence)) {
  subseq <- substr(sequence, i, i + 10)
  subseqs <- insert(subseqs, i, key = subseq)
}

subseqs
```

Now we can search for all subsequences starting with the query. First, split to find
the first entry where the measure is `>=` the query:

```{r}
first_split <- split_by_predicate(subseqs, function(m) query <= m, "max_seq")
first_split$left
first_split$right
```

An empty `$right` indicates no matches (all measures are less than the query).
If it's not empty, but the first measure in `$right` does not start with the
query, then there again no matches (the key is between the left and right
but doesn't match). We can use the `get_measures()` function to return a `flexseq` 
of measure values by name and index with `[[1]]` to get the first measure.

If the first key in `$right` *does* start with the query, then there is at least one 
match, and all matches would be in `$right`. We can split again to find where
matching keys end with a different predicate, one that fires on the first measure
that is strictly *greater* than the query *and* no longer starts with it (resulting
in the index of the first non-match in the run).


```{r}
if(length(first_split$right) > 0) {
  right_measures = get_measures(first_split$right, "max_seq")
  first_measure = right_measures[[1]]

  if(startsWith(first_measure, query)) {
    second_split <- split_by_predicate(
      first_split$right,
      function(m) query < m & !startsWith(m, query),
      "max_seq"
    )
    matches <- second_split$left
  }
}

matches
```

Finally we can extract the original string positions from the matching entries:

```{r}
positions <- fapply(matches, function(value, key) value) |> unlist()
positions
```

## Validation helpers

`validate_tree()` checks structural invariants (monoid consistency, node
structure) across the entire tree. `validate_name_state()` verifies that a tree
is either fully unnamed or fully named with unique, non-empty names. Both are
$O(n)$ and intended for debugging and tests.

```{r}
validate_tree(x)
validate_name_state(flexseq(a = 1, b = 2, c = 3))
```

---
title: "Priority Queues"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Priority Queues}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
if(requireNamespace("pkgload", quietly = TRUE)) {
  pkgload::load_all(".", quiet = TRUE)
} else if(requireNamespace("Immutables", quietly = TRUE)) {
  library(Immutables)
} else {
  stop("Need either installed 'Immutables' or the 'pkgload' package to render this vignette.")
}
```


## Priority queue basics

`priority_queue` is a persistent structure that associates elements with priority values, typically numeric or character (string). They provide
fast access to elements by min or max priority, and insertion of new 
elements with given priorities. 

All operations are persistent, and return modified copies. Peeking at the min/max returns the stored element (which may be any type), 
popping returns a list with `$value` containing the stored element, `$priority` the priority,
and `$remaining` the rest of the priority queue without those elements.

```{r}
x <- priority_queue("task_a", "task_b", "task_c", priorities = c(3, 1, 2))
x

peek_min(x)
res <- pop_max(x)

res$value
res$priority
res$remaining
```

The `as_priority_queue()` variant builds a queue from a vector or list of elements paired with a priority vector. useful when priorities are already in a separate vector.

```{r}
q <- as_priority_queue(letters[1:4], priorities = c(3, 1, 2, 1))
q
```

## Priority ties

In the case of multiple identical priorities, peek/pop for min/max return a *single element*, in their original insertion order. Their "all" counterparts (`peek_all_min`, `pop_all_max`, etc.) return two priorities: in `$elements` the elements peeked/popped and in `$remaining` the other elements. `peek_all_max` and `pop_all_max` behave symmetrically on the maximum-priority tie run.
```{r}
y <- priority_queue() |>
  insert("Jones", priority = "B") |>
  insert("Smith", priority = "A") |>
  insert("Adams", priority = "A")

peek_min(y)
peek_all_min(y)

res1 <- pop_min(y)
res1$value
res1$priority
res1$remaining

res2 <- pop_all_min(y)

res2$elements
res2$remaining
```
Priority queues can be converted to R lists with `as.list()`. The `drop_meta` parameter defaulting to `FALSE` controls whether priority values are included.

```{r}
as.list(res2$elements) |> str()
as.list(res2$elements, drop_meta = TRUE) |> str()
```

## Empty queues

Peek/pop helpers are non-throwing on empty queues, `length()` allows for empty checking.

```{r}
empty_q <- priority_queue()
length(empty_q)
peek_min(empty_q)
pop_min(empty_q)
```

## Priority values

`min_priority()` and `max_priority()` return the current extrema *priority* (not the stored element), in O(1) via cached monoid state.

```{r}
q <- priority_queue("a", "b", "c", priorities = c(3, 1, 2))
min_priority(q)
max_priority(q)
min_priority(priority_queue())  # NULL when empty
```

## Named priorities

Priority queues can carry names, set either at construction or on `insert()`. Names support read-only indexing via `[`, `[[`, and `$`. This is the *only* indexing form `priority_queue` supports. Positional indexing and all replacement forms (`[<-`, `[[<-`, `$<-`) error; cast to `as_flexseq()` first to mutate.

```{r}
q <- priority_queue(a = "task-a", b = "task-b", priorities = c(2, 1)) |>
  insert("task-c", priority = 3, name = "c")

q[["b"]]
q[c("a", "c")]

try(q[1])         # positional indexing blocked
try(q$a <- "!!")  # replacement blocked
```

## Transforming, iterating, merging

`fapply()` maps a function over queue elements while preserving priorities. The function receives `(value, priority)`, or `(value, priority, name)` if it accepts a third argument — priorities and names are passed in read-only, and the return value replaces the stored element.

```{r}
q <- priority_queue("alice", "bob", "carol", priorities = c(3, 1, 2))
fapply(q, function(value, priority) toupper(value))
```

`loop()` (re-exported from the **coro** package) walks the queue in priority-ascending order, yielding bare values. Traversal is driven by repeated `pop_min()`, so full traversal is $O(n \log n)$ ($O(\log n)$ per step); ties within equal priority follow insertion order.

```{r}
q <- priority_queue("task_a", "task_b", "task_c", priorities = c(3, 1, 2))
loop(for (v in q) print(v))
```

Use `fapply()` if your callback needs the priority alongside the value, or cast with `as_flexseq()` for insertion-order iteration. Plain `for (v in q)` (without `loop()`) yields raw finger-tree internals rather than elements — always wrap with `loop()`.

`merge(x, y)` combines two priority queues into a new one containing every entry from both, in O(log(min(m, n))). The `.pq_min` / `.pq_max` monoids recompute automatically on the merged tree, so extremum queries work immediately on the result.

```{r}
a <- priority_queue("x", "y", priorities = c(5, 1))
b <- priority_queue("z", priorities = 3)
m <- merge(a, b)
length(m)
peek_min(m)
peek_max(m)
```

Both queues must share the same priority type and monoid set; mismatches error (rather than being silently harmonized). Both inputs are left unmodified.

For full sequence-style operations, cast with `as_flexseq()`. This returns
payload items only; priority metadata and any custom monoids are dropped. To
instead obtain a `flexseq` of entry records (`value` + `priority`), compose
with `as.list()`.

```{r}
y <- priority_queue() |>
  insert("Jones", priority = "B") |>
  insert("Smith", priority = "A") |>
  insert("Adams", priority = "A")

as_flexseq(y)
as_flexseq(as.list(y))
```

---
title: "Interval Indices"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Interval Indices}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
if(requireNamespace("pkgload", quietly = TRUE)) {
  pkgload::load_all(".", quiet = TRUE)
} else if(requireNamespace("Immutables", quietly = TRUE)) {
  library(Immutables)
} else {
  stop("Need either installed 'Immutables' or the 'pkgload' package to render this vignette.")
}
```

## Interval index basics

`interval_index` is a persistent structure that associates elements with interval endpoints and supports fast interval-relation queries. Intervals are stored sorted by start endpoint, with stable first-in-first-out (FIFO) ordering for ties.

All operations are persistent, and return modified copies. Peeking returns the stored element (which may be any type), popping returns a list with `$value` containing the stored element, `$start` and `$end` the interval endpoints, and `$remaining` the rest of the index without that element.

```{r}
ix <- interval_index(
  "A", "B", "C",
  start = c(1, 2, 4),
  end = c(3, 4, 5)
)
ix
```

The `as_interval_index()` variant builds an index from a vector or list of elements paired with start/end vectors, useful when endpoints are already in separate vectors.

```{r}
ix2 <- as_interval_index(c("phase1", "phase2", "phase3"), start = c(1, 3, 2), end = c(4, 5, 6))
ix2
```

`as_flexseq(ix)` returns a payload-only `flexseq`; interval bounds and any custom monoids are dropped (see `as.list` to convert and preserve interval metadata).


## Point and interval queries

`peek_point()` returns the *first* element (in canonical interval order — smallest start, with insertion/FIFO order breaking ties among equal starts) whose interval contains a query point; `peek_all_point()` returns all matches as an `interval_index` slice.

```{r pointops}
peek_point(ix, point = 2)
peek_all_point(ix, point = 2)
```

`pop_point()` removes and returns the first match as `$value`/`$start`/`$end`/`$remaining`. Its "all" counterpart `pop_all_point()` returns `$elements` (an interval_index of removed matches) and `$remaining`.

```{r}
res <- pop_point(ix, 2)
res$value
res$start
res$end
res$remaining
```

By default the four `*_point()` functions test interval containment. The `match_at` parameter lets them match instead on coordinate equality at the entry's start, end, or either endpoint. This is useful for sweep-line patterns where you need to identify intervals arriving or retiring at a specific event point. `bounds` is ignored for these modes (coordinate equality is structural).

```{r}
# Entries whose start coordinate equals 2
peek_all_point(ix, point = 2, match_at = "start")

# Entries whose end coordinate equals 3 — the sweep-line retirement query
peek_all_point(ix, point = 3, match_at = "end")

# Either endpoint equals 3
peek_all_point(ix, point = 3, match_at = "either")
```

Interval-vs-interval queries select overlaps of different kinds. First we'll build a new index to query:

```{r}
ix3 <- interval_index(
  "narrow", "wide", "right", "inner",
  start = c(2, 1, 4, 3),
  end   = c(3, 6, 5, 4)
)
ix3
```

`peek_overlaps()` returns any interval that shares at least one point with the query range. `peek_containing()` returns intervals that fully contain the query range. `peek_within()` returns intervals fully contained by the query range.

```{r}
# overlaps [2, 5]: any shared point
peek_all_overlaps(ix3, start = 2, end = 5)

# containing [3, 4]: index interval must enclose query
peek_all_containing(ix3, start = 3, end = 4)

# within [1, 5]: index interval must fit inside query
peek_all_within(ix3, start = 1, end = 5)
```

All relation queries have symmetric `pop_*` and `pop_all_*` variants following the same return contract as `pop_point()` and `pop_all_point()`.


## Boundary modes

Endpoint inclusion is controlled by two related parameters. The constructors take `default_query_bounds` — one of `"[)"` (default, left-closed right-open), `"[]"` (closed), `"()"` (open), or `"(]"` (left-open right-closed) — which sets the convention used when queries on this index don't specify their own. Query functions (`peek_*` / `pop_*`) then accept a `bounds` override for the duration of a single call.

```{r}
edge <- interval_index("E", start = 1, end = 3, default_query_bounds = "[)")

# right endpoint excluded by default
peek_point(edge, 3)

# override to closed bounds for this query
peek_point(edge, 3, bounds = "[]")
```

Note that bounds affect what counts as a valid interval: under `"[)"`, an interval with `start == end` like [2, 2) is empty and will never match any query. Use `"[]"` if you need point intervals where `start == end`.

```{r}
# [2, 2) contains nothing under half-open bounds
pt_open <- interval_index("X", start = 2, end = 2, default_query_bounds = "[)")
peek_point(pt_open, point = 2)

# [2, 2] contains exactly the point 2
pt_closed <- interval_index("X", start = 2, end = 2, default_query_bounds = "[]")
peek_point(pt_closed, point = 2)
```


## Insertion and endpoint extrema

`insert()` adds an element preserving interval start order.

```{r}
ix4 <- insert(ix, "D", start = 2, end = 6)
ix4
```

`min_endpoint()` and `max_endpoint()` return the current smallest start and largest end endpoints (not the stored elements).

```{r}
min_endpoint(ix)
max_endpoint(ix)
min_endpoint(interval_index())
```


## Empty indices

Query and endpoint helpers are non-throwing on empty indices, and `length()` allows for empty checking.

```{r}
empty_ix <- interval_index()
length(empty_ix)
peek_point(empty_ix, point = 1)
pop_overlaps(empty_ix, start = 1, end = 5)
```


## Named interval indices

Interval indices can carry names, set either at construction or through `as_interval_index()` on a named list. Names support read-only indexing via `[`, `[[`, and `$`; all replacement forms (`[<-`, `[[<-`, `$<-`) error, because index-based assignment may break the ordering invariant. Named and unnamed elements cannot be mixed within one index.

```{r}
ix_named <- as_interval_index(
  setNames(list("alice", "bob", "carol"), c("a", "b", "c")),
  start = c(1, 3, 2),
  end = c(4, 5, 6)
)
ix_named

ix_named[["b"]]
ix_named[c("a", "c")]
ix_named[1]

try(ix_named$a <- "!!")
```


## Transforming, iterating, merging

`fapply()` maps a function over elements while preserving intervals and order. The function receives `(value, start, end)`, or `(value, start, end, name)` if it accepts a fourth argument. Endpoints and names are passed in read-only, and the return value replaces the stored element (similar to `lapply()` for named R lists).

```{r}
fapply(ix, function(value, start, end) paste0(value, "[", start, ",", end, "]"))
```

`loop()` (re-exported from the **coro** package) walks the index in interval start-position order, yielding bare values. Interval endpoints are dropped from each yield; use `fapply()` if your callback needs `(value, start, end)`.

```{r}
loop(for (v in ix) print(v))
```

Plain `for (v in ix)` (without `loop()`) does *not* dispatch to the iteration protocol, it walks the underlying structure and yields internal nodes rather than elements. Always wrap with `loop()`.

`merge(x, y)` combines two interval indices into a new one in start-position order, preserving left-biased FIFO on tied starts. The general case runs in O(m + n); disjoint start ranges collapse to O(log(min(m, n))) via concat. The reserved `.ivx_min_end` / `.ivx_max_end` monoids recompute automatically, so interval-relation queries work immediately on the result.

```{r}
a <- as_interval_index(c("A1", "A2"), start = c(1, 5), end = c(4, 8))
b <- as_interval_index(c("B1", "B2"), start = c(3, 7), end = c(6, 10))
m <- merge(a, b)
peek_all_point(m, 3)
```

Both indices must share the same endpoint type, the same `bounds` convention, and the same monoid set; mismatches error. Both inputs are left unmodified.

---
title: "flexseq"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{flexseq}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
if(requireNamespace("pkgload", quietly = TRUE)) {
  pkgload::load_all(".", quiet = TRUE)
} else if(requireNamespace("Immutables", quietly = TRUE)) {
  library(Immutables)
} else {
  stop("Need either installed 'Immutables' or the 'pkgload' package to render this vignette.")
}
```


## `flexseq` basics

`flexseq` is a persistent sequence type, with many of the same features as an R `list()`, including storing arbitrary items.
All updates return a new object and keep the original unchanged.

```{r}
x <- flexseq(1, 2, 3)
x

x2 <- as_flexseq(letters[1:5])
x2
```
## Names and indexing

Flexseqs may be named, but if any are named they must all have unique non-`NULL` names. Large named flexseqs are somewhat slower than unnamed ones; for efficient key/value stores consider using an `ordered_sequence()`.

```{r}
x <- flexseq(a = 1, b = 2, c = 3)
x
x[c("c", "b")]

x$b <- NULL # delete b
x
```

In general indexing and slicing works as it does with lists: `[]` returns a sublist indexed by integer (by position), character vector (by name if named), or logical vector (by inclusion). Note that these operations are $O(k\log n)$ where $k$ is the length of the indexing vector.

```{r}
x <- as_flexseq(letters[1:6])

x[[3]]

x[c(1, 3, 5)]
```

## End operations

Flexseqs support pushing new elements onto the front or back of the sequences, peeking at the front or back (i.e., getting a copy of the first or last item), and popping from the front or back, which returns the removed element and a copy of the flexseq input with that item removed. `pop_front()` and `pop_back()` both return a list with fields `$value` and `$remaining`; `peek_back()` is the symmetric counterpart to `peek_front()`.

```{r}
x <- as_flexseq(4:6)

x <- x |>
  push_back(100) |>
  push_front(50)

x

xpopped <- pop_front(x)

xpopped$value
xpopped$remaining
```
## Positional operations

Beyond the ends, flexseqs support arbitrary-position persistent operations. `peek_at()` is the non-throwing positional read (compare with `[[`, which errors when out of bounds); `pop_at()` removes an element at a given index and returns `$value`/`$remaining`; `insert_at()` inserts before a given index, or at `length(x) + 1` to append.

```{r}
x <- flexseq("a", "b", "c", "d")

peek_at(x, 10)  # NULL, no error

out <- pop_at(x, 2)
out$value
out$remaining

insert_at(x, 3, c("x", "y"))
```

## Transforming and combining

`c()` concatenates flexseqs, and `fapply()` applies a function to each element of one returning the result as a flexseq, analygous to `lapply()` for R lists.
```{r}
x <- as_flexseq(4:6)
x2 <- as_flexseq(8:10)

c(x, x2)

x <- as_flexseq(1:3)
fapply(x, function(el) el * 10)
```

`merge(x, y)` is provided as an alias for `c(x, y)` so that `merge()` works uniformly across the package's four structure types. For `flexseq`, `c()` and `merge()` are equivalent; the distinction matters for `ordered_sequence`, `interval_index`, and `priority_queue`, where `merge()` performs a proper sorted/priority-aware combine.

`loop()` (re-exported from the **coro** package) enables `for`-loop traversal, yielding elements left-to-right lazily without materializing a list.

```{r}
x <- flexseq("a", "b", "c", "d")

loop(for (el in x) {
  print(el)
})
```

For named flexseqs, `loop()` yields bare values (matching `peek_front()`); use `as.list()` when you need names alongside values.

```{r}
x <- flexseq(a = 1, b = 2, c = 3)
loop(for (el in x) print(el))
```

Plain `for (el in x)` (without `loop()`) does *not* dispatch to the iteration protocol — it walks the underlying finger-tree storage and yields internal nodes rather than sequence elements. Always wrap with `loop()`.

Flexseqs convert back to standard R structures with `as.list()` (preserves names and list semantics) and `unlist()` (atomic vector when possible). `length()` returns the element count in O(1), and `str()` gives a compact diagnostic display.

```{r}
x <- flexseq(a = 1, b = 2, c = 3)

as.list(x)
unlist(x)
length(x)
```

---
title: "Ordered Sequences"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Ordered Sequences}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
if(requireNamespace("pkgload", quietly = TRUE)) {
  pkgload::load_all(".", quiet = TRUE)
} else if(requireNamespace("Immutables", quietly = TRUE)) {
  library(Immutables)
} else {
  stop("Need either installed 'Immutables' or the 'pkgload' package to render this vignette.")
}
```

## Ordered sequence basics

`ordered_sequence` is a persistent structure that associates elements with keys, typically numeric or character (string). They store elements sorted by key, with stable first-in-first-out (FIFO) behavior for duplicate keys, and provide fast key-based lookup, range queries over keys, and insertion.

All operations are persistent, and return modified copies. Peeking by key returns the stored element (which may be any type), popping returns a list with `$value` containing the stored element, `$key` the key, and `$remaining` the rest of the sequence without that element.

```{r}
xs <- ordered_sequence("a1", "b1", "b2", "c1", keys = c(1, 2, 2, 3))
xs
```

The `as_ordered_sequence()` variant builds a sequence from a vector or list of elements paired with a key vector, useful when keys are already in a separate vector.

```{r}
xs2 <- as_ordered_sequence(c(3, 1, 2, 1), keys = letters[1:4])
xs2
```

`as_flexseq(xs)` returns a payload-only `flexseq`; keys and any custom monoids are dropped (see `as.list` to convert and 
preservekey metadata).


## Peek, pop, insert, and range query by key

Insertion is by key, and *keys may be duplicated*. 

```{r}
seq <- as_ordered_sequence(1:3, keys = letters[1:3])
seq2 <- insert(seq, 10, key = "b")
seq2
```

`pop_key()` removes and returns the first match for a key as `$value`/`$key`/`$remaining`. Its "all" counterpart `pop_all_key()` removes the entire tie run for that key, returning `$elements` (an ordered sequence of removed matches) and `$remaining`.

```{r}
one <- pop_key(seq, key = "b")
one$value
one$key
one$remaining

all_two <- pop_all_key(seq, "b")
all_two$elements
all_two$remaining
```

Keys may be counted, and elements accessed; `peek_key()` returns one element, the
first in insertion order. `peek_all_key()` returns an ordered sequence with all
matching keys.

```{r}
count_key(seq, key = "b")
peek_key(seq, key = "b")
peek_all_key(seq, key = "b")
```

Counting and accessing elements over query ranges support inclusive/exclusive on both sides (defaulting to `TRUE`).

```{r}
elements_between(seq, from_key = "b", to_key ="c", include_from = TRUE, include_to = TRUE)
count_between(seq, from_key = "b", to_key = "c", include_from = TRUE, include_to = TRUE)

# exclude "b" keys
elements_between(seq, from_key = "b", to_key ="c", include_from = FALSE, include_to = TRUE)
```


## Key boundaries, and extrema

`lower_bound()` finds the first element with key `>=` a query key; `upper_bound()` finds the first with key strictly `>`. Both return a list with `$found`, `$index`, `$value`, and `$key` (`NULL` fields when no match exists). Together they support successor queries ("find the nearest entry at or above this key") and duplicate counting via index arithmetic.

```{r}
seq <- as_ordered_sequence(1:4, keys = c("b", "d", "d", "f"))

lower_bound(seq, key = "d") |> str()
upper_bound(seq, key = "d") |> str()
```

When the query key falls between or outside existing keys, `lower_bound()` returns the next entry at or above, useful for nearest-match lookups. Both return `found = FALSE` when no keys satisfy the condition.

```{r}
lower_bound(seq, key = "a") |> str()

upper_bound(seq, key = "g") |> str()
```

The difference `upper_bound(x, k)$index - lower_bound(x, k)$index` gives the count of entries with key `k` (this is what `count_key()` does internally).

```{r}
upper_bound(seq, key = "d")$index - lower_bound(seq, key = "d")$index
count_key(seq, key = "d")
```


`min_key()` and `max_key()` return the current minimum and maximum *keys* (not the stored elements).

```{r}
min_key(xs)
max_key(xs)
min_key(ordered_sequence())  # NULL when empty
```

## Empty sequences

Key and range helpers are non-throwing on empty sequences, and `length()` allows for empty checking.

```{r}
empty_os <- ordered_sequence()
length(empty_os)
peek_key(empty_os, 1)
count_between(empty_os, 1, 5)
```

## Named ordered sequences

Ordered sequences can carry names, set either at construction or through `as_ordered_sequence()` on a named list. Names and integer positions both support read-only indexing via `[`, `[[`, and `$`. All replacement forms (`[<-`, `[[<-`, `$<-`) error, because
index-based assignment may break the ordering invariant. Ordered sequences may be cast down with `as_flexseq()` or `as.list()`. Named and unnamed elements cannot be mixed within one sequence.

```{r}
xs_named <- as_ordered_sequence(
  setNames(list("alice", "bob", "carol"), c("a", "b", "c")),
  keys = c(3, 1, 2)
)
xs_named

xs_named[["b"]]
xs_named[c("a", "c")]
xs_named[1]            # positional read also works

try(xs_named$a <- "!!")  # replacement blocked
```

## Transforming, aterating, merging

`fapply()` maps a function over elements while preserving keys and order. The function receives `(value, key)`, or `(value, key, name)` if it accepts a third argument. Keys and names are passed in read-only, and the return value replaces the stored element at the associated key (and name if named).

```{r}
xs_t <- ordered_sequence("alice", "bob", "carol", keys = c(3, 1, 2))
fapply(xs_t, function(value, key) toupper(value))
```

`loop()` (re-exported from the **coro** package) walks the sequence in key-ascending order, yielding bare values. Keys are dropped from each yield; use `fapply()` if your callback needs the key alongside the value, or iterate over `as.list()` when you want a list keyed by name.

```{r}
loop(for (v in xs_t) print(v))
```

Plain `for (v in xs_t)` (without `loop()`) does *not* dispatch to the iteration protocol, it walks the underlying internal structure and yields those rather than sequence elements. Always wrap with `loop()`.

`merge(x, y)` combines two ordered sequences into a new one in key order, preserving left-biased FIFO on duplicate keys (all of `x`'s entries at a tied key precede `y`'s). The general case runs in O(m + n); when the key ranges are disjoint it collapses to O(log(min(m, n))) via concat.

```{r}
a <- as_ordered_sequence(c("a1", "a2", "a3"), keys = c(1, 3, 5))
b <- as_ordered_sequence(c("b1", "b2", "b3"), keys = c(2, 3, 6))
merge(a, b)
```

Both sequences must share the same key type and monoid set; mismatches error. Both inputs are left unmodified.

---
title: "Benchmarking immutables Collections"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Benchmarking immutables Collections}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

loaded <- tryCatch({
  pkgload::load_all(quiet = TRUE)
  TRUE
}, error = function(e) FALSE)

if(!loaded) {
  library(Immutables)
}

library(microbenchmark)
library(bench)
library(ggplot2)
library(dplyr)

# Two env-var gates control this vignette:
#   IMMUTABLES_RUN_SLOW=true       -> re-run all benchmarks and save per-batch
#                                     results to inst/extdata/benchmarks-*.rds
# When neither is set (CRAN, ordinary pkgdown build) the vignette loads the
# cached results and renders plots without re-timing.
#
# Results are cached per batch (one .rds per section) so a single section can
# be re-timed in isolation; each bench chunk writes its own .rds; each plot chunk 
# writes its own .pdf.
run_benchmarks <- identical(Sys.getenv("IMMUTABLES_RUN_SLOW"), "true")
repeats <- 10L

proj_root <- tryCatch(
  rprojroot::find_root(rprojroot::is_r_package),
  error = function(e) "."
)

pow2_labels <- function(x) parse(text = paste0("2^", log2(as.numeric(x))))

# scales::label_timespan() prints microseconds with a Greek mu (U+03BC), which
# the default pdf() device can't render (it drops the glyph -> ".s"). Swap it
# for the micro sign (U+00B5), which lives in the device's ISO-Latin-1
# encoding and renders as a proper "us".
label_time <- function(x) sub("μ", "µ", scales::label_timespan()(x))

# Accumulator populated either by the benchmark chunks (when
# run_benchmarks) or by the loader chunk below (otherwise). One named slot
# per batch: sequence / pq / ordered / ivx.
results_list <- list()

bench_one <- function(rows, impl, op, n, repeats, setup, bench) {
  state <- setup()
  gc(full = TRUE)

  # warmup
  for(i in 1:3) bench(state)

  res <- bench::mark(bench(state),
                     min_iterations = repeats,
                     max_iterations = repeats + 0)

  gcs <- res$gc[[1]] # tibble w/ 3 colums; level0, level1, level2
  gcs$times <- unlist(res$time[[1]])
  # keeping gc-triggering measures rather than bumping them out
  #non_gc_times <- gcs |>
  #  filter(level0 == 0 & level1 == 0 & level2 == 0) |>
  #  pull(times)
  gc_times <- gcs |> pull(times)

  push_back(rows, data.frame(impl = impl, op = op, n = n, time_s = gc_times))
}

# Per-batch cache helpers. Each batch is one data.frame persisted to its own
# file so a single section can be re-benched without touching the others.
.batch_path <- function(batch) {
  file.path(proj_root, "inst", "extdata", paste0("benchmarks-", batch, ".rds"))
}
save_batch <- function(batch) {
  dir.create(file.path(proj_root, "inst", "extdata"),
             showWarnings = FALSE, recursive = TRUE)
  saveRDS(results_list[[batch]], .batch_path(batch))
}
save_figure <- function(plot, file, width, height) {
    out_dir <- file.path(proj_root, "paper", "figures")
    dir.create(out_dir, showWarnings = FALSE, recursive = TRUE)
    ggsave(file.path(out_dir, file), plot, width = width, height = height)
}
```

```{r load-cached, eval=!run_benchmarks, include=FALSE}
for(batch in c("sequence", "pq", "ordered", "ivx")) {
  p <- system.file("extdata", paste0("benchmarks-", batch, ".rds"),
                   package = "Immutables")
  if(nzchar(p) && file.exists(p)) results_list[[batch]] <- readRDS(p)
}
```

## Methods

This article measures individual operation times for each collection type
provided by `Immutables`. Each recorded repetition rebuilds the setup for
size *n*, then times one operation on that fresh state with `bench::mark()`.

**NOTE**: This script takes >48 hours to run in total as a result of the 
large number of tests and repetitions, run serially with full garbage collection 
between each. The numbers shown below are loaded from cached results shipped with the 
package; re-running the cells in this document regenerates them, as does
executing the script version `data-raw/generate_publication_results.R`.

## Sequence operations


```{r sequence-bench, eval=run_benchmarks, cache=FALSE}
sequence_sizes <- 2^(12 + 0:6) # 2^12 up to 2^18
rows <- flexseq()

for(n in sequence_sizes) {
  cat("Sequence ops, size ", n, "\n")
  vals <- function() as.list(sprintf("v_%06d", seq_len(n)))
  mid <- as.integer(n / 2)
  flex_setup <- function() list(s = as_flexseq(vals()), mid = mid)
  list_setup <- function() list(s = vals(), mid = mid)
  pair_flex  <- function() list(a = as_flexseq(vals()), b = as_flexseq(vals()))
  pair_list  <- function() list(a = vals(), b = vals())

  rows <- bench_one(rows, "flexseq", "enqueue", n, repeats, flex_setup,
    function(st) push_back(st$s, "d"))
  rows <- bench_one(rows, "flexseq", "dequeue", n, repeats, flex_setup,
    function(st) pop_front(st$s)$remaining)
  rows <- bench_one(rows, "flexseq", "replace middle", n, repeats, flex_setup,
    function(st) { s <- st$s; s[[st$mid]] <- "y"; s })
  rows <- bench_one(rows, "flexseq", "remove middle", n, repeats, flex_setup,
    function(st) pop_at(st$s, st$mid)$remaining)
  rows <- bench_one(rows, "flexseq", "concatenate", n, repeats, pair_flex,
    function(st) c(st$a, st$b))
  rows <- bench_one(rows, "flexseq", "split at middle", n, repeats, flex_setup,
    function(st) split_at(st$s, st$mid))

  rows <- bench_one(rows, "base R", "enqueue", n, repeats, list_setup,
    function(st) c(st$s, list("d")))
  rows <- bench_one(rows, "base R", "dequeue", n, repeats, list_setup,
    function(st) st$s[-1L])
  rows <- bench_one(rows, "base R", "replace middle", n, repeats, list_setup,
    function(st) { s <- st$s; s[[st$mid]] <- "y"; s })
  rows <- bench_one(rows, "base R", "remove middle", n, repeats, list_setup,
    function(st) st$s[-st$mid])
  rows <- bench_one(rows, "base R", "concatenate", n, repeats, pair_list,
    function(st) c(st$a, st$b))
  rows <- bench_one(rows, "base R", "split at middle", n, repeats, list_setup,
    function(st) list(
      left = st$s[seq_len(st$mid - 1L)],
      value = st$s[[st$mid]],
      right = st$s[(st$mid + 1L):n]
    ))
}

results_list$sequence <- do.call(rbind, as.list(rows))
save_batch("sequence")
```


```{r sequence-plot, fig.width=9, fig.height=5}
if(!is.null(results_list$sequence)) {
  seq_results <- results_list$sequence
  sorted_sizes <- sort(unique(seq_results$n))
  seq_results$n_cat <- factor(seq_results$n, levels = sorted_sizes)

  p_sequence <- ggplot(seq_results, aes(x = n_cat, y = as.numeric(time_s), color = impl)) +
    geom_boxplot() +
    facet_wrap(~ op, scales = "free_y") +
    scale_x_discrete(labels = pow2_labels) +
    labs(
      title = "Sequence Operations",
      x = "Number of elements",
      y = "Time",
      color = "Implementation"
    ) +
    scale_y_log10(labels = label_time, guide = "axis_logticks") +
    scale_color_manual(values = c("base R" = "#fc8d62", "flexseq" = "#66c2a5")) +
    #scale_y_continuous(labels = label_time, trans = "log10") +
    theme_bw() +
    theme(plot.title = element_text(hjust = 0.5), legend.position = "bottom")
  print(p_sequence)
  save_figure(p_sequence, "benchmarks-sequence.pdf", width = 9, height = 5)
} else {
  knitr::asis_output("*Benchmark results not yet generated. Run `data-raw/generate_publication_results.R` to populate.*")
}
```



## Priority queue operations



```{r pq-bench, eval=run_benchmarks, cache=FALSE}
pq_sizes <- 2^(12 + 0:6)
rows <- flexseq()

set.seed(42)
max_pq <- max(pq_sizes)
all_pq_vals <- sprintf("val_%06d", seq_len(max_pq))
all_pq_pri  <- runif(max_pq)

for(n in pq_sizes) {
  cat("Priority queue ops, size ", n, "\n")
  pv <- as.list(all_pq_vals[seq_len(n)])
  pw <- all_pq_pri[seq_len(n)]
  pq_setup   <- function() list(pq = as_priority_queue(pv, priorities = pw))
  # Values in a list (arbitrary payloads, as the priority_queue stores);
  # priorities in a numeric vector.
  base_setup <- function() list(v = as.list(all_pq_vals[seq_len(n)]), p = pw)

  rows <- bench_one(rows, "priority_queue", "insert",   n, repeats, pq_setup,
    function(st) insert(st$pq, "val_new", 0.5))
  rows <- bench_one(rows, "priority_queue", "peek max", n, repeats, pq_setup,
    function(st) peek_max(st$pq))
  rows <- bench_one(rows, "priority_queue", "pop max",  n, repeats, pq_setup,
    function(st) pop_max(st$pq)$remaining)

  rows <- bench_one(rows, "base R", "insert",   n, repeats, base_setup,
    function(st) list(values = c(st$v, list("val_new")), priorities = c(st$p, 0.5)))
  rows <- bench_one(rows, "base R", "peek max", n, repeats, base_setup,
    function(st) st$v[[which.max(st$p)]])
  rows <- bench_one(rows, "base R", "pop max",  n, repeats, base_setup,
    function(st) { i <- which.max(st$p); list(values = st$v[-i], priorities = st$p[-i]) })
}

results_list$pq <- do.call(rbind, as.list(rows))
save_batch("pq")
```


```{r pq-plot, fig.width=9, fig.height=3.2}
if(!is.null(results_list$pq)) {
  pq_results <- results_list$pq
  sorted_sizes <- sort(unique(pq_results$n))
  pq_results$n_cat <- factor(pq_results$n, levels = sorted_sizes)

  p_pq <- ggplot(pq_results, aes(x = n_cat, y = as.numeric(time_s), color = impl)) +
    geom_boxplot() +
    facet_wrap(~ op, scales = "free_y") +
    scale_x_discrete(labels = pow2_labels) +
    labs(
      title = "Priority Queue Operations",
      x = "Number of elements",
      y = "Time",
      color = "Implementation"
    ) +
    theme_bw() +
    scale_color_manual(values = c("base R" = "#fc8d62", "priority_queue" = "#66c2a5")) +
    scale_y_log10(labels = label_time, guide = "axis_logticks") +
    #scale_y_continuous(labels = label_time) +
    theme(plot.title = element_text(hjust = 0.5), legend.position = "bottom")
  print(p_pq)
  save_figure(p_pq, "benchmarks-pq.pdf", width = 9, height = 3.2)
} else {
  knitr::asis_output("*Benchmark results not yet generated.*")
}
```




## Ordered sequence operations

```{r ordered-bench, eval=run_benchmarks, cache=FALSE}
ord_sizes <- 2^(14 + 0:6) # crossover is at 2^23 for peek_key
rows <- flexseq()

set.seed(99)
max_ord <- max(ord_sizes)
all_ord_keys <- sample.int(max_ord, max_ord, replace = TRUE) # duplicates present
all_ord_vals <- sprintf("e_%06d", seq_len(max_ord))

for(n in ord_sizes) {
  cat("Ordered sequence ops, size ", n, "\n")
  keys <- all_ord_keys[seq_len(n)]
  vals <- all_ord_vals[seq_len(n)]
  sk   <- sort(keys)

  # Query key and range bounds derive from the size-n subset so they always
  # land inside the key range; the range window spans a fixed number of keys
  # so result-set size k stays roughly constant (~50) as n grows.
  qk    <- sk[n %/% 2L]
  rlo   <- sk[n %/% 2L]
  rhi   <- sk[min(n, n %/% 2L + 50L)]
  ins_k <- qk

  ord_setup  <- function() list(os = as_ordered_sequence(as.list(vals), keys = keys))
  # Naive base R ordered map: a numeric key vector and a parallel value list,
  # sorted by key. All queries are plain linear scans over the real key vector
  # (no binary search, no key encoding), so the baseline is O(n) per op and
  # key-type-agnostic. Values live in a list, as the ordered_sequence stores
  # arbitrary element payloads.
  base_setup <- function() { o <- order(keys); list(k = keys[o], v = as.list(vals[o])) }

  rows <- bench_one(rows, "ordered_sequence", "insert", n, repeats, ord_setup,
    function(st) insert(st$os, "e_new", ins_k))
  rows <- bench_one(rows, "ordered_sequence", "peek key", n, repeats, ord_setup,
    function(st) peek_key(st$os, qk))
  rows <- bench_one(rows, "ordered_sequence", "pop key", n, repeats, ord_setup,
    function(st) pop_key(st$os, qk)$remaining)
  rows <- bench_one(rows, "ordered_sequence", "count key", n, repeats, ord_setup,
    function(st) count_key(st$os, qk))
  rows <- bench_one(rows, "ordered_sequence", "range query", n, repeats, ord_setup,
    function(st) elements_between(st$os, rlo, rhi))
  rows <- bench_one(rows, "ordered_sequence", "lower bound", n, repeats, ord_setup,
    function(st) lower_bound(st$os, qk))

  rows <- bench_one(rows, "base R", "insert", n, repeats, base_setup,
    function(st) { p <- sum(st$k <= ins_k)
      list(k = append(st$k, ins_k, p), v = append(st$v, list("e_new"), after = p)) })
  rows <- bench_one(rows, "base R", "peek key", n, repeats, base_setup,
    function(st) st$v[[match(qk, st$k)]])
  rows <- bench_one(rows, "base R", "pop key", n, repeats, base_setup,
    function(st) { i <- match(qk, st$k); list(k = st$k[-i], v = st$v[-i]) })
  rows <- bench_one(rows, "base R", "count key", n, repeats, base_setup,
    function(st) sum(st$k == qk))
  rows <- bench_one(rows, "base R", "range query", n, repeats, base_setup,
    function(st) st$v[st$k >= rlo & st$k <= rhi])
  rows <- bench_one(rows, "base R", "lower bound", n, repeats, base_setup,
    function(st) which.max(st$k >= qk))
}

results_list$ordered <- do.call(rbind, as.list(rows))
save_batch("ordered")
```


```{r ordered-plot, fig.width=9, fig.height=5}
if(!is.null(results_list$ordered)) {
  ord_results <- results_list$ordered
  sorted_sizes <- sort(unique(ord_results$n))
  ord_results$n_cat <- factor(ord_results$n, levels = sorted_sizes)

  p_ordered <- ggplot(ord_results, aes(x = n_cat, y = as.numeric(time_s), color = impl)) +
    geom_boxplot() +
    facet_wrap(~ op, scales = "free_y") +
    scale_x_discrete(labels = pow2_labels) +
    labs(
      title = "Ordered Sequence Operations",
      x = "Number of elements",
      y = "Time",
      color = "Implementation"
    ) +
    theme_bw() +
    #scale_y_continuous(labels = label_time) +
    scale_color_manual(values = c("base R" = "#fc8d62", "ordered_sequence" = "#66c2a5")) +
    scale_y_log10(labels = label_time, guide = "axis_logticks") +
    theme(plot.title = element_text(hjust = 0.5), legend.position = "bottom")
  print(p_ordered)
  save_figure(p_ordered, "benchmarks-ordered.pdf", width = 9, height = 5)
} else {
  knitr::asis_output("*Benchmark results not yet generated.*")
}
```




## Interval index operations

```{r ivx-bench, eval=run_benchmarks, cache=FALSE}
ivx_sizes <- 2^(12 + 0:6)
rows <- flexseq()

set.seed(123)
max_ivx <- max(ivx_sizes)
all_starts <- sort(sample.int(max_ivx * 10L, max_ivx))
all_widths <- sample.int(100L, max_ivx, replace = TRUE)
all_ends   <- all_starts + all_widths
all_vals   <- sprintf("interval_%06d", seq_len(max_ivx))

has_iranges <- requireNamespace("IRanges", quietly = TRUE) &&
  requireNamespace("S4Vectors", quietly = TRUE)

for(n in ivx_sizes) {
  cat("Interval ops, size ", n, "\n")
  starts <- all_starts[seq_len(n)]
  ends   <- all_ends[seq_len(n)]
  vals   <- all_vals[seq_len(n)]

  # Query and insertion points are derived from the size-n subset so they fall
  # inside the indexed coordinate range at every size, and the overlap window
  # spans a fixed number of starts so result-set size k stays constant
  # (~50 matches) as n grows. (Fixed global points computed from the largest
  # size would fall beyond the data for smaller n, timing only the
  # empty-result path.)
  qpt   <- starts[n %/% 2L] + 10L
  qlo   <- starts[n %/% 2L]
  qhi   <- starts[n %/% 2L + 50L]
  ins_s <- starts[n %/% 2L]
  ins_e <- ins_s + 50L

  ivx_setup  <- function() list(ix = as_interval_index(as.list(vals), start = starts, end = ends, default_query_bounds = "[]"))
  # Endpoints in a data.frame (the natural interval table); payloads in a
  # parallel list, as the interval_index stores arbitrary value objects.
  df_setup   <- function() list(df = data.frame(start = starts, end = ends), value = as.list(vals))

  rows <- bench_one(rows, "interval_index", "insert", n, repeats, ivx_setup,
    function(st) insert(st$ix, "interval_new", ins_s, ins_e))
  rows <- bench_one(rows, "interval_index", "point query", n, repeats, ivx_setup,
    function(st) peek_point(st$ix, qpt, bounds = "[]"))
  rows <- bench_one(rows, "interval_index", "all point matches", n, repeats, ivx_setup,
    function(st) peek_all_point(st$ix, qpt, bounds = "[]", as_list = TRUE))
  rows <- bench_one(rows, "interval_index", "overlap query", n, repeats, ivx_setup,
    function(st) peek_all_overlaps(st$ix, qlo, qhi, bounds = "[]", as_list = TRUE))
  rows <- bench_one(rows, "interval_index", "within query", n, repeats, ivx_setup,
    function(st) peek_all_within(st$ix, qlo, qhi, bounds = "[]", as_list = TRUE))
  rows <- bench_one(rows, "interval_index", "remove by overlap", n, repeats, ivx_setup,
    function(st) pop_all_overlaps(st$ix, qlo, qhi, bounds = "[]")$remaining)

  rows <- bench_one(rows, "base R", "insert", n, repeats, df_setup,
    function(st) list(
      df = rbind(st$df, data.frame(start = ins_s, end = ins_e)),
      value = c(st$value, list("interval_new"))
    ))
  rows <- bench_one(rows, "base R", "point query", n, repeats, df_setup,
    function(st) {
      hits <- which(st$df$start <= qpt & qpt <= st$df$end)
      if(length(hits)) st$value[[hits[1L]]] else NULL
    })
  rows <- bench_one(rows, "base R", "all point matches", n, repeats, df_setup,
    function(st) st$value[st$df$start <= qpt & qpt <= st$df$end])
  rows <- bench_one(rows, "base R", "overlap query", n, repeats, df_setup,
    function(st) st$value[st$df$start <= qhi & st$df$end >= qlo])
  rows <- bench_one(rows, "base R", "within query", n, repeats, df_setup,
    function(st) st$value[st$df$start >= qlo & st$df$end <= qhi])
  rows <- bench_one(rows, "base R", "remove by overlap", n, repeats, df_setup,
    function(st) { keep <- !(st$df$start <= qhi & st$df$end >= qlo)
      list(df = st$df[keep, , drop = FALSE], value = st$value[keep]) })

  if(has_iranges) {
    ir_setup <- function() list(
      ir = IRanges::IRanges(start = starts, end = ends),
      v  = as.list(vals)
    )

    rows <- bench_one(rows, "IRanges", "insert", n, repeats, ir_setup,
      function(st) list(
        ir = c(st$ir, IRanges::IRanges(start = ins_s, end = ins_e)),
        v  = c(st$v, list("interval_new"))
      ))
    rows <- bench_one(rows, "IRanges", "point query", n, repeats, ir_setup,
      function(st) {
        hits <- S4Vectors::subjectHits(IRanges::findOverlaps(IRanges::IRanges(start = qpt, width = 1L), st$ir))
        if(length(hits)) st$v[[hits[1L]]] else NULL
      })
    rows <- bench_one(rows, "IRanges", "all point matches", n, repeats, ir_setup,
      function(st) {
        st$v[S4Vectors::subjectHits(IRanges::findOverlaps(IRanges::IRanges(start = qpt, width = 1L), st$ir))]
      })
    rows <- bench_one(rows, "IRanges", "overlap query", n, repeats, ir_setup,
      function(st) {
        st$v[S4Vectors::subjectHits(IRanges::findOverlaps(IRanges::IRanges(start = qlo, end = qhi), st$ir))]
      })
    rows <- bench_one(rows, "IRanges", "within query", n, repeats, ir_setup,
      function(st) {
        st$v[S4Vectors::queryHits(IRanges::findOverlaps(st$ir, IRanges::IRanges(start = qlo, end = qhi), type = "within"))]
      })
    rows <- bench_one(rows, "IRanges", "remove by overlap", n, repeats, ir_setup,
      function(st) {
        hits <- S4Vectors::subjectHits(IRanges::findOverlaps(IRanges::IRanges(start = qlo, end = qhi), st$ir))
        keep <- setdiff(seq_along(st$v), hits)
        list(ir = st$ir[keep], v = st$v[keep])
      })
  }
}

results_list$ivx <- do.call(rbind, as.list(rows))
save_batch("ivx")
```


```{r ivx-plot, fig.width=9, fig.height=5}
if(!is.null(results_list$ivx)) {
  ivx_results <- results_list$ivx
  sorted_sizes <- sort(unique(ivx_results$n))
  ivx_results$n_cat <- factor(ivx_results$n, levels = sorted_sizes)
  ivx_results$impl <- factor(ivx_results$impl, levels = c("base R", "IRanges", "interval_index"))

  p_ivx <- ggplot(ivx_results, aes(x = n_cat, y = as.numeric(time_s), color = impl)) +
    geom_boxplot(position = position_dodge()) +
    facet_wrap(~ op, scales = "free_y") +
    scale_x_discrete(labels = pow2_labels) +
    labs(
      title = "Interval Index Queries",
      x = "Number of elements",
      y = "Time",
      color = "Implementation"
    ) +
    theme_bw() +
    scale_color_manual(values = c("base R" = "#fc8d62", "IRanges" = "#8da0cb", "interval_index" = "#66c2a5")) +
    scale_y_log10(labels = label_time, guide = "axis_logticks") +
    #scale_y_continuous(labels = label_time) +
    theme(plot.title = element_text(hjust = 0.5), legend.position = "bottom")
  print(p_ivx)
  save_figure(p_ivx, "benchmarks-ivx.pdf", width = 9, height = 5)
} else {
  knitr::asis_output("*Benchmark results not yet generated.*")
}
```

---
title: "Immutables"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{immutables}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
if(requireNamespace("pkgload", quietly = TRUE)) {
  pkgload::load_all(".", quiet = TRUE)
} else if(requireNamespace("Immutables", quietly = TRUE)) {
  library(Immutables)
} else {
  stop("Need either installed 'Immutables' or the 'pkgload' package to render this vignette.")
}
```

## Overview

`immutables` provides four fast data structures:

[`flexseq()`](flexseq.html) - Provides list-like operations including indexed and named element
   access, push/pop/peek from either end for double-ended queue behavior,
   insertion, splitting, and concatenation.

[`priority_queue()`](priority-queues.html) - Associates items with priority values and provides
  min and max peek/pop by priority and fast insertion.

[`ordered_sequence()`](ordered-sequences.html) - Associates items with key values and keeps the elements in
  sorted order by key. These may be similarly be inserted/popped/peeked by key
  value as well as position. Keys may be duplicated, with first-in-first-out
  order within key groups.

[`interval_index()`](interval-indices.html) - Stores items associated with interval ranges, supporting
  point and interval overlap/containment/subsumption queries. Items are kept
  in interval-start-order.

A [developer API](developer-api.html) exposes the underlying monoid-annotated
finger tree primitives (custom monoids, predicate-based locate/split,
validation helpers) for building new structures and indexes.


## Quick Tour

### flexseq

Construct, then push or pop at either end:

```{r}
s <- flexseq("a", "b", "c", "d")
s |> push_front("z") |> push_back("e")
pop_front(s)$remaining
```

Random-access and positional edits:

```{r}
s[[3]]
peek_back(s)
insert_at(s, 3, c("x", "y"))
```

Map, concatenate, and iterate:

```{r}
fapply(s, toupper)
c(s, as_flexseq(letters[10:12]))
loop(for (el in s) print(el))
```

### priority_queue

Construct with priorities, optionally with names, and insert more entries:

```{r}
q <- priority_queue(a = "task_a", b = "task_b", c = "task_c", priorities = c(3, 1, 2)) |>
  insert("task_d", priority = 1, name = "d")
```

Peek/pop by min or max priority; the `*_all_*` variants handle ties:

```{r}
peek_min(q)
pop_max(q)$value
pop_all_min(q)$elements
```

Read extrema and merge two queues into one:

```{r}
min_priority(q); max_priority(q)
merge(q, priority_queue("task_e", priorities = 0))
```

### ordered_sequence

Construct sorted by key (duplicates allowed, ties FIFO):

```{r}
xs <- ordered_sequence("a1", "b1", "b2", "c1", keys = c(1, 2, 2, 3))
peek_key(xs, 2)
peek_all_key(xs, 2)
```

Range queries and successor lookup:

```{r}
count_between(xs, 1, 2)
elements_between(xs, 2, 3, include_to = FALSE)
lower_bound(xs, 2)$index
```

Extrema and merge:

```{r}
min_key(xs); max_key(xs)
merge(xs, ordered_sequence("d1", keys = 4))
```

### interval_index

Construct with start/end coordinates; entries stay in start order:

```{r}
ix <- interval_index("A", "B", "C", start = c(1, 2, 4), end = c(3, 4, 5))
peek_point(ix, 2)
```

Interval-relation queries (overlap, containment), plus sweep-line endpoint matches via `match_at`:

```{r}
peek_all_overlaps(ix, start = 2, end = 5)
peek_all_containing(ix, start = 2, end = 3)
peek_all_point(ix, point = 3, match_at = "end")
```

Endpoint extrema and merge:

```{r}
min_endpoint(ix); max_endpoint(ix)
merge(ix, interval_index("D", start = 6, end = 8))
```

### developer API

Define a custom monoid (here, a running sum) and attach it to any structure:

```{r}
sum_monoid <- measure_monoid(f = `+`, i = 0, measure = function(el) el)
x <- as_flexseq(c(3, 1, 4, 1, 5, 9, 2, 6)) |>
  add_monoids(list(sum = sum_monoid))
get_measure(x, "sum")
```

Use a monotonic predicate over the cached measure for $O(\log n)$ locate and split:

```{r}
locate_by_predicate(x, function(v) v > 10, "sum")
split_around_by_predicate(x, function(v) v > 10, "sum")$value
validate_tree(x)
```

# print.flexseq handles empty, single, and small named trees

    Code
      print(flexseq())
    Output
      Unnamed flexseq with 0 elements.

---

    Code
      print(as_flexseq(1))
    Output
      Unnamed flexseq with 1 element.
      
      Elements:
      
      [[1]]
      [1] 1
      

---

    Code
      print(as_flexseq(setNames(as.list(1:4), letters[1:4])))
    Output
      Named flexseq with 4 elements.
      
      Elements:
      
      $a
      [1] 1
      
      $b
      [1] 2
      
      $c
      [1] 3
      
      $d
      [1] 4
      

# print.flexseq truncates long unnamed trees with head/tail preview

    Code
      print(as_flexseq(1:12), max_elements = 4)
    Output
      Unnamed flexseq with 12 elements.
      
      Elements:
      
      [[1]]
      [1] 1
      
      [[2]]
      [1] 2
      
      ... (skipping 8 elements)
      
      [[11]]
      [1] 11
      
      [[12]]
      [1] 12
      

---

    Code
      print(as_flexseq(1:12), max_elements = 0)
    Output
      Unnamed flexseq with 12 elements.

# print.flexseq shows custom monoid aggregates when requested

    Code
      print(x, max_elements = 2, show_custom_monoids = TRUE)
    Output
      Unnamed flexseq with 5 elements.
      Custom monoids + measures:
        sum: 15 (aggregate)
      
      Elements:
      
      [[1]]
      [1] 1
        sum measure: 1
      
      ... (skipping 3 elements)
      
      [[5]]
      [1] 5
        sum measure: 5
      

---

    Code
      print(as_flexseq(1:3), max_elements = 0, show_custom_monoids = TRUE)
    Output
      Unnamed flexseq with 3 elements.
      Custom monoids + measures: <none>

# print.priority_queue handles empty, named, unnamed, and truncated queues

    Code
      print(priority_queue())
    Output
      Unnamed priority_queue with 0 elements.

---

    Code
      print(priority_queue(one = 1, two = 2, three = 3, priorities = c(20, 30, 10)))
    Output
      Named priority_queue with 3 elements.
      Minimum priority: 10, Maximum priority: 30
      
      Elements (by priority):
      
      $three (priority 10)
      [1] 3
      
      $one (priority 20)
      [1] 1
      
      $two (priority 30)
      [1] 2
      

---

    Code
      print(priority_queue(1, 2, 3, priorities = c(2, 1, 3)))
    Output
      Unnamed priority_queue with 3 elements.
      Minimum priority: 1, Maximum priority: 3
      
      Elements (by priority):
      
      (priority 1)
      [1] 2
      
      (priority 2)
      [1] 1
      
      (priority 3)
      [1] 3
      

---

    Code
      print(big_q, max_elements = 4)
    Output
      Unnamed priority_queue with 10 elements.
      Minimum priority: 1, Maximum priority: 10
      
      Elements (by priority):
      
      (priority 1)
      [1] "b"
      
      (priority 2)
      [1] "e"
      
      ... (skipping 6 elements)
      
      (priority 9)
      [1] "f"
      
      (priority 10)
      [1] "j"
      

# print.priority_queue shows custom monoid aggregates

    Code
      print(q, max_elements = 0, show_custom_monoids = TRUE)
    Output
      Unnamed priority_queue with 3 elements.
      Minimum priority: 1, Maximum priority: 3
      Custom monoids + measures:
        sum_item: 6 (aggregate)

# print.ordered_sequence handles empty, named, unnamed, and truncated sequences

    Code
      print(ordered_sequence())
    Output
      Unnamed ordered_sequence with 0 elements.

---

    Code
      print(ordered_sequence(one = "a", two = "b", three = "c", keys = c(20, 30, 10)))
    Output
      Named ordered_sequence with 3 elements.
      
      Elements (by key order):
      
      $three (key 10)
      [1] "c"
      
      $one (key 20)
      [1] "a"
      
      $two (key 30)
      [1] "b"
      

---

    Code
      print(ordered_sequence("x", "y", "z", keys = c(2, 1, 3)))
    Output
      Unnamed ordered_sequence with 3 elements.
      
      Elements (by key order):
      
      [[1]] (key 1)
      [1] "y"
      
      [[2]] (key 2)
      [1] "x"
      
      [[3]] (key 3)
      [1] "z"
      

---

    Code
      print(as_ordered_sequence(letters[1:10], keys = 10:1), max_elements = 4)
    Output
      Unnamed ordered_sequence with 10 elements.
      
      Elements (by key order):
      
      [[1]] (key 1)
      [1] "j"
      
      [[2]] (key 2)
      [1] "i"
      
      ... (skipping 6 elements)
      
      [[9]] (key 9)
      [1] "b"
      
      [[10]] (key 10)
      [1] "a"
      

# print.ordered_sequence shows custom monoid aggregates

    Code
      print(ys, max_elements = 0, show_custom_monoids = TRUE)
    Output
      Unnamed ordered_sequence with 2 elements.
      Custom monoids + measures:
        sum_key: 3 (aggregate)

# print.interval_index handles empty, named, unnamed, and truncated indices

    Code
      print(interval_index())
    Output
      Unnamed interval_index with 0 elements, default query bounds [start, end).

---

    Code
      print(interval_index(one = 1, two = 2, three = 3, start = c(20, 30, 10), end = c(
        25, 37, 24)))
    Output
      Named interval_index with 3 elements, default query bounds [start, end).
      
      Elements (by interval start order):
      
      $three (interval 10 - 24)
      [1] 3
      
      $one (interval 20 - 25)
      [1] 1
      
      $two (interval 30 - 37)
      [1] 2
      

---

    Code
      print(interval_index(1, 2, 3, start = c(2, 4, 6), end = c(3, 5, 8),
      default_query_bounds = "[]"))
    Output
      Unnamed interval_index with 3 elements, default query bounds [start, end].
      
      Elements (by interval start order):
      
      [[1]] (interval 2 - 3)
      [1] 1
      
      [[2]] (interval 4 - 5)
      [1] 2
      
      [[3]] (interval 6 - 8)
      [1] 3
      

---

    Code
      print(big_ix, max_elements = 4)
    Output
      Unnamed interval_index with 10 elements, default query bounds [start, end).
      
      Elements (by interval start order):
      
      [[1]] (interval 1 - 2)
      [1] "a"
      
      [[2]] (interval 2 - 3)
      [1] "b"
      
      ... (skipping 6 elements)
      
      [[9]] (interval 9 - 10)
      [1] "i"
      
      [[10]] (interval 10 - 11)
      [1] "j"
      

# print.interval_index shows custom monoid aggregates

    Code
      print(ix, max_elements = 0, show_custom_monoids = TRUE)
    Output
      Unnamed interval_index with 2 elements, default query bounds [start, end).
      Custom monoids + measures:
        width_sum: 3 (aggregate)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-methods.R
\name{as.list.ordered_sequence}
\alias{as.list.ordered_sequence}
\title{Coerce Ordered Sequence to List}
\usage{
\method{as.list}{ordered_sequence}(x, ...)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{...}{Unused.}
}
\value{
A plain list of elements in key order.
}
\description{
Coerce Ordered Sequence to List
}
\details{
Returns payload elements only (keys are omitted) in canonical key order.
If entries are named, names are preserved on the returned list.
}
\examples{
x <- ordered_sequence("a", "b", "c", keys = c(2, 1, 3))
as.list(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-ends.R
\name{peek_at}
\alias{peek_at}
\title{Peek at an Element by Position}
\usage{
peek_at(x, index)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{index}{One-based position to read.}
}
\value{
Element at \code{index}, or \code{NULL} when \code{index} is out of bounds.
}
\description{
Returns the element at a one-based index without modifying the sequence.
}
\details{
Positive integer indices beyond \code{length(x)} return \code{NULL}.
Invalid indices (\code{NA}, non-integer, \verb{<= 0}, or length not equal to 1) error.
}
\examples{
x <- flexseq("a", "b", "c")
peek_at(x, 2)
peek_at(x, 10)

try(peek_at(x, 0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-constructors.R
\name{as_interval_index}
\alias{as_interval_index}
\title{Build an Interval Index from \code{x}, \code{start}, and \code{end}}
\usage{
as_interval_index(x, start, end, default_query_bounds = "[)")
}
\arguments{
\item{x}{Elements to add.}

\item{start}{Start endpoints with the same length as \code{x}.}

\item{end}{End endpoints with the same length as \code{x}.}

\item{default_query_bounds}{Boundary convention used as the default for
query operations on this index: one of \code{"[)"}, \code{"[]"}, \code{"()"}, \code{"(]"}.
Per-query \verb{peek_*} / \verb{pop_*} calls may override via their own \code{bounds}
argument.}
}
\value{
An \code{interval_index}.
}
\description{
Constructs an \code{interval_index} by pairing each element of \code{x} with
corresponding \code{start} and \code{end} endpoints.
}
\details{
Output is ordered by interval \code{start}.

Names on \code{x} are preserved as element names.
}
\examples{
ix <- as_interval_index(c("a", "b", "c"), start = c(1, 2, 2), end = c(3, 2, 4))
ix
as.list(peek_all_point(ix, 2))

# Endpoints can be other comparable types
ix_date <- as_interval_index(
  c("phase1", "phase2"),
  start = as.Date(c("2024-01-01", "2024-01-10")),
  end = as.Date(c("2024-01-05", "2024-01-15"))
)
ix_date
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-queue-ops.R
\name{peek_all_min}
\alias{peek_all_min}
\title{Peek All Minimum-Priority Elements}
\usage{
peek_all_min(x)
}
\arguments{
\item{x}{A \code{priority_queue}.}
}
\value{
A \code{priority_queue} containing all minimum-priority elements in stable
queue order. Returns an empty queue when \code{x} is empty.
}
\description{
Returns the full minimum-priority tie run as a \code{priority_queue}.
}
\details{
The return is another \code{priority_queue()}, use \code{as.list()} to convert
the result to a standard R list.
}
\examples{
x <- priority_queue("a", "b", "c", priorities = c(2, 1, 1))
peek_all_min(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-methods.R
\name{length.ordered_sequence}
\alias{length.ordered_sequence}
\title{Ordered Sequence Length}
\usage{
\method{length}{ordered_sequence}(x)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}
}
\value{
Integer length.
}
\description{
Ordered Sequence Length
}
\details{
Uses cached size metadata and runs in O(1).
}
\examples{
x <- ordered_sequence("a", "b", keys = c(2, 1))
length(x)

length(ordered_sequence())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/20-api-generics.R
\name{split_by_predicate}
\alias{split_by_predicate}
\title{Split into Left and Right Parts}
\usage{
split_by_predicate(x, predicate, monoid_name)
}
\arguments{
\item{x}{A \code{flexseq} (or subclass).}

\item{predicate}{Function applied to accumulated monoid values.}

\item{monoid_name}{Name of the monoid used for scanning.}
}
\value{
A list with fields:
\itemize{
\item \code{left}: elements before the split point.
\item \code{right}: elements from the split point onward.
}
}
\description{
Splits at the first point where \code{predicate} becomes \code{TRUE} while scanning
the sequence.
}
\details{
This is the two-way variant of \code{\link[=split_around_by_predicate]{split_around_by_predicate()}}.

As with \code{\link[=split_around_by_predicate]{split_around_by_predicate()}}, a common setup is a custom monoid
created with \code{\link[=measure_monoid]{measure_monoid()}} and attached via \code{\link[=add_monoids]{add_monoids()}}.

\code{left} and \code{right} preserve subclass when the input is a subclass of
\code{flexseq}.
}
\examples{
x <- flexseq("a", "b", "c", "d")
size_monoid <- measure_monoid(`+`, 0L, function(e) 1L)
x2 <- add_monoids(x, list(size = size_monoid))
split_by_predicate(x2, function(v) v >= 3L, "size")

# Split at the first element
split_by_predicate(x2, function(v) v >= 1L, "size")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-iter.R
\name{as_iterator.priority_queue}
\alias{as_iterator.priority_queue}
\title{Iterate over a \code{priority_queue} (coro iterator)}
\usage{
\method{as_iterator}{priority_queue}(x)
}
\arguments{
\item{x}{A \code{priority_queue}.}
}
\value{
A \code{coro} iterator function.
}
\description{
Returns a lazy iterator that yields payload elements in priority-ascending
order. Use with \code{\link[=loop]{loop()}}:
}
\details{
\if{html}{\out{<div class="sourceCode">}}\preformatted{loop(for (x in pq) print(x))
}\if{html}{\out{</div>}}

Traversal is driven by repeated \code{\link[=pop_min]{pop_min()}}: each step is O(log n), so full
traversal is O(n log n). Ties within equal priorities are yielded in FIFO
insertion order (inherited from \code{\link[=pop_min]{pop_min()}}).

The original \code{x} is not modified; the iterator holds a private cursor and
partial iteration (e.g. via \code{break}) leaves the source intact.

Each yielded value is the bare payload (matching \code{\link[=peek_min]{peek_min()}}). Use
\code{\link[=fapply]{fapply()}} if your callback needs the priority alongside the value, or cast
with \code{\link[=as_flexseq]{as_flexseq()}} for insertion-order iteration.
}
\examples{
pq <- priority_queue("a", "b", "c", priorities = c(3, 1, 2))
loop(for (x in pq) print(x))  # "b", "c", "a"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-query-api.R
\name{peek_all_within}
\alias{peek_all_within}
\title{Peek All Intervals Within a Query Interval}
\usage{
peek_all_within(x, start, end, bounds = NULL)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{start}{Query interval start.}

\item{end}{Query interval end.}

\item{bounds}{Optional boundary override. One of \code{"[)"}, \code{"[]"}, \code{"()"}, \code{"(]"}.}
}
\value{
An \code{interval_index} slice of all matches (possibly empty).
}
\description{
Peek All Intervals Within a Query Interval
}
\details{
The returned \code{interval_index} can be inspected with \code{\link[=as.list]{as.list()}}.
}
\examples{
ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 3, 5))
as.list(peek_all_within(ix, 1, 4))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-iter.R
\name{as_iterator.flexseq}
\alias{as_iterator.flexseq}
\title{Iterate over a \code{flexseq} (coro iterator)}
\usage{
\method{as_iterator}{flexseq}(x)
}
\arguments{
\item{x}{A \code{flexseq}.}
}
\value{
A \code{coro} iterator function.
}
\description{
Returns a lazy iterator that yields payload elements left-to-right.
Use with \code{\link[=loop]{loop()}} as the canonical iteration form:
}
\details{
\if{html}{\out{<div class="sourceCode">}}\preformatted{loop(for (x in s) print(x))
}\if{html}{\out{</div>}}

Iteration uses repeated left-view (\code{viewL}) and is O(n) total, O(1) amortized
per step. The original \code{x} is not modified; the iterator holds a private
cursor over progressively-smaller tails.

For named sequences, internal name metadata is stripped from yielded values
to match \code{peek_front(s)} semantics. Access names via \code{\link[=as.list]{as.list()}} when needed.

Inherited by \code{ordered_sequence} and \code{interval_index}: for those subclasses
the yielded value is the unwrapped payload (keys / interval endpoints
dropped), in key-ascending / start-position order respectively. See
\code{\link[=as_iterator.priority_queue]{as_iterator.priority_queue()}} for the priority-order override.
}
\section{Do not use plain \code{for} directly}{
Writing \code{for (x in s) ...} (without \code{\link[=loop]{loop()}}) will not dispatch this method.
R's \code{for} walks the object's underlying list storage at the C level and
bypasses S3 \code{length}/\code{[[}, so it silently yields raw finger-tree internals
(Digit/Empty/Deep nodes) rather than sequence elements. Always wrap with
\code{\link[=loop]{loop()}}, or call \code{\link[=as.list]{as.list()}} first for an eager copy.
}

\examples{
s <- flexseq("a", "b", "c")
loop(for (x in s) print(x))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/20-api-generics.R
\name{split_around_by_predicate}
\alias{split_around_by_predicate}
\title{Split Around First Predicate Match}
\usage{
split_around_by_predicate(t, predicate, monoid_name, accumulator = NULL)
}
\arguments{
\item{t}{A \code{flexseq} (or subclass).}

\item{predicate}{Function applied to accumulated monoid values.}

\item{monoid_name}{Name of the monoid used for scanning.}

\item{accumulator}{Optional starting accumulator value.}
}
\value{
A list with fields:
\itemize{
\item \code{left}: elements before the split point.
\item \code{value}: the matched element at the split point.
\item \code{right}: elements after the split point.
}
}
\description{
Splits at the first point where a \code{predicate} function becomes \code{TRUE}
while scanning the sequence.
}
\details{
This function generally requires the sequence be annotated with
a \code{measure_monoid()}; see the examples and \code{measure_monoid()}
for more information.

\code{value} is the matched leaf entry for the input structure:
\itemize{
\item \code{flexseq}: stored user element.
\item \code{ordered_sequence}: \code{list(value, key)}.
\item \code{priority_queue}: \code{list(value, priority)}.
\item \code{interval_index}: \code{list(value, start, end)}.
}

\code{left} and \code{right} preserve subclass when the input is a subclass of
\code{flexseq}.
}
\examples{
x <- flexseq("a", "b", "c", "d")

# Each element e has measure 1; the accumulated measure for
# a set of elements is computed by the associative function `+`
# along with the identity value 0.
size_monoid <- measure_monoid(`+`, 0L, function(e) 1L)
x2 <- add_monoids(x, list(size = size_monoid))

# the first time the measure stored in the size monoid
# accumulates to greater than or equal to 3 is the 3rd
# element in sequence.
split_around_by_predicate(x2, function(v) v >= 3L, "size")

# Split at the first element
split_around_by_predicate(x2, function(v) v >= 1L, "size")

# Split at the last element
split_around_by_predicate(x2, function(v) v >= 4L, "size")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-measure-monoid.R
\name{measure_monoid}
\alias{measure_monoid}
\title{Construct a Measure Monoid Specification}
\usage{
measure_monoid(f, i, measure)
}
\arguments{
\item{f}{Associative binary function over measure values.}

\item{i}{Identity value for \code{f}.}

\item{measure}{Function mapping one element to its measure value.}
}
\value{
An object of class \code{measure_monoid}.
}
\description{
Defines how element-level values are measured and combined as accumulated
tree metadata.
}
\details{
A measure monoid has three parts:
\itemize{
\item \code{measure(entry)}: maps each stored leaf entry to a measure value.
\item \code{f(left, right)}: combines two measure values.
\item \code{i}: identity value for \code{f}.
}

Requirements:
\itemize{
\item \code{f} should be associative.
\item \code{i} should satisfy \code{f(i, x) == x} and \code{f(x, i) == x}.
\item \code{measure()} outputs must be compatible with \code{f} and \code{i}.
}

Developer APIs are leaf-entry oriented:
\itemize{
\item \code{flexseq}: entry is the stored user element.
\item \code{ordered_sequence}: entry is \code{list(value, key)}.
\item \code{priority_queue}: entry is \code{list(value, priority)}.
\item \code{interval_index}: entry is \code{list(value, start, end)}.
}

\code{measure_monoid()} only constructs the specification; it becomes active after
being attached to a structure via \code{\link[=add_monoids]{add_monoids()}}.
}
\examples{
sum_m <- measure_monoid(`+`, 0, as.numeric)
x <- as_flexseq(1:5)
x2 <- add_monoids(x, list(sum = sum_m))
attr(x2, "measures")$sum
split_around_by_predicate(x2, function(v) v >= 6, "sum")

# Count elements
count_m <- measure_monoid(`+`, 0L, function(el) 1L)
x3 <- add_monoids(x, list(count = count_m))
attr(x3, "measures")$count

# Character-width accumulation
width_m <- measure_monoid(`+`, 0L, function(el) nchar(as.character(el)))
s <- as_flexseq(c("aa", "b", "cccc"))
s2 <- add_monoids(s, list(width = width_m))
attr(s2, "measures")$width
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-concat-trees.R
\name{concat_trees}
\alias{concat_trees}
\title{Concatenate Two Structural Trees}
\usage{
concat_trees(x, y)
}
\arguments{
\item{x}{A \code{flexseq} (left side).}

\item{y}{A \code{flexseq} (right side).}
}
\value{
Concatenated tree.
}
\description{
Same-name monoids are assumed equivalent; left-tree definitions win.
Missing monoids are added to each side before concatenation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-methods.R
\name{merge.interval_index}
\alias{merge.interval_index}
\title{Merge Two Interval Indices}
\usage{
\method{merge}{interval_index}(x, y, ...)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{y}{An \code{interval_index}.}

\item{...}{Unused.}
}
\value{
A new \code{interval_index} of size \code{length(x) + length(y)}.
}
\description{
Returns a new \code{interval_index} containing every entry from both inputs,
preserving start-position order. On tied start positions, \code{x}'s entries
precede \code{y}'s (left-biased FIFO).
}
\details{
The merge runs in O(m + n) via a zipper-style traversal on interval
starts, with a fast path to O(log(min(m, n))) when the start ranges are
disjoint.

Both indices must share the same endpoint type and the same \code{bounds}
convention (e.g. \code{"[)"} half-open vs. \code{"[]"} closed), and the same
monoid set. Mismatches error.

The reserved monoids \code{.ivx_min_end} / \code{.ivx_max_end} recompute
automatically on the merged tree, so \code{min_endpoint()} / \code{max_endpoint()}
and interval-relation queries work immediately on the result.

Both inputs are left unmodified.
}
\examples{
a <- interval_index("A1", "A2", start = c(1, 5), end = c(4, 8))
b <- interval_index("B1", "B2", start = c(3, 7), end = c(6, 10))
m <- merge(a, b)
as.list(m)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-core.R
\name{c.flexseq}
\alias{c.flexseq}
\title{Concatenate Sequences}
\usage{
\method{c}{flexseq}(..., recursive = FALSE)
}
\arguments{
\item{...}{\code{flexseq} objects.}

\item{recursive}{Unused; must be \code{FALSE}.}
}
\value{
A concatenated \code{flexseq}.
}
\description{
Concatenate Sequences
}
\details{
\code{c()} is supported for \code{flexseq} and returns a new concatenated \code{flexseq}.

For \code{priority_queue}, \code{ordered_sequence}, and \code{interval_index}, \code{c()} is not
supported because concatenation can violate structure-specific invariants.
Cast first with \code{\link[=as_flexseq]{as_flexseq()}} when sequence-style concatenation is intended,
noting that this drops ordering and priority metadata.
}
\examples{
x <- flexseq("a", "b")
y <- flexseq("c", "d")
c(x, y)

q1 <- priority_queue("a", priorities = 2)
q2 <- priority_queue("b", priorities = 1)
try(c(q1, q2))
c(as_flexseq(q1), as_flexseq(q2))

o1 <- ordered_sequence("a", keys = 1)
o2 <- ordered_sequence("b", keys = 2)
try(c(o1, o2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/20-api-generics.R
\name{add_monoids}
\alias{add_monoids}
\title{Add or Merge Measure Monoids}
\usage{
add_monoids(t, monoids, overwrite = FALSE)
}
\arguments{
\item{t}{Immutable structure (\code{flexseq} and subclasses).}

\item{monoids}{Named list of \code{\link[=measure_monoid]{measure_monoid()}} objects.}

\item{overwrite}{Logical; if \code{TRUE}, replace existing monoids with the same
names. If \code{FALSE}, existing names are kept.}
}
\value{
A persistent copy with updated monoid definitions and cached measures.
}
\description{
Attaches one or more named \code{\link[=measure_monoid]{measure_monoid()}} definitions to an existing
immutable structure.
}
\details{
Mechanics:
\itemize{
\item Each monoid name defines an independent accumulated measure over elements in the tree.
\item New monoids are computed for all elements and cached in the returned object.
\item Existing monoids are unchanged unless \code{overwrite = TRUE}.
\item Structural/reserved monoid names cannot be replaced.
}

Measure-function signatures:
\itemize{
\item \code{flexseq}: \code{measure(entry)} where \code{entry} is the stored element.
\item \code{ordered_sequence}: \code{measure(entry)} where \code{entry} is \code{list(value, key)}.
\item \code{priority_queue}: \code{measure(entry)} where \code{entry} is \code{list(value, priority)}.
\item \code{interval_index}: \code{measure(entry)} where \code{entry} is \code{list(value, start, end)}.
}

This operation is persistent: \code{t} is not modified.

Use this when you want fast predicate scans (for example with
\code{\link[=locate_by_predicate]{locate_by_predicate()}}, \code{\link[=split_by_predicate]{split_by_predicate()}}, \code{\link[=split_around_by_predicate]{split_around_by_predicate()}})
driven by domain-specific accumulated values.
}
\examples{
x <- flexseq(10, 20, 30)

running_sum <- measure_monoid(`+`, 0, as.numeric)
x2 <- add_monoids(x, list(sum = running_sum))
attr(x2, "measures")$sum

# Use the monoid in a split query
split_around_by_predicate(x2, function(v) v >= 30, "sum")

# Overwrite an existing monoid definition
running_count <- measure_monoid(`+`, 0L, function(e) 1L)
x3 <- add_monoids(x2, list(sum = running_count), overwrite = TRUE)
attr(x3, "measures")$sum
}
\seealso{
\code{\link[=measure_monoid]{measure_monoid()}}, \code{\link[=get_measure]{get_measure()}}, \code{\link[=get_measures]{get_measures()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-queue-ops.R
\name{pop_all_max}
\alias{pop_all_max}
\title{Pop All Maximum-Priority Elements}
\usage{
pop_all_max(x)
}
\arguments{
\item{x}{A \code{priority_queue}.}
}
\value{
A list with fields:
\itemize{
\item \code{elements}: \code{priority_queue} of removed maximum-priority elements.
\item \code{remaining}: queue after removal.
}
}
\description{
Removes the full maximum-priority tie run.
}
\details{
The return \code{elements} is another \code{priority_queue()}, use \code{as.list()} to
convert the result to a standard R list.
}
\examples{
x <- priority_queue("a", "b", "c", priorities = c(2, 3, 3))
out <- pop_all_max(x)
out$elements
out$remaining
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{count_between}
\alias{count_between}
\title{Count Elements in a Key Range}
\usage{
count_between(x, from_key, to_key, include_from = TRUE, include_to = TRUE)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{from_key}{Lower bound key.}

\item{to_key}{Upper bound key.}

\item{include_from}{Include lower bound when \code{TRUE}.}

\item{include_to}{Include upper bound when \code{TRUE}.}
}
\value{
Integer count of matches.
}
\description{
Count Elements in a Key Range
}
\details{
Uses the same range semantics as \code{\link[=elements_between]{elements_between()}} but returns only the
count:
\itemize{
\item \code{include_from = TRUE} uses \code{key >= from_key}; otherwise \code{key > from_key}.
\item \code{include_to = TRUE} uses \code{key <= to_key}; otherwise \code{key < to_key}.
}
}
\examples{
x <- ordered_sequence("a", "b", "c", "d", keys = c(1, 2, 2, 3))
count_between(x, 2, 3)
count_between(x, 2, 2, include_to = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-methods.R
\name{length.priority_queue}
\alias{length.priority_queue}
\title{Priority Queue Length}
\usage{
\method{length}{priority_queue}(x)
}
\arguments{
\item{x}{A \code{priority_queue}.}
}
\value{
Integer length.
}
\description{
Priority Queue Length
}
\details{
Uses cached size metadata and runs in O(1).
}
\examples{
q <- priority_queue("a", "b", priorities = c(2, 1))
length(q)

length(priority_queue())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-queue-ops.R
\name{merge.priority_queue}
\alias{merge.priority_queue}
\title{Merge Two Priority Queues}
\usage{
\method{merge}{priority_queue}(x, y, ...)
}
\arguments{
\item{x}{A \code{priority_queue}.}

\item{y}{A \code{priority_queue}.}

\item{...}{Unused.}
}
\value{
A new \code{priority_queue} of size \code{length(x) + length(y)}.
}
\description{
Returns a new \code{priority_queue} containing every entry from both inputs,
preserving each queue's internal insertion order (entries of \code{x} come
first, then entries of \code{y}).
}
\details{
The cached \code{.pq_min} / \code{.pq_max} monoids recompute automatically on the
merged tree, so \code{peek_min()} / \code{peek_max()} reflect the combined extremum
immediately.

Both queues must share the same priority type and the same monoid set;
mismatches error rather than being silently harmonized. Merging an
empty queue with a non-empty queue returns the non-empty queue unchanged.

Both inputs are left unmodified.
}
\examples{
a <- priority_queue("x", "y", priorities = c(5, 1))
b <- priority_queue("z", priorities = 3)
m <- merge(a, b)
peek_min(m)
length(m)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-print.R
\name{print.priority_queue}
\alias{print.priority_queue}
\title{Print a Priority Queue Summary}
\usage{
\method{print}{priority_queue}(x, max_elements = 4L, show_custom_monoids = FALSE, ...)
}
\arguments{
\item{x}{A \code{priority_queue}.}

\item{max_elements}{Maximum number of elements shown in the preview.}

\item{show_custom_monoids}{Logical; show attached non-default monoids and
their root cached measures.}

\item{...}{Passed through to per-element \code{print()}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Prints a compact summary with priority range and a head/tail preview.
}
\examples{
q <- priority_queue(one = 1, two = 2, three = 3, priorities = c(20, 30, 10))
print(q, max_elements = 4)
sum_item <- measure_monoid(`+`, 0, function(entry) as.numeric(entry$value))
q3 <- add_monoids(priority_queue(1, 2, priorities = c(2, 1)), list(sum_item = sum_item))
print(q3, max_elements = 0, show_custom_monoids = TRUE)

q2 <- priority_queue(1, 2, 3, priorities = c(2, 1, 3))
print(q2, max_elements = 3)

print(priority_queue())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-constructors.R
\name{priority_queue}
\alias{priority_queue}
\title{Construct a Priority Queue}
\usage{
priority_queue(..., priorities)
}
\arguments{
\item{...}{Elements to enqueue.}

\item{priorities}{Priorities with the same length as \code{...}.}
}
\value{
A \code{priority_queue}.
}
\description{
Creates a \code{priority_queue} from elements in \code{...} and matching
\code{priorities}.
}
\details{
Empty construction is supported: \code{priority_queue()} returns an empty queue.

If elements are named, names are preserved for name-based reads.

Queue operations are exposed through \code{insert()}, \verb{peek_*()}, \verb{pop_*()},
and \code{fapply()}.
}
\examples{
x <- priority_queue("a", "b", "c", priorities = c(2, 1, 2))
x
peek_min(x)

empty_q <- priority_queue()
peek_min(empty_q)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-ends.R
\name{pop_at}
\alias{pop_at}
\title{Pop an Element by Position}
\usage{
pop_at(x, index)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{index}{One-based position to remove.}
}
\value{
A list with fields:
\itemize{
\item \code{value}: the element at \code{index}, or \code{NULL} when \code{index} is out of bounds.
\item \code{remaining}: the sequence after removing the selected element.
}
}
\description{
Returns the selected element and the remaining sequence.
}
\details{
This operation is persistent: \code{x} is not modified.

Positive integer indices beyond \code{length(x)} return a non-throwing miss object
with \code{value = NULL} and \code{remaining = x}.
Invalid indices (\code{NA}, non-integer, \verb{<= 0}, or length not equal to 1) error.
}
\examples{
x <- flexseq("a", "b", "c", "d")
out <- pop_at(x, 3)
out$value
out$remaining
x  # unchanged

pop_at(x, 10)
try(pop_at(x, 0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/70-plotting.R
\name{plot_structure}
\alias{plot_structure}
\title{Plot the Internal Finger-Tree Structure}
\usage{
plot_structure(
  t1,
  vertex.size = 15,
  vertex.shape = "rounded_rect",
  edge.width = 1,
  label_edges = FALSE,
  title = NULL,
  node_label = "value",
  label.cex = NULL,
  asp = NA,
  legend = TRUE,
  ...
)
}
\arguments{
\item{t1}{A finger-tree-backed immutables structure.}

\item{vertex.size}{Passed to \code{igraph::plot.igraph}. Ignored by the default
\code{"rounded_rect"} shape, which sizes each box to its own label text.}

\item{vertex.shape}{Vertex shape. Defaults to \code{"rounded_rect"}, a custom
shape registered on first call that auto-sizes every box to fit its
label (so multi-line labels render cleanly without tuning
\code{vertex.size}). Any built-in igraph shape (\code{"circle"}, \code{"rectangle"},
\code{"none"}, ...) also works and is passed through unchanged.}

\item{edge.width}{Edge width passed to \code{igraph::plot.igraph}.}

\item{label_edges}{If \code{TRUE}, draw the child slot name on each edge
(\code{"prefix"}/\code{"middle"}/\code{"suffix"} for \code{Deep} nodes, or the numeric
position for \code{Digit}/\code{Node2}/\code{Node3}).}

\item{title}{Optional plot title.}

\item{node_label}{Either one of the preset modes \code{"value"} (default;
payload for elements, blank for structural), \code{"type"} (node class
name), \code{"both"} (type + value on separate lines), \code{"none"}, or a
user-supplied function \verb{function(node)} returning a single character
string per vertex. The \code{node} argument is a list with fields:
\describe{
\item{\code{id}}{Internal graph vertex id (string).}
\item{\code{type}}{Node class: \code{"Element"}, \code{"Digit"}, \code{"Deep"},
\code{"Node2"}, \code{"Node3"}, \code{"Single"}, or \code{"Empty"}.}
\item{\code{label}}{The default label string.}
\item{\code{measures}}{Named list of accumulated monoid values for the
subtree rooted at this node. For structural nodes these are the
cached values; for element leaves, each entry is the monoid's
\code{measure()} applied to the leaf entry. Keys include built-ins
(\code{.size}, \code{.named_count}, and any structure-specific ones like
\code{.pq_min}) plus any custom name added via \code{\link[=add_monoids]{add_monoids()}}.}
\item{\code{element}}{For element nodes, the raw leaf entry. Shape
depends on the structure type; see \code{\link[=measure_monoid]{measure_monoid()}} for the
entry contract. \code{NULL} for structural nodes.}
}
Measure values are exposed as-is, including list-valued measures
(e.g. the built-in \code{.pq_min} is \code{list(has, priority)}).}

\item{label.cex}{Numeric scalar controlling label text size (passed as
\code{vertex.label.cex} to igraph). If \code{NULL} (default), scales
automatically from ~1.0 on small trees down to ~0.55 on dense ones so
the auto-sized \code{"rounded_rect"} boxes don't collide. Override with a
fixed value for finer control.}

\item{asp}{Plot aspect ratio (physical y-unit / physical x-unit).
Default \code{NA} lets the tree fill the device without aspect constraint.
Set to a numeric (e.g. \code{0.4}) to enforce horizontal stretching,
or \code{1} for a square plot region.}

\item{legend}{If \code{TRUE} (default), draws a horizontal legend below the
tree mapping node-type names to their fill colors, restricted to
types actually present.}

\item{...}{Additional arguments passed to \code{igraph::plot.igraph} (e.g.
\code{vertex.label.color}, \code{vertex.frame.color}).}
}
\value{
Invoked for its side effect (draws to the active graphics
device). Returns \code{NULL} invisibly.
}
\description{
Developer-facing visualizer for the finger-tree backing any immutables
structure (\code{flexseq}, \code{ordered_sequence}, \code{priority_queue},
\code{interval_index}). The S3 \code{plot()} methods on those classes forward to
this function, but \code{plot_structure()} is also callable directly and gives
access to the full \code{node_label} API for custom label formatting. Requires
the \code{igraph} package (listed in \code{Suggests}).
}
\examples{
if (requireNamespace("igraph", quietly = TRUE)) {
\donttest{
t <- as_flexseq(letters[1:8])
plot_structure(t, title = "Finger tree")

# Label every node with its subtree size (leaves contribute 1).
plot_structure(as_flexseq(1:10), node_label = function(node) {
  paste0(node$type, "\n.size=", node$measures$.size)
})

# Custom monoid: subtree sum of numeric payloads. Structural nodes show
# the accumulated total; leaves show their own contribution.
sum_monoid <- measure_monoid(`+`, 0, function(el) el)
xs <- add_monoids(as_flexseq(c(3, 1, 4, 1, 5, 9, 2, 6)),
                  list(sum = sum_monoid))
plot_structure(xs, node_label = function(node) {
  if(node$type == "Element") sprintf("\%g\nsum=\%g", node$element, node$measures$sum)
  else sprintf("\%s\nsum=\%g", node$type, node$measures$sum)
})

# List-valued built-in measure: priority_queue's .pq_min tracks the min
# priority seen in a subtree as list(has, priority). Unpack in the label.
pq <- priority_queue("task-a", "task-b", "task-c",
                     priorities = c(5, 1, 3))
plot_structure(pq, node_label = function(node) {
  m <- node$measures$.pq_min
  if(node$type == "Element") {
    sprintf("\%s\np=\%g", node$element$value, node$element$priority)
  } else if(isTRUE(m$has)) {
    sprintf("\%s\nmin=\%g", node$type, m$priority)
  } else {
    node$type
  }
})
}
}
}
\seealso{
\code{\link[=measure_monoid]{measure_monoid()}}, \code{\link[=add_monoids]{add_monoids()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-get-measure.R
\name{get_measure}
\alias{get_measure}
\title{Read a Cached Subtree Measure}
\usage{
get_measure(x, monoid_name)
}
\arguments{
\item{x}{A flexseq (or any immutables subclass).}

\item{monoid_name}{Name of an attached monoid (e.g. \code{.size}, a custom
name from \code{\link[=add_monoids]{add_monoids()}}, or a built-in like \code{.pq_min} /
\code{.oms_max_key}).}
}
\value{
The cached monoid value at the root of \code{x}. Shape depends on
the monoid — typically atomic (e.g. a numeric sum) but may be a list
for built-ins that carry auxiliary state.
}
\description{
Returns the monoid's accumulated measure across the whole (sub)tree at
the root of \code{x}. This is the cached value that internal operations use
for O(log n) locate and split.
}
\details{
On the full tree \code{x}, this is the aggregate over every element. After a
split — e.g. \code{s <- split_by_predicate(x, p, "sum")} — call
\code{get_measure(s$left, "sum")} to read the aggregate of just the left
side in O(1).

For per-element measures, see \code{\link[=get_measures]{get_measures()}}.
}
\examples{
sum_m <- measure_monoid(`+`, 0, function(el) el)
x <- add_monoids(as_flexseq(c(3, 1, 4, 1, 5, 9, 2, 6)),
                 list(sum = sum_m))
get_measure(x, "sum")         # total across the whole tree
get_measure(x, ".size")       # built-in: element count

q <- priority_queue("a", "b", "c", priorities = c(5, 1, 3))
get_measure(q, ".pq_min")     # built-in, list-valued
}
\seealso{
\code{\link[=get_measures]{get_measures()}}, \code{\link[=measure_monoid]{measure_monoid()}}, \code{\link[=add_monoids]{add_monoids()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-tree-from.R
\name{tree_from}
\alias{tree_from}
\title{Build a Structural Tree from a Vector or List}
\usage{
tree_from(x, monoids = NULL)
}
\arguments{
\item{x}{Elements to insert.}

\item{monoids}{Optional named list of \code{measure_monoid} objects.}
}
\value{
A finger tree with cached measures for all monoids.
If \code{x} has names, they are used for name-based indexing and must be
complete (no missing/empty names) and unique.
}
\description{
Build a Structural Tree from a Vector or List
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-query-api.R
\name{peek_within}
\alias{peek_within}
\title{Peek First Interval Within a Query Interval}
\usage{
peek_within(x, start, end, bounds = NULL)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{start}{Query interval start.}

\item{end}{Query interval end.}

\item{bounds}{Optional boundary override. One of \code{"[)"}, \code{"[]"}, \code{"()"}, \code{"(]"}.}
}
\value{
The payload value from the first match, or \code{NULL} on no match.
}
\description{
Peek First Interval Within a Query Interval
}
\details{
Returns the first match in canonical interval order. Use \code{\link[=peek_all_within]{peek_all_within()}}
to retrieve all matches as an \code{interval_index} slice.
}
\examples{
ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 3, 5))
peek_within(ix, 1, 4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-indexing.R
\name{sub-.ordered_sequence}
\alias{sub-.ordered_sequence}
\alias{[.ordered_sequence}
\alias{[[.ordered_sequence}
\alias{[<-.ordered_sequence}
\alias{[[<-.ordered_sequence}
\alias{$.ordered_sequence}
\alias{$<-.ordered_sequence}
\title{Indexing for Ordered Sequences}
\usage{
\method{[}{ordered_sequence}(x, i, ...)

\method{[[}{ordered_sequence}(x, i, ...)

\method{[}{ordered_sequence}(x, i) <- value

\method{[[}{ordered_sequence}(x, i) <- value

\method{$}{ordered_sequence}(x, name)

\method{$}{ordered_sequence}(x, name) <- value
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{i}{Index input.}

\item{...}{Unused.}

\item{value}{Replacement value (unsupported).}

\item{name}{Element name (for \code{$} and \verb{$<-}).}
}
\value{
Read methods return ordered payload values/subsets; replacement forms
always error.
}
\description{
Read indexing treats vectors as selectors and returns subsets in key order.
Out-of-order selectors are canonicalized with a warning.
Replacement indexing is blocked to prevent order-breaking writes.
}
\details{
Vector selectors are treated as membership selectors, not output-order
instructions.
\itemize{
\item Integer/character vectors are normalized to unique positions and returned in
canonical sequence order.
\item Out-of-order selector vectors trigger a warning and are canonicalized.
\item Duplicate selectors are rejected.
\item Replacement indexing (\verb{[<-}, \verb{[[<-}, \verb{$<-}) is unsupported.
}
}
\examples{
x <- ordered_sequence(a = "A", b = "B", c = "C", keys = c(1, 2, 3))

x[c(3, 1)]          # warning; result returned in key order
x[c("c", "a")]      # warning; result returned in key order
x[c(TRUE, FALSE, TRUE)]
x[["b"]]
x$b

try(x[c(2, 2)])
try(x$b <- "updated")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-indexing.R
\name{sub-.interval_index}
\alias{sub-.interval_index}
\alias{[.interval_index}
\alias{[[.interval_index}
\alias{[<-.interval_index}
\alias{[[<-.interval_index}
\alias{$.interval_index}
\alias{$<-.interval_index}
\title{Indexing for Interval Indexes}
\usage{
\method{[}{interval_index}(x, i, ...)

\method{[[}{interval_index}(x, i, ...)

\method{[}{interval_index}(x, i) <- value

\method{[[}{interval_index}(x, i) <- value

\method{$}{interval_index}(x, name)

\method{$}{interval_index}(x, name) <- value
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{i}{Index input.}

\item{...}{Unused.}

\item{value}{Replacement value (unsupported).}

\item{name}{Element name (for \code{$} and \verb{$<-}).}
}
\value{
Read methods return interval payload values/subsets; replacement forms
always error.

For \code{$}: the matched payload element.

No return value; always errors because replacement indexing is unsupported.
}
\description{
Read indexing returns \code{interval_index} subsets while preserving interval/key
order; out-of-order selectors are canonicalized with a warning. Replacement
indexing is blocked.
}
\details{
Read indexing preserves canonical interval order in returned subsets.
\itemize{
\item Integer/character vectors are treated as selectors and canonicalized to
interval-order output.
\item Out-of-order selector vectors trigger a warning and are canonicalized.
\item Duplicate selectors are rejected.
\item \code{[[} and \code{$} return payload values.
\item Replacement indexing (\verb{[<-}, \verb{[[<-}, \verb{$<-}) is unsupported.
}
}
\examples{
ix <- interval_index(a = "A", b = "B", c = "C", start = c(1, 3, 5), end = c(2, 4, 6))

ix[c(3, 1)]         # warning; result returned in interval order
ix[c("c", "a")]     # warning; result returned in interval order
ix[c(TRUE, FALSE, TRUE)]
ix[["b"]]
ix$b

try(ix[c(2, 2)])
try(ix$b <- "updated")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-ends.R
\name{push_front}
\alias{push_front}
\title{Push an Element to the Front}
\usage{
push_front(x, value)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{value}{Element to prepend.}
}
\value{
Updated \code{flexseq}.
}
\description{
Returns a new sequence with \code{value} prepended at the left end.
}
\details{
This operation is persistent: \code{x} is not modified.

Elements can be named, but only if all are uniquely named (no missing
names).
}
\examples{
s <- as_flexseq(letters[2:4])
s2 <- push_front(s, "a")
s2
s  # unchanged

n <- as_flexseq(list(two = 2, three = 3))
new_el <- 1
names(new_el) <- "one"
push_front(n, new_el)

# Named/unnamed mixes are rejected
try(push_front(n, 0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-query-api.R
\name{pop_all_point}
\alias{pop_all_point}
\title{Pop All Intervals Matching a Point}
\usage{
pop_all_point(
  x,
  point,
  bounds = NULL,
  match_at = c("interval", "start", "end", "either")
)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{point}{Query point.}

\item{bounds}{Optional boundary override. One of \code{"[)"}, \code{"[]"}, \code{"()"}, \code{"(]"}.
Ignored when \code{match_at} is not \code{"interval"}.}

\item{match_at}{How the query point is matched against each entry. One of
\code{"interval"} (default; containment under \code{bounds}), \code{"start"}, \code{"end"},
or \code{"either"}. See \code{\link[=peek_point]{peek_point()}} for details. The \code{"end"} mode is the
standard primitive for retiring active intervals in a sweep-line.}
}
\value{
A list with \code{elements} and \code{remaining}, both \code{interval_index} objects.
}
\description{
Pop All Intervals Matching a Point
}
\details{
Use \code{\link[=as.list]{as.list()}} to convert \code{elements} to a standard R list.
}
\examples{
ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(3, 2, 5))
out <- pop_all_point(ix, 2)
as.list(out$elements)

# Sweep-line retirement: remove everything ending at x = 3
retired <- pop_all_point(ix, 3, match_at = "end")
as.list(retired$elements)
as.list(retired$remaining)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-print.R
\name{print.interval_index}
\alias{print.interval_index}
\title{Print an Interval Index Summary}
\usage{
\method{print}{interval_index}(x, max_elements = 4L, show_custom_monoids = FALSE, ...)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{max_elements}{Maximum number of elements shown in the preview.}

\item{show_custom_monoids}{Logical; show attached non-default monoids and
their root cached measures.}

\item{...}{Passed through to per-element \code{print()}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Prints a compact summary with interval bounds and a head/tail preview of
payload elements.
}
\examples{
ix <- interval_index(
  one = 1, two = 2, three = 3,
  start = c(20, 30, 10), end = c(25, 37, 24)
)
print(ix, max_elements = 4)
width_sum <- measure_monoid(
  `+`, 0, function(entry) as.numeric(entry$end - entry$start)
)
ix3 <- add_monoids(
  interval_index(1, 2, start = c(1, 3), end = c(2, 5)),
  list(width_sum = width_sum)
)
print(ix3, max_elements = 0, show_custom_monoids = TRUE)

ix2 <- interval_index(1, 2, 3, start = c(2, 4, 6), end = c(3, 5, 8), default_query_bounds = "[]")
print(ix2, max_elements = 3)

print(interval_index())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-constructors.R
\name{ordered_sequence}
\alias{ordered_sequence}
\title{Construct an Ordered Sequence}
\usage{
ordered_sequence(..., keys)
}
\arguments{
\item{...}{Elements to add.}

\item{keys}{Key values with the same length as \code{...}.}
}
\value{
An \code{ordered_sequence}.
}
\description{
Convenience constructor from \code{...} and matching \code{keys}.
}
\details{
Empty construction is supported: \code{ordered_sequence()} returns an empty
ordered sequence.

Output is always sorted by key, with stable order across duplicate keys.
}
\examples{
xs <- ordered_sequence("bb", "a", "ccc", keys = c(2, 1, 3))
xs
lower_bound(xs, 2)

num_by_chr <- ordered_sequence(20, 10, 30, keys = c("b", "a", "c"))
num_by_chr

ordered_sequence()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-methods.R
\name{as.list.interval_index}
\alias{as.list.interval_index}
\title{Coerce Interval Index to List}
\usage{
\method{as.list}{interval_index}(x, ...)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{...}{Unused.}
}
\value{
A plain list of payload elements in interval order.
}
\description{
Coerce Interval Index to List
}
\details{
This returns payload values only.
}
\examples{
ix <- interval_index("a", "b", "c", start = c(3, 1, 2), end = c(4, 2, 3))
as.list(ix)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{merge.ordered_sequence}
\alias{merge.ordered_sequence}
\title{Merge Two Ordered Sequences}
\usage{
\method{merge}{ordered_sequence}(x, y, ...)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{y}{An \code{ordered_sequence}.}

\item{...}{Unused.}
}
\value{
A new \code{ordered_sequence} of size \code{length(x) + length(y)}.
}
\description{
Returns a new \code{ordered_sequence} containing every entry from both inputs,
preserving key order. On duplicate keys, \code{x}'s entries precede \code{y}'s
(left-biased FIFO).
}
\details{
The merge runs in O(m + n) via a zipper-style traversal, with a fast path
to O(log(min(m, n))) when the key ranges are disjoint (all of \code{x}'s keys
\code{<=} all of \code{y}'s keys, or vice versa with a strict \code{<} to keep
left-biased FIFO intact on equal boundary keys).

Both sequences must share the same key type and the same monoid set;
mismatches error rather than being silently harmonized. Merging an
empty sequence with a non-empty sequence returns the non-empty one
unchanged.

Both inputs are left unmodified.
}
\examples{
a <- ordered_sequence("a1", "a2", "a3", keys = c(1, 3, 5))
b <- ordered_sequence("b1", "b2", "b3", keys = c(2, 3, 6))
m <- merge(a, b)
as.list(m)
# At the tied key 3, "a2" precedes "b2".
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/20-api-generics.R
\name{split_at}
\alias{split_at}
\title{Split at a Position or Name}
\usage{
split_at(x, at, pull_index = FALSE)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{at}{A single positive integer position or a single character name.}

\item{pull_index}{Controls output shape:
\itemize{
\item \code{FALSE} (default): returns \code{list(left, value, right)}.
\item \code{TRUE}: returns \code{list(left, right)}.
}}
}
\value{
A split result with shape controlled by \code{pull_index}.
}
\description{
Splits by a single one-based position or a single element name.
}
\details{
\code{split_at(x, at, pull_index = FALSE)} is a convenience wrapper around
\code{\link[=split_around_by_predicate]{split_around_by_predicate()}} using positional scanning.

\code{split_at(x, at, pull_index = TRUE)} is the two-way variant using
\code{\link[=split_by_predicate]{split_by_predicate()}}.
}
\examples{
x <- flexseq("a", "b", "c", "d")
split_at(x, 3)
split_at(x, 3, pull_index = TRUE)

n <- flexseq(a = 1, b = 2, c = 3)
split_at(n, "b")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/20-api-generics.R
\name{fapply}
\alias{fapply}
\title{Fapply with S3 dispatch}
\usage{
fapply(X, FUN, ...)
}
\arguments{
\item{X}{Object to apply over.}

\item{FUN}{Function to apply.}

\item{...}{Method-specific arguments.}
}
\value{
Method-dependent result.
}
\description{
\code{fapply()} is an S3 generic for applying functions over immutable
structures with type-specific dispatch.
}
\details{
\code{fapply()} returns a new object and preserves class semantics.

Method signatures:
\itemize{
\item \code{fapply.flexseq(X, FUN, ..., preserve_custom_monoids = TRUE)}
\item \code{fapply.priority_queue(X, FUN, ..., preserve_custom_monoids = TRUE)}
\item \code{fapply.ordered_sequence(X, FUN, ..., preserve_custom_monoids = TRUE)}
\item \code{fapply.interval_index(X, FUN, ..., preserve_custom_monoids = TRUE)}
}

For \code{priority_queue}, \code{ordered_sequence}, and \code{interval_index},
\code{FUN} receives structured fields (\code{value} plus metadata) and should return the
new payload value only; ordering metadata is preserved. The name argument is
optional: callbacks that only take value/metadata also work.

If supported by the method, \code{preserve_custom_monoids = TRUE} keeps added user
monoids; \code{FALSE} rebuilds with required structural monoids only.
}
\examples{
x <- flexseq("a", "b", "c")
fapply(x, toupper)

q <- priority_queue(one = "a", two = "b", priorities = c(2, 1))
fapply(q, function(value, priority, name) paste0(value, priority))

o <- ordered_sequence(one = "a", two = "b", keys = c(2, 1))
fapply(o, function(value, key, name) paste0(value, "_", key))

ix <- interval_index(one = "a", two = "b", start = c(1, 3), end = c(2, 4))
fapply(ix, function(value, start, end, name) paste0(value, "[", start, ",", end, "]"))
}
\seealso{
\code{\link[=flexseq]{flexseq()}}, \code{\link[=priority_queue]{priority_queue()}}, \code{\link[=ordered_sequence]{ordered_sequence()}}, \code{\link[=interval_index]{interval_index()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-queue-ops.R
\name{peek_min}
\alias{peek_min}
\title{Peek Minimum-Priority Element}
\usage{
peek_min(x)
}
\arguments{
\item{x}{A \code{priority_queue}.}
}
\value{
Element at minimum priority, or \code{NULL} when \code{x} is empty.
}
\description{
Returns the element at the minimum priority without modifying the queue.
}
\details{
Ties are stable: when multiple elements share minimum priority, this returns
the earliest element in queue order.
}
\examples{
x <- priority_queue("a", "b", "c", priorities = c(2, 1, 1))
peek_min(x)
peek_min(priority_queue())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/20-api-generics.R
\name{locate_by_predicate}
\alias{locate_by_predicate}
\title{Locate First Predicate Match}
\usage{
locate_by_predicate(
  t,
  predicate,
  monoid_name,
  accumulator = NULL,
  include_metadata = FALSE
)
}
\arguments{
\item{t}{A \code{flexseq} (or subclass).}

\item{predicate}{Function applied to accumulated monoid values.}

\item{monoid_name}{Name of the monoid used for scanning.}

\item{accumulator}{Optional starting accumulator value.}

\item{include_metadata}{Logical; include scan metadata.}
}
\value{
If \code{include_metadata = FALSE}, a list with:
\itemize{
\item \code{found}: logical flag.
\item \code{value}: matched element when found, otherwise \code{NULL}.
}

If \code{include_metadata = TRUE}, adds \code{metadata} with:
\itemize{
\item \code{left_measure}
\item \code{hit_measure}
\item \code{right_measure}
\item \code{index}
}
}
\description{
Scans accumulated monoid values and returns the first element where
\code{predicate} becomes \code{TRUE}, without rebuilding split trees.
}
\details{
This is the read-only analogue of \code{\link[=split_around_by_predicate]{split_around_by_predicate()}}.

As with split helpers, a common setup is a custom monoid created with
\code{\link[=measure_monoid]{measure_monoid()}} and attached via \code{\link[=add_monoids]{add_monoids()}}.

\code{value} is the matched leaf entry for the input structure:
\itemize{
\item \code{flexseq}: stored user element.
\item \code{ordered_sequence}: \code{list(value, key)}.
\item \code{priority_queue}: \code{list(value, priority)}.
\item \code{interval_index}: \code{list(value, start, end)}.
}
}
\examples{
x <- flexseq("a", "b", "c", "d")
size_monoid <- measure_monoid(`+`, 0L, function(e) 1L)
x2 <- add_monoids(x, list(size = size_monoid))

locate_by_predicate(x2, function(v) v >= 3L, "size")
locate_by_predicate(x2, function(v) v >= 3L, "size", include_metadata = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-query-api.R
\name{min_endpoint}
\alias{min_endpoint}
\title{Minimum Left Endpoint}
\usage{
min_endpoint(x)
}
\arguments{
\item{x}{An \code{interval_index}.}
}
\value{
Minimum left endpoint, or \code{NULL} when \code{x} is empty.
}
\description{
Returns the smallest left endpoint (\code{start}) currently present.
}
\details{
Because intervals are stored in start order, this reads the first entry's
\code{start}.
}
\examples{
ix <- interval_index("a", "b", start = c(3, 1), end = c(4, 2))
min_endpoint(ix)
min_endpoint(interval_index())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-dollar.R,
%   R/30-api-flexseq-indexing.R
\name{$.flexseq}
\alias{$.flexseq}
\alias{$<-.flexseq}
\alias{[.flexseq}
\alias{[[.flexseq}
\alias{[<-.flexseq}
\alias{[[<-.flexseq}
\title{Flexseq Indexing}
\usage{
\method{$}{flexseq}(x, name)

\method{$}{flexseq}(x, name) <- value

\method{[}{flexseq}(x, i, ...)

\method{[[}{flexseq}(x, i, ...)

\method{[}{flexseq}(x, i) <- value

\method{[[}{flexseq}(x, i) <- value
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{name}{Element name (for \code{$} and \verb{$<-}).}

\item{value}{Replacement values; recycled to selected index length.}

\item{i}{Positive integer indices, character element names, or logical mask.
For \code{[[}, a single integer or character name.}

\item{...}{Unused.}
}
\value{
For \code{$}: the matched element.

For \verb{$<-}: updated tree with one named element replaced.

For \code{[}: a new \code{flexseq} containing selected elements in query order.
For character indexing, missing names are represented as \code{NULL} elements.

For \code{[[}: the extracted element (internal name metadata is removed).

For \verb{[<-}: a new \code{flexseq} with selected elements replaced.

For \verb{[[<-}: a new \code{flexseq} with one element replaced.
}
\description{
Index, replace, and extract elements of a \code{flexseq} by position or name.
}
\details{
Selector behavior:
\itemize{
\item Integer indexing (\code{[}) returns elements in the requested order.
\item Character indexing (\code{[}) returns elements in requested name order; unknown
names become \code{NULL} elements in the output.
\item Logical indexing (\code{[}) is positional and follows logical-mask selection.
}

Replacement behavior:
\itemize{
\item \verb{[<-} is persistent and recycles \code{value} to the number of selected
positions (with standard R recycling warnings where applicable).
\item \verb{[[<-} replaces one element; assigning \code{NULL} removes that element.
}
}
\examples{

# $ extracts by name
x <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))
x$b

# $<- replaces by name
x <- as_flexseq(setNames(as.list(1:3), c("a", "b", "c")))
x$b <- 20
x$b
x <- as_flexseq(letters[1:6])
x

x2 <- x[c(2, 4, 6)]
x2

# named lookups return NULL for missing names
x3 <- as_flexseq(setNames(as.list(letters[1:4]), c("w", "x", "y", "z")))
x4 <- x3[c("y", "missing", "w")]
x4

# logical indexing supports recycling
x[c(TRUE, FALSE)]

# [[ extracts one element
x <- as_flexseq(letters[1:5])
x[[3]]

x2 <- as_flexseq(setNames(as.list(letters[1:3]), c("a1", "a2", "a3")))
x2[["a2"]]

# [<- replaces selected elements
x <- as_flexseq(1:6)
x

x2 <- x
x2[c(2, 5)] <- list(20, 50)
x2

# character replacement uses element names
x3 <- as_flexseq(setNames(as.list(1:4), c("a", "b", "c", "d")))
x3[c("d", "a")] <- list(40, 10)
x3

# logical replacement supports recycling
x4 <- x
x4[c(TRUE, FALSE, TRUE, FALSE, TRUE, FALSE)] <- list(1)
x4

# [[<- replaces one element
x <- as_flexseq(letters[1:4])
x2 <- x
x2[[2]] <- "ZZ"
x2

x3 <- as_flexseq(setNames(as.list(1:3), c("x", "y", "z")))
x3[["y"]] <- 99
x3

# assigning NULL removes one element
x4 <- as_flexseq(letters[1:4])
x4[[2]] <- NULL
x4
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-constructors.R
\name{as_ordered_sequence}
\alias{as_ordered_sequence}
\title{Build an Ordered Sequence from \code{x} and \code{keys}}
\usage{
as_ordered_sequence(x, keys)
}
\arguments{
\item{x}{Elements to add.}

\item{keys}{Key values with the same length as \code{x}.}
}
\value{
An \code{ordered_sequence}.
}
\description{
Constructs an \code{ordered_sequence} by pairing each element of \code{x} with the
corresponding key in \code{keys}.
}
\details{
Output is always sorted by key.

Duplicate keys are allowed; ties preserve input order (stable/FIFO within the
same key).

Names on \code{x} are preserved as element names.
}
\examples{
xs <- as_ordered_sequence(c("d", "a", "b", "a2"), keys = c(4, 1, 2, 1))
xs
length(elements_between(xs, 1, 1))

n <- as_ordered_sequence(setNames(as.list(c("a", "b")), c("ka", "kb")), keys = c(2, 1))
n[["kb"]]

# Keys can be other comparable types
num_by_chr <- as_ordered_sequence(c(20, 10, 30), keys = c("b", "a", "c"))
num_by_chr
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-query-api.R
\name{max_endpoint}
\alias{max_endpoint}
\title{Maximum Right Endpoint}
\usage{
max_endpoint(x)
}
\arguments{
\item{x}{An \code{interval_index}.}
}
\value{
Maximum right endpoint, or \code{NULL} when \code{x} is empty.
}
\description{
Returns the largest right endpoint (\code{end}) currently present.
}
\details{
Uses cached \code{.ivx_max_end} monoid state.
}
\examples{
ix <- interval_index("a", "b", start = c(3, 1), end = c(4, 2))
max_endpoint(ix)
max_endpoint(interval_index())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-core.R
\name{flexseq}
\alias{flexseq}
\title{Construct a Persistent Flexible Sequence}
\usage{
flexseq(...)
}
\arguments{
\item{...}{Sequence elements.}
}
\value{
A \code{flexseq} object.
}
\description{
\code{flexseq(...)} creates an immutable sequence from \code{...}, preserving element
order and optional names, with efficient persistent updates.
}
\details{
It is list-like in payload flexibility (any R object per element), but
sequence-oriented in API (\verb{push_*}, \verb{peek_*}, \verb{pop_*}, indexing, split/concat).

\code{flexseq} is the base general-purpose structure in this package.
Specialized structures such as \code{priority_queue}, \code{ordered_sequence}, and
\code{interval_index} build on related internals but expose narrower semantics.

\code{flexseq} operations are persistent: updates return new objects and do not
mutate prior versions.
}
\examples{
x <- flexseq(1, 2, 3)
x

y <- push_front(x, 0)
y
x  # unchanged

named <- flexseq(a = 1, b = 2)
named
named[["a"]]
}
\seealso{
\code{\link[=as_flexseq]{as_flexseq()}}, \code{\link[=priority_queue]{priority_queue()}}, \code{\link[=ordered_sequence]{ordered_sequence()}}, \code{\link[=interval_index]{interval_index()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{elements_between}
\alias{elements_between}
\title{Return Elements in a Key Range}
\usage{
elements_between(x, from_key, to_key, include_from = TRUE, include_to = TRUE)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{from_key}{Lower bound key.}

\item{to_key}{Upper bound key.}

\item{include_from}{Include lower bound when \code{TRUE}.}

\item{include_to}{Include upper bound when \code{TRUE}.}
}
\value{
An \code{ordered_sequence} of matched elements, in key order. Use
\code{\link[=as.list]{as.list()}} to convert to a plain list.
}
\description{
Return Elements in a Key Range
}
\details{
Range membership is controlled by \code{include_from} and \code{include_to}:
\itemize{
\item \code{include_from = TRUE} uses \code{key >= from_key}; otherwise \code{key > from_key}.
\item \code{include_to = TRUE} uses \code{key <= to_key}; otherwise \code{key < to_key}.
}

If no elements fall in the range, returns an empty \code{ordered_sequence}.
}
\examples{
x <- ordered_sequence("a", "b", "c", "d", keys = c(1, 2, 2, 3))
elements_between(x, 2, 3)
as.list(elements_between(x, 2, 3))
elements_between(x, 2, 2, include_to = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-query-api.R
\name{pop_overlaps}
\alias{pop_overlaps}
\title{Pop First Overlapping Interval}
\usage{
pop_overlaps(x, start, end, bounds = NULL)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{start}{Query interval start.}

\item{end}{Query interval end.}

\item{bounds}{Optional boundary override. One of \code{"[)"}, \code{"[]"}, \code{"()"}, \code{"(]"}.}
}
\value{
A list with \code{element}, \code{start}, \code{end}, and \code{remaining}.
On miss: \code{element}, \code{start}, and \code{end} are \code{NULL}.
}
\description{
Pop First Overlapping Interval
}
\details{
Removes the first match in canonical interval order. On miss, returns a
non-throwing miss object with \code{remaining = x}.
Use \code{\link[=pop_all_overlaps]{pop_all_overlaps()}} to remove all matches.
}
\examples{
ix <- interval_index("a", "b", "c", start = c(1, 3, 5), end = c(2, 4, 6))
pop_overlaps(ix, 2, 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-print.R
\name{print.flexseq}
\alias{print.flexseq}
\title{Print a flexseq}
\usage{
\method{print}{flexseq}(x, max_elements = 4L, show_custom_monoids = FALSE, ...)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{max_elements}{Maximum number of elements shown in preview (\code{head + tail}).}

\item{show_custom_monoids}{Logical; show attached non-default monoids and
their root cached measures.}

\item{...}{Passed through to per-element \code{print()}.}
}
\value{
The input \code{x}, returned invisibly. Called for its side effect
of printing a formatted preview of the sequence to the console.
}
\description{
Print a flexseq
}
\examples{
x <- as_flexseq(setNames(as.list(1:6), letters[1:6]))
print(x, max_elements = 4)

y <- as_flexseq(as.list(1:6))
print(y, max_elements = 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/70-plotting.R
\name{get_graph_df}
\alias{get_graph_df}
\title{Build graph data frames for a finger tree}
\usage{
get_graph_df(t)
}
\arguments{
\item{t}{FingerTree.}
}
\value{
A list with three elements: \code{edge_df} (parent/child/label),
\code{node_df} (node/type/label), and \code{node_data} — a named list keyed by node
id whose values are lists with fields \code{type}, \code{label}, \code{measures} (cached
monoid values for structural nodes; \code{NULL} for element leaves), and
\code{element} (the raw leaf entry for element nodes; \code{NULL} for structural
nodes).
}
\description{
Build graph data frames for a finger tree
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-ends.R
\name{push_back}
\alias{push_back}
\title{Push an Element to the Back}
\usage{
push_back(x, value)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{value}{Element to append.}
}
\value{
Updated \code{flexseq}.
}
\description{
Returns a new sequence with \code{value} appended at the right end.
}
\details{
This operation is persistent: \code{x} is not modified.

Elements can be named, but only if all are uniquely named (no missing
names).
}
\examples{
s <- as_flexseq(letters[1:3])
s2 <- push_back(s, "d")
s2
s  # unchanged

n <- as_flexseq(list(two = 2, three = 3))
new_el <- 4
names(new_el) <- "four"
push_back(n, new_el)

# Named/unnamed mixes are rejected
try(push_back(n, 5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-validate.R
\name{validate_name_state}
\alias{validate_name_state}
\title{Validate name-state invariants only (debug/test utility)}
\usage{
validate_name_state(t)
}
\arguments{
\item{t}{FingerTree.}
}
\value{
\code{TRUE} invisibly; errors if name invariants are violated.
}
\description{
Checks that trees are either fully unnamed or fully named with unique,
non-empty names.
}
\details{
Intended for debugging and tests, not hot runtime paths.
}
\examples{
x <- as_flexseq(setNames(as.list(letters[1:4]), letters[1:4]))
validate_name_state(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-ends.R
\name{pop_front}
\alias{pop_front}
\title{Pop the Front Element}
\usage{
pop_front(x)
}
\arguments{
\item{x}{A \code{flexseq}.}
}
\value{
A list with fields:
\itemize{
\item \code{value}: the first element, or \code{NULL} when \code{x} is empty.
\item \code{remaining}: the sequence after removing the first element.
}
}
\description{
Returns the first element and the remaining sequence.
}
\details{
This operation is persistent: \code{x} is not modified.

On empty input, returns a non-throwing miss object with
\code{value = NULL} and \code{remaining = x}.
}
\examples{
s <- flexseq("a", "b", "c")
out <- pop_front(s)
out$value
out$remaining
s  # unchanged

pop_front(flexseq())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-print.R
\name{print.FingerTree}
\alias{print.FingerTree}
\alias{print.Deep}
\alias{print.Single}
\alias{print.Empty}
\title{Print a compact summary of a finger tree}
\usage{
\method{print}{FingerTree}(x, max_elements = 4L, show_custom_monoids = FALSE, ...)

\method{print}{Deep}(x, ...)

\method{print}{Single}(x, ...)

\method{print}{Empty}(x, ...)
}
\arguments{
\item{x}{FingerTree.}

\item{max_elements}{Maximum number of elements shown in preview (\code{head + tail}).
Default \code{4}.}

\item{show_custom_monoids}{Logical; show attached non-default monoids and
their root cached measures. Default \code{FALSE}.}

\item{...}{Passed through to \code{print()} for preview elements.}
}
\value{
\code{x}, invisibly.
}
\description{
Print a compact summary of a finger tree
}
\examples{
x <- as_flexseq(setNames(as.list(1:6), letters[1:6]))
print(x, max_elements = 4)
sum_m <- measure_monoid(`+`, 0, function(el) as.numeric(el))
print(add_monoids(as_flexseq(1:3), list(sum = sum_m)), max_elements = 0, show_custom_monoids = TRUE)

y <- as_flexseq(as.list(1:6))
print(y, max_elements = 3)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-query-api.R
\name{pop_all_containing}
\alias{pop_all_containing}
\title{Pop All Containing Intervals}
\usage{
pop_all_containing(x, start, end, bounds = NULL)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{start}{Query interval start.}

\item{end}{Query interval end.}

\item{bounds}{Optional boundary override. One of \code{"[)"}, \code{"[]"}, \code{"()"}, \code{"(]"}.}
}
\value{
A list with \code{elements} and \code{remaining}, both \code{interval_index} objects.
}
\description{
Pop All Containing Intervals
}
\details{
Use \code{\link[=as.list]{as.list()}} to convert \code{elements} to a standard R list.
}
\examples{
ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(6, 5, 7))
out <- pop_all_containing(ix, 2, 4)
as.list(out$elements)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/immutables-package.R
\docType{package}
\name{Immutables-package}
\alias{Immutables}
\alias{Immutables-package}
\title{Immutables: Fast and Functional Data Structures}
\description{
The \code{Immutables} R package implements several immutable, or persistent, data
structures: operations return modified copies while remaining fast and true
to R's side-effect-free functional nature.
}
\details{
\itemize{
\item \code{flexseq}s provide list-like operations: indexed and named element access;
push/pop/peek from either end for double-ended queue behavior; insertion,
splitting, and concatenation.
\item \code{priority_queue}s associate items with priority values and provide
min and max peek/pop by priority and fast insertion.
\item \code{ordered_sequence}s associate items with key values and keep the elements in
sorted order by key. These may be similarly be inserted/popped/peeked by key
value as well as position. Keys may be duplicated, with first-in-first-out order within key groups.
\item \code{interval_index}es store items associated with interval ranges, supporting
point as well as interval overlaps/contains/within queries. Items are kept
in start-order enabling ordered sequence operations and sweep-line
algorithms.
}

Backed by monoid-annotated 2-3 fingertrees as described by
Hinze and Paterson, most operations are constant time, amortized constant time, or $O(\\log(n))$. Core functions are implemented in C++ (via Rcpp) for speed, with matching pure-R reference implementations using \code{lambda.r} syntax to match the paper.

Finally, the developer API supports the addition of custom structures via combinations of monoids and measures.
\itemize{
\item Hinze, R. and Paterson, R. (2006), \emph{Finger trees: a simple general-purpose data structure}. \doi{10.1017/S0956796805005769}
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://oneilsh.github.io/immutables/}
  \item \url{https://github.com/oneilsh/immutables}
  \item Report bugs at \url{https://github.com/oneilsh/immutables/issues}
}

}
\author{
\strong{Maintainer}: Shawn T. O'Neil \email{shawn@tislab.org}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-str.R
\name{str.flexseq}
\alias{str.flexseq}
\title{Display Internal Structure of a flexseq}
\usage{
\method{str}{flexseq}(object, ...)
}
\arguments{
\item{object}{A \code{flexseq}.}

\item{...}{Passed to \code{\link[utils:str]{utils::str()}}.}
}
\value{
\code{NULL}, invisibly.
}
\description{
Display Internal Structure of a flexseq
}
\examples{
x <- flexseq(a = 1, b = list(k = 2))
str(x)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-core.R
\name{as.list.flexseq}
\alias{as.list.flexseq}
\title{Coerce a Sequence to Base List}
\usage{
\method{as.list}{flexseq}(x, ...)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{...}{Unused.}
}
\value{
A base R list of sequence elements.
}
\description{
Returns elements in left-to-right sequence order.
}
\details{
Returns payload elements in sequence order. If the sequence is fully named,
those names are preserved on the returned list.
}
\examples{
x <- flexseq("a", "b", "c")
as.list(x)

n <- flexseq(a = 1, b = 2)
as.list(n)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-query-api.R
\name{peek_point}
\alias{peek_point}
\title{Peek First Interval Matching a Point}
\usage{
peek_point(
  x,
  point,
  bounds = NULL,
  match_at = c("interval", "start", "end", "either")
)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{point}{Query point.}

\item{bounds}{Optional boundary override. One of \code{"[)"}, \code{"[]"}, \code{"()"}, \code{"(]"}.
Ignored when \code{match_at} is not \code{"interval"}.}

\item{match_at}{How the query point is matched against each entry. One of
\code{"interval"} (default; entry interval contains \code{point}, under \code{bounds}),
\code{"start"} (entry's start coordinate equals \code{point}), \code{"end"} (entry's end
equals \code{point}), or \code{"either"} (start or end equals \code{point}). The three
coordinate-equality modes are structural and ignore \code{bounds}.}
}
\value{
The payload value from the first match, or \code{NULL} on no match.
}
\description{
Peek First Interval Matching a Point
}
\details{
Returns the first match in canonical interval order. Use \code{\link[=peek_all_point]{peek_all_point()}} to
retrieve all matches as an \code{interval_index} slice.
}
\examples{
ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(3, 2, 5))
peek_point(ix, 2)

# Boundary override at an endpoint
edge <- interval_index("a", start = 1, end = 3, default_query_bounds = "[)")
peek_point(edge, 3)                # default "[)": no match at right endpoint
peek_point(edge, 3, bounds = "[]") # closed bounds: endpoint matches

# Coordinate-equality modes (bounds irrelevant)
peek_point(ix, 2, match_at = "start")  # entry starting at 2
peek_point(ix, 3, match_at = "end")    # entry ending at 3
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-queue-ops.R
\name{max_priority}
\alias{max_priority}
\title{Maximum Priority Value}
\usage{
max_priority(x)
}
\arguments{
\item{x}{A \code{priority_queue}.}
}
\value{
Maximum priority value, or \code{NULL} when \code{x} is empty.
}
\description{
Returns the current maximum priority scalar in the queue.
}
\details{
Uses cached \code{.pq_max} monoid state.
}
\examples{
q <- priority_queue("a", "b", priorities = c(2, 1))
max_priority(q)
max_priority(priority_queue())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{pop_all_key}
\alias{pop_all_key}
\title{Pop All Elements for One Key}
\usage{
pop_all_key(x, key)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{key}{Query key.}
}
\value{
A list with fields:
\itemize{
\item \code{elements}: \code{ordered_sequence} of removed matches.
\item \code{remaining}: \code{ordered_sequence} after removal.
}
}
\description{
Removes and returns all elements whose key equals \code{key}.
}
\details{
Use \code{\link[=as.list]{as.list()}} to convert \code{elements} to a standard R list.
}
\examples{
x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
out <- pop_all_key(x, 2)
as.list(out$elements)
out$remaining
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-methods.R
\name{length.interval_index}
\alias{length.interval_index}
\title{Interval Index Length}
\usage{
\method{length}{interval_index}(x)
}
\arguments{
\item{x}{An \code{interval_index}.}
}
\value{
Number of indexed intervals.
}
\description{
Interval Index Length
}
\details{
Uses cached size metadata and runs in O(1).
}
\examples{
ix <- interval_index("a", "b", start = c(1, 3), end = c(2, 5))
length(ix)

length(interval_index())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{count_key}
\alias{count_key}
\title{Count Elements Matching One Key}
\usage{
count_key(x, key)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{key}{Query key.}
}
\value{
Integer count of matches.
}
\description{
Count Elements Matching One Key
}
\details{
Counts multiplicity for a single key. Returns \code{0L} when the key is not
present.
}
\examples{
x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
count_key(x, 2)
count_key(x, 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-query-api.R
\name{peek_all_point}
\alias{peek_all_point}
\title{Peek All Intervals Matching a Point}
\usage{
peek_all_point(
  x,
  point,
  bounds = NULL,
  match_at = c("interval", "start", "end", "either")
)
}
\arguments{
\item{x}{An \code{interval_index}.}

\item{point}{Query point.}

\item{bounds}{Optional boundary override. One of \code{"[)"}, \code{"[]"}, \code{"()"}, \code{"(]"}.
Ignored when \code{match_at} is not \code{"interval"}.}

\item{match_at}{How the query point is matched against each entry. One of
\code{"interval"} (default; containment under \code{bounds}), \code{"start"}, \code{"end"},
or \code{"either"}. See \code{\link[=peek_point]{peek_point()}} for details.}
}
\value{
An \code{interval_index} slice of all matches (possibly empty).
}
\description{
Peek All Intervals Matching a Point
}
\details{
The returned \code{interval_index} can be inspected with \code{\link[=as.list]{as.list()}}.
}
\examples{
ix <- interval_index("a", "b", "c", start = c(1, 2, 4), end = c(3, 2, 5))
as.list(peek_all_point(ix, 2))

# Entries ending at 3
as.list(peek_all_point(ix, 3, match_at = "end"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{min_key}
\alias{min_key}
\title{Minimum Key Value}
\usage{
min_key(x)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}
}
\value{
Minimum key, or \code{NULL} when \code{x} is empty.
}
\description{
Returns the smallest key currently present in the ordered sequence.
}
\details{
This follows sequence key order directly.
}
\examples{
x <- ordered_sequence("a", "b", keys = c(2, 1))
min_key(x)
min_key(ordered_sequence())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-validate.R
\name{validate_tree}
\alias{validate_tree}
\title{Validate full tree invariants (debug/test utility)}
\usage{
validate_tree(t)
}
\arguments{
\item{t}{FingerTree.}
}
\value{
\code{TRUE} invisibly; errors if invariant violations are found.
}
\description{
Performs expensive full-tree auditing of:
\itemize{
\item structural attributes (\code{monoids}/\code{measures}) consistency
\item global name-state invariants
}
}
\details{
Intended for debugging and tests, not hot runtime paths.
}
\examples{
x <- as_flexseq(letters[1:10])
validate_tree(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-iter.R
\name{loop}
\alias{loop}
\title{Iterate over an iterator (re-exported from coro)}
\usage{
loop(loop)
}
\arguments{
\item{loop}{A \code{for} loop expression.}
}
\value{
\code{NULL}, invisibly. Called for side effects.
}
\description{
Re-exported \code{\link[coro:collect]{coro::loop()}}. Enables \code{for}-loop-style iteration over
\code{immutables} structures without needing to load \code{coro} separately.
}
\examples{
s <- flexseq(1, 2, 3)
loop(for (x in s) print(x))
}
\seealso{
\code{\link[coro:collect]{coro::loop()}} for full documentation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/20-api-generics.R
\name{insert}
\alias{insert}
\title{Insert an Element}
\usage{
insert(x, ...)
}
\arguments{
\item{x}{Object to insert into.}

\item{...}{Method-specific arguments.}
}
\value{
Updated object of the same class as \code{x}.
}
\description{
Inserts an element into a structure-specific position according to class
semantics.
}
\details{
\code{insert()} is an S3 generic. Required arguments in \code{...} depend on \code{x}:
\itemize{
\item \code{priority_queue}: \code{element}, \code{priority} (optional \code{name})
\item \code{ordered_sequence}: \code{element}, \code{key} (optional \code{name})
\item \code{interval_index}: \code{element}, \code{start}, \code{end} (optional \code{name})
}

This operation is persistent: \code{x} is not modified.
}
\examples{
q <- priority_queue("a", "b", priorities = c(2, 1))
insert(q, "c", priority = 3)

o <- ordered_sequence("a", "c", keys = c(1, 3))
insert(o, "b", key = 2)

iv <- interval_index("A", "B", start = c(1, 5), end = c(3, 8))
insert(iv, "C", start = 2, end = 6)
}
\seealso{
\code{\link[=priority_queue]{priority_queue()}}, \code{\link[=ordered_sequence]{ordered_sequence()}}, \code{\link[=interval_index]{interval_index()}},
\code{\link[=insert_at]{insert_at()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/60-interval_index-constructors.R
\name{interval_index}
\alias{interval_index}
\title{Construct an Interval Index}
\usage{
interval_index(..., start, end, default_query_bounds = "[)")
}
\arguments{
\item{...}{Elements to add.}

\item{start}{Start endpoints matching \code{...}.}

\item{end}{End endpoints matching \code{...}.}

\item{default_query_bounds}{Boundary convention used as the default for
query operations on this index: one of \code{"[)"}, \code{"[]"}, \code{"()"}, \code{"(]"}.
Per-query \verb{peek_*} / \verb{pop_*} calls may override via their own \code{bounds}
argument.}
}
\value{
An \code{interval_index}.
}
\description{
Convenience constructor from \code{...}, \code{start}, and \code{end}.
}
\details{
Empty construction is supported: \code{interval_index()} returns an empty index.

Output is ordered by interval \code{start}.
}
\examples{
ix <- interval_index("a", "b", "c", start = c(1, 2, 2), end = c(3, 2, 4))
ix

interval_index()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{pop_key}
\alias{pop_key}
\title{Pop First Element for One Key}
\usage{
pop_key(x, key)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{key}{Query key.}
}
\value{
A list with fields:
\itemize{
\item \code{value}: removed element, or \code{NULL} on miss.
\item \code{key}: removed key, or \code{NULL} on miss.
\item \code{remaining}: ordered sequence after removal.
}
}
\description{
Removes and returns the first element whose key equals \code{key}.
}
\details{
For duplicate keys, the first element in stable sequence order is removed.
}
\examples{
x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
out <- pop_key(x, 2)
out$value
out$remaining
pop_key(x, 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{peek_key}
\alias{peek_key}
\title{Peek First Element for One Key}
\usage{
peek_key(x, key)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{key}{Query key.}
}
\value{
Matched element, or \code{NULL} when no matching key exists.
}
\description{
Returns the first element whose key equals \code{key}.
}
\details{
For duplicate keys, this returns the first element in stable sequence order.
}
\examples{
x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
peek_key(x, 2)
peek_key(x, 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-empty-tree.R
\name{empty_tree}
\alias{empty_tree}
\title{Create an Empty Structural Tree}
\usage{
empty_tree(monoids = NULL)
}
\arguments{
\item{monoids}{Optional named list of \code{measure_monoid} objects.}
}
\value{
An empty finger tree with structural \code{monoids} and \code{measures} attrs.
}
\description{
Create an Empty Structural Tree
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-core.R
\name{unlist.flexseq}
\alias{unlist.flexseq}
\title{Coerce a Sequence to an Atomic Vector}
\usage{
\method{unlist}{flexseq}(x, recursive = TRUE, use.names = TRUE)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{recursive}{Passed through to \code{\link[base:unlist]{base::unlist()}}.}

\item{use.names}{Passed through to \code{\link[base:unlist]{base::unlist()}}.}
}
\value{
An atomic vector built from \code{\link[=as.list.flexseq]{as.list.flexseq()}}.
}
\description{
Convenience wrapper around \code{\link[base:unlist]{base::unlist()}} over \code{\link[=as.list]{as.list()}}.
}
\details{
For \code{priority_queue}, this unwraps queue entries to payload items before
unlisting (equivalent to \code{unlist(as.list(x, drop_meta = TRUE), ...)}).

Inherited by \code{ordered_sequence} and \code{interval_index} through the shared
class stack.
}
\examples{
x <- flexseq(1, 2, 3)
unlist(x)

q <- priority_queue("a", "b", priorities = c(2, 1))
unlist(q)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-core.R
\name{length.flexseq}
\alias{length.flexseq}
\title{Sequence Length}
\usage{
\method{length}{flexseq}(x)
}
\arguments{
\item{x}{A \code{flexseq}.}
}
\value{
Number of elements in the sequence.
}
\description{
Sequence Length
}
\details{
Uses cached size metadata and runs in O(1).
}
\examples{
x <- flexseq("a", "b")
length(x)

length(flexseq())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{upper_bound}
\alias{upper_bound}
\title{Find First Element with Key \code{>} Query}
\usage{
upper_bound(x, key)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{key}{Query key.}
}
\value{
A list with fields:
\itemize{
\item \code{found}: logical flag.
\item \code{index}: one-based position of the first match, or \code{NULL}.
\item \code{value}: matched element, or \code{NULL}.
\item \code{key}: matched key, or \code{NULL}.
}
}
\description{
Find First Element with Key \code{>} Query
}
\details{
\code{upper_bound()} finds the first element with key \verb{> key}. This skips exact
key matches, which is useful for exclusive range endpoints and for finding
the position immediately after a duplicate-key run.
}
\examples{
x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
upper_bound(x, 2)
upper_bound(x, 10)
}
\seealso{
\code{\link[=lower_bound]{lower_bound()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-ops.R
\name{peek_all_key}
\alias{peek_all_key}
\title{Peek All Elements for One Key}
\usage{
peek_all_key(x, key)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{key}{Query key.}
}
\value{
An \code{ordered_sequence} containing all matches; empty on miss.
}
\description{
Returns all elements whose key equals \code{key}.
}
\details{
The returned \code{ordered_sequence} can be inspected with \code{\link[=as.list]{as.list()}}.
}
\examples{
x <- ordered_sequence("a", "b", "c", keys = c(1, 2, 2))
out <- peek_all_key(x, 2)
as.list(out)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-ends.R
\name{peek_front}
\alias{peek_front}
\title{Peek at the Front Element}
\usage{
peek_front(x)
}
\arguments{
\item{x}{A \code{flexseq}.}
}
\value{
First element, or \code{NULL} when \code{x} is empty.
}
\description{
Returns the first element without modifying the sequence.
}
\details{
Returns the payload element without modifying \code{x}.
}
\examples{
x <- flexseq("a", "b", "c")
peek_front(x)

peek_front(flexseq())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/50-ordered_sequence-print.R
\name{print.ordered_sequence}
\alias{print.ordered_sequence}
\title{Print an Ordered Sequence Summary}
\usage{
\method{print}{ordered_sequence}(x, max_elements = 4L, show_custom_monoids = FALSE, ...)
}
\arguments{
\item{x}{An \code{ordered_sequence}.}

\item{max_elements}{Maximum number of elements shown in the preview.}

\item{show_custom_monoids}{Logical; show attached non-default monoids and
their root cached measures.}

\item{...}{Passed through to per-element \code{print()}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Prints a compact summary and head/tail preview in key order.
}
\examples{
xs <- ordered_sequence(one = "a", two = "b", three = "c", keys = c(20, 30, 10))
print(xs, max_elements = 4)
sum_key <- measure_monoid(`+`, 0, function(entry) as.numeric(entry$key))
ys2 <- add_monoids(ordered_sequence("a", "b", keys = c(2, 1)), list(sum_key = sum_key))
print(ys2, max_elements = 0, show_custom_monoids = TRUE)

ys <- ordered_sequence("x", "y", "z", keys = c(2, 1, 3))
print(ys, max_elements = 3)

print(ordered_sequence())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-constructors.R
\name{as_priority_queue}
\alias{as_priority_queue}
\title{Build a Priority Queue from \code{x} and \code{priorities}}
\usage{
as_priority_queue(x, priorities)
}
\arguments{
\item{x}{Elements to enqueue.}

\item{priorities}{Priorities with the same length as \code{x}.}
}
\value{
A \code{priority_queue}.
}
\description{
Constructs a queue by pairing each element of \code{x} with the corresponding
value in \code{priorities}.
}
\details{
\code{x} is interpreted element-wise (via list coercion). Names on \code{x} are
preserved as queue element names.

All priorities must be non-missing and mutually comparable.
}
\examples{
x <- as_priority_queue(letters[1:4], priorities = c(3, 1, 2, 1))
x
peek_min(x)
peek_max(x)

# Names are preserved
n <- as_priority_queue(setNames(as.list(1:3), c("a", "b", "c")), priorities = c(2, 1, 3))
n[["b"]]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-core.R
\name{merge.flexseq}
\alias{merge.flexseq}
\title{Merge Two Sequences}
\usage{
\method{merge}{flexseq}(x, y, ...)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{y}{A \code{flexseq}.}

\item{...}{Unused.}
}
\value{
A new \code{flexseq}.
}
\description{
Returns a new \code{flexseq} containing all elements of \code{x} followed by all
elements of \code{y}. Thin wrapper over \code{\link[=c]{c()}} for API uniformity across the
package's merge methods; \code{c(x, y)} and \code{merge(x, y)} are equivalent for
\code{flexseq}.
}
\details{
For ordered types (\code{ordered_sequence}, \code{interval_index}), \code{merge()} performs
a proper sorted merge respecting keys/intervals — see \code{\link[=merge.ordered_sequence]{merge.ordered_sequence()}}
and \code{\link[=merge.interval_index]{merge.interval_index()}}. For \code{priority_queue}, see
\code{\link[=merge.priority_queue]{merge.priority_queue()}}.
}
\examples{
x <- flexseq("a", "b")
y <- flexseq("c", "d")
merge(x, y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-indexing.R
\name{sub-.priority_queue}
\alias{sub-.priority_queue}
\alias{[.priority_queue}
\alias{[[.priority_queue}
\alias{[<-.priority_queue}
\alias{[[<-.priority_queue}
\alias{$.priority_queue}
\alias{$<-.priority_queue}
\title{Indexing for Priority Queues}
\usage{
\method{[}{priority_queue}(x, i, ...)

\method{[[}{priority_queue}(x, i, ...)

\method{[}{priority_queue}(x, i) <- value

\method{[[}{priority_queue}(x, i) <- value

\method{$}{priority_queue}(x, name)

\method{$}{priority_queue}(x, name) <- value
}
\arguments{
\item{x}{A \code{priority_queue}.}

\item{i}{Index input. For reads, must be a character name (scalar for \code{[[}).}

\item{...}{Unused.}

\item{value}{Replacement value (unsupported).}

\item{name}{Element name (for \code{$} and \verb{$<-}).}
}
\value{
For \code{$}/\code{[[}/\code{[}: queue payload values or queue subsets by name.
Replacement forms always error.
}
\description{
Name-based indexing is supported for reads only. Positional indexing and all
replacement indexing are intentionally blocked to preserve queue-first UX.
}
\details{
\code{priority_queue} supports name-based read indexing only.
\itemize{
\item \code{[}: character vector of names, returns a \code{priority_queue} subset.
\item \code{[[} and \code{$}: scalar name, return the payload value.
\item Positional indexing and all replacement indexing forms are unsupported.
}
}
\examples{
q <- priority_queue(a = "task-a", b = "task-b", priorities = c(2, 1))

q["a"]
q[["b"]]
q$b

try(q[1])
try(q$a <- "updated")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-get-measure.R
\name{get_measures}
\alias{get_measures}
\title{Read Per-Element Monoid Measures}
\usage{
get_measures(x, monoid_name)
}
\arguments{
\item{x}{A flexseq (or any immutables subclass).}

\item{monoid_name}{Name of an attached monoid.}
}
\value{
A plain unnamed \code{flexseq} of length \code{length(x)}. Entry \code{i} is
the monoid's \code{measure()} applied to the \code{i}-th leaf entry.
}
\description{
Applies the named monoid's \code{measure()} function to each leaf entry in
sequence order, returning the results as a new \code{flexseq}.
}
\details{
Each leaf in a finger tree stores a structure-specific \emph{entry}
(flexseq: the raw element; ordered_sequence: \code{list(value, key)};
priority_queue: \code{list(value, priority)}; interval_index:
\code{list(value, start, end)}). A monoid's \code{measure()} function is defined
over that entry shape, and this accessor exposes the per-element
values that would be combined to produce the cached aggregate.

Returns a \code{flexseq} rather than a base list so results can be piped
back into other immutables operations; use \code{\link[=as.list]{as.list()}} or \code{\link[=unlist]{unlist()}}
to convert.
}
\examples{
sum_m <- measure_monoid(`+`, 0, function(el) el)
x <- add_monoids(as_flexseq(c(3, 1, 4, 1, 5, 9)), list(sum = sum_m))
get_measures(x, "sum") |> unlist()

q <- priority_queue("a", "b", "c", priorities = c(5, 1, 3))
# Per-element .pq_min measures — each is list(has, priority).
get_measures(q, ".pq_min")
}
\seealso{
\code{\link[=get_measure]{get_measure()}}, \code{\link[=measure_monoid]{measure_monoid()}}, \code{\link[=fapply]{fapply()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/30-api-flexseq-ends.R
\name{insert_at}
\alias{insert_at}
\title{Insert Elements at a Position}
\usage{
insert_at(x, index, values)
}
\arguments{
\item{x}{A \code{flexseq}.}

\item{index}{One-based insertion position in \verb{[1, length(x) + 1]}.}

\item{values}{Values to insert.}
}
\value{
Updated sequence with inserted values.
}
\description{
Inserts \code{values} before the current element at \code{index}.
}
\details{
\code{values} is interpreted as a collection of elements to splice in.

Common cases:
\itemize{
\item Atomic vector (\code{c("x", "y")}): inserts one element per vector entry.
\item List (\code{list("x", "y")}): inserts one element per list entry.
\item \code{flexseq}: inserts all of its elements.
\item Empty input (\code{list()} or \code{flexseq()}): no change.
}

To insert one composite object (for example, a vector or a list) as a single
element, wrap it in \code{list(...)}.

This operation is persistent: \code{x} is not modified.
}
\examples{
x <- flexseq("a", "b", "c", "d")
insert_at(x, 3, c("x", "y"))
insert_at(x, 1, "start")
insert_at(x, length(x) + 1, "end")

# Insert one vector as a single element
insert_at(x, 3, list(c("u", "v")))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/20-api-generics.R
\name{as_flexseq}
\alias{as_flexseq}
\title{Coerce Objects to \code{flexseq}}
\usage{
as_flexseq(x)
}
\arguments{
\item{x}{Input object.}
}
\value{
A plain \code{flexseq}.
}
\description{
\code{as_flexseq()} is the canonical way to obtain a plain \code{flexseq} for full
sequence-style operations.
}
\details{
For base vectors/lists, this builds a new \code{flexseq} preserving element order
and names.

For specialized immutable subclasses (\code{priority_queue},
\code{ordered_sequence}, \code{interval_index}), this intentionally drops subclass
semantics and returns a plain \code{flexseq}.

This is an S3 generic. Notable method behavior:
\itemize{
\item \code{as_flexseq.flexseq(x)} returns \code{x} unchanged.
\item \code{as_flexseq.priority_queue(x)} returns payload items.
\item \code{as_flexseq.ordered_sequence(x)} returns payload items.
\item \code{as_flexseq.interval_index(x)} returns payload items.
}

For advanced types, custom monoids are dropped and the rebuilt \code{flexseq}
keeps only structural monoids (\code{.size}, \code{.named_count}). For
\code{priority_queue}, a flexseq of full entry records can be obtained by
composing with \code{\link[=as.list]{as.list()}}: \code{as_flexseq(as.list(x))}. For
\code{ordered_sequence} and \code{interval_index}, \code{as.list()} also returns
payload-only lists, so no direct record-preserving cast is provided.
}
\examples{
x <- as_flexseq(1:3)
x

q <- priority_queue("a", "b", priorities = c(2, 1))
as_flexseq(q)

o <- ordered_sequence("a", "b", keys = c(2, 1))
as_flexseq(o)
}
\seealso{
\code{\link[=flexseq]{flexseq()}}, \code{\link[=priority_queue]{priority_queue()}}, \code{\link[=ordered_sequence]{ordered_sequence()}}, \code{\link[=interval_index]{interval_index()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-methods.R
\name{as.list.priority_queue}
\alias{as.list.priority_queue}
\title{Coerce Priority Queue to List}
\usage{
\method{as.list}{priority_queue}(x, ..., drop_meta = FALSE)
}
\arguments{
\item{x}{A \code{priority_queue}.}

\item{...}{Unused.}

\item{drop_meta}{Logical scalar. When \code{FALSE} (default), returns full queue
entry records (\code{value} + \code{priority}). When \code{TRUE}, returns payload values
only.}
}
\value{
A plain list of queue entry records (\code{drop_meta = FALSE}) or payload
values (\code{drop_meta = TRUE}).
}
\description{
Returns queue entries as a plain list of records with fields \code{value} and
\code{priority}, in queue sequence order.
}
\details{
Each returned entry is a record with fields \code{value} and \code{priority}.
Entry names (when present) are preserved on the returned list.
}
\examples{
q <- priority_queue("a", "b", priorities = c(2, 1))
as.list(q)
as.list(q, drop_meta = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/40-priority_queue-queue-ops.R
\name{pop_max}
\alias{pop_max}
\title{Pop Maximum-Priority Element}
\usage{
pop_max(x)
}
\arguments{
\item{x}{A \code{priority_queue}.}
}
\value{
A list with fields:
\itemize{
\item \code{value}: removed element, or \code{NULL} when \code{x} is empty.
\item \code{priority}: removed priority, or \code{NULL} when \code{x} is empty.
\item \code{remaining}: queue after removal.
}
}
\description{
Removes one maximum-priority element and returns it with the remaining queue.
}
\details{
Ties are stable: when multiple elements share maximum priority, the earliest
element in queue order is removed.
}
\examples{
x <- priority_queue("a", "b", "c", priorities = c(2, 3, 3))
out <- pop_max(x)
out$value
out$priority
out$remaining
pop_max(priority_queue())
}
